//  The Worldforge Project
//  Copyright (C) 1998, 1999,2000 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 07/2000

#ifndef _DII_METHOD_H_
#define _DII_METHOD_H_

#include "object.h"
#include <sigc++/slot.h>

namespace DII {

///
class AbstractFunction
{
 public:
  ///
  virtual DII::Object operator() (const DII::Object&) = 0;
};
 
/**FunctionHandle.
   Holds the pointer to a memberfunction of class T.
   I is the type of the parameter, O the output's type.
 */
template<class T> class FunctionHandle : public AbstractFunction
{
 
 protected:
    ///
    typedef DII::Object (T::*Func_ptr)(const DII::Object&);
    ///
    T* obj;
    ///
    Func_ptr func_ptr;

 public:
    ///
    FunctionHandle(T* o, Func_ptr f) : obj(o), func_ptr(f) {};
    /// 
    virtual ~FunctionHandle() {};
    ///
    Func_ptr ptr() const { return func_ptr; };
    ///
    virtual DII::Object operator () (const DII::Object& in) 
      { 
	return (obj->*func_ptr)(in); 
      }    
};

template<class R, class P>
class SlotHandle : public AbstractFunction
{
 public:
  SlotHandle(SigC::Slot1<R,P> slot) : m_slot(slot) {}
  virtual ~SlotHandle() {}

  virtual DII::Object operator () (const DII::Object& in) 
    {return m_slot.call(ObjConv<P>::convert(in));}

 private:
  SigC::Slot1<R,P> m_slot;
};

template<class R>
class SlotHandle<R,void> : public AbstractFunction
{
 public:
  SlotHandle(SigC::Slot0<R> slot) : m_slot(slot) {}
  virtual ~SlotHandle() {}

  virtual DII::Object operator () (const DII::Object& in) 
    {return m_slot.call();}

 private:
  SigC::Slot0<R> m_slot;
};

template<class P>
class SlotHandle<void,P> : public AbstractFunction
{
 public:
  SlotHandle(SigC::Slot1<void,P> slot) : m_slot(slot) {}
  virtual ~SlotHandle() {}

  virtual DII::Object operator () (const DII::Object& in) 
    {m_slot.call(ObjConv<P>::convert(in)); return true;}

 private:
  SigC::Slot1<void,P> m_slot;
};

template<>
class SlotHandle<void,void> : public AbstractFunction
{
 public:
  SlotHandle(SigC::Slot0<void> slot) : m_slot(slot) {}
  virtual ~SlotHandle() {}

  virtual DII::Object operator () (const DII::Object& in) 
    {m_slot.call(); return true;}

 private:
  SigC::Slot0<void> m_slot;
};

template<class C, class R, class P>
inline SlotHandle<R,P>* GetSlotHandle(C *c, R(C::*func)(P))
{
  return new SlotHandle<R,P>(SigC::slot(c, func));
}

template<class C, class R>
inline SlotHandle<R,void>* GetSlotHandle(C *c, R(C::*func)())
{
  return new SlotHandle<R,void>(SigC::slot(c, func));
}

template<class R, class P>
inline SlotHandle<R,P>* GetSlotHandle(R(*func)(P))
{
  return new SlotHandle<R,P>(SigC::slot(func));
}

template<class R>
inline SlotHandle<R,void>* GetSlotHandle(R(*func)())
{
  return new SlotHandle<R,void>(SigC::slot(func));
}

/**Method.
   Is a function handle for member functions of class Module,
   which takes an constDII::Object& and returnsDII::Object.
*/ 
typedef AbstractFunction* Method;

///some standard return values
//enum { FAILURE = 0, SUCCESS = 1 };
extern DII::Object FAILURE;
extern DII::Object SUCCESS;
}


#endif // !_DII_METHOD_H_
