//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 12/1999

#include "module.h"
#include "config.h"
#include <varconf/Config.h>

#include <string.h>
#include <list>

using std::cerr;
using std::endl;

namespace DII {

#ifdef ATLAS_OBJECT_HAS_LONG
Object FAILURE = 0l;
Object SUCCESS = 1l;
#else
Object FAILURE = 0;
Object SUCCESS = 1;
#endif

/* If we use const char* as the key we run into problems, as usually the const char* pointers
 * passed to the functions are not really "const", but temporary variables.
 * At least temporary when compared to the lifetime of this databases...
 * 
 */
typedef STRING_HASH_MAP(Module*) moduleDB;

typedef struct {
  Method service;
  Module* module;
} ServiceEntry;

typedef LOCAL_SLIST(ServiceEntry) ServiceEntrySList;
typedef STRING_HASH_MAP(ServiceEntrySList) serviceDB;

moduleDB modules;
serviceDB services;
  
StringSList queryServices()
{
  StringSList List;
  serviceDB::const_iterator itr = services.begin();
  while(itr != services.end())
    {
      List.push_front(std::string(itr->first));
      itr++;
    }
  return List;
}

Module* lookupModule(const char* modname)
{
  moduleDB::iterator itr = modules.find(modname);
  if(itr != modules.end())
    return itr->second;

  cerr << "DII failed to lookup module "<<modname<<endl;
  return NULL;
}

Method lookupMethod(Module* mod, const char* methname)
{
  if(!mod)
    {
      cerr << "DII was passed a NULL pointer as module reference !"<<endl;
      return NULL;
    }

  return mod->lookupMethod(methname);
}

Method lookupMethod(const char* modname, const char* methname)
{
  moduleDB::iterator itr = modules.find(modname);
  if(itr != modules.end())
    return itr->second->lookupMethod(methname);
  return NULL;
}

Method lookupService(const char* methname)
{
  serviceDB::iterator itr = services.find(methname);
  if(itr != services.end())
    if(!itr->second.empty())
      return itr->second.front().service;

  cerr << "DII failed to lookup service "<<methname<<endl;
  return NULL;
}


void registerService(Module* mod, const char* methname)
{
  if(!mod || !methname)
    { 
      cerr << "DII failed to register service "<<methname<<endl;
      return;
    }
  ServiceEntry entry;
  entry.module = mod;
  entry.service = mod->lookupMethod(methname);
  
  services[methname].push_front(entry);
}



DII::Object callMethod(const char* modname,
		  const char* methname,
		  const DII::Object& in)
{
  try
    {
      moduleDB::iterator itr = modules.find(modname);
      if(itr != modules.end())
	return itr->second->callMethod(methname, in);
    }
  catch(...)
    {
#ifdef DEBUG
      std::cerr << "DII: calling "<<modname<<"::"<<methname<<" caused exception !"
	   << std::endl;
#endif
    }

  return DII::FAILURE;
}


DII::Object callService(const char* servicename, const DII::Object& in)
{
  try
    {
      serviceDB::iterator itr = services.find(servicename);
      if(itr != services.end())
	{ 
	  if(!itr->second.empty())
	    {
	      return (*(itr->second.front().service))(in);
	    }
	}
      cerr << "DII failed to call service \""<<servicename<<"\"."<<endl;
    }
  catch(...)
    {
#ifdef DEBUG
      std::cerr << "DII: calling service "<<servicename<<" caused exception !"
	   << std::endl;
#endif
    }

  return DII::FAILURE;
}


//******************************************************
// private functions

class service_match : public std::unary_function<ServiceEntry,bool>
{
public:
  service_match(Module* mod) : mod_(mod) {};
  ~service_match() {};
  bool operator()(const ServiceEntry& entry)
  { 
    return (entry.module == mod_); 
  }
  Module* mod_;

};

struct slist_empty : public std::unary_function<ServiceEntrySList, bool>
{
  bool operator() (const ServiceEntrySList& entry)
  {
    return entry.empty();
  }
};

void installModule(Module* mod)
{
  if(mod)
    modules[mod->name().c_str()]=mod;
  else
    cerr << "DII was passed a NULL pointer as module reference !"<<endl;
}


void removeModule(Module* mod)
{
  moduleDB::iterator itr = modules.find(mod->name().c_str());
  if(itr != modules.end())
    {
      std::list<std::string> toDelete;
      //remove this module from all services
      serviceDB::iterator itr2 = services.begin();
      while(itr2 != services.end())
	{
	  itr2->second.remove_if(service_match(mod));
	  if(itr2->second.empty())
	    toDelete.push_back(itr2->first);
	  itr2++;
	}

      //remove empty entries
      std::list<std::string>::iterator p = toDelete.begin();
      while(p != toDelete.end())
	{
	  itr2 = services.find(*p);
	  services.erase(itr2);
	  p++;
	}

      //remove the module from the database
      modules.erase(itr); 
    }
}

//****************************************************

//the Module class implementation
Module::Module(const std::string& name):
  name_(name),
  sem_(0)
{
  DII::installModule(this);

  varconf::Config::inst()->
    sigv.connect(SigC::slot(this, &Module::configCallback));

  /*  REGISTER_METHOD(Module, getOptionsMap); */
  methods_["getOptionsMap"] = 
    new DII::FunctionHandle<Module>(this,
				    &Module::getOptionsMap);

}

Module::~Module()
{
  DII::removeModule(this);
}


Method
Module::lookupMethod(const char* name)
{
  Methods::iterator itr = methods_.find(name);
  if(itr != methods_.end())
    {
      return itr->second;
    }
  
  cerr << "Module \""<<name_<<"\" does not provide method \""<<name
       <<"\"."<<endl;

  return NULL;

}


DII::Object
Module::callMethod(const char* name, const DII::Object& input)
{
  try
    { 
      Methods::iterator itr = methods_.find(name);
      if(itr != methods_.end())
        {
          Module::lock();
          DII::Object result = (*(itr->second))(input);
          Module::release();
          return result;
        }
    }
  catch(...) {} // probably passed in the wrong type of DII::Object
  return DII::FAILURE;
}

std::string lowcase(const std::string& in)
{
  std::string result = in;
  std::string::iterator itr = result.begin();
  while(itr != result.end())
    {
      *itr = tolower(*itr);
      itr++;
    }
  return result;
}

void
Module::configCallback(const std::string& name, const std::string&)
{
  // if an option of this module's section has been changed, call 
  // Module::configure

  // have to use strcasecmp as varconf converts all names to all lower case
  if(lowcase(name) == lowcase(name_))
    configure();
}


  
void
Module::setOption(const std::string& optname, double value)
{
  varconf::Config::inst()->setItem(lowcase(name_), lowcase(optname), value);
}

void
Module::setOption(const std::string& optname, int value)
{
  varconf::Config::inst()->setItem(lowcase(name_), lowcase(optname), value);
}

void
Module::setOption(const std::string& optname, bool value)
{
  varconf::Config::inst()->setItem(lowcase(name_), lowcase(optname), value);
}

void
Module::setOption(const std::string& optname, const std::string& value)
{
  varconf::Config::inst()->setItem(lowcase(name_), lowcase(optname), value);
}

bool
Module::getOption(const std::string& optname, double* value)
{
  if(!varconf::Config::inst()->findItem(lowcase(name_), lowcase(optname)))
    return false;
  varconf::Variable var = varconf::Config::inst()->getItem(lowcase(name_), lowcase(optname));
//   if(!var.is_double())
//     return false;
  *value = double ( var );
  return true;
}

bool
Module::getOption(const std::string& optname, int* value)
{
  if(!varconf::Config::inst()->findItem(lowcase(name_), lowcase(optname)))
    return false;
  varconf::Variable var = varconf::Config::inst()->getItem(lowcase(name_), lowcase(optname));
//   if(!var.is_int())
//     return false;
  *value = int ( var );
  return true;
}
 
bool 
Module::getOption(const std::string& optname, bool* value)
{
  if(!varconf::Config::inst()->findItem(lowcase(name_), lowcase(optname)))
    return false;
  varconf::Variable var = varconf::Config::inst()->getItem(lowcase(name_), lowcase(optname));
//   if(!var.is_bool())
//     return false;
  *value = bool ( var );

  return true;
}

bool
Module::getOption(const std::string& optname, std::string* value)
{
   if(!varconf::Config::inst()->findItem(lowcase(name_), lowcase(optname)))
    return false;
  varconf::Variable var = varconf::Config::inst()->getItem(lowcase(name_), lowcase(optname));
//   if(!var.is_string())
//     return false;
  *value = std::string ( var );
  return true;
}

DII::Object 
Module::getOptionsMap(const DII::Object&)
{
  DII::Object::MapType result;
  OptionsMap::const_iterator itr = optionsMap_.begin();
  while(itr != optionsMap_.end())
    {
      DII::Object::MapType entry;
      entry["description"] = itr->second.second;
      switch(itr->second.first)
	{
	case Module::BOOL:
	  {
	    bool tmp = false;
	    getOption(itr->first, &tmp);
	    entry[itr->first] = tmp;
	  }
	  break;

	case Module::INT:
	  {
	    int tmp = 0;
	    getOption(itr->first, &tmp);
	    entry[itr->first] = tmp;
	  }
	  break;
	case Module::FLOAT:
	  {
	    double tmp = 0.0;
	    getOption(itr->first, &tmp);
	    entry[itr->first] = tmp;
	  }
	  break;
	case Module::STRING:
	  {
	    std::string tmp = "";
	    getOption(itr->first, &tmp);
	    entry[itr->first] = tmp;
	  }
	  break;
	}
      result[itr->first] = entry;
      itr++;
    }
  return result;
}

}  // namespace DII




