//  The Worldforge Project
//  Copyright (C) 1998, 1999,2000 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 12/1999


#ifndef _MODULE_H_
#define _MODULE_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string>
#include <map>
#include "toolbox/local_slist.h"
#include "toolbox/local_hash_map.h"

#include <sigc++/signal_system.h>
#include <varconf/Config.h>

#include "method.h"
#include "object.h"
#include "message.h"

namespace DII {

std::string lowcase(const std::string& in);

template<class T>
void SetOption(const std::string &modname, const std::string &optname, T value)
{
  varconf::Config::inst()->setItem(lowcase(modname), lowcase(optname), value);
}

template<class T>
bool GetOption(const std::string &modname, const std::string &optname, T& value)
{
  if(!varconf::Config::inst()->findItem(lowcase(modname), lowcase(optname)))
    return false;
  varconf::Variable var = varconf::Config::inst()->getItem(lowcase(modname),
    lowcase(optname));
  value = T(var);
  return true;
}

/**Modul class.
   This class stores all the module specific data, like name, dependencies,
   method repository. Each module must contain an implementation of an
   superclass with the same name as the module.
*/
class Module : virtual public SigC::Object
{
 public:
  ///
  Module(const std::string& name);
  ///
  virtual ~Module();
  ///
  virtual DII::Object init(const DII::Object&) {return DII::SUCCESS;}
  ///
  virtual DII::Object deinit(const DII::Object&) {return DII::SUCCESS;}
  ///
  Method lookupMethod(const char* name);
  ///
  DII::Object callMethod(const char* name, const DII::Object& input = 0);
  /**send out a signal carrying a Object.
     The signal is send to all connected slots until a slot returns true.
  */ 
  SigC::Signal1<bool,const DII::Object&,SigC::FixedMarshal<bool, false> > 
    signal;

  ///the counterpart to signal 
  virtual bool message(const DII::Object&) { return false; };

  ///
  bool isLocked() const { return sem_ > 0; };

  ///
  const std::string& name() const { return name_; };
  
  ///
  StringSList dependencies() { return dependencies_; };
  

  /**Lock this module.
     Locked modules cannot be unloaded. Every method call locks the providing
     module.
  */
  void lock() { sem_++; };
  ///
  void release() { if(sem_ > 0) sem_--;};
  
  ///
  struct eqstr
  {
    bool operator()(const char* s1, const char* s2) const
    {
      return strcmp(s1, s2) == 0;
    }
  };
  ///
#if defined(HAVE_STD_HASHMAP) || defined(HAVE_EXT_STD_HASHMAP)
  typedef std::hash_map<const char*, Method, std::hash<const char*>, eqstr> Methods;
#elif defined(HAVE_GNU_CXX_HASHMAP)
  typedef __gnu_cxx::hash_map<const char*, Method, __gnu_cxx::hash<const char*>, eqstr> Methods;
#else
#error "hash map not available"
#endif
  ///
  const Methods& methods() const { return methods_; };
  ///
  DII::Object getOptionsMap(const DII::Object&);

 protected:

  ///
  enum option_t { BOOL, INT, FLOAT, STRING };
  /** An option is described by its type and a textual description
   *  of its meaning.
   */
  typedef std::pair<option_t, std::string> Option;
  /** Each exported option is registered within this map.
   *  The index the the option name as used in the option database
   *  which is realized by libvarconf.
   */
  typedef std::map<std::string, Option> OptionsMap;
  ///
  OptionsMap optionsMap_;

  /**Modulename.
     The name must be identical to the actual class name. Otherwise
     dependency checking won't work.
  */
  std::string name_;
  
  ///Semaphore, increased by every method call, decreased by every return
  unsigned int sem_;
  
  /**dependencies.
     just a list of names.
  */
  StringSList dependencies_;
  ///
  Methods methods_;
  ///
  void configCallback(const std::string&, const std::string&);

  /** Configure an option.
      This method gets called whenever one of this module's options
      gets changed.
      The changed value can be retrieved to a call to one of the 
      getOption() methods.
  */
  virtual void configure() {};
  /** Get module option.
      Each module own a own set of settings which are stored persistent and
      can get changed by the user using the /set command:
      /set <modulename> <optionname> 
  */
  //@{
  ///
  bool getOption(const std::string& optname, double* val);
  ///
  bool getOption(const std::string& optname, int* val);
  ///
  bool getOption(const std::string& optname, bool* val);
  ///
  bool getOption(const std::string& optname, std::string* val);
  //@}
  /** Set module option.
   */
  //@{
  ///
  void setOption(const std::string& optname, double val);
  ///
  void setOption(const std::string& optname, int val);
  ///
  void setOption(const std::string& optname, bool val);
  ///
  void setOption(const std::string& optname, const std::string& val);
  //@}
  /**Register the option with given type.
     A registered option is meant to get changed by the user.
     The module can be queried for a list of the registered options.

   */
  void registerOption(const std::string& optname, option_t type, const std::string& desc)
    { optionsMap_[optname] = Option(type, desc); };
};

//@}

} //namespace DII

#endif
