//  The Worldforge Project
//  Copyright (C) 1998, 1999,2000,2001 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 03/2001

#define NEED_DII_SERVICES

#include "script_mod.h"
#include "dii.h"
#include <iostream>

using namespace Lua;
using namespace DII;
using std::string;
using std::cerr;
using std::endl;

///
class ScriptFunction : public AbstractFunction
{
protected:
  ///
  std::string funcname_;
  ///
  ScriptModule* script_;
  
public:
  ///
  ScriptFunction(const std::string& name, ScriptModule* script): 
    funcname_(name), script_(script) {};
  ///
  virtual ~ScriptFunction() {};
  ///
  virtual Object operator() (const Object& obj) 
  {
    Lua::Script::ListType args(1,script_->toLuaValue(obj));
    Lua::Value result;
    try
      {
	result = script_->call(funcname_, args);
      }
    catch(...)
      {
        std::string msg = std::string("caught exception while calling script function ")+
	  script_->name()+"::"+funcname_+" !\n";
	DII::callService("std_err",msg);
	  cerr << msg << endl;
      }
    return script_->toDIIObject(result);
  };
};

ScriptModule::ScriptModule(const std::string& name, const std::string& filename, 
			   const std::string& mode) :
  Module(name),
  Script(INIT_BASELIB | INIT_STRINGS | INIT_MATH)
{
  // register functions to be called from the scripts

  Script::registerFunc("REGISTER_METHOD",&ScriptModule::registerMethod);
  Script::registerFunc("REGISTER_AS_SERVICE",&ScriptModule::registerAsService);
  Script::registerFunc("NEEDS_MODULE",&ScriptModule::needsModule);
  
  Script::registerFunc("_ALERT",&ScriptModule::_ALERT);
  
  Script::registerFunc("Module_getOption",&ScriptModule::Module_getOption);
  Script::registerFunc("Module_registerOption",&ScriptModule::Module_registerOption);

  if(mode == "trusted" || mode == "default")
    {
      if(mode == "trusted")
	// enabled socketIO and fileIO in trusted mode
	Script::initExtensions(INIT_SOCKETS | INIT_FILEIO);
      else
	// enable only socketIO in default mode
	Script::initExtensions(INIT_SOCKETS);
      
      Script::registerFunc("loadExtension",&ScriptModule::loadExtension);
      Script::registerFunc("DII_registerMsgHandler",&ScriptModule::DII_registerMsgHandler);
      Script::registerFunc("DII_callMethod",&ScriptModule::DII_callMethod);
      Script::registerFunc("DII_callService",&ScriptModule::DII_callService);
      Script::registerFunc("Module_setOption",&ScriptModule::Module_setOption);
      
    }
  else
    {
      // neither socket io nor file io enabled (sandbox mode)
      
      // script cannot load additional extension from lua files

      // script cannot call method or services from other modules
      
    }
    
  // register new data types with LUA
  Script::registerType("List");
  Script::registerType("Map");

  //tell script about its name
  Script::setGlobal("Name",name);

  doFile(filename.c_str());
  
}

ScriptModule::~ScriptModule()
{

}

void 
ScriptModule::configure()
{
  try
    {
      //Lua::Script::ListType args;
      call("configure");
    }
  catch(...)
    {
      // we do not care
    }
}

bool 
ScriptModule::message(const DII::Object& msg)
{
  if(msgHandler_.empty())
    return false;
  
  bool result = false;

  try
    {
      Lua::Script::ListType args(1,toLuaValue(msg));
      result = (int)(call(msgHandler_, args).asNumber());
    }
  catch(...)
    {
      std::cerr << "ScriptError("<<Module::name_<<") while calling msgHandler !" << std::endl;
    }

  return result;
}

int
ScriptModule::loadExtension(Script::Handle handle)
{
  ScriptModule* script = dynamic_cast<ScriptModule*>(Script::getRef(handle));
  assert(script);
  int argcount = script->getStacksize();  
  if(argcount != 1)
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter count to loadExtension " 
	   << std::endl;
      return 0;
    }
  
  Lua::Value filename = script->popValue();
  if(!filename.isString())
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter to loadExtension " 
	   << std::endl;
      return 0;
    }

  std::string fullname = script->scriptPath_+std::string("/")+filename.asString();
  //  std::cerr << "Script \""<<script->name()<<"\" loads extensions from "<<fullname<<std::endl;

  int result = script->doFile(fullname.c_str());
  script->pushValue(result);

  return 1;
}

int
ScriptModule::registerMethod(Script::Handle handle)
{
  ScriptModule* script = dynamic_cast<ScriptModule*>(Script::getRef(handle));
  assert(script);
  int argcount = script->getStacksize();  

  if(argcount != 1)
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter count to registerMethod " 
	   << std::endl;
      return 0;
    }
  
  Lua::Value func = script->popValue();
  if(!func.isString())
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter to registerMethod " 
	   << std::endl;
      return 0;
    }

  script->methods_[func.asString().c_str()]
    = new ScriptFunction(func.asString(), script);

  return 0;
}
int
ScriptModule::registerAsService(Script::Handle handle)
{
  ScriptModule* script = dynamic_cast<ScriptModule*>(Script::getRef(handle));
  assert(script);
  int argcount = script->getStacksize();  
 
  if(argcount != 1)
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter count to registerAsService" 
	   << std::endl;
      return 0;
    }
  
  Lua::Value func = script->popValue();
  if(!func.isString())
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter to registerAsService" 
	   << std::endl;
      return 0;
    }

  DII::registerService(script,func.asString().c_str()); 

  return 0;
}

int
ScriptModule::needsModule(Script::Handle handle)
{
  ScriptModule* script = dynamic_cast<ScriptModule*>(Script::getRef(handle));
  assert(script);
  int argcount = script->getStacksize();  
  if(argcount != 1)
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter count to needsModule " 
	   << std::endl;
      return 0;
    }
  
  Lua::Value modname = script->popValue();
  if(!modname.isString())
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter to needsModule" 
	   << std::endl;
      return 0;
    }
  
  script->dependencies_.push_front(modname.asString());  
   
  return 0;
}

int
ScriptModule::DII_emitMessage(Script::Handle handle)
{
  ScriptModule* script = dynamic_cast<ScriptModule*>(Script::getRef(handle));
  assert(script);
  int argcount = script->getStacksize();  
  if(argcount != 1)
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter count to DII_emitMessage " 
	   << std::endl;
      return 0;
    }

  Lua::Value msg = script->popValue();
  
  int result = (int)(script->signal.emit(script->toDIIObject(msg)));

  script->pushValue(result);
  
  return 1;
}

int
ScriptModule::DII_registerMsgHandler(Script::Handle handle)
{
  ScriptModule* script = dynamic_cast<ScriptModule*>(Script::getRef(handle));
  assert(script);
  int argcount = script->getStacksize();  
  if(argcount != 2)
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter count to DII_registerMsgHandler " 
	   << std::endl;
      return 0;
    }
  
  Lua::Value handler = script->popValue();
  Lua::Value modname = script->popValue();
  
  if(!modname.isString() || !handler.isString())
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameters to DII_registerMsgHandler" 
	   << std::endl;
      return 0;
    }
  
  DII::Module* mod = DII::lookupModule(modname.asString().c_str());
  if(mod)
    {
      script->msgHandler_ = handler.asString();
      mod->signal.connect(SigC::slot(script, &ScriptModule::message));
      script->pushValue(1);
      return 1;
    }
  return 0;
}
int
ScriptModule::DII_callMethod(Script::Handle handle)
{
  ScriptModule* script = dynamic_cast<ScriptModule*>(Script::getRef(handle));
  assert(script);
  int argcount = script->getStacksize();  
  if(argcount != 3)
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter count to DII_callMethod " 
	   << std::endl;
      return 0;
    }
  
  Lua::Value arg = script->popValue();
  Lua::Value method = script->popValue();
  Lua::Value modname = script->popValue();
  
  if(!modname.isString() || !method.isString())
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter to DII_callMethod" 
	   << std::endl;
      return 0;
    }
  
  script->pushValue(script->toLuaValue(DII::callMethod(modname.asString().c_str(),
						       method.asString().c_str(),
						       script->toDIIObject(arg))));
  // we always return 1 result
  return 1;
}

int
ScriptModule::_ALERT(Script::Handle handle)
{
  ScriptModule* script = dynamic_cast<ScriptModule*>(Script::getRef(handle));
  assert(script);
  int argcount = script->getStacksize();  
  if(argcount != 1)
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter count to _ALERT" 
	   << std::endl;
      return 0;
    }
  
  Lua::Value arg = script->popValue();

  if(arg.isString())
    DII::callService("std_err",arg.asString());

  return 0;
}

int
ScriptModule::DII_callService(Script::Handle handle)
{
  ScriptModule* script = dynamic_cast<ScriptModule*>(Script::getRef(handle));
  assert(script);
  int argcount = script->getStacksize();  
  if(argcount != 2)
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter count to DII_callService" 
	   << std::endl;
      return 0;
    }
  
  Lua::Value arg = script->popValue();
  Lua::Value service = script->popValue();
  
  if(!service.isString())
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter to DII_callService" 
	   << std::endl;
      return 0;
    }
  
  script->pushValue(script->toLuaValue(DII::callService(service.asString().c_str(),
							script->toDIIObject(arg))));

  // we always return 1 result
  return 1;
}

int
ScriptModule::Module_setOption(Script::Handle handle)
{
  ScriptModule* script = dynamic_cast<ScriptModule*>(Script::getRef(handle));
  assert(script);
  int argcount = script->getStacksize();  
  if(argcount != 2)
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter count to Module_setOption " 
	   << std::endl;
      return 0;
    }
  
  Lua::Value value = script->popValue();
  Lua::Value option = script->popValue();

  assert(option.isString());

  if(value.isString())
    script->Module::setOption(option.asString(), value.asString());
  else if(value.isNumber())
    script->Module::setOption(option.asString(), (double)value.asNumber());

  // returning void
  return 0;
}

int
ScriptModule::Module_registerOption(Script::Handle handle)
{
  ScriptModule* script = dynamic_cast<ScriptModule*>(Script::getRef(handle));
  assert(script);
  int argcount = script->getStacksize(); 
  if(argcount != 3)
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter count to Module_registerOption" 
	   << std::endl;
      return 0;
    }
  Lua::Value desc = script->popValue();
  Lua::Value type = script->popValue();
  Lua::Value name = script->popValue();
  
  int typeID = (int)type.asNumber();
  if(typeID != Module::INT && typeID != Module::FLOAT &&
     typeID != Module::BOOL && typeID != Module::STRING)
    typeID = Module::STRING;

  script->Module::registerOption(name.asString(), (option_t)type.asNumber(), desc.asString());

  // returning void
  return 0;
}

int
ScriptModule::Module_getOption(Script::Handle handle)
{
  ScriptModule* script = dynamic_cast<ScriptModule*>(Script::getRef(handle));
  assert(script);
  int argcount = script->getStacksize();  
  if(argcount != 1)
    {
      std::cerr << "ScriptError("<<script->name()<<"): wrong parameter count to Module_getOption" 
	   << std::endl;
      return 0;
    }
  
  Lua::Value option = script->popValue();
  Lua::Value value;

  assert(option.isString());

  int i;
  double d;
  std::string s;
  bool b;

  if(script->Module::getOption(option.asString(), &b))
    value = (int)b;
  if(script->Module::getOption(option.asString(), &d))
    value = d;
  if(script->Module::getOption(option.asString(), &i))
    value = i;
  if(script->Module::getOption(option.asString(), &s))
    value = s;
  // if option is not known we return a NIL value
  
  // push result onto the stack
  script->pushValue(value);

  // we return 1 result
  return 1;
}


DII::Object 
ScriptModule::toDIIObject(const Lua::Value& arg)
{
  DII::Object result;
  if(arg.isNil())
    result = DII::FAILURE;
  else if(arg.isString())
    result = arg.asString();
  else if(arg.isNumber())
    result = arg.asNumber();
  else if(arg.isTable())
    {
      if(arg.tag() == getTypeTag("List"))
	{
	  /* FIXME !!
	   * this conversion is not 100% correct, as LUA permits 
	   * lists which are not complete filled: e.g. { [0] = "Hello", [10] = "World" }
	   */
	  DII::Object::ListType tmp;
	  Lua::Value::TableType::const_iterator itr = arg.asTable().begin();
	  while(itr != arg.asTable().end())
	    {
	      tmp.push_back(toDIIObject(itr->second));
	      itr++;
	    }
	  result = tmp;
	}
      else if(arg.tag() == getTypeTag("Map"))
	{
	  DII::Object::MapType tmp;
	  Lua::Value::TableType::const_iterator itr = arg.asTable().begin();
	  try
	    {
	      while(itr != arg.asTable().end())
		{
		  tmp[itr->first.asString()] = toDIIObject(itr->second);
		  itr++;
		}
	    }
	  catch(...)
	    {
	      std::cerr << "DII::Object does only support strings as index keys !"<< std::endl;
	    }
	  result = tmp;
	}
    }
  else
    {
      std::cerr << "cannot convert this type of Lua::Value to DII::Object "<< std::endl;
    }
  return result;
}


Lua::Value 
ScriptModule::toLuaValue(const DII::Object& arg)
{
  Lua::Value result;

  // this returns a NIL value
  if(arg == DII::FAILURE)
    return result;

  if(arg.IsFloat())
    result = arg.AsFloat();
  else if(arg.IsInt())
    result =(Lua::Value::NumberType)arg.AsInt();
  else if(arg.IsString())
    result = arg.AsString();
  else if(arg.IsList())
    {
      Lua::Value::TableType tmp;
      DII::Object::ListType::const_iterator itr = arg.AsList().begin();
      tmp.resize(arg.AsList().size());
      int n = 0;
      while(itr != arg.AsList().end())
	{
	  // note that lua indexes run from 1 to n (and not from 0 to n-1 like in C)
	  tmp[n]=std::pair<Lua::Value, Lua::Value>(n+1, toLuaValue(*itr));
	  n++;
	  itr++;
	}
      result = Script::createObject("List", tmp);
    }
  else if(arg.IsMap())
    {
      Lua::Value::TableType tmp;
      DII::Object::MapType::const_iterator itr = arg.AsMap().begin();
      tmp.resize(arg.AsMap().size());
      int n = 0;
      while(itr != arg.AsMap().end())
	{
	  tmp[n]=std::pair<Lua::Value, Lua::Value>(itr->first, toLuaValue(itr->second));
	  n++;
	  itr++;
	}
      result = Script::createObject("Map", tmp);
    }
  else
    {
      std::cerr << "failed to convert from Lua::Value to DII::Object "<< std::endl;
    }	  

  return result;
}
