//  The Worldforge Project
//  Copyright (C) 1998, 1999, 2000 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 08/2000


#include "serialize.h"

#include <fstream>

//libxml includes:
#include <gnome-xml/tree.h>
#include <gnome-xml/parser.h>
#include <gnome-xml/xmlIO.h>
#include <gnome-xml/parserInternals.h>

#include <stack>

#undef DEBUG

using std::endl;

namespace DII { namespace Serialize {

std::stack<std::string> _names;
std::stack<DII::Object> _stack;
DII::Object _root;
  std::string indent;
  int errors;

//-------- Callbacks -------------------------
void startDocument (void *ctx)
{
#ifdef DEBUG
    fprintf(stderr,"DII: --------\nDII: Starting Document\n");
#endif   
}

void endDocument (void *ctx)
{
#ifdef DEBUG
    fprintf(stderr,"DII: Ending Document\n --------\n");
#endif   
}

void startElement (void *ctx, 
		   const xmlChar *name,
		   const xmlChar **attr)
{
  DII::Object elem;

#ifdef DEBUG
  if(name)
    fprintf(stderr,"DII: --- %s ---\n",name);
  else
    fprintf(stderr,"DII: --empty tag --\n");
#endif

  // FIXME Legacy support for old format of file
  // <obj> was replace with <atlas> for compatability
  // with other tools
  if(!strcmp((const char*)name,"obj"))
    {
      return;
    }  
  if(!strcmp((const char*)name,"atlas"))
    {
      return;
    }  
  if(!strcmp((const char*)name,"map"))
    {
      elem = DII::Object::MapType();
    }
  else if(!strcmp((const char*)name,"list"))
    {
      elem = DII::Object::ListType();
    }
  else if(!strcmp((const char*)name,"string"))
    {
      elem = DII::Object::StringType();
    }
  else if(!strcmp((const char*)name,"int")) 
    {
      elem = DII::Object::IntType();
    }
//  else if(!strcmp((const char*)name,"ptr"))
//     {
//       elem = DII::Object::PtrType();
//     }
  else if(!strcmp((const char*)name,"float"))
    {
      elem = DII::Object::FloatType();
    }
  else
    {
      fprintf(stderr,"DII::Serialize unknown tag: %s\n", name);
    }
  
  std::string typname;
  if(attr)
    {
      typname = std::string((const char*)attr[1]);
    }
  
  _stack.push(elem);
  _names.push(typname);

}

void endElement (void *ctx, const xmlChar *tagname)
{
  if(_stack.empty())
    return;

  DII::Object elem = _stack.top();
  _stack.pop();
  std::string name = _names.top();
  _names.pop();

#ifdef DEBUG
  if(tagname)
    fprintf(stderr,"DII: --- /%s ---\n",tagname);
#endif

  if(!_stack.empty())
    {
      switch(_stack.top().GetType())
	{
	case DII::Object::TYPE_MAP:
	  _stack.top().AsMap()[name] = elem;
	  break;
	case DII::Object::TYPE_LIST: 	  
	  _stack.top().AsList().push_back(elem);
	  break;
	default:
	  fprintf(stderr,"DII::Serialize illegal nested tags !\n");
	  break;
	}
    }
  else
    {
      _root = elem;
    }
}

void characters (void *ctx, const xmlChar *ch, int len)
{
  if(_stack.empty())
    return;
  
  char* buffer;
  
  if(!ch)
    {
      fprintf(stderr, "DII: -empty string-\n");
      return;
    }

  buffer = new char[len+1];
  strncpy(buffer,(const char*)ch, len);

  buffer[len] = '\0';
#ifdef DEBUG
  fprintf(stderr, "DII: \"%s\"\n", buffer);
#endif
  
  switch(_stack.top().GetType())
    {
    case DII::Object::TYPE_INT:
      _stack.top() = atoi(buffer);
      break;
//     case DII::Object::TYPE_PTR:
//       _stack.top() = (void*)(strtoul(buffer,0,0));
//       break;
    case DII::Object::TYPE_FLOAT: 	  
      _stack.top() = atof(buffer);
      break;
    case DII::Object::TYPE_STRING:
      _stack.top() = std::string(buffer,len);
      break;
    default:
      break;
    }

  delete [] buffer;

}

void warning (void *ctx, const char *msg, ...)
{
    fprintf(stderr, "DII: ***WARNING*** %s\n", msg);
}

void error (void *ctx, const char *msg, ...)
{
    fprintf(stderr, "DII: ***ERROR*** %s\n", msg);
    errors++;
}

void fatalError (void *ctx, const char *msg, ...)
{
    fprintf(stderr, "DII: ***FATAL ERROR*** %s\n", msg);
    errors++;
}


  
bool read(const char* filename, DII::Object& obj)
{
  xmlSAXHandler m_saxfuncs;

  // Initialize all function pointers to NULL
  memset (&m_saxfuncs, 0, sizeof (xmlSAXHandler));
  
  // Parsing callbacks
  m_saxfuncs.startDocument = startDocument;
  m_saxfuncs.endDocument = endDocument;
  m_saxfuncs.startElement = startElement;
  m_saxfuncs.endElement = endElement;
  m_saxfuncs.characters = characters;
  
  // Formatting callbacks
  m_saxfuncs.warning = warning;
  m_saxfuncs.error = error;
  m_saxfuncs.fatalError = fatalError;

#ifdef DEBUG
  fprintf(stderr, "DII::Serialize: Loading %s...\n", filename);
#endif

//   struct stat statbuf;
//   if (stat(filename, &statbuf) != 0)
//     {
//       fprintf(stderr, ("DII: Warning, unable to find filename %s", filename);
//       return false;
//     }
  
  bool ret = 0;
  xmlParserCtxtPtr context;
  
  context = xmlCreateFileParserCtxt(filename);
  if (context == NULL)
    {
      return false;
    }
  context->sax = &m_saxfuncs;

  errors = 0;

  xmlParseDocument(context);
  
  if (context->wellFormed)
    ret = true;
  else
    ret = false;

  if(errors)
    ret = false;

  context->sax = NULL;
  xmlFreeParserCtxt(context);
  
  if(ret)
    obj = _root;

  return ret;
}
  
void
walker(std::ostream& out, const DII::Object& obj, const std::string& name = "")
{
  indent += "  ";

  out << indent;

  switch(obj.GetType())
    {
//     case DII::Object::TYPE_PTR: 
//       if(name.empty())
// 	out << "<ptr>";
//       else
// 	out << "<ptr name=\""<<name<<"\">";
//       out <<obj.AsPtr()<<"</ptr>";

//       break;
    case DII::Object::TYPE_INT: 
      if(name.empty())
	out << "<int>";
      else
	out << "<int name=\""<<name<<"\">";
      out <<obj.AsInt()<<"</int>";

      break;
    case DII::Object::TYPE_FLOAT:       
      if(name.empty())
	out << "<float>";
      else
	out << "<float name=\""<<name<<"\">";
      out <<obj.AsFloat()<<"</float>";

      break;
    case DII::Object::TYPE_STRING: 
      if(name.empty())
	out << "<string>";
      else
	out << "<string name=\""<<name<<"\">";
      out <<obj.AsString()<<"</string>";
      break;      
    case DII::Object::TYPE_LIST:
      {
	DII::Object::ListType::const_iterator itr = obj.AsList().begin();
	if(name.empty())
	  out << "<list>"<<endl;
	else
	  out << "<list name=\""<<name<<"\">"<<endl;;
	while(itr != obj.AsList().end())
	  {
	    walker(out, *itr);
	    out << endl;
	    itr++;
	  }
	out << indent << "</list>" << endl;
      }
      break;

    case DII::Object::TYPE_MAP: 
      {
	DII::Object::MapType::const_iterator itr = obj.AsMap().begin();
	if(name.empty())
	  out << "<map>"<<endl;
	else
	out << "<map name=\""<<name<<"\">"<<endl;
	while(itr != obj.AsMap().end())
	  {
	    walker(out, itr->second, itr->first);
	    out << endl ;
	    itr++;
	  }
	out << indent << "</map>" << endl;
      }
      break;
    default:
      break;
    }

  indent.erase(indent.end() - 2, indent.end());
}
  
bool write(const DII::Object& obj, const char* filename)
{
  std::ofstream out;
  out.open(filename);
  if(!out)
    return false;
  
  out << "<atlas>" << endl;
  indent = "";

  walker(out, obj);

  out << "</atlas>" << endl;
  
  out.close();

  return true;
}

bool write(const Object& obj, std::ostream& out)
{
  walker(out, obj);
  
  return true;
}

} }
