//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 01/2000

#define NEED_DII_SERVICES

#include "environment.h"
#include "dii/dii.h"

#include "../modules/viewport.h"
#include "../renderer/renderer.h"
#include "../registry/registry.h"

#include <uta/debug.h>
#include <uta/painter.h>
#include <uta/directsurface.h>
#include <uta/resources.h>

#ifndef HAVE_DRAND48
#define drand48() rand()/((double)RAND_MAX)
#endif

using namespace uta;

#define RAIN_DROPCOUNT 200
#define SNOW_STEPS 5  

Environment* Environment::instance_ = 0;

Environment::Environment():
  weather_(SUNNY),
  daytime_(DAY),
  timer_(5000),
  fasttimer_(100),
  target_(NULL),
  cache_(NULL)
{
  assert(!instance_);
  instance_ = this;

  timer_.halt();
  fasttimer_.halt();

  timer_.alarm.connect(slot(this, &Environment::tick));
  fasttimer_.alarm.connect(slot(this, &Environment::fasttick));
}

void
Environment::deinit()
{
  
  timer_.halt();
  fasttimer_.halt();

  for(unsigned int i=0; i < birds_.size(); i++)
    {
      //      delete birds_[i];
    }
  birds_.clear();

  if(rain_)
    {
      // delete rain_;
      rain_ = NULL;
    }

  //  delete thunder_;

  target_ = NULL;
  if(cache_)
    delete cache_;

  DII::SetOption("Environment", "lightning_start", lightning_.start);
  DII::SetOption("Environment", "lightning_max", lightning_.max);
  DII::SetOption("Environment", "lightning_raise", lightning_.raise);
  DII::SetOption("Environment", "lightning_fall", lightning_.fall);
  DII::SetOption("Environment", "lightning_prob", lightning_.probability);
}


bool
Environment::init(DII::Module*)
{  

  std::string prefix = Registry::instance()->readInstallDir()
    + "/environment/";
  
  birds_.resize(6);
  
  birds_[0] = SAMPLE_RES("bird_0");
  birds_[1] = SAMPLE_RES("bird_1");
  birds_[2] = SAMPLE_RES("bird_2");
  birds_[3] = SAMPLE_RES("bird_3");
  birds_[4] = SAMPLE_RES("bird_4");
  birds_[5] = SAMPLE_RES("bird_5");

  rain_ = SAMPLE_RES("rain");

  //thunder_ = new uta::Music(prefix+"thunder_1_us.mp3"); 
  
  //install event handler:
  Viewport* viewport_ = Viewport::instance();
  if(!viewport_)
    {
      cerr << "Environment: Error, viewport not found !";
      return false;
    }
 
  viewport_->signalExposeHide.connect(SigC::slot(this, &Environment::eventHandler));

  //check if viewport is there... if so enable timer
  if(viewport_->getScreen())
    timer_.run();
  
  target_ = NULL;
  cache_ = NULL;

  lightlevel_ = 0;
  lightdelta_ = 0;
  
  lightning_.start = 35;
  lightning_.max = 35;
  lightning_.raise = 50;
  lightning_.fall = -50;
  lightning_.probability = 0.01;

  configure();

  return true;
}

void 
Environment::configure()
{
  DII::GetOption("Environment", "lightning_start", lightning_.start);
  DII::GetOption("Environment", "lightning_max", lightning_.max);
  DII::GetOption("Environment", "lightning_raise", lightning_.raise);
  DII::GetOption("Environment", "lightning_fall", lightning_.fall);
  DII::GetOption("Environment", "lightning_prob", lightning_.probability);
}

void
Environment::fasttick()
{
  if(weather_ == RAIN)
    {
      //restart sounds if necessary
      if(rain_)
	if(!rain_->playing())
	  rain_->play();
      //if(!thunder_->playing())
      //thunder_->play();
    }
  else if(weather_ == THUNDER)
    {
      lightlevel_ += lightdelta_;

      if(lightdelta_ == 0)
	{
	  if(drand48() > (1 - lightning_.probability))
	    {
	      lightlevel_ = lightning_.start;
	      lightdelta_ = lightning_.raise;
	    }
	}
      
      if(lightlevel_ >= lightning_.max)
	{
	  lightdelta_ = lightning_.fall;
	}
      
      if(lightlevel_ < 0)
	{
	  lightlevel_ = 0;
	  lightdelta_ = 0;
	}

      if(lightlevel_ > 0)
	{
	  cerr << "lightlevel: " << lightlevel_ << endl;
          if(Renderer::instance())
	    Renderer::instance()->setClouds(0);
	  DII::callMethod("daylight","dim", lightlevel_);
	}
      else
	{
          if(Renderer::instance())
	    Renderer::instance()->setClouds(1);
	}
    }
  else
    {

    }
}

void
Environment::tick()
{
  if(weather_ == SUNNY)
    {
     
      unsigned index = (unsigned)(drand48() * (double)birds_.size());
      debugN(6,cerr << "Environment playing index " << index << endl;);
      
      //play sample once
      if(birds_[index])
	birds_[index]->play();
      
      unsigned interval = (unsigned)(drand48() * 15000.0) + 5000;
      debugN(6,cerr << "      setting interval to " << interval << endl;);
      timer_.setInterval(interval);
    }

  debugN(6,cerr << "ok." << endl;);

}

bool
Environment::eventHandler(bool expose, uta::DirectSurface *surface)
{
  if(surface != NULL)
    resume();
  else 
    pause();

  return false;
}

void
Environment::setTarget(uta::Surface *target)
{
  target_ = target;
  prepare();
//   if(target_)
//     {
//       if(cache_)
// 	delete cache_;
      
//       //create identically surface
//       //same size, same pixelformat as target
//       cache_ = new Surface(*target_);
//       cache_->useRLEAcceleration(true);
//       cache_->setTransparency(true);
//       cache_->fill(transparent);

//     }
}

bool
Environment::render()
{
  if(!target_)
    return false;

  if(daytime_ == DAY && weather_ == SUNNY)
    return true;

  //  cache_->blit(target_);
  
  if(weather_ == RAIN || weather_ == THUNDER )
    {   
      Point shift = Point(+2, +12);
      Point drop = Point(+1, +8);
      Point drop_small = Point(0, 4);
 
      Painter* painter = new Painter(target_);
 
      for(int i=0; i < RAIN_DROPCOUNT; i++)
        {
          painter->setColor(Color(200,200,230));
          if(drand48() > 0.9)
            {
              //this raindrop has already reached the ground
              painter->line(raindrops_[i],raindrops_[i]+Point(-2,-5));
              painter->line(raindrops_[i],raindrops_[i]+Point(+3,-4));
              painter->line(raindrops_[i],raindrops_[i]+Point(+1,-3));
 
              raindrops_[i].y = (int)(drand48()*10);
            }
          else
            {
              raindrops_[i] += shift;
 
              if(drand48() > 0.5)
                {
                  raindrops_[i] += shift;
                  painter->setColor(Color(180,180,200));
                }       
              if(raindrops_[i].y > target_->height())
                raindrops_[i].y = (int)(drand48()*10);
 
              if(raindrops_[i].x > target_->width())
                raindrops_[i].x = (int)(drand48()*10);
 
              if(drand48() > 0.5)
                painter->line(raindrops_[i],raindrops_[i]+drop);
              else
                painter->line(raindrops_[i],raindrops_[i]+drop_small);
            }
        }
 
      delete painter;                                                          
    }
  else if(weather_ == SNOW)
    {
      Point shift_left = Point(-4, +2);
      Point shift_right = Point(+5, +2);
 
      Painter* painter = new Painter(target_);
 
      for(int i=0; i < RAIN_DROPCOUNT; i++)
        {
          painter->setColor(Color(250,250,250));
 
          if(drand48() > 0.999)
            {
              //this snowflake has already reached the ground
              painter->hLine(raindrops_[i],raindrops_[i]+Point(+2,0));
              painter->vLine(raindrops_[i]+Point(+1,-1),raindrops_[i]+Point(+1,+1));
              raindrops_[i].y = (int)(drand48()*10);
            }
          else
            {
              flakeState_[i]++;                                
              if(flakeState_[i] >= 2 * SNOW_STEPS)
                flakeState_[i] = 0;
 
              if(flakeState_[i] < SNOW_STEPS)
                raindrops_[i] += shift_right;
              else
                raindrops_[i] += shift_left;
 
              if(raindrops_[i].y > target_->height())
                raindrops_[i].y = (int)(drand48()*10);
 
              if(raindrops_[i].x < 0)
                raindrops_[i].x = (int)(drand48()*100);
 
              if(raindrops_[i].x > target_->width())
                raindrops_[i].x = (int)(drand48()*10);
 
              painter->hLine(raindrops_[i],raindrops_[i]+Point(+2,0));
              painter->vLine(raindrops_[i]+Point(+1,-1),raindrops_[i]+Point(+1,+1));
            }
        }
 
      delete painter;
    }   

  return true;
}

bool
Environment::weather(const std::string& whatNow)
{
  if(whatNow == "none")
    {
      weather_ = SUNNY;
      daytime_ = DAY;
    }
  
  if(whatNow == "rain")
    weather_ = RAIN;
  else if(whatNow == "snow")
    weather_ = SNOW;
//   else if(whatNow == "fog")
//     weather_ = FOG;
//   else if(whatNow == "night")
//     daytime_ = NIGHT;
//   else if(whatNow == "day")
//     daytime_ = DAY;
  else if (whatNow == "sunny")
    weather_ = SUNNY;
  else if (whatNow == "thunder")
    weather_ = THUNDER;
  else
    return false;

  prepare();

  return true;
}

void
Environment::prepare()
{  
  if(weather_ == RAIN || weather_ == SNOW || weather_ == THUNDER)
    {
      if(weather_ == RAIN)
	{
	  DII::callMethod("daylight","dim", -8);
	  if(Renderer::instance())
	    Renderer::instance()->setClouds(0);
	}
      lightlevel_ = 0;
      lightdelta_ = 0;

      if(target_)
	{          
	  int w = target_->width();
          int h = target_->height();
	  
          raindrops_.clear();
          flakeState_.clear();
          raindrops_.resize(RAIN_DROPCOUNT);
          flakeState_.resize(RAIN_DROPCOUNT);
 
          for(int x=0; x < RAIN_DROPCOUNT; x++)
            {
              raindrops_[x] = Point((int)(drand48() * w), (int)(drand48() * h));
              flakeState_[x] = (int)(2*SNOW_STEPS*drand48());
            }
	  
	  if(rain_)
	    {
	      if(weather_ == RAIN)
		if(!rain_->playing())
		  rain_->play(255);
		else
		  rain_->stop();
	    }
	  //we need the music channel for the thunder, so free it first !
	  // DII::callMethod("Music","pause");
	  //	  thunder_->play();
	}
    }
  else
    {
      
      if(rain_)
	rain_->stop();
      if(Renderer::instance())
        Renderer::instance()->setClouds(0);
      DII::callMethod("daylight","dim", 0);
      //if(thunder_->playing())
      //  thunder_->stop();
      //ok, let the music play again ...
      //DII::callMethod("Music","resume");  
    }

  
}

void
Environment::pause()
{
  timer_.halt();
  fasttimer_.halt();

  if(weather_ == RAIN || weather_ == THUNDER)
    {
      if(rain_)
	rain_->stop();
      //    thunder_->pause();
    }
}

void
Environment::resume()
{
  timer_.run();
  fasttimer_.run();

  if(weather_ == RAIN || weather_ == THUNDER)
    {
      if(rain_)
	if(!rain_->playing())
	  rain_->play();
//       if(thunder_->paused())
// 	thunder_->resume();
    }
}
