//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 12/1999

#define NEED_DII_SERVICES

#include "loader.h"
#include "config.h"
#include "script_mod.h"

#ifdef WITH_PYTHON
#include "pthn_module.h"
#endif

#ifndef NDEBUG
#define UCLIENT_DEBUG(a) a;
#else
#define UCLIENT_DEBUG(a) ;
#endif

using namespace DII;
using std::endl;

#include <fstream>

std::ofstream logfile;

std::map<std::string, Loader::ModuleCB> Loader::modules_;
std::string Loader::path_;
std::string Loader::scriptPath_;

std::set<std::string> Loader::uninitialized_;

/** we need this static instance so that the
    destructor Loader::~Loader() gets called !
*/
static Loader theLoader;

Loader::Loader()
{
    UCLIENT_DEBUG(logfile.open("loader.log"););
}

Loader::~Loader()
{
  releaseAll();

  UCLIENT_DEBUG(logfile.close(););
}

void
Loader::dump(std::ostream &s)
{
  modulDB::iterator itr = modules_.begin();
  StringSList::iterator itr2;


  s << "loaded module               used by" << endl;
  s << "-----------------------------------------"<<endl;

  while(itr != modules_.end())
    {
      s << itr->first << " ";
      if(itr->second.mod->isLocked())
	{
	  s << "(locked)";
	}
      s << "               ";

      itr2 = itr->second.usedBy.begin();
      while(itr2 != itr->second.usedBy.end())
	{
	  if(itr2 == itr->second.usedBy.begin())
	    s << "[";
	  s << *itr2;
	  itr2++;
	  if(itr2 == itr->second.usedBy.end())
	    s << "]";
	  else
	    s << ", ";
	}
      s << endl;
      itr++;
    }
  s << "-----------------------------------------"<<endl;
}


bool
Loader::loadScript(const std::string& scriptname, const std::string& mode)
{
  //if the module is already loaded, ignore this request
  if(modules_.find(scriptname) != modules_.end())
    return true;

  std::string filename;
  std::ifstream test;
  Module* mod = NULL;

  // first try to load LUA script ...
  filename = std::string(scriptPath_+"/"+scriptname+".lua");

  UCLIENT_DEBUG(logfile << "DII::Loader: searching for " << filename << endl;);

  test.open(filename.c_str());
  if(test)
    {
      test.close();
      UCLIENT_DEBUG(logfile << "DII::Loader: loading LUA script from " << filename << endl;);

      ScriptModule* _mod = new ScriptModule(scriptname, filename, mode);
      assert(_mod);

      // tell script from where it may load extensions
      _mod->setExtensionPath(scriptPath_);
      mod = _mod;
    }
#ifdef WITH_PYTHON
  else
    {
      filename = std::string(scriptPath_);

      UCLIENT_DEBUG(logfile << "DII::Loader: searching for " << filename << endl;);
      test.open(filename.c_str());
      if(test)
	{
	  test.close();
	  UCLIENT_DEBUG(logfile << "DII::Loader: loading Python module from " << filename << endl;);
	  PthnModule* _mod = new PthnModule(scriptname, filename, mode);
	  assert(_mod);
	  mod = _mod;
	}
    }
#endif

  if(!mod)
    {
        UCLIENT_DEBUG(logfile << "DII::Loader: failed to loaded script !"<<endl;);
      return false;
    }

#ifdef DEBUG
  UCLIENT_DEBUG(logfile << "DII::Loader: succesfully loaded script "<<mod->name()<<endl;);
#endif


  // register this module to the loader's module DB
  modules_[mod->name()].mod = mod;

  /* resolve dependencies, register its services, init the module */
  return init_module(mod);
}

bool
Loader::loadModule(const std::string& modulname)
{
  //if the module is already loaded, ignore this request
  if(modules_.find(modulname) != modules_.end())
    return true;

  std::string filename = modulname;

  UCLIENT_DEBUG(logfile << "DII::Loader: trying to load " << filename << endl;);

  Module* mod = bind_module(filename);
  if(!mod)
    return false;


  /* resolve dependencies, register its services, init the module */
  return init_module(mod);
}

bool
Loader::init_module(Module* mod)
{
  bool success = false;

  if(mod)
    {
      std::string modulname = mod->name();
      uninitialized_.insert(modulname);
      success = true;

      //first try to load all modules this module depends on
      StringSList depmodules = mod->dependencies();
      StringSList::iterator itr = depmodules.begin();
      while(itr != depmodules.end())
	{
	  if(uninitialized_.find(*itr) != uninitialized_.end())
	    {
	      UCLIENT_DEBUG(logfile << "DII::Loader: circular dependency between modules "
                           <<modulname<<" and "<<*itr<<" detected !" << endl;);
              UCLIENT_DEBUG(logfile << "             current loaded-module stack will get unloaded." << endl;);
	      success = false;
	    }
	  else
	    {
	      if(modules_.find(*itr) == modules_.end())
		{
		  //only try to load if not already loaded
		  if(!loadModule(*itr))
		    {
		      UCLIENT_DEBUG(logfile << "DII::Loader: failed dependency for \""<<mod->name()
                                   << "\": \""<<*itr<<"\" could not be loaded."<< endl;);
		      success = false;
		    }
		  else
		    {
		      //keep track of dependencies
		      modules_[*itr].usedBy.push_front(modulname);
		    }
		}
	      else
		{
		  //keep track of dependencies
		  modules_[*itr].usedBy.push_front(modulname);
		}
	    }
	  itr++;
	}

      //then call init !
      if(success)
	if(mod->init(0) == DII::FAILURE)
	  {

              UCLIENT_DEBUG(logfile << "DII::Loader: initializing of module "<<mod->name()<<" failed. " << endl;);

	    success = false;
	  }

      if(!success)
	{
	  uninitialized_.erase(modulname);
	  releaseModule(modulname);
	  return false;
	}

//       DII::Object::MapType info;
//       info["event"]=std::string("loaded");
//       info["module"]=modulname;

//       Broadcast msg("Loader",info);
//       Broadcaster::send(msg);

      uninitialized_.erase(modulname);
      success = true;

    } //if mod


  return success;
}

bool
Loader::releaseModule(const std::string& modulname)
{
  bool success = false;

  if(modulname.empty())
    return false;

  modulDB::iterator modItr = modules_.find(modulname);

  //remove the module and after this the entry
  if(modItr != modules_.end())
    {
      //are there any other modules using this one ?
      if(!modItr->second.usedBy.empty())
	{
#ifdef DEBUG
	  UCLIENT_DEBUG(logfile << "DII::Loader: cannot unload "<<modulname<<" because it is still used !"
                       <<endl;);
#endif
	}
      else
	{
	  //only unload module if not used anymore !

	  if(modItr->second.mod->isLocked())
	    {
#ifdef DEBUG
	      UCLIENT_DEBUG(logfile << "DII::Loader: cannot release locked module \""
                           << modulname << "\"." << endl;);
#endif
	      return false;
	    }

	  //remove any entries referencing this module:
	  removeDeps(modulname);

	  modItr->second.mod->deinit(0);
	  //then delete and unload the module
	  delete modItr->second.mod;
	  // the module has been successfully released
	  modules_.erase(modItr);
	  // 	  DII::Object::MapType info;
	  // 		  info["event"]=std::string("released");
	  // 		  info["module"]=modulname;

	  // 		  Broadcast msg("Loader",info);
	  // 		  Broadcaster::send(msg);
	  success = true;

	}
    }

  return success;
}


void
Loader::removeDeps(const std::string& name)
{
  modulDB::iterator itr = modules_.begin();
  while(itr != modules_.end())
    {
      itr->second.usedBy.remove(name);
      itr++;
    }
}


void
Loader::getModuleList(std::list<std::string>& names)
{
  modulDB::iterator itr = modules_.begin();
  while(itr != modules_.end())
    {
      names.push_back(itr->first);
      itr++;
    }
}

void
Loader::releaseAll()
{
#ifdef DEBUG
    UCLIENT_DEBUG(logfile << "DII::Loader releasing all loaded modules ..."<<endl;);
#endif

  unsigned int n = modules_.size()+1;
  std::list<std::string> names;

  /* while any modules are loaded _and_ while the number
     of loaded modules keeps getting lower each iteration !
     (to avoid any deadlocks because of locked modules)
  */
  while(n > modules_.size() && !modules_.empty())
    {
      n = modules_.size();

      getModuleList(names);

      std::list<std::string>::iterator p = names.begin();
      while(p != names.end())
	{
	  releaseModule(*p);
	  p++;
	}
      names.clear();
    }

  if(!modules_.empty())
  {
#ifdef DEBUG
      UCLIENT_DEBUG(logfile << "DII::Loader failed to unload all modules." << endl;);
      UCLIENT_DEBUG(logfile << "listing loaded module dependencies ..." << endl;);
      UCLIENT_DEBUG(Loader::dump(logfile););
#endif
    }
}


Module* Loader::
bind_module(const std::string& filename)
{
  Module* module_ = get_bind_module(filename);

  if(module_)
    {
      modules_[module_->name()].mod = module_;
#ifdef DEBUG
      UCLIENT_DEBUG(logfile << "DII::Loader succesfully loaded module "<<module_->name()<<endl;);
#endif
    }
  else
    {
      UCLIENT_DEBUG(logfile << "DII::Loader failed to lookup module " << filename << endl
                   << "     probably an appropriate entry in dii/loader.cc is missing !" << endl;);
    }

  return module_;
}
