// The Worldforge Project
// Copyright (C) 1998-2002 The Worldforge Project
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// 
// For information about Worldforge and its authors, please contact
// the Worldforge Web Site at http://www.worldforge.org.
// 
// created by Michael Koch 07/2002

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include "content_request.h"

#include "sprite_factory.h"

#include <uta/debug.h>

#ifdef __WIN32__
#define SLASH '\\'
#else
#define SLASH '/'
#endif

std::string ContentRequest::s_baseurl = "";
std::string ContentRequest::s_cachedir = "";
std::string ContentRequest::s_proxy_host = "";
unsigned int ContentRequest::s_proxy_port = 0;

void ContentRequest::setBaseURL(const std::string& baseurl)
{
    s_baseurl = baseurl;

    if(baseurl != "") {
        if( s_baseurl[ s_baseurl.length() - 1 ] != '/' ) {
          s_baseurl += "/";
        }
    }
}

void ContentRequest::setCacheDir(const std::string& cachedir)
{
    s_cachedir = cachedir;

    if(cachedir != "") {
        if( s_cachedir[ s_cachedir.length() - 1 ] != SLASH ) {
            s_cachedir += SLASH;
        }
    }
}

void ContentRequest::setProxy(const std::string& host, unsigned int port)
{
    s_proxy_host = host;
    s_proxy_port = port;
}

ContentRequest::ContentRequest(Mode mode)
    : m_state(DONE)
{
    m_request = ghttp_request_new();
    ghttp_set_sync(m_request, mode == SYNC ? ghttp_sync : ghttp_async);
}

ContentRequest::~ContentRequest()
{
    assert(m_request);

    ghttp_request_destroy(m_request);
}

void ContentRequest::create(const std::string& filename)
{
    assert(m_request);

    // already request in progress ?
    if(m_state == PROCESSING) {
      debugN(4,std::cerr << "ContentRequest: already processing a request"
	                 << std::endl;);
      return;
    }
    
    // this two values have to be set before using this class
    if(s_baseurl == "" || s_cachedir == "") {
      debugN(4,std::cerr << "ContentRequest: BaseURL and/or CacheDir not set"
	                 << std::endl;);
      debugN(4,std::cerr << "ContentRequest: file: " << filename << std::endl;);
      return;
    }

    // test if file already exists
    struct stat dummy_buf;
    if(stat(fullFilename(filename).c_str(),&dummy_buf) == 0) {
      return;
    }
  
    // validate URL
    if(ghttp_uri_validate(const_cast<char*>((s_baseurl + filename).c_str())) == -1)
    {
      debugN(4,std::cerr << "ContentRequest: invalid URL: "
	                 << s_baseurl + filename << std::endl;);
      return;
    }
    
    debugN(4,std::cout << "ContentRequest: request for "
	               << s_baseurl + filename << std::endl;);
    
    // set request
    m_filename = filename;
    ghttp_set_header(m_request, http_hdr_Connection, "close");
    ghttp_set_uri(m_request, const_cast<char*>(fullURL(filename).c_str()));

    // set proxy
    if( s_proxy_host != "") {
      char buf[10];
      sprintf(buf, "%i", s_proxy_port);
      std::string proxy_url = "http://" + s_proxy_host + ":" + buf;
      ghttp_set_proxy(m_request, const_cast<char*>(proxy_url.c_str()));
    }
    
    // prepare request
    ghttp_prepare(m_request);
    m_state = PROCESSING;
}

void ContentRequest::process()
{
    assert(m_request);

    int http_status;
   
    if(m_state == PROCESSING) {
        ghttp_status result = ghttp_process(m_request);

	switch(result) {
          case ghttp_not_done:
	    break;

	  case ghttp_done:
	    m_state = DONE;
	    // check if file was downloaded or an http error occured
	    http_status = ghttp_status_code(m_request);
	    if (http_status != 200) {
              debugN(4,std::cerr << "ContentRequest: could not download media file, HTTP-code:" << http_status << std::endl;);
            } else {
	      saveFile();
            }
            ghttp_clean(m_request);
            ghttp_request_destroy(m_request);
            m_request = ghttp_request_new();
	    break;
	    
	  case ghttp_error:
            m_state = ERROR;
            debugN(4,std::cerr << "ContentRequest: error: " << std::string(ghttp_get_error(m_request)) << std::endl;);
            ghttp_clean(m_request);
            ghttp_request_destroy(m_request);
            m_request = ghttp_request_new();
	    break;
	    
	  default:
	    debugN(4,std::cerr << "ContentRequest: unknown result while processing" << std::endl;);
            ghttp_clean(m_request);
            ghttp_request_destroy(m_request);
            m_request = ghttp_request_new();
	    break;
	}
    }
}

void ContentRequest::saveFile()
{
    assert(m_request);
    assert(m_state == DONE);

    std::string file = fullFilename(m_filename);
    std::string dir = file.substr(0,file.rfind(SLASH));

    // TODO: replace
    system(std::string("mkdir -p " + dir).c_str());
 
    int fd = creat(file.c_str(), S_IRUSR|S_IWUSR);
    if(fd != -1) {
        write(fd, ghttp_get_body(m_request), ghttp_get_body_len(m_request));
    	close(fd);

	signalMediaAvailable.emit(m_filename);
    }
    else {
	m_state = ERROR;
        debugN(4,std::cerr << "ContentRequest: error during write into cache (" << errno << ")" <<  std::endl;);
    }
}

std::string ContentRequest::fullFilename(std::string relFilename)
{
    return s_cachedir + relFilename;
}

std::string ContentRequest::fullURL(std::string relURL)
{
    if (relURL != "") {
        while ((relURL.length() > 0) && (relURL[0] == '/')) {
            // Gobble up any leading slashes
            //relURL = relURL.substr(1, relURL.length() - 1);
            relURL.erase(0, 1);
	}
    }
    return s_baseurl + relURL;
}
