// The Worldforge Project
// Copyright (C) 1998, 1999 The Worldforge Project
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// 
// For information about Worldforge and its authors, please contact
// the Worldforge Web Site at http://www.worldforge.org.
// 
// created by Al Riddoch 07/2002

#include "media_server.h"
#include "sprite_factory.h"
#include "content_request.h"

#include <uta/debug.h>

using Atlas::Objects::Operation::Get;

MediaServer * MediaServer::instance_ = 0;

MediaServer::MediaServer():
factory_(NULL),
timer_(100),
connection_(&dispatcher_),
state_(INIT),
port_(-1),
mediaRequest_(NULL)
{
    assert(!instance_);
    instance_ = this;

    timer_.alarm.connect(slot(this, &MediaServer::tick));

    dispatcher_.OpInfoArrived.connect(SigC::slot(this, &MediaServer::OpInfo));
    dispatcher_.OpErrorArrived.connect(SigC::slot(this, &MediaServer::OpError));

    connection_.disconnected.connect(slot(this, &MediaServer::disconnected));

    mediaRequest_ = new ContentRequest(ContentRequest::ASYNC);
    mediaRequest_->signalMediaAvailable.connect(slot(this, &MediaServer::mediaAvailable));
}

MediaServer::~MediaServer()
{
    delete mediaRequest_;

    assert(instance_);
    instance_ = 0;
}

void MediaServer::tick()
{
    if (connection_.connected())
        connection_.poll();

    // This code could be adapted to handle a number of parallel requests
    if(mediaRequest_->isReady())
      {
        if (!contentQueue.empty())
          {
            std::cout << "FETCHING " <<contentQueue.front()
                      << std::endl << std::flush;
            mediaRequest_->create(contentQueue.front());
            contentQueue.pop();
          }
      }
    else
      {
        mediaRequest_->process();
      }
}

void MediaServer::setServer(const std::string & host, unsigned int port,
                            SpriteFactory * sf)
{
    hostname_ = host;
    port_ = port;
    factory_ = sf;
}

void MediaServer::setContentBaseURL(const std::string& baseurl)
{
    contentBaseURL_ = baseurl;
}

void MediaServer::request(const SpriteDesc & sd)
{
    if (state_ != CONNECTED)
      {
        if (!connect())
          return;
      }
    debugN(8,std::cerr<<"MediaServer: Requesting "<<sd.type<<std::endl;);

    Get g = Get::Instantiate();
    Atlas::Message::Object::MapType smap;
    smap["id"] = sd.type;
    smap["version"] = 0;
    g.SetArgs(Atlas::Message::Object::ListType(1,smap));
    connection_.send(&g);

}

bool MediaServer::requestMediaFile(const std::string& filename)
{
    if(mediaRequest_->isReady())
        mediaRequest_->create(filename);
    else
        return false;

    return true;
}

bool MediaServer::registerCacheReplacement(SpriteID sprite)
{
    if (contentSet.empty())
      {
        return false;
      }
    
    std::set<std::string>::const_iterator I = contentSet.begin();
    SpriteFileSet * set = new SpriteFileSet();
    set->sprite = sprite;
    for(; I != contentSet.end(); ++I)
      {
        fileSetMap[*I] = set;
        ++set->count;
        contentQueue.push(*I);
      }
    contentSet.clear();
    return true;
}

bool MediaServer::connect()
{
    if ((state_ != INIT) || hostname_.empty() || (port_ == -1))
        return false;

    if(connection_.state() == Connection::CONNECTING
      || connection_.state() == Connection::NEGOTIATING)
      return false;

    connection_.connect(hostname_, port_);

    if (!connection_.connected())
        return false;

    state_ = CONNECTED;
    debugN(8,std::cerr<<"MediaServer: Connected to "<<hostname_<<std::endl;);
    return true;
}

void MediaServer::disconnected()
{
    state_ = INIT;

    std::cout << "MediaServer: Lost connection to " << hostname_ << std::endl;
}

void
 MediaServer::OpError(const Atlas::Objects::Operation::Error & info)
{

}

void
 MediaServer::OpInfo(const Atlas::Objects::Operation::Info & op)
{
    debugN(8,std::cerr<<"MediaServer: Got info from"<<hostname_<<std::endl;);
    if (factory_ == NULL)
      {
        debugN(8,std::cerr<<"MediaServer: No factory"<<std::endl;);
        return;
      }

    if (op.GetArgs().empty() || !op.GetArgs().front().IsMap())
      {
        debugN(8,std::cerr<<"MediaServer: ERROR: Got info op with no args"
                          <<std::endl<<std::flush;);
        return;
      }
    DII::Object::MapType arg = op.GetArgs().front().AsMap();
    DII::Object::MapType::iterator itr = arg.find("id");
    if ((itr == arg.end()) || (!itr->second.IsString()))
      {
        debugN(8,std::cerr<<"MediaServer: No id"<<std::endl;);
        return;
      }

    std::string type = itr->second.AsString();
    arg.erase("id");
    factory_->installSpriteDefinition(type, arg);
    debugN(8,std::cerr<<"MediaServer: "<<type<<" installed"<<std::endl;);
}

void MediaServer::mediaAvailable(const std::string & filename)
{
    std::map<std::string, SpriteFileSet *>::iterator I = fileSetMap.find(filename);
    if (I == fileSetMap.end())
      {
        std::cerr << "Informed that unknown file " << filename
                  << " is now ready" << std::endl << std::flush;
        return;
      }

    SpriteFileSet * set = I->second;
    fileSetMap.erase(I);
    if (--set->count > 0)
      {
        return;
      }
    SpriteCache::inst()->reload(set->sprite);
    delete set;
}
