#include "rle.h"

#include <cassert>
#include <uta/debug.h>

namespace RLE {
void dump(const Surface* source)
{
  int w = source->width;
  int h = source->height;
  int x,y;
  SliceHeader* sliceHeader;
  LineHeader* lineHeader;
  
  lineHeader = (LineHeader*)source->data;
  y = 0;
  while(y<h)
    {
      std::cerr << ">";
      if(lineHeader->type == EMPTY)
        {
          std::cerr <<std::endl;
          lineHeader++;
        }
      else
        {
          x=0;
          sliceHeader = (SliceHeader*)((Uint8*)lineHeader + sizeof(LineHeader));
          while(x < w)
            {
              if(sliceHeader->type == TRANSPARENT)
                {
                  //  std::cerr << "T"<<(int)sliceHeader->count;
                  for(int n=0; n < sliceHeader->count; n++)
                    std::cerr << ".";
                  x+=sliceHeader->count;
                  /* the next slice header immediately follows: */
                  sliceHeader ++;
                }
              else
                {
                  //std::cerr << "O"<<(int)sliceHeader->count;
                  for(int n=0; n < sliceHeader->count; n++)
                    std::cerr << "#";
                  /* jump to the next slice header */
                  x+=sliceHeader->count;
                  sliceHeader = (SliceHeader*)((Uint8*)sliceHeader+sliceHeader->count*2+sizeof(SliceHeader));
                }          
            } // while(x<w)
          std::cerr << std::endl;
          lineHeader = (LineHeader*)((Uint8*)lineHeader + lineHeader->skip); 
        }
      y++;
    }
}

Surface* encode(const uta::Surface* source)
{
  assert(source);
  assert(source->pixelformat().bpp() == 2);
  
  debugN(8,std::cerr<<"RLE: encoding "<<source->width()<<"*"<<source->height()
         <<" pixels ... ("<<source->width()*source->height()*2<<" bytes)" << std::endl;);

  Surface* target = (Surface*)malloc(sizeof(Surface));
  assert(target);

  /* prealloc memory of 2 times the size of the original image
   */
  long allocated_size = source->width() * source->height() * 2 * 2;
//  long current_size = 0;

  target->data = malloc(allocated_size);
  assert(target->data);
  
  debugN(8,std::cerr<<"RLE: preallocated "<<allocated_size<<" bytes."<<std::endl;);

  /* remember original dimensions
   */
  target->width = source->width();
  target->height = source->height();
  
  /* _pixels_ to skip from end to the beginning of the next line
   * (inside the source surface)
   */
  int srcSkip = source->pitch()/2 - source->width();

  
  source->lock();
  
  Uint16* sPtr = (Uint16*)source->pixels();
  Uint8* tPtr = (Uint8*)target->data;
  bool transparentLine = false;
  LineHeader* lineHeader = NULL;
  SliceHeader* sliceHeader = NULL;
  
  /* extract transparent color from pixel[0,0]
   */
  Uint16 transColor = *((Uint16*)sPtr);

  int count;
  int x,y;
  y = 0;
  while(y < source->height())
    {
      /* checking for transparentLine is OK at this point,
       * as it can only be true, if lineHeader is not NULL
       */
      if(transparentLine)
        {
          lineHeader->type = EMPTY;
          lineHeader->skip += 1;
          lineHeader++;

          /* reset tPtr ... thus erasing the unused SliceHeader */
          tPtr = (Uint8*)lineHeader + sizeof(LineHeader);
        }
      else
        {
          /* Store lineheader at current position,
           * write pointer to this new one into the last one,
           * and set its type to NORMAL
          */
          if(lineHeader)
            {
              lineHeader->type = NORMAL;
              lineHeader->skip = (size_t)tPtr - (size_t)lineHeader;
            }
          /* let lineHeader point to the position of the new lineHeader */
          lineHeader = (RLE::LineHeader*)tPtr;
          /* move target pointer past this LineHeader struct
           */
          tPtr += sizeof(LineHeader);
          
        }

      /* we assume an empty line ... */
      lineHeader->type = EMPTY;
      lineHeader->skip = 0;
      transparentLine = true;

      x = 0;
      while(x < source->width())
        {
          /* leave space for SliceHeader at current pos */
          sliceHeader = (RLE::SliceHeader*)tPtr;
          tPtr+= sizeof(RLE::SliceHeader);
          /* reset slice size counter */
          count = 0;
          /* check for transparent pixels ... */
          if(*sPtr == transColor)
            {
              while(*sPtr == transColor && 
                    count < SLICE_MAXLEN &&
                    x+count < source->width() )
                {
                  count++;
                  sPtr++;    // step to next pixel
                };
              
              sliceHeader->type = TRANSPARENT;
              sliceHeader->count = count;
              x+= count;

              // let tPtr point to next sliceHeader
              tPtr = (Uint8*)(sliceHeader + 1);                
            }
          else
            {
              /* check for opaque pixels and copy the pixel values */
              while(*sPtr != transColor &&
                    count < SLICE_MAXLEN &&
                    x+count < source->width() )
                {
                  count ++;
                  *((Uint16*)tPtr) = *sPtr;
                  tPtr+=2;
                  sPtr++;
                }
              sliceHeader->type = OPAQUE;
              sliceHeader->count = count;
              x += count;        

              transparentLine = false;
            }

        } // while(x ..

      sPtr += srcSkip;
      y++;

    } // while (y...
  
  source->unlock();

  /* we are done ...
   * now resize the target memory to the needed minimum ...
   */
  target->length = (size_t)tPtr - (size_t)target->data;
  target->data = realloc(target->data, target->length);
  assert(target->data);
  debugN(8,std::cerr<<"RLE: encoded to "<<target->length<<" bytes."<<std::endl;);

  //  dump (target);

  return target;
};

uta::Surface* decode(const Surface* source)
{
  assert(source);
  uta::Surface* target = new uta::Surface(source->width, 
                                          source->height,
                                          uta::Pixelformat::RGB565);

  blit(source, target, 0, 0);
  return target;
}



void blit(const Surface* source, uta::Surface* target, int dst_x, int dst_y)
{
  /* check for right/bottom clipping */
  int w = (source->width + dst_x < target->width()) ? 
    source->width : target->width() - dst_x;
  int h = (source->height + dst_y < target->height()) ?
    source->height : target->height() - dst_y;

  if(dst_x > target->width())
    return;
  if(dst_y > target->height())
    return;
  if(dst_x + w <= 0)
    return;
  if(dst_y + h <= 0)
    return;

  /* take into account the top/left clipping borders */
  int dx = (dst_x >= 0) ? 0 : -dst_x;
  int dy = (dst_y >= 0) ? 0 : -dst_y;
  h-= dy;
  w-= dx;

  target->lock();
    
  Uint8* dPtr = (Uint8*)target->pixels() + 
    ((dst_x > 0) ? dst_x*2 : 0) + 
    ((dst_y > 0) ? dst_y*target->pitch() : 0);
  
  int dstSkip = target->pitch() - w*2;

  Uint8* sPtr = (Uint8*)source->data;
  
  SliceHeader* sliceHeader;
  LineHeader* lineHeader;
  int count;
  int y;
  int x;
  int dx_;

  /* clipping dy lines */
  lineHeader = (LineHeader*)sPtr;
  while(dy > 0)
    {
      if(lineHeader->type == NORMAL)
        {
          dy--;
          // here ->skip is distance in bytes:
          lineHeader = (LineHeader*)((Uint8*)lineHeader + lineHeader->skip);
        }
      else
        {
          dy --;
          lineHeader++;
        }
    }

  y = 0;
  /* go for the remaining (visible) lines ...*/
  while(y < h)
    {
      //    std::cerr << "->";
        if(lineHeader->type == EMPTY)
        {
          /* skip empty lines */
          dPtr += target->pitch();
          y++;
          /* the next line header follows immediately */
          lineHeader++;
          //          std::cerr << std::endl;
        }
      else
        {
          /* this lines contains at least one opaque pixel 
             SliceHeader immediately follows the LineHeader
           */
          sliceHeader = (SliceHeader*)((Uint8*)lineHeader + sizeof(LineHeader));
            
          /* first care for left border clipping :*/
          dx_ = dx;
          while(dx_ > 0)
            {
              if(sliceHeader->type == TRANSPARENT)
                {
                  dx_ -= sliceHeader->count;
                  if(dx_ < 0)
                    dPtr += (-dx_)*2;

//                   for(int n=0; n < (-dx_); n++)
//                     std::cerr << ".";

                  // next SliceHeader immediately follows
                  sliceHeader++;
                }
              else
                {
                  dx_ -= sliceHeader->count;
                  if(dx_ < 0)
                    {
                      sPtr = (Uint8*)sliceHeader+sizeof(SliceHeader)+(sliceHeader->count+dx_)*2;
                      for(int i=0; i < (-dx_); i++)
                        {
//                           std::cerr << "#";
//                          *((Uint16*)dPtr)++ = *((Uint16*)sPtr++);

                          *((Uint16*)dPtr) = *((Uint16*)sPtr);
                          dPtr+=2;
                          sPtr+=2;
                        }
                    }
                  // skip count*2 bytes to find next SliceHeader
                  sliceHeader = (SliceHeader*)((Uint8*)sliceHeader+sliceHeader->count*2+sizeof(SliceHeader)); 
                }
            }
          x = -dx_;
          while(x < w)
            {
              if(sliceHeader->type == TRANSPARENT)
                {
                  /* skip transparent pixels */
                  count = (sliceHeader->count > w-x) ? w-x : sliceHeader->count;                  
//                   for(int n=0; n < count; n++)
//                     std::cerr << ".";

                  dPtr += count*2;
                  x+=count;
                  /* the next slice header immediately follows: */
                  sliceHeader ++;
                }
              else
                {
                  /* copy count pixels (checking for clipping as well)
                   */
                  count = (sliceHeader->count > w-x) ? w-x : sliceHeader->count;
                  x += count;
                  sPtr =  (Uint8*)sliceHeader+sizeof(SliceHeader);
                  
//                   for(int n=0; n < count; n++)
//                     std::cerr << "#";

                  while(count--)
                    {
                      *((Uint16*)dPtr) = *((Uint16*)sPtr);
                      dPtr+= 2;
                      sPtr+= 2;
                    }
                  /* jump to the next slice header */
                  sliceHeader = (SliceHeader*)((Uint8*)sliceHeader+sliceHeader->count*2+sizeof(SliceHeader));
                }

            } // while(x<w)
          /* jump to the header of the next line*/
          //          lineHeader = lineHeader->next;
          lineHeader = (LineHeader*)((Uint8*)lineHeader + lineHeader->skip);
          //          std::cerr << std::endl;
          /* move destination pointer to dest pixel at beginning
           * of next row
           */
          dPtr += dstSkip;
          y++;
        } // if(lineHeader->type == EMPTY)        
    
    }

  target->unlock();
}

/* Fast blend between two pixels of the same format.  Output to
 * 16bpp
 */
#define BLEND16_50(dst, src, mask)                               \
    do {                                                         \
        Uint32 s = *((Uint16*)src);                              \
        Uint32 d = *((Uint16*)dst);                              \
        *((Uint16*)dst) = (((s & mask) + (d & mask)) >> 1)       \
                 + (s & d & (~mask & 0xffff));                   \
    } while(0)

/* Customized version of the above macro.  Assumes that source
 * pixel is 565 RGB and that the destination is 555 RGB.  As expected,
 * the green channel is the one that's diluted.
 */
#define BLEND16_555_50(dst, src, mask)                           \
    do {                                                         \
        Uint32 s = *((Uint16*)src);                              \
        s = (s & 0x1f) | ((s & 0xffc0) >> 1);                    \
        Uint32 d = *((Uint16*)dst);                              \
        *((Uint16*)dst) = (((s & mask) + (d & mask)) >> 1)       \
                 + (s & d & (~mask & 0x7fff));                   \
    } while(0)
/* Blitting to 888 pixel-format.  Mask is hardcoded.  Target is 
 * at 32bpp, source is 16 (565)
 */
#define BLEND16_888_50(dst, src)                                \
    do {                                                        \
        Uint32 s = *((Uint16*)src);                             \
        Uint32 d = *((Uint32*)dst);                             \
        s = ((s & 0xf800) << 5) | ((s & 0x07e0) << 3) |         \
            (s & 0x001f);                                       \
        *dst = (((s & 0x00fefefe) + (d & 0x00fefefe)) >> 1)     \
               + (s & d & 0x00010101);                          \
    } while(0)
             

void blit_semitrans(const Surface* source, uta::Surface* target, int dst_x, int dst_y)
{
  /* check for right/bottom clipping */
  int w = (source->width + dst_x < target->width()) ? 
    source->width : target->width() - dst_x;
  int h = (source->height + dst_y < target->height()) ?
    source->height : target->height() - dst_y;

  if(dst_x > target->width())
    return;
  if(dst_y > target->height())
    return;
  if(dst_x + w <= 0)
    return;
  if(dst_y + h <= 0)
    return;

  /* take into account the top/left clipping borders */
  int dx = (dst_x >= 0) ? 0 : -dst_x;
  int dy = (dst_y >= 0) ? 0 : -dst_y;
  h-= dy;
  w-= dx;

  /* bpp is BYTES per pixel, not bits... we want 2 or 4 ! */
  int bpp = uta::Pixelformat::displayFormat.bpp();

  target->lock();
    
  Uint8* dPtr = (Uint8*)target->pixels() + 
    ((dst_x > 0) ? dst_x * bpp : 0) + 
    ((dst_y > 0) ? dst_y*target->pitch() : 0);
  
  int dstSkip = target->pitch() - w*bpp;

  Uint8* sPtr = (Uint8*)source->data;
  
  SliceHeader* sliceHeader;
  LineHeader* lineHeader;
  int count;
  int y;
  int x;
  int dx_;

  /* clipping dy lines */
  lineHeader = (LineHeader*)sPtr;
  while(dy > 0)
    {
      if(lineHeader->type == NORMAL)
        {
          dy--;
          // here ->skip is distance in bytes:
          lineHeader = (LineHeader*)((Uint8*)lineHeader + lineHeader->skip);
        }
      else
        {
          dy --;
          lineHeader++;
        }
    }

  y = 0;
  /* go for the remaining (visible) lines ...*/
  while(y < h)
    {
      //    std::cerr << "->";
        if(lineHeader->type == EMPTY)
        {
          /* skip empty lines */
          dPtr += target->pitch();
          y++;
          /* the next line header follows immediately */
          lineHeader++;
          //          std::cerr << std::endl;
        }
      else
        {
          /* this lines contains at least one opaque pixel 
             SliceHeader immediately follows the LineHeader
           */
          sliceHeader = (SliceHeader*)((Uint8*)lineHeader + sizeof(LineHeader));
            
          /* first care for left border clipping :*/
          dx_ = dx;
          while(dx_ > 0)
            {
              if(sliceHeader->type == TRANSPARENT)
                {
                  dx_ -= sliceHeader->count;
                  if(dx_ < 0)
                    dPtr += (-dx_)*bpp;

//                   for(int n=0; n < (-dx_); n++)
//                     std::cerr << ".";

                  // next SliceHeader immediately follows
                  sliceHeader++;
                }
              else
                {
                  dx_ -= sliceHeader->count;
                  if(dx_ < 0)
                    {
                      sPtr = (Uint8*)sliceHeader+sizeof(SliceHeader)+(sliceHeader->count+dx_)*2;
                      if (uta::Pixelformat::displayFormat ==
                          uta::Pixelformat::RGB555) {
                        for(int i=0; i < (-dx_); i++)
                        {
                          
                          BLEND16_555_50(dPtr, sPtr, 0x7BDE);
                          dPtr += bpp;
                          sPtr += 2;
                        }
                      }
		      else if (bpp == 4) {
                        // Assuming RGB0888 !!
                        for (int i = 0; i < (-dx_); ++i) 
                        {
                           BLEND16_888_50(dPtr, sPtr);
                           dPtr += bpp;
                           sPtr += 2;
                        }
                      }
                      else {  // assume display is RGB565
                        if (uta::Pixelformat::displayFormat != uta::Pixelformat::RGB565)
                        {
                          std::cerr << "Warning: unhandled format # " << uta::Pixelformat::displayFormat << std::endl;
                        }
                        for(int i=0; i < (-dx_); i++)
                        {
                          BLEND16_50(dPtr, sPtr, 0xF7DE);
                          dPtr += bpp;
                          sPtr += 2;
                        }

                      }
                    }
                  // skip count*2 bytes to find next SliceHeader
                  sliceHeader = (SliceHeader*)((Uint8*)sliceHeader+sliceHeader->count*2+sizeof(SliceHeader)); 
                }
            }
          x = -dx_;
          while(x < w)
            {
              if(sliceHeader->type == TRANSPARENT)
                {
                  /* skip transparent pixels */
                  count = (sliceHeader->count > w-x) ? w-x : sliceHeader->count;                  
//                   for(int n=0; n < count; n++)
//                     std::cerr << ".";

                  dPtr += count*bpp;
                  x+=count;
                  /* the next slice header immediately follows: */
                  sliceHeader ++;
                }
              else
                {
                  /* copy count pixels (checking for clipping as well)
                   */
                  count = (sliceHeader->count > w-x) ? w-x : sliceHeader->count;
                  x += count;
                  sPtr =  (Uint8*)sliceHeader+sizeof(SliceHeader);
                  
//                   for(int n=0; n < count; n++)
//                     std::cerr << "#";

                  if (uta::Pixelformat::displayFormat ==
                      uta::Pixelformat::RGB555) {
                    while(count--)
                    {
                      BLEND16_555_50(dPtr, sPtr, 0x7BDE);
                      dPtr += bpp;
                      sPtr += 2;
                    }
                  }
                  else if (bpp == 4) {
                    while(count--)
                    {
                      BLEND16_888_50(dPtr, sPtr);
                      dPtr += bpp;
                      sPtr += 2;
                    }
                  }
                   else {
                    if (uta::Pixelformat::displayFormat != uta::Pixelformat::RGB565)
                    {
                      std::cerr << "Warning: unhandled format # " << uta::Pixelformat::displayFormat << std::endl;
                    }
while(count--)
                    {
                      BLEND16_50(dPtr, sPtr, 0xF7DE);
                      dPtr += bpp;
                      sPtr += 2;
                    }

                  }
                  /* jump to the next slice header */
                  sliceHeader = (SliceHeader*)((Uint8*)sliceHeader+sliceHeader->count*2+sizeof(SliceHeader));
                }

            } // while(x<w)
          /* jump to the header of the next line*/
          //          lineHeader = lineHeader->next;
          lineHeader = (LineHeader*)((Uint8*)lineHeader + lineHeader->skip);
          //          std::cerr << std::endl;
          /* move destination pointer to dest pixel at beginning
           * of next row
           */
          dPtr += dstSkip;
          y++;
        } // if(lineHeader->type == EMPTY)        
    
    }

  target->unlock();
}

/*
 * For 16bpp pixels we can go a step further: put the middle component
 * in the high 16 bits of a 32 bit word, and process all three RGB
 * components at the same time. Since the smallest gap is here just
 * 5 bits, we have to scale alpha down to 5 bits as well.
 *
 * These macros are stolen from SDL.
 */
#define ALPHA_BLIT16_565(to, from, alpha)  \
    do {                                                \
                                      \
        Uint16 *src = (Uint16 *)(from);                 \
        Uint16 *dst = (Uint16 *)(to);                   \
            Uint32 s = *src++;                          \
            Uint32 d = *dst;                            \
            s = (s | s << 16) & 0x07e0f81f;             \
            d = (d | d << 16) & 0x07e0f81f;             \
            d += (s - d) * alpha >> 5;                  \
            d &= 0x07e0f81f;                            \
            *dst++ = d | d >> 16;                       \
    } while(0)

/* This special-case 555 macro may NOT work, since I think
 * we are blitting from a 565 source to a 555 target!
 */
#define ALPHA_BLIT16_555(to, from, alpha)  \
    do {                                                \
                                                        \
        Uint16 *src = (Uint16 *)(from);                 \
        Uint16 *dst = (Uint16 *)(to);                   \
        Uint32 s = *src++;                              \
        Uint32 d = *dst;                                \
        s = ((s >> 1) & 0x7c00) | (s & 0x3ff);          \
        s = (s | s << 16) & 0x03e07c1f;                 \
        d = (d | d << 16) & 0x03e07c1f;                 \
        d += (s - d) * alpha >> 5;                      \
        d &= 0x03e07c1f;                                \
        *dst++ = d | d >> 16;                           \
     } while(0)

/* TODO: a alpha blit to 888 target! */

void blit_alpha(const Surface* source, uta::Surface* target, int dst_x, int dst_y, unsigned char alpha)
{
  /* check for right/bottom clipping */
  int w = (source->width + dst_x < target->width()) ? 
    source->width : target->width() - dst_x;
  int h = (source->height + dst_y < target->height()) ?
    source->height : target->height() - dst_y;

  if(dst_x > target->width())
    return;
  if(dst_y > target->height())
    return;
  if(dst_x + w <= 0)
    return;
  if(dst_y + h <= 0)
    return;

  /* take into account the top/left clipping borders */
  int dx = (dst_x >= 0) ? 0 : -dst_x;
  int dy = (dst_y >= 0) ? 0 : -dst_y;
  h-= dy;
  w-= dx;

  target->lock();
    
  Uint8* dPtr = (Uint8*)target->pixels() + 
    ((dst_x > 0) ? dst_x*2 : 0) + 
    ((dst_y > 0) ? dst_y*target->pitch() : 0);
  
  int dstSkip = target->pitch() - w*2;

  Uint8* sPtr = (Uint8*)source->data;
  
  SliceHeader* sliceHeader;
  LineHeader* lineHeader;
  int count;
  int y;
  int x;
  int dx_;

  /* clipping dy lines */
  lineHeader = (LineHeader*)sPtr;
  while(dy > 0)
    {
      if(lineHeader->type == NORMAL)
        {
          dy--;
          // here ->skip is distance in bytes:
          lineHeader = (LineHeader*)((Uint8*)lineHeader + lineHeader->skip);
        }
      else
        {
          dy --;
          lineHeader++;
        }
    }

  y = 0;
  /* go for the remaining (visible) lines ...*/
  while(y < h)
    {
      //    std::cerr << "->";
        if(lineHeader->type == EMPTY)
        {
          /* skip empty lines */
          dPtr += target->pitch();
          y++;
          /* the next line header follows immediately */
          lineHeader++;
          //          std::cerr << std::endl;
        }
      else
        {
          /* this lines contains at least one opaque pixel 
             SliceHeader immediately follows the LineHeader
           */
          sliceHeader = (SliceHeader*)((Uint8*)lineHeader + sizeof(LineHeader));
            
          /* first care for left border clipping :*/
          dx_ = dx;
          while(dx_ > 0)
            {
              if(sliceHeader->type == TRANSPARENT)
                {
                  dx_ -= sliceHeader->count;
                  if(dx_ < 0)
                    dPtr += (-dx_)*2;

//                   for(int n=0; n < (-dx_); n++)
//                     std::cerr << ".";

                  // next SliceHeader immediately follows
                  sliceHeader++;
                }
              else
                {
                  dx_ -= sliceHeader->count;
                  if(dx_ < 0)
                    {
                      sPtr = (Uint8*)sliceHeader+sizeof(SliceHeader)+(sliceHeader->count+dx_)*2;
                      // Need to match up to destination surface format.
                      if (uta::Pixelformat::displayFormat == uta::Pixelformat::RGB555) {
                         for(int i=0; i < (-dx_); i++)
                         {
                            ALPHA_BLIT16_555(dPtr, sPtr, alpha);
                         }
                      }
                      else {  // format is RGB565, we hope !
                         for(int i=0; i < (-dx_); i++)
                         {
                            ALPHA_BLIT16_565(dPtr, sPtr, alpha);
                         }
                      }
                    }
                  // skip count*2 bytes to find next SliceHeader
                  sliceHeader = (SliceHeader*)((Uint8*)sliceHeader+sliceHeader->count*2+sizeof(SliceHeader)); 
                }
            }
          x = -dx_;
          while(x < w)
            {
              if(sliceHeader->type == TRANSPARENT)
                {
                  /* skip transparent pixels */
                  count = (sliceHeader->count > w-x) ? w-x : sliceHeader->count;                  
//                   for(int n=0; n < count; n++)
//                     std::cerr << ".";

                  dPtr += count*2;
                  x+=count;
                  /* the next slice header immediately follows: */
                  sliceHeader ++;
                }
              else
                {
                  /* copy count pixels (checking for clipping as well)
                   */
                  count = (sliceHeader->count > w-x) ? w-x : sliceHeader->count;
                  x += count;
                  sPtr =  (Uint8*)sliceHeader+sizeof(SliceHeader);
                  
                  if (uta::Pixelformat::displayFormat == uta::Pixelformat::RGB555) {
                    while(count--) {
                      ALPHA_BLIT16_555(dPtr, sPtr, alpha);
                    }
                  }
                  else {
                     while(count--) {
                       ALPHA_BLIT16_565(dPtr, sPtr, alpha);
                     }
                  }
                  /* jump to the next slice header */
                  sliceHeader = (SliceHeader*)((Uint8*)sliceHeader+sliceHeader->count*2+sizeof(SliceHeader));
                }

            } // while(x<w)
          /* jump to the header of the next line*/
          //          lineHeader = lineHeader->next;
          lineHeader = (LineHeader*)((Uint8*)lineHeader + lineHeader->skip);
          //          std::cerr << std::endl;
          /* move destination pointer to dest pixel at beginning
           * of next row
           */
          dPtr += dstSkip;
          y++;
        } // if(lineHeader->type == EMPTY)        
    
    }

  target->unlock();
}

void blit_highlighted(const Surface* source, uta::Surface* target, int dst_x, int dst_y, Color col)
{
    /* check for right/bottom clipping */
  int w = (source->width + dst_x < target->width()) ? 
    source->width : target->width() - dst_x;
  int h = (source->height + dst_y < target->height()) ?
    source->height : target->height() - dst_y;

  if(dst_x > target->width())
    return;
  if(dst_y > target->height())
    return;
  if(dst_x + w <= 0)
    return;
  if(dst_y + h <= 0)
    return;

  /* take into account the top/left clipping borders */
  int dx = (dst_x >= 0) ? 0 : -dst_x;
  int dy = (dst_y >= 0) ? 0 : -dst_y;
  h-= dy;
  w-= dx;

  target->lock();
    
  Uint8* dPtr = (Uint8*)target->pixels() + 
    ((dst_x > 0) ? dst_x*2 : 0) + 
    ((dst_y > 0) ? dst_y*target->pitch() : 0);
  
  int dstSkip = target->pitch() - w*2;

  Uint8* sPtr = (Uint8*)source->data;
  
  SliceHeader* sliceHeader;
  LineHeader* lineHeader;
  int count;
  int y;
  int x;
  int dx_;

  /* clipping dy lines */
  lineHeader = (LineHeader*)sPtr;
  while(dy > 0)
    {
      if(lineHeader->type == NORMAL)
        {
          dy--;
          // here ->skip is distance in bytes:
          lineHeader = (LineHeader*)((Uint8*)lineHeader + lineHeader->skip);
        }
      else
        {
          dy --;
          lineHeader++;
        }
    }

  y = 0;
  /* go for the remaining (visible) lines ...*/
  while(y < h)
    {
      //    std::cerr << "->";
        if(lineHeader->type == EMPTY)
        {
          /* skip empty lines */
          dPtr += target->pitch();
          y++;
          /* the next line header follows immediately */
          lineHeader++;
          //          std::cerr << std::endl;
        }
      else
        {
          /* this lines contains at least one opaque pixel 
             SliceHeader immediately follows the LineHeader
           */
          sliceHeader = (SliceHeader*)((Uint8*)lineHeader + sizeof(LineHeader));
            
          /* first care for left border clipping :*/
          dx_ = dx;
          while(dx_ > 0)
            {
              if(sliceHeader->type == TRANSPARENT)
                {
                  dx_ -= sliceHeader->count;
                  if(dx_ < 0)
          {
                    dPtr += (-dx_)*2;
             // there is at least a one pixel gap !
             *((Uint16*)(dPtr-2)) = col;
          }

//                   for(int n=0; n < (-dx_); n++)
//                     std::cerr << ".";

                  // next SliceHeader immediately follows
                  sliceHeader++;
                }
              else
                {
                  dx_ -= sliceHeader->count;
                  if(dx_ < 0)
                    {
                      sPtr = (Uint8*)sliceHeader+sizeof(SliceHeader)+(sliceHeader->count+dx_)*2;
                      for(int i=0; i < (-dx_); i++)
                        {
//                           std::cerr << "#";
//                          *((Uint16*)dPtr)++ = *((Uint16*)sPtr++);

                          *((Uint16*)dPtr) = *((Uint16*)sPtr);
                          dPtr+=2;
                          sPtr+=2;
                        }
                    }
                  // skip count*2 bytes to find next SliceHeader
                  sliceHeader = (SliceHeader*)((Uint8*)sliceHeader+sliceHeader->count*2+sizeof(SliceHeader)); 
                }
            }
          x = -dx_;
          while(x < w)
            {
              if(sliceHeader->type == TRANSPARENT)
                {
                  /* skip transparent pixels */
                  count = (sliceHeader->count > w-x) ? w-x : sliceHeader->count;                  
//                   for(int n=0; n < count; n++)
//                     std::cerr << ".";

                  dPtr += count*2;
                  x+=count;
                  /* the next slice header immediately follows: */
                  sliceHeader ++;
          if(x<w && sliceHeader->type != TRANSPARENT)
          {
             // this as well is safe since we had at least a single transparent pixel inside the clipping range !
             *((Uint16*)(dPtr-2)) = col;
          }
                }
              else
                {
                  /* copy count pixels (checking for clipping as well)
                   */
                  count = (sliceHeader->count > w-x) ? w-x : sliceHeader->count;
                  x += count;
                  sPtr =  (Uint8*)sliceHeader+sizeof(SliceHeader);
                  
//                   for(int n=0; n < count; n++)
//                     std::cerr << "#";

                  while(count--)
                    {
                      *((Uint16*)dPtr) = *((Uint16*)sPtr);
                      dPtr+= 2;
                      sPtr+= 2;
                    }
                  /* jump to the next slice header */
                  sliceHeader = (SliceHeader*)((Uint8*)sliceHeader+sliceHeader->count*2+sizeof(SliceHeader));
          if(sliceHeader->type == TRANSPARENT)
          {
             if(x < w)
               *((Uint16*)(dPtr)) = col;
          }
                }

            } // while(x<w)
          /* jump to the header of the next line*/
          //          lineHeader = lineHeader->next;
          lineHeader = (LineHeader*)((Uint8*)lineHeader + lineHeader->skip);
          //          std::cerr << std::endl;
          /* move destination pointer to dest pixel at beginning
           * of next row
           */
          dPtr += dstSkip;
          y++;
        } // if(lineHeader->type == EMPTY)        
    
    }

  target->unlock();

}

bool test_pixel(const Surface* source, int x, int y)
{
  int w = source->width;
  int h = source->height;
 
  if( x >= w || y >= h)
    return false;

  int _x,_y;
  SliceHeader* sliceHeader;
  LineHeader* lineHeader;
  
  lineHeader = (LineHeader*)source->data;
  
  _y = 0;
  while(_y < y)
    {
      if(lineHeader->type == EMPTY)
        {
          lineHeader++;
        }
      else
        {
          lineHeader = (LineHeader*)((Uint8*)lineHeader + lineHeader->skip);
        }
      _y++;
    } // while (_y <= y)

  /* now we have reached the scanline containing the pixel to test ... */

  _x=0;
  sliceHeader = (SliceHeader*)((Uint8*)lineHeader + sizeof(LineHeader));
  while(_x < x)
    {
      if(sliceHeader->type == TRANSPARENT)
        {
          _x+=sliceHeader->count;
          /* the next slice header immediately follows: */
          sliceHeader ++;

          if(x < _x)
            return false;
        }
      else
        {
          /* jump to the next slice header */
          _x+=sliceHeader->count;
          sliceHeader =(SliceHeader*)((Uint8*)sliceHeader+sliceHeader->count*2+sizeof(SliceHeader));

          if(x < _x)
            return true;
        }          
    } // while(_x<w)
          
  return false;
}


void free(Surface* data)
{
  assert(data);
  assert(data->data);

  ::free(data->data);
  ::free(data);
}

}
