/* RLE encoding/decoding/blitting ...
 *
 * written by Karsten-O. Laux <lonewolf@ardanoldo.org>
 * This code did not work in windows, and provided strange
 * results in screens that were not RGB565 (e.g. RGB555)
 *  
 */

#ifndef _MEDIA_RLE_H_
#define _MEDIA_RLE_H_

#include <uta/surface.h>
#include <uta/pixelformat.h>

// Now talking directly to SDL ... 
#include <SDL/SDL_video.h>

/* wingdi.h defines these */
#ifdef TRANSPARENT
#undef TRANSPARENT
#endif
#ifdef OPAQUE
#undef OPAQUE
#endif

/* Yes, we want the RLE compression !   */
/* ...broken on win32 */
#ifndef __WIN32__
#define USE_COMPRESSION
#endif

namespace RLE {

/* The max. count of pixels per slice. 
 * This value is only used while encoding.
 */
#define SLICE_MAXLEN  255 

/* Slice types.
 * TRANSPARENT: count transparent pixels are to get drawn, 
 *              the next SliceHeader follows directly
 * OPAQUE: count opaque pixels follow, the next SliceHeader is at
 *         ptr + sizeof(Pixel) * count
 */
enum slice_t { TRANSPARENT, OPAQUE };

/* Header of a Slice.
 */
typedef struct
{
  slice_t type;        // slice type
  unsigned char count; // how many pixels are in this slice ?
} SliceHeader;

/* Pixel scanline types.
 * EMPTY : Stands for <skip> complete transparent lines (skip > 0). The next
 *         LineHeader struct follows immediately
 * NORMAL: Well a normal line, with at least one opaque pixel. For
 *         faster transversal next points to the next LineHeader.
 */
enum scanline_t { EMPTY, NORMAL };

/* Header for a Scanline.
 * A scanline stands for a line in the decoded bitmap.
 */
typedef struct _LineHeader
{
  scanline_t type;
  size_t skip;
} LineHeader;

/** Color */
typedef Uint16 Color; // ...only works for 16bpp!

typedef struct
{
  size_t length;
  void* data;
  int width;
  int height;
  
} Surface;

//typedef struct SDL_Surface Surface;

/** RLE encode the given surface.
 *  Returns a new malloced RLE::Surface or NULL on failure.
 */
Surface* encode(const uta::Surface* source);

/** Decode.
 *  Creates a new uta::Surface and blits the decoded surface
 *  into it.
 */
uta::Surface* decode(const Surface* source);

/** Blit into target surface.
 *  Clipping is treated completely: x,y may also be negative in order to 
 *  blit across the left/top border.
 */
void blit(const Surface* source, uta::Surface* target, int x, int y);

/** Blit semitransparent into target surface.
 *  Clipping is treated completely: x,y may also be negative in order to 
 *  blit across the left/top border.
 */
void blit_semitrans(const Surface* source, uta::Surface* target, int x, int y);

/** Blit into target surface using alpha 
 *  Clipping is treated completely: x,y may also be negative in order to 
 *  blit across the left/top border.
 */
void blit_alpha(const Surface* source, uta::Surface* target, int x, int y, unsigned char alpha);


/** Blit into target surface adding a hightlighting border around the sprite.
 *  Clipping is treated completely: x,y may also be negative in order to 
 *  blit across the left/top border.
 */
void blit_highlighted(const Surface* source, uta::Surface* target, int x, int y, Color col);

/** Test pixel at given position.
 *  Return true if pixel is not transparent (=it is existant)
 */
bool test_pixel(const Surface* source, int x, int y);

/** Free RLE encoded surface.
 */
void free(Surface* data);

} // namespace RLE

#endif
