//
//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//

//  written by Karsten <klaux@rhrk.uni-kl.de> in May 2000


#include "sprite.h"
#include <uta/debug.h>
#include <uta/painter.h>

Sprite::Sprite(void* pixels, unsigned _w, unsigned _h, unsigned pitch, int pixelformat,
	       type_t type, unsigned char frames, unsigned char layers):
  syncNow_(0ul), 
  contentLayer_(0), 
  roofLayer_(Sprite::MAX_LAYER), 
  sfx_(NULL),
  delay_(0), 
  loopable_(true),
  isContainer_(false),
  isIdleAnimation_(false),
  reversable_(false),
  incomplete_(false),
  rawpixels_(pixels),
  numFrames_(frames),
  numLayers_(layers),
  cache_(NULL)
{
  assert(_w && _h && frames && layers);

  if(type == OCT)
    numOrientations_ = 8;
  else if (type == QUAD)
    numOrientations_ = 4;
  else
    numOrientations_ = 1;

#ifdef USE_COMPRESSION

  spriteWidth_ = _w;
  spriteHeight_ = _h;

  rawpixels_ = pixels;
  frames_ = new RLE::Surface*[numOrientations_*numFrames_*numLayers_];

  //int offset;
  int index; 
  for(int l=0; l < numLayers_; l++)
    for(int o = 0; o < numOrientations_; o++)
      for(int f = 0; f < numFrames_; f++)
	{
	  index = getIndex(o,f,l);
	  frames_[index] = 
	    (RLE::Surface*)((unsigned char*)pixels+index*sizeof(RLE::Surface));
// 	  assert(frames_[index]);
// 	  assert(frames_[index]->width == spriteWidth_);
// 	  assert(frames_[index]->height == spriteHeight_);
	} 
#else

  spriteWidth_ = _w / numFrames_;
  spriteHeight_ = _h / (numLayers_ * numOrientations_);

  // create image to hold the pixels
  pixels_ = new Image(pixels,_w,_h,pitch,pixelformat);

  layers_ = new Image*[numOrientations_*numFrames_*numLayers_];
   
  int bpp = pixels_->pixelformat().bpp(); // bytes per pixel 
  int format = pixels_->pixelformat()();

  int offset;
  int index;
  unsigned char* pixel_data = (unsigned char*)pixels_->pixels();

  uta::Painter* painter = new uta::Painter(pixels_);
  uta::Color transcolor = painter->getPixel(uta::Point(0,0));
  delete painter;

  for(int l=0; l < numLayers_; l++)
    for(int o = 0; o < numOrientations_; o++)
      for(int f = 0; f < numFrames_; f++)
	{
	  index = getIndex(o,f,l);
	  offset = spriteWidth_*f*bpp + spriteHeight_*(numOrientations_*l + o)*pitch;
	  // create surface using (_not copying_) the pixel data
	  layers_[index] = new Image(pixel_data+offset,
				     spriteWidth_,spriteHeight_,pitch,format);

	  layers_[index]->setTransColor(transcolor);
	  layers_[index]->setTransparency(true);
	}
#endif // ifdef USE_COMPRESSION

}

Sprite::~Sprite()
{
  debugN(8,std::cerr << "Sprite::~Sprite() ...";);
#ifdef USE_COMPRESSION
  delete [] frames_;

  debugN(8,std::cerr << "...";);	

  free(rawpixels_);

  debugN(8,std::cerr << "...";);

  if(cache_)
    delete cache_;

#else
  // delete the images ...
  for(int l=0; l < numLayers_; l++)
    for(int o = 0; o < numOrientations_; o++)
      for(int f = 0; f < numFrames_; f++)
	if(layers_[getIndex(o,f,l)])
	  delete layers_[getIndex(o,f,l)];
 
  debugN(8,std::cerr << "...";);	

  delete [] layers_;

  debugN(8,std::cerr << "...";);

  delete pixels_;

  debugN(8,std::cerr << "...";);

  free(rawpixels_);
#endif

  debugN(8, std::cerr << "survived !"<<std::endl;);
}

const Image* 
Sprite::getLayer(unsigned char orientation, unsigned char frameNo,
		 unsigned char layerNo)
{
  if(frameNo < numFrames_ && 
     layerNo < numLayers_ &&
     orientation < INVALID )
    {
      unsigned int orient = 0;
      
      if(numOrientations_ == 8)
	orient = orientation;
      else if(numOrientations_ == 4)
	orient = orientation/2;
    
      //assuming that layer 0 is always blitted
      if(sfx_)
	if(layerNo == 0)
	  if(syncNow_.test((size_t)frameNo))
	    sfx_->play(0);
#ifdef USE_COMPRESSION
      if(!cache_)
	cache_ = new Image(spriteWidth_, spriteHeight_, uta::Pixelformat::RGB565);

      blit(orientation, frameNo, layerNo, cache_, 0, 0);
	
      return cache_;
#else
      return layers_[getIndex(orient,frameNo,layerNo)];
#endif
    }  
  else
    {
      std::cerr << "Sprite::getLayer(): parameters out of range:" << std::endl;
      std::cerr << "         frame="<<(int)frameNo
	   << ",layer="<<(int)layerNo
	   << ",orientation="<<(int)orientation<<std::endl;

      return NULL;
    }
}


void
Sprite::blit(unsigned char orientation, unsigned char frameNo,
	     unsigned char layerNo, 
	     Image* target, int x, int y) 
{
#ifdef USE_COMPRESSION
    if(frameNo < numFrames_ && 
     layerNo < numLayers_ &&
     orientation < INVALID )
    {
      unsigned int orient = 0;
      if(numOrientations_ == 8)
	orient = orientation;
      else if(numOrientations_ == 4)
	orient = orientation/2;
      RLE::blit(frames_[getIndex(orient,frameNo,layerNo)], target, x, y);
    }
  else
    {
      std::cerr << "Sprite::blit(): parameters out of range:" << std::endl;
      std::cerr << "         frame="<<(int)frameNo
	   << ",layer="<<(int)layerNo
	   << ",orientation="<<(int)orientation<<std::endl;
    }
#else
  const Image* img = getLayer(orientation, frameNo, layerNo);
  if(img)
    img->blit(target, uta::Rect(x,y,spriteWidth_, spriteHeight_));
#endif
}

void
Sprite::blit_semitrans(unsigned char orientation, unsigned char frameNo,
		       unsigned char layerNo, 
		       Image* target, int x, int y) 
{
#ifdef USE_COMPRESSION
    if(frameNo < numFrames_ && 
     layerNo < numLayers_ &&
     orientation < INVALID )
    {
      unsigned int orient = 0;
      if(numOrientations_ == 8)
	orient = orientation;
      else if(numOrientations_ == 4)
	orient = orientation/2;
      RLE::blit_semitrans(frames_[getIndex(orient,frameNo,layerNo)], target, x, y);
    }
  else
    {
      std::cerr << "Sprite::blit_semitrans(): parameters out of range:" << std::endl;
      std::cerr << "         frame="<<(int)frameNo
	   << ",layer="<<(int)layerNo
	   << ",orientation="<<(int)orientation<<std::endl;
    }
#else
  const Image* img = getLayer(orientation, frameNo, layerNo);
  if(img)
    img->blit(target, uta::Rect(x,y,spriteWidth_, spriteHeight_));
#endif
}

void
Sprite::blit_alpha(unsigned char orientation, unsigned char frameNo,
		   unsigned char layerNo, 
		   Image* target, int x, int y,
		   unsigned char alpha) 
{
#ifdef USE_COMPRESSION
    if(frameNo < numFrames_ && 
     layerNo < numLayers_ &&
     orientation < INVALID )
    {
      unsigned int orient = 0;
      if(numOrientations_ == 8)
	orient = orientation;
      else if(numOrientations_ == 4)
	orient = orientation/2;
      RLE::blit_alpha(frames_[getIndex(orient,frameNo,layerNo)], target, x, y, alpha);
    }
  else
    {
      std::cerr << "Sprite::blit_alpha(): parameters out of range:" << std::endl;
      std::cerr << "         frame="<<(int)frameNo
	   << ",layer="<<(int)layerNo
	   << ",orientation="<<(int)orientation<<std::endl;
    }
#else
  const Image* img = getLayer(orientation, frameNo, layerNo);
  if(img)
    img->blit(target, uta::Rect(x,y,spriteWidth_, spriteHeight_));
#endif
}

void
Sprite::blit_highlighted(unsigned char orientation, unsigned char frameNo,
		   unsigned char layerNo, 
		   Image* target, int x, int y,
		   RLE::Color col) 
{
#ifdef USE_COMPRESSION
    if(frameNo < numFrames_ && 
     layerNo < numLayers_ &&
     orientation < INVALID )
    {
      unsigned int orient = 0;
      if(numOrientations_ == 8)
	orient = orientation;
      else if(numOrientations_ == 4)
	orient = orientation/2;
      RLE::blit_highlighted(frames_[getIndex(orient,frameNo,layerNo)], target, x, y, col);
    }
  else
    {
      std::cerr << "Sprite::blit_highlighted(): parameters out of range:" << std::endl;
      std::cerr << "         frame="<<(int)frameNo
	   << ",layer="<<(int)layerNo
	   << ",orientation="<<(int)orientation<<std::endl;
    }
#else
  const Image* img = getLayer(orientation, frameNo, layerNo);
  if(img)
    img->blit(target, uta::Rect(x,y,spriteWidth_, spriteHeight_));
#endif
}

bool 
Sprite::isOpaquePixel(int x, int y)
{
  if(x < 0 || y < 0)
    return false;

#ifdef USE_COMPRESSION
  return RLE::test_pixel(frames_[0],x,y);
#else
  return true;
#endif
  
}
