//
//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//

//  written by Karsten <klaux@rhrk.uni-kl.de> in March 2001

#ifndef _SPRITE_H
#define _SPRITE_H

#include "rle.h"
#include <uta/sample.h>

#define MAX_FRAMES 32

#include <uta/surface.h>
typedef uta::Surface Image;

#include <bitset>
class SpriteFactory;

/** Sprite
 *  Animated sprite with optinally different sequences for
 *  each orientation. All sequences are of same length. 
*/
class Sprite
{ 
  friend class SpriteFactory;

  ///possible orientations
  enum { NE=0, N=1, NW=2, W=3, SW=4, S=5, SE=6, E=7, INVALID=8 };

  ///possible type
  enum type_t 
  { 
    OCT,   // 8 orientations (dif. frames for NE,N,NW,W,SW,S,SE,E)
    QUAD,  // 4 orientations (dif. frames for NE||N, NW||W, SW||S, SE||E)
    SINGLE // only one sight
  };

  ///
  enum { MAX_LAYER = 1000 };

 public:
  /** Draw sprite frame in target at given position.
   *  Preferable method to draw a sprite; this might make use of
   *  online sprite decompression.
   */
  void blit(unsigned char orientation, unsigned char frameNo,
	    unsigned char layerNo, 
	    Image* target, int x, int y);
  /** Blit semitransparently.
   */
  void blit_semitrans(unsigned char orientation, unsigned char frameNo,
		      unsigned char layerNo, 
		      Image* target, int x, int y);
  /** Blit using arbitrary alpha
   */
  void blit_alpha(unsigned char orientation, unsigned char frameNo,
		  unsigned char layerNo, 
		  Image* target, int x, int y,
		  unsigned char alpha);  
          
 /** Blit adding a highlighting border */
  void blit_highlighted(unsigned char orientation, unsigned char frameNo,
		  unsigned char layerNo, 
		  Image* target, int x, int y,
		  RLE::Color col);

  /** Test if given pixel inside sprite is opaque.
      This method is used for collision detection.
   */
  bool isOpaquePixel(int x, int y);

  /** Get image for the requested frame.
   *  
   */
  const Image* getLayer(unsigned char orientation = 0,
			unsigned char frameNo = 0,
			unsigned char layerNo = 0);
  ///
  unsigned char numFrames() const { return numFrames_; };
  ///
  unsigned char numLayers() const { return numLayers_; };
  ///
  unsigned char numOrientations() const { return numOrientations_; };
  ///
  unsigned char contentLayer() const { return contentLayer_; };
  ///
  unsigned char roofLayer() const { return roofLayer_; };
  ///
  bool loopable() const { return loopable_; };
  ///
  bool isContainer() const { return isContainer_; };
  ///
  bool isIdleAnimation() const { return isIdleAnimation_;};
  ///
  bool reversable() const { return reversable_; };
  //
  bool incomplete() const { return incomplete_; };
  ///
  unsigned getDelay() const { return delay_; };
  ///
  int width() const { return spriteWidth_; };
  ///
  int height() const { return spriteHeight_; };
  //
  void setIncomplete() { incomplete_ = true; };

  ///
  ~Sprite();

 protected:
  ///
  Sprite(void* pixels, unsigned w, unsigned h, unsigned pitch, int pixelformat,
	 type_t type = Sprite::SINGLE, 
	 unsigned char frames = 1, 
	 unsigned char layers = 1);

  ///
  std::bitset<MAX_FRAMES> syncNow_;
  ///
  unsigned char contentLayer_;
  ///
  unsigned char roofLayer_;
  ///
  const uta::Sample* sfx_;
  ///
  unsigned delay_;
  ///
  bool loopable_;
  ///
  bool isContainer_;
  ///
  bool isIdleAnimation_;
  ///
  bool reversable_;
  ///
  bool incomplete_;
  ///
  Image* image() { return pixels_; };
  ///
  void *rawpixels_;

 private:
  ///
  int getIndex(int orient, int frame, int layer) const
    {
      return frame + (orient + layer*numOrientations_) * numFrames_; 
    }
  /** layers_[orientation][frame][layer].
   */
  Image** layers_;
  /** 
   */
  RLE::Surface** frames_;
  ///
  Image* pixels_;
  ///
  unsigned char numOrientations_;
  ///
  unsigned char numFrames_;
  ///
  unsigned char numLayers_;
  ///
  int spriteWidth_;
  ///
  int spriteHeight_;
  ///
  Image* cache_;
};

#endif
