//
//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//

//  written by Karsten <klaux@rhrk.uni-kl.de> in May 2000

#include "sprite_cache.h"
#include "media_server.h"
#include <uta/debug.h>
#include <varconf/Config.h>
#include "dii/object.h"
#include "dii/serialize.h"
#include "toolbox/operation.h"

#include "../registry/registry.h"

//how many unlocked sprites shall we cache ?
#define SPRITES_TO_KEEP 16
//how often shall we check the cache (every CHECK_INTERVAL ms)
#define CHECK_INTERVAL 1000


SpriteID SpriteCache::id_ = 1;
SpriteCache* SpriteCache::instance_ = NULL;

SpriteCache*
SpriteCache::inst()
{
  if(!instance_)
    instance_ = new SpriteCache();
    
  return instance_;
}

SpriteCache::SpriteCache() :
  factory_(),
  timer_(CHECK_INTERVAL),
  maxMemCacheSize_(SPRITES_TO_KEEP),
  maxDiskCacheSize_(0)
{
	// This will fire off the check() on every alarm sync.
	// (once a second by default)
  timer_.alarm.connect(slot(this, &SpriteCache::check));
 
  varconf::Config::inst()->
    sigv.connect(SigC::slot(this, &SpriteCache::configCallback));  

  std::string indexname = std::string(VARDIR)+"/uclient/cache_index.xml";
  
  DII::Object index;
  DII::Object::MapType entry;
  DII::Object::MapType description;
  SpriteID id;

  // read cache index back ... _no_ sanity checks here !
  if(DII::Serialize::read(indexname.c_str(), index))
    {
      assert(index.IsList());
      
      DII::Object::ListType::iterator itr = index.AsList().begin();
      while(itr != index.AsList().end())
	{
	  if(itr->IsMap())
	    {
	      entry = itr->AsMap();
	      CacheEntry cacheEntry;
	      /* set these for all restored entries:
		 These mean a known, but flushed sprite 
	      */
	      cacheEntry.ref = 0;
	      cacheEntry.age = 0;
	      cacheEntry.sprite = NULL;
	      cacheEntry.meta.valid = true;

	      if(doesContain(entry,"description"))
		{	 
		  description = entry["description"].AsMap();
		 
		  if(doesContain(description,"type"))
		    cacheEntry.description.type = description["type"].AsString();
		  if(doesContain(description,"mode")) 
		    cacheEntry.description.mode = description["mode"].AsString();
		  if(doesContain(description,"sex"))
		    cacheEntry.description.sex = description["sex"].AsString();
		  if(doesContain(description,"status"))
		    cacheEntry.description.status = description["status"].AsString();
		  if(doesContain(description,"size"))
		    cacheEntry.description.size = description["size"].AsString();
		}
	      if(doesContain(entry,"id"))
		id = entry["id"].AsInt();
	      if(doesContain(entry,"width"))
		cacheEntry.meta.width = entry["width"].AsInt();	      
	      if(doesContain(entry,"height"))
		cacheEntry.meta.height = entry["height"].AsInt();	      
	      if(doesContain(entry,"layers"))
		cacheEntry.meta.layers = entry["layers"].AsInt();	      
	      if(doesContain(entry,"frames"))
		cacheEntry.meta.frames = entry["frames"].AsInt();	      
	      requests_[cacheEntry.description] = id;
	      sprites_[id] = cacheEntry;
	      
	      /* Ensure that new ids are generated consistently 
	       * to restored ones.
	       */
	      if(id > SpriteCache::id_)
		SpriteCache::id_ = id+1;
	    }
	  itr++;
	}
      
    }
  
}

SpriteCache::~SpriteCache()
{
  debugN(8,cerr<<"SpriteCache::~SpriteCache() ... "<<endl;);

  debugN(8,cerr<<" - deleting sprite data  ..."<<endl;);
  SpriteDB::iterator itr = sprites_.begin();
  while(itr != sprites_.end())	  
    {
      if(itr->second.sprite)
	{
	  // tell factory to destroy this sprite
	  factory_.destroy(itr->second.sprite, 
			   itr->second.description);
	  
	  //set pointer to NULL
	  itr->second.sprite=NULL;
	}
      itr++;
    }
  debugN(8,cerr<<"   Done deleting sprite data."<<endl;);
  
  try
    {
      DII::Object::ListType index;
      DII::Object::MapType entry;
      DII::Object::MapType description;
      
      SpriteDB::iterator itr = sprites_.begin();
      while(itr != sprites_.end())
	{
	  description["type"] = itr->second.description.type;
	  description["mode"] = itr->second.description.mode;
	  description["sex"] = itr->second.description.sex;
	  description["status"] = itr->second.description.status;
	  description["size"] = itr->second.description.size;
	  
	  entry["id"] = (long)itr->first;
	  entry["description"]=description;
	  entry["width"]=(int)itr->second.meta.width;
	  entry["height"]=(int)itr->second.meta.height;
	  entry["frames"]=itr->second.meta.frames;
	  entry["layers"]=itr->second.meta.layers;
	  
	  index.push_back(entry);


	  itr++;
	}
      
      std::string indexname = std::string(VARDIR)+"/uclient/cache_index.xml";
      
      debugN(8,cerr << " dumping  "<<index.size()<<" index entry to "<<indexname;);
      if(DII::Serialize::write(index, indexname.c_str()))
	{
	  debugN(8,cerr << " [ok]" << endl;);
	}
      else
	{
	  debugN(8,cerr << " [FAILED]" << endl;);
	}

    }
  catch(...)
    {
      cerr << "SpriteCache failed to save index file ! " << endl;
    }
}

#define ASSIGN_IF_KNOWN_INT(var, section, key, default)\
{  if(varconf::Config::inst()->findItem(section,key)) \
    var =(int)varconf::Config::inst()->getItem(section,key); \
   else\
    var = default; }

void
SpriteCache::configCallback(const std::string& section, const std::string& key)
{
  if(section != "spritecache")   
    return;
  
  ASSIGN_IF_KNOWN_INT(maxMemCacheSize_,"spritecache","max_memcache_size", SPRITES_TO_KEEP);
  ASSIGN_IF_KNOWN_INT(maxDiskCacheSize_,"spritecache","max_diskcache_size",0);

  if(varconf::Config::inst()->findItem("spritecache","check_interval"))
    timer_.setInterval((int)varconf::Config::inst()->getItem("spritecache","check_interval"));
}

SpriteID
SpriteCache::createID()
{
  id_++;
  return id_;
}


SpriteID
SpriteCache::request(const SpriteDesc& desc)
{
  SpriteID id;
  RequestDB::iterator req;

  req = requests_.find(desc);
  if(req != requests_.end())
    {
      //this sprite has been requested before:
      //thus there is already a valid id
      return req->second;
    }

  //ask SpriteFactory for bestmatch:
  SpriteDesc matched = factory_.findMatch(desc);

  //check if the result for this request is already here:
  req = requests_.find(matched);

  if(req != requests_.end())
    {
      //remember the orginal request (creating an alias)
      requests_[desc] = req->second;
      //this sprite has been requested before:
      //thus there is already a valid id
      return req->second;
    }

  //first time we get this request !
  
  //create new unique id for this request
  id = createID();
  
  
  //store request-id mapping
  requests_[matched] = id;
  // but also store the orignal description
  requests_[desc] = id;
  
  //create "empty" cache entry
  sprites_[id].description = matched;
  sprites_[id].sprite = NULL;
  sprites_[id].age = 0;
  sprites_[id].ref = 0;
  // we have no meta data available yet
  sprites_[id].meta.valid = false;
  
  //return the new id
  return id;
    
}

Sprite*
SpriteCache::lookup(SpriteID id)
{
  SpriteDB::iterator itr;
  //search sprite in local cache:
  itr = sprites_.find(id);
  if(itr == sprites_.end())
    {
      cerr << "SpriteCache::sprite() called with invalid ID !" << endl;
      return NULL;
    }
  if(!itr->second.ref)
    {
      cerr << "SpriteCache::sprite() ID references unlocked sprite !" << endl;
      return NULL;
    } 

  return itr->second.sprite;
}

bool
SpriteCache::queryInfo(SpriteID id, SpriteInfo& info)
{
  SpriteDB::iterator itr;
  //search sprite in local cache:
  itr = sprites_.find(id);

  if(itr == sprites_.end())
    return false;
     
  if(itr->second.ref)
    info.state = LOCKED;
  else if(itr->second.sprite)
    info.state = CACHED;
  else if(itr->second.meta.valid)
    info.state = FLUSHED;
  else 
    info.state = UNKNOWN;

  info.frames = itr->second.meta.frames;
  info.layers = itr->second.meta.layers;
  info.width = itr->second.meta.width;
  info.height = itr->second.meta.height;
      
  return true;
}

void
SpriteCache::lock(SpriteID id)
{
  SpriteDB::iterator itr;
  
  //search sprite in local cache:
  itr = sprites_.find(id);
  
  if(itr == sprites_.end())
    {
      cerr << "SpriteCache::lock() on invalid ID !" << endl;
      return;
    }
  
  if(!itr->second.sprite)
    {
      //the sprite is not cached (anymore)
      // -> call factory to create this sprite
      //    (according to the definition)
      itr->second.sprite = factory_.create(itr->second.description);
      if(!itr->second.sprite)
	{
	  cerr << "SpriteCache::lock() failed to get sprite from SpriteFactory ! " << endl
	       << "               Very likely your installation is not complete. " << endl;
	  return;
 	}
      if (itr->second.sprite->incomplete())
        {
          MediaServer::instance()->registerCacheReplacement(id);
        }

      //set reference counter
      itr->second.ref = 1;
      // set meta information:
      itr->second.meta.valid = true;
      itr->second.meta.frames = itr->second.sprite->numFrames();
      itr->second.meta.layers = itr->second.sprite->numLayers();
      itr->second.meta.width = itr->second.sprite->width();
      itr->second.meta.height = itr->second.sprite->height();
    
    }
  else
    {
      //sprite is still cached !
      // -> increment reference count
      itr->second.ref++;
    }

  //reset age
  itr->second.age = 0;

}


void
SpriteCache::release(SpriteID id)
{
  SpriteDB::iterator itr;
 
  //search sprite in local cache:
  itr = sprites_.find(id);
  
  if(itr == sprites_.end())
    {
      //nothing known about this id
      cerr << "SpriteCache::release() on invalid ID !" << endl;
      return;
    }

  if(!itr->second.sprite)
    {
      //already purged sprite...
      cerr << "SpriteCache::release() on already purged ID" << endl;
      return;
    }
  
  if(itr->second.ref > 0)
    itr->second.ref--;
  
}

void
SpriteCache::check()
{
  //count cached and locked sprites
  unsigned locked = 0;
  unsigned cached = 0;
  unsigned flushed = 0;

 //flush purgeList
  while(!purgeList_.empty())
    purgeList_.pop();

  SpriteDB::iterator itr = sprites_.begin();
  while(itr != sprites_.end())
    {
      if(itr->second.ref)
	locked++;
      else
	{
	  //increase age
	  itr->second.age++;

	  if(itr->second.sprite != NULL)
	    { 
	      //add to purgeList
	      purgeList_.push(&(itr->second));
	      cached++;
	    }
	  else
	    flushed++;

	}
      itr++;
    }
  
  //print statistics
  debugN(8,
	 cerr << "SpriteCache: " 
	 << sprites_.size() << " sprites. ("
	 << locked << " locked, "
	 << cached << " cached, "
	 << flushed << " flushed)"
	 << endl;
	 );

  //check if we need to discard some sprites
  if(cached > maxMemCacheSize_)
    {
      //how many ?
      unsigned toDel = cached - maxMemCacheSize_;
      debugN(8,
	     cerr << "             discarding " << toDel << " sprites...";
	     );

      while(toDel > 0)
	{
	  //actually delete sprite
	  assert (purgeList_.top()->sprite != NULL);

	  // tell factory to destroy this sprite
	  factory_.destroy(purgeList_.top()->sprite, 
			   purgeList_.top()->description);

	  //set pointer to NULL
	  purgeList_.top()->sprite=NULL;
     
	  //ok, done
	  purgeList_.pop();
	  toDel--;
	}
      debugN(8,cerr << "[ok.]" << endl;);
    }
  
}

void
SpriteCache::reload(SpriteID id)
{
  SpriteDB::iterator itr;
 
  //search sprite in local cache:
  itr = sprites_.find(id);

  if(itr == sprites_.end())
    {
      //nothing known about this id
      cerr << "SpriteCache::reload() on invalid ID !" << endl;
      return;
    }

  assert (itr->second.sprite != NULL);

  // tell factory to destroy this sprite
  factory_.destroy(itr->second.sprite, 
		   itr->second.description);

  // the sprite is to be reloaded from the newly downloaded files.
  // -> call factory to create this sprite
  //    (according to the definition)
  itr->second.sprite = factory_.create(itr->second.description);
  if(!itr->second.sprite)
    
  cerr << "SpriteCache::reload() failed to get sprite from SpriteFactory ! "
       << endl;
  return;

  //set reference counter
  itr->second.ref = 1;
  // set meta information:
  itr->second.meta.valid = true;
  itr->second.meta.frames = itr->second.sprite->numFrames();
  itr->second.meta.layers = itr->second.sprite->numLayers();
  itr->second.meta.width = itr->second.sprite->width();
  itr->second.meta.height = itr->second.sprite->height();
    
}
