//
//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//

//  written by Karsten <klaux@rhrk.uni-kl.de> in May 2000

#ifndef _SPRITE_CACHE_H
#define _SPRITE_CACHE_H

#include "sprite_factory.h"
#include <uta/timer.h>

#include <sigc++/signal_system.h>
#include <queue>

///
typedef unsigned long SpriteID;



/** SpriteCache.
    Uses a MRU-algorithm to provide fast access to a high number of
    sprites, while conserving memory.

Usage:
    1) id = request(<definition>);
    
    Note: id != 0 
    Now you can always call queryInfo(id, info) to get the state of the sprite
    (CACHED, LOCKED, FLUSHED) and to get info about the sprite (width, height, frames, ..)

    2) lock(id);
    3) use sprite=lookup(id) to access the sprite

     ... when done using the sprite
    4) release(id);
    
    Sprites can get locked again, the request id is valid during the whole lifetime of
    the instance of SpriteCache.
 
*/
class SpriteCache : virtual public SigC::Object
{
 public:

  ///
  enum state_t { UNKNOWN, LOCKED, CACHED, FLUSHED };

  ///
  typedef struct
  {
    unsigned width;
    unsigned height;
    unsigned char frames;
    unsigned char layers;
    state_t state;
  }
  SpriteInfo;

  ///
  static SpriteCache* inst();

  ///
  ~SpriteCache();
  
  ///
  SpriteID request(const SpriteDesc&);
  ///
  Sprite* lookup(SpriteID); 
  ///
  bool queryInfo(SpriteID, SpriteInfo&);
  ///
  void lock(SpriteID);
  ///
  void release(SpriteID);
  ///
  void check();
  ///
  void reload(SpriteID);

 protected:

  ///
  SpriteCache();
  ///
  void configCallback(const std::string& section, const std::string& key);   
  ///
  SpriteFactory factory_;
  ///
  uta::Timer timer_;

  ///
  typedef struct 
  {
    unsigned width;
    unsigned height;
    unsigned char frames;
    unsigned char layers;
    bool valid;
  } MetaData;

  ///
  typedef struct 
  {
    SpriteDesc description;
    Sprite* sprite;
    unsigned int ref;
    unsigned age;
    MetaData meta;
  }
  CacheEntry;

  ///
  struct compEntry
  {
    bool operator()(CacheEntry* d1, CacheEntry* d2) const
    {
      return (d1->age < d2->age);
    }
  };

  ///
  std::priority_queue<CacheEntry*,std::vector<CacheEntry*>,compEntry> purgeList_;

  ///
  typedef std::map<SpriteID, CacheEntry> SpriteDB;
  ///
  typedef std::map<SpriteDesc, SpriteID, compDef> RequestDB;
  
  ///
  RequestDB requests_;
  ///
  SpriteDB sprites_;
  /// 
  SpriteID createID();

  ///static counter used to create unique sprite ids 
  static SpriteID id_;  
  ///
  static SpriteCache* instance_;
  ///
  unsigned long maxMemCacheSize_;
  ///
  unsigned long maxDiskCacheSize_;
  ///
  std::string installDir_;
};

#endif
