//
//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
//  written by Karsten <klaux@rhrk.uni-kl.de> in May 2000
//     rewritten in March 2001 


#include "sprite_factory.h"
#include "media_server.h"
#include "content_request.h"
#include "dii/object.h"
#include "dii/serialize.h"

#include "settings.h"
#include "media_index.h"
#include "my_time.h"

#include "../registry/registry.h"

#include <uta/progressdialog.h>
#include <uta/debug.h>
#include <uta/painter.h>
#include <uta/resources.h>

#include <varconf/Config.h>
#include <fstream>

#include "operation.h"
#include "raw.h"
#include "rle.h"

#include <set>

#include <sys/stat.h>

SpriteFactory::SpriteFactory() :
  spriteFormat_(DISPLAY)
{
  std::string mode = varconf::Config::inst()->getItem("media","spritemode");
  
  if(mode == "native")
    {
      spriteFormat_ = NATIVE;
      debugN(8,cerr << "SpriteFactory: using sprites in native mode." << endl;);
    }
  else if(mode == "indexed")
    {
      spriteFormat_ = INDEXED;
      debugN(8,cerr << "SpriteFactory: using sprites in indexed mode." << endl;);
    }
  else
    {
      spriteFormat_ = DISPLAY;
      debugN(8,cerr << "SpriteFactory: using sprites in display mode." << endl;);
    }


  //load sprite definition database 
  loadSpriteDefinitions();


}

SpriteFactory::~SpriteFactory()
{

}

unsigned
SpriteFactory::classify(const SpriteDesc& desc)
{
  unsigned n=0;
  if(!desc.type.empty())
    n++;
  if(!desc.mode.empty())
    n++;
  if(!desc.sex.empty())
    n++;
  if(!desc.status.empty())
    n++;
  if(!desc.size.empty())
    n++;
  return n;
}

bool
SpriteFactory::walker(SpriteDesc desc, SpriteDesc& result)
{
  debugN(8,cerr << "...searching "<<desc.asString()<<" ... ";);

  DefinitionDB::iterator def = definitions_.find(desc.asString());
  //found an exact match
  if(def != definitions_.end())
    {
      debugN(8,cerr<<" found it !" << endl;);
      result = desc;
      matches_[classify(result)]=result;
      //return true;
    }
  else
    {
      debugN(8,cerr<<"not found."<<endl;);
    }

  // size, mode, sex, status
   
  SpriteDesc tmp;
  
  tmp = desc;
  if(!tmp.size.empty())
    { 
      tmp.size="";
      walker(tmp, result);
    }
  tmp = desc;
  if(!tmp.mode.empty())
    { 
      tmp.mode="";
      walker(tmp, result);
    }

  tmp = desc;
  if(!tmp.sex.empty())
    { 
      tmp.sex="";
      walker(tmp, result);
    }

  tmp = desc;
  if(!tmp.status.empty())
    { 
      tmp.status="";
      walker(tmp, result);
    }  

  return false;
}

SpriteDesc
SpriteFactory::findMatch(const SpriteDesc& _desc)
{
  SpriteDesc desc = _desc;
  SpriteDesc result;
  
  std::map<std::string, double>::const_iterator ptr;
     
  //check if a floating status has been provided
  if(desc.status.empty())
    {
      ptr = desc.values.find("status");
      if(ptr != desc.values.end())
        {
          //got one !
          //          double status = ptr->second;

          //calculate desc.status from this...
        }
    }

  //check if a floating size has been provided
  if(desc.size.empty())
    {
      //FIXME weight, size ??!!!
      ptr = desc.values.find("mass");
      if(ptr != desc.values.end())
        {
          //got one !
          double mass = ptr->second;

          //calculate desc.size from this...
          if(mass < 10.0)
            desc.size = "small";
          else if(mass < 20.0)
            desc.size = "medium";
          else
            desc.size = "big";
        }
    }

  matches_.clear();
  walker(desc,result);
  if(!matches_.empty())
    {
      //use match with highes rating
      result = matches_.rbegin()->second;
      debugN(8,cerr << "... using "<<result.asString()<<endl;);
      return result;
    }

  MediaServer::instance()->request(desc);

#if 0 // The media index is useless arse
  //check if we may find a image for the given type inside the media index
  std::string filename = MediaIndex::lookupFilename(desc.type);
  if(!filename.empty())
    {
      // got one ! now we need to add a sprite definition
      SpriteDefinition def;

      //get filename from the media index
      def.filenameBase = filename;

      //set all other entries to sensible values 
      def.filenameSuffix=""; // the filename is already complete !
      def.loopable = false;
      def.container = false;
      def.first_frame = 1;
      def.last_frame = 1;
      def.skip =0;
      def.layers = 0;
      def.directions = 1;
      def.contentLayer = 1;
      def.delay = 100;
      def.soundFile = "";
      def.soundSyncFrames.clear();      
      
      //finally add the definition to the database
      definitions_[desc.asString()] = def;
      
      debugN(8,cerr<<"... found "<<desc.asString()<<" in media archive !"<<endl;);    
      return desc;
    }
#endif

  result.mode="";
  result.size="";
  result.status="";
  result.sex="";
  result.type="entity";
  
  debugN(8,cerr << "... using "<<result.asString()<<" instead."<<endl;);
  return result;
}

void
SpriteFactory::destroy(Sprite* sprite, const SpriteDesc& desc)
{
  debugN(8,cerr << "SpriteFactory::destroy("<<desc.asString()<<") ..." << endl;);

  if(!sprite)
    return;

  if(sprite->incomplete())
    {
      debugN(8,std::cerr << "SpriteFactory::destroy: destroying incomplete sprite. Will not create raw file." << endl;);
      delete sprite;
      return;
    }

  //search a sprite definition for given name
  DefinitionDB::iterator def = definitions_.find(desc.asString());
 
  if(def == definitions_.end())
    {
      cerr << "SpriteFactory::destroy got an unknown sprite description ... it cannot create raw file !" 
           << endl;
      delete sprite;
      return;
    }
 
  std::string filename = spriteFileRoot_+def->second.filenameBase+"_"+desc.asString()+"_"+std::string(".raw");

  std::ifstream raw;
  raw.open(filename.c_str());
  if(!raw)
    {
      raw.close();
#ifdef USE_COMPRESSION
      std::string magic = RAWRLE_MAGIC;
#else
      std::string magic = RAW_MAGIC;
#endif

      RawFormat::Header header;
      
      my_time now;
      header.timestamp = now();

#ifdef USE_COMPRESSION
      header.width = sprite->width();
      header.height = sprite->height();
      header.pixelformat = uta::Pixelformat::RGB565;
#else
      header.width = sprite->image()->width();
      header.height = sprite->image()->height();
      header.pixelformat = sprite->image()->pixelformat()();
#endif
      header.spriteFormat = SpriteFactory::spriteFormat_;
      header.layers = sprite->numLayers();
      header.frames = sprite->numFrames();
      switch(sprite->numOrientations())
        {
        case 8:
          header.spriteType = Sprite::OCT;
          break;
        case 4:
          header.spriteType = Sprite::QUAD;
          break;
        case 1:
          header.spriteType = Sprite::SINGLE;
          break;
        default:
          assert(false); // should not get reached !
          break;
        }
      
      std::ofstream output;
      output.open(filename.c_str());
      
      if(output)
        {
          debugN(8,cerr << " opened file ... ";);
          output.write(magic.c_str(), magic.size());
          output.put('\0');
          output.write((const char*)&header, sizeof(RawFormat::Header));

#ifdef USE_COMPRESSION      
          size_t entries = 
            sprite->numFrames()*sprite->numLayers()*sprite->numOrientations();          
          unsigned char* start_address = (unsigned char*)(sprite->rawpixels_);

          /* substract start_address from all pointers in index
             ... thus creating a reloctable map
          */
          RLE::Surface* index = (RLE::Surface*)start_address;
          for(size_t n=0; n < entries; n++)
            {
              index[n].data = 
                (void*)((size_t)(index[n].data) - (size_t)start_address);
            }

          /* search for end of last frame: */
          unsigned char* end_address = 
            start_address +
            (size_t)(index[entries-1].data)+index[entries-1].length;

          unsigned char* ptr = start_address;
          size_t pixels_size = (size_t)end_address-(size_t)start_address; 

#else
          size_t pixels_size = 
            sprite->image()->width() * 
            sprite->image()->height() *
            sprite->image()->pixelformat().bpp();
          
          unsigned char* ptr = (unsigned char*)(sprite->image()->pixels());
#endif
          
          debugN(8,cerr<<"SpriteFactory::destroy() going to dump "
                 <<pixels_size+sizeof(RawFormat::Header)+10
                 <<" bytes "<<endl<<"to file " << filename << " ... ";);
          
          output.write((char *)ptr, pixels_size);
 
          output.close();
          
          debugN(8,cerr<<" done." << endl;);
        }
      else
        {
          debugN(8,cerr << "failed to open " << filename <<" for writing !" <<endl
                 << " Maybe you are not allowed to write to target directory !?" << endl;);
        }
    }
  else
    {
      debugN(8,cerr<<" will not dump to "<<filename<<". File already exists :)" << endl;);
      raw.close();
    }

  delete sprite;
}

Sprite*
SpriteFactory::buildSprite(DefinitionDB::iterator def)
{
  std::vector<std::string> directionTags;  
  std::bitset<8> mirrored(0UL);

  Sprite::type_t sprite_type;
  int orientations;
  if(def->second.directions == 8)
    {
      directionTags.push_back("NE");
      directionTags.push_back("N");
      directionTags.push_back("NW");
      directionTags.push_back("W");
      directionTags.push_back("SW");
      directionTags.push_back("S");
      directionTags.push_back("SE");
      directionTags.push_back("E");
      orientations = 8;
      sprite_type = Sprite::OCT;
    }
  else if(def->second.directions == 5)
    {
      directionTags.push_back("NE");
      directionTags.push_back("N");
      directionTags.push_back("NE");  // mirrored
      directionTags.push_back("E");   // mirrored
      directionTags.push_back("SE");  // mirrored
      directionTags.push_back("S");
      directionTags.push_back("SE");
      directionTags.push_back("E");
      // mirror these directions
      mirrored.set(2);
      mirrored.set(3);
      mirrored.set(4);

      orientations = 8;
      sprite_type = Sprite::OCT;
    }
  else if(def->second.directions == 4)
    {
      directionTags.push_back("NE");
      directionTags.push_back("NW");
      directionTags.push_back("SW");
      directionTags.push_back("SE");
      sprite_type = Sprite::QUAD;
      orientations = 4;
    }
  else
    {
      orientations = 1;
      sprite_type = Sprite::SINGLE;
    }

  
  int layers = def->second.layers;
  int frames = (def->second.last_frame - def->second.first_frame)/(def->second.skip+1)+1;
  int o,f,l;

  if(!layers)
    layers = 1;
  if(!frames)
    frames = 1;

  /* create an array of Image* to hold pointer to all loaded frames */
  std::vector< std::vector< std::vector<Image*> > > images;  
  images.resize(orientations);
  for(o = 0; o < orientations; o++)
    {
      images[o].resize(frames);
      for(f = 0; f < frames; f++)
        {
          images[o][f].resize(layers);
          // initialize all pointers to NULL
          for(l=0; l<layers; l++)
            images[o][f][l] = NULL;
        }
    }
  
  std::string filename;
  std::string base = spriteFileRoot_+def->second.filenameBase;
  std::string directionTag;
  std::string frameTag;
  std::string layerTag;
  std::string suffix = def->second.filenameSuffix;
  Image* img = NULL;

  int __w = 0;
  int __h = 0;
  std::set<std::string> missingFrames;

  /* load the frames ... */
  for(o= 0; o < orientations; o++)
    {
      if(directionTags.size())
        directionTag = std::string("_")+directionTags[o];
      int frame = def->second.first_frame;
      for(f=0; f < frames; f++)
        { 
          if(frames > 1)
            frameTag = itos(frame);

          frame += def->second.skip + 1;

          for(l=0; l < layers; l++)
            {
              if(layers > 1)
                layerTag = std::string("l")+itos(l);
              
              if(!directionTag.empty() && !layerTag.empty())
                layerTag = std::string("_")+layerTag;

              if(!frameTag.empty())
                if(!layerTag.empty() ||
                   (layerTag.empty() && !directionTag.empty()))
                  frameTag = std::string("_")+frameTag;

              filename = 
                base+directionTag+layerTag+frameTag+def->second.filenameSuffix;

              debugN(8, cerr << base <<"#"<<directionTag<<"#"
                     <<layerTag<<"#"<<frameTag
                     <<"#"<<def->second.filenameSuffix<<endl;);

              img = new Image();
              // TODO TODO TODO
              // check if we can stat() the file - if not, 
              // try to stat() a cached file in the user's ~/.uclient/cache
              struct stat imgFileStats;
              if (stat(filename.c_str(), &imgFileStats) != 0)
                {
                  std::cout << "Could not stat RF " << filename
                            << std::endl << std::flush;
                  // Check if the file is in the cache, and if not, get
                  // it from the server.
                  std::string request = def->second.filenameBase+directionTag
                                                                +layerTag
                                                                +frameTag
                                                                +suffix;
                  std::string cachedFile = spriteCacheRoot_ + request;
                  if (stat(cachedFile.c_str(), &imgFileStats) != 0)
                    {
                      std::cout << "Could not stat CF " << cachedFile
                                << std::endl << std::flush;
                      missingFrames.insert(request);
                      // MediaServer::instance()->addContentRequest(request);
                      // Give the media server an idea of how many
                      // files we need, and get it to signal us back
                      // when all the files are in place, then we
                      // can create a completely new sprite. Adding
                      // the images one by one as they arrive looks to
                      // be far too involved.
                      if(!__w || !__h)
                        {
                          __w = 32;
                          __h = 32;
                        }
                    }
                  else
                    {
                      filename = cachedFile;
                    }
                }
              //
              // In the event of No File, contact the media server and try
              // to download the file.
              if(img->load(filename))
                {
#ifdef USE_COMPRESSION
                  // hardwired ... as the RLE stuff only handles 16bit surfaces
                  img->convert(uta::Pixelformat::RGB565);
                  uta::Painter* painter = new uta::Painter(img);
                  img->setTransColor(painter->getPixel(uta::Point(0,0)));
                  img->setTransparency(true);
                  delete painter;
#else
                  /* change pixelformat */
                  convertFormat(img);
#endif
                  /* check if we need to mirror this frame
                     (depending on the orientation) 
                  */
                  if(mirrored.test(o))
                     img->mirror();
                  
                  __w = img->width();
                  __h = img->height();

                  // store reference to this new image 
                  images[o][f][l] = img;
                  debugN(8,cerr << " OK." <<endl;);
                }
              else
                {
                  // else discard
                  delete img;
                  debugN(8, cerr << " FAILED ! creating empty surface ...";);
                  if(!__w || !__h)
                    {
                      cerr << "CRITICAL: SpriteFactory::buildSprite() cannot access \""<<filename<<"\" !" 
                           << endl
                           << "          ... but we NEED this file. Please check your installation !" 
                           << endl;

                      __w = 32;
                      __h = 32;
                    }
                    
                  img = new uta::Surface(__w, __h);
                  assert(img);
#ifdef USE_COMPRESSION
                  // hardwired ... as the RLE stuff only handles 16bit surfaces
                  img->convert(uta::Pixelformat::RGB565);
                  uta::Painter* painter = new uta::Painter(img);
                  img->setTransColor(painter->getPixel(uta::Point(0,0)));
                  img->setTransparency(true);
                  delete painter;
#else
                  /* change pixelformat */
                  convertFormat(img);
#endif
                  // store reference to this new image 
                  images[o][f][l] = img;
                  debugN(8,cerr << " OK." << endl;);
                
                }
            }
        }
    }
  
  assert(images[0][0][0]);

  int sprite_width = images[0][0][0]->width();
  int sprite_height = images[0][0][0]->height();
  
  int pixelformat;
  
#ifdef USE_COMPRESSION

  // hardwired ... as the RLE stuff only handles 16 bit surfaces
  pixelformat = uta::Pixelformat::RGB565;

  int w = sprite_width;
  int h = sprite_height;
  int bpp = 2;
  int pitch = w * bpp;

  size_t entries = orientations*frames*layers;
  size_t compressed_size = 0;
  size_t uncompressed_size = entries*sprite_width*sprite_height*2;
  size_t index_size = entries*sizeof(RLE::Surface);

  debugN(8, cerr << "   going to compress "<<entries<<" frame(s) ...";);
  
  /* create an array of RLE::Surface* to hold pointer to all 
     compressed frames */
  RLE::Surface* comp_images = (RLE::Surface*)malloc(index_size);
  assert(comp_images);

  /* RL encode every frame and store pointer */
  RLE::Surface* _img = NULL;
  for(o = 0; o<orientations; o++)
    for(f = 0; f <frames; f++)
      for(l = 0; l < layers; l++)
        {
          _img = RLE::encode(images[o][f][l]);
          delete images[o][f][l];
          compressed_size += _img->length;
          comp_images[f + (o + l*orientations) * frames] = *_img;
        }

  debugN(4,cerr << "compressed sprite "<<def->first
                <<" :  compression ratio = "<<uncompressed_size
                <<" / "<<compressed_size
                <<" ("<<(float)uncompressed_size/compressed_size<<")"<<endl;);

  /* now we now how much space the compressed images will need,
     reallocate appropriate memory area:
  */
  unsigned char* pixels = 
    (unsigned char*)realloc((void*)comp_images,index_size+compressed_size);
  assert(pixels);
  /* reassign index table pointer to the new memory position */
  comp_images = (RLE::Surface*)pixels;

  /* set ptr to address where the compressed pixels are meant to start*/
  unsigned char* ptr = pixels + index_size;

  debugN(8,cerr << "    building compound sprite and freeing temp frames ...";);
  /* now copy over the compressed data to the new place */
  for(size_t n=0; n < entries; n++)
    {
      //copy compressed pixels
      memcpy((void*)ptr, (void*)comp_images[n].data, comp_images[n].length);
      //free old data
      free(comp_images[n].data);
      //update pixel pointer in index
      comp_images[n].data = (void*)ptr;
      //advance dest pointer
      ptr += comp_images[n].length;
    }
  
  debugN(8,cerr << " DONE !" << endl;);

#else
  switch(spriteFormat_)
    {
    case DISPLAY:
      pixelformat = uta::Pixelformat::DISPLAY;
      break;
    case INDEXED:
      pixelformat = uta::Pixelformat::IND8;
      break;
    case NATIVE:
    default:
      pixelformat = uta::Pixelformat::RGBA8888;
    }

  int w = sprite_width*frames;
  int h = sprite_height*orientations*layers;
  int bpp = uta::Pixelformat(pixelformat).bpp();
  int pitch = w * bpp;

  debugN(8,cerr<<"   creating a compound surface of "
         <<w<<"*"<<h<<"pixels to hold all"
         <<" frames of the new sprite."<<endl;);

  /* create big surface to contain all frames */
  unsigned char* pixels = (unsigned char*)malloc(w*h*bpp);
  assert(pixels);

  /* create tempory surface to allow copying */
  Image* temp = new Image(pixels,w,h,pitch,pixelformat);

  uta::Rect dest = uta::Rect(0,0,sprite_width, sprite_height);

  /* copy all frames to this surface 
   */
  unsigned char *srcPtr, *dstPtr;
  int dstSkip = (w - sprite_width)*bpp;
  int srcSkip = 0;
  Image* _img;
  for(o = 0; o<orientations; o++)
    for(f = 0; f <frames; f++)
      for(l = 0; l < layers; l++)
        {
          _img = images[o][f][l];
          _img->lock();
          srcSkip = _img->pitch() - _img->width()*bpp;
          srcPtr = (unsigned char*)_img->pixels();
          dstPtr = (unsigned char*) temp->pixels() + 
            sprite_width*f*bpp + sprite_height*(orientations*l + o)*w*bpp;
          int x = 0;
          int y = sprite_height;
          while(y--)
            {
              x = sprite_width * bpp;
              while(x--)
                {
                  *dstPtr++ = *srcPtr++;
                }
              dstPtr += dstSkip;
              srcPtr += srcSkip;
            }
          _img->unlock();
        }

  /* delete temp surface, this does not delete the pixels ! */
  delete temp;

#endif

  /* create sprite object from the pixel data
     the sprite takes over the ownership for the newly created surface,
     so pixels is _not_ deleted here
  */
  Sprite* sprite = new Sprite(pixels,w,h,pitch,pixelformat,
                              sprite_type, frames, layers);
  if (!missingFrames.empty())
    {
      MediaServer::instance()->addFileSet(missingFrames);
      sprite->setIncomplete();
    }
  
  assert(sprite);
  return sprite;
}

Sprite*
SpriteFactory::create(const SpriteDesc& desc)
{
  //search a sprite definition for given name
  DefinitionDB::iterator def = definitions_.find(desc.asString());
  if(def == definitions_.end())
    {
#if 0 // Not the media index _AGAIN_
      std::string filename = MediaIndex::lookupFilename(desc.type);
      if(!filename.empty())
        {
          // got one ! now we need to add a sprite definition
          SpriteDefinition _def;

          //get filename from the media index
          _def.filenameBase = filename;
          
          //set all other entries to sensible values 
          _def.filenameSuffix=""; // the filename is already complete !
          _def.loopable = false;
          _def.container = false;
          _def.first_frame = 1;
          _def.last_frame = 1;
          _def.skip =0;
          _def.layers = 0;
          _def.directions = 1;
          _def.contentLayer = 1;
          _def.delay = 100;
          _def.soundFile = "";
          _def.soundSyncFrames.clear();      
          
          //finally add the definition to the database
          definitions_[desc.asString()] = _def;
        }
#endif
      def = definitions_.find(desc.asString());
      
      if(def == definitions_.end())
        {
          cerr << "SpriteFactory::create() completely failed to find a sprite for "<<desc.asString() << " ! " << endl
               << "     Very likely your installation is not complete. " << endl;
          return NULL;
        }
    }

  
  debugN(8, cerr << "SpriteFactory: going to load "<<desc.asString()<<endl;);

  Sprite* sprite = NULL;

  /* check there exists a raw file ...*/
  std::ifstream raw;
  std::string rawfilename = spriteFileRoot_+def->second.filenameBase+"_"+desc.asString()+"_"+std::string(".raw");
  raw.open(rawfilename.c_str());
  if(raw)
    {
      char magic[10];
      RawFormat::Header header;
      debugN(8,cerr<<"SpriteFactory::create() found raw sprite file !"<<endl
             <<" ... going to check it ... ";);
 
      raw.read(magic, 10);
#ifdef USE_COMPRESSION
      if(!strcmp(magic, RAWRLE_MAGIC))
        {
          debugN(8, cerr << "OK, found a RLE compressed raw file." << endl;);
#else
      if(!strcmp(magic,RAW_MAGIC))
        {
          debugN(8, cerr << "OK, found a valid raw file." << endl;);
#endif
          my_time now;

          raw.read((char*)&header,sizeof(RawFormat::Header));
        
                // I wonder if I can put this into a debugN?
          debugN(8, long int deltaT = now() - header.timestamp; cerr << "  - raw file age: "<<deltaT/1000<<" seconds" << endl;);

          if(header.spriteFormat != SpriteFactory::spriteFormat_)
            debugN(8,cerr<<"  - raw file uses different sprite format ... this may result in low quality rendering ... "<<endl<<"    maybe we should reconstruct raw sprite file ..."<<endl;);

          uta::Pixelformat pixelformat(header.pixelformat);
          assert(pixelformat.valid());

#ifdef USE_COMPRESSION 
          int orients = 1;
          if(header.spriteType == Sprite::OCT)
            orients = 8;
          else if (header.spriteType == Sprite::QUAD)
            orients = 4;
          
          size_t entries = header.frames*header.layers*orients;
          size_t index_size = entries*sizeof(RLE::Surface);
          RLE::Surface* index = (RLE::Surface*)malloc(index_size);
          assert(index);

          // read index
          raw.read((char*)index, index_size);
          
          size_t pixels_size = 
            (long)index[entries-1].data+index[entries-1].length;

          // reallocate memory to hold index and pixel data:
          unsigned char* data = 
            (unsigned char*)realloc((void*)index, index_size+pixels_size);
          assert(data);
          index = (RLE::Surface*)data;

          // change relative offsets to absolute address:
          for(size_t n=0; n < entries; n++)
            index[n].data = data + (long)index[n].data;

          debugN(8,cerr<<"SpriteFactory::create() going to read "<<pixels_size
                 <<" bytes pixeldata from "<<rawfilename << " ...";);

          // finally read the compressed pixels:
          raw.read((char*)data+index_size, pixels_size);

          debugN(8,cerr << " OK !" << endl;);
#else
          size_t pixels_size = header.width * header.height * pixelformat.bpp();

          debugN(8,cerr<<"SpriteFactory::create() going to read "<<pixels_size
                 <<" bytes pixeldata from "<<rawfilename << " ...";);
          
          unsigned char* data = (unsigned char*)malloc(pixels_size);
#endif
          raw.read((char*)data, pixels_size);

          debugN(8,cerr << " OK !" << endl;);
          
          /* now create a sprite from the data */

          sprite = new Sprite(data,
                              header.width,
                              header.height,
                              header.width*pixelformat.bpp(),
                                   header.pixelformat,
                              (Sprite::type_t)header.spriteType, 
                              header.frames, 
                              header.layers);                 
        }
      else
        {
          debugN(8, cerr << "ERROR. not a raw file !" << endl;);
        } 
      raw.close();
        } // if(raw)
    
  /* Did we succeed to reconstruct sprite from raw file ?
   * If not we have to build it from single frames ...
   * ... the hard and _slow_ way:
   */
  if(!sprite)
    {
      debugN(8,cerr<<rawfilename << " not found. Building sprite from original pngs ... " << endl;);
      sprite = buildSprite(def);
    }

  /* buildSprite() is meant to _always_ return a valid sprite */
  if(!sprite)
    {
      cerr << "SpriteFactory::create() completely failed to load a sprite for "<<desc.asString() << " ! " << endl
           << "     Very likely your installation is not complete. " << endl;
      return NULL;
    }   

  /// set sprite parameters
  sprite->isContainer_ = def->second.container;
  sprite->isIdleAnimation_ = def->second.isIdleAnimation;
  sprite->reversable_ = def->second.reversable;
  sprite->contentLayer_ = def->second.contentLayer;
  sprite->roofLayer_ = def->second.roofLayer;
  sprite->loopable_ = def->second.loopable;
  sprite->delay_ = def->second.delay;
          
  
  //sound fx stuff
  if(!def->second.soundFile.empty())
    {
      //get sound resource
      // hmmmmmmm ... why not also loading the sample here using SAMPLE_RES()
      // ...
      sprite->sfx_ = SAMPLE_RES(def->second.soundFile.c_str());
      
      //mark sync frames
      std::list<unsigned int>::const_iterator itr = 
        def->second.soundSyncFrames.begin();
      while(itr !=  def->second.soundSyncFrames.end())
        {
          sprite->syncNow_.set(*itr);
          itr++;
        }
    }
  
  
  debugN(8, cerr << "SpriteFactory: done loading "<<desc.asString()<<endl;);

  //return pointer
  return sprite;
}

void
SpriteFactory::convertFormat(Image* frame)
{
  switch(spriteFormat_)
    {
    case DISPLAY:
      {
        frame->convert(uta::Pixelformat::DISPLAY);
        //use RLE acceleration, which is very fast !
        uta::Painter* painter = new uta::Painter(frame);
        frame->setTransColor(painter->getPixel(uta::Point(0,0)));
        frame->useRLEAcceleration(true);
        frame->setTransparency(true);
        delete painter;
      }
      break;
      
    case INDEXED:
      {
        frame->convert(uta::Pixelformat::IND8);
        //use RLE acceleration, which is very fast !
        uta::Painter* painter = new uta::Painter(frame);
        frame->setTransColor(painter->getPixel(uta::Point(0,0)));
        frame->useRLEAcceleration(true);
        frame->setTransparency(true);
        delete painter;
      }
      break;
    case NATIVE:
      {
        if(!frame->hasAlphaChannel())
          {        
            //use RLE acceleration, which is very fast !
            uta::Painter* painter = new uta::Painter(frame);
            frame->setTransColor(painter->getPixel(uta::Point(0,0)));
            frame->useRLEAcceleration(true);
            frame->setTransparency(true);
            delete painter;
          }
      }
    default:
      break;
    }
}

void
SpriteFactory::installSpriteDefinition(const std::string & type,
                                       const DII::Object & spriteDef)
{
  std::string install_dir = Registry::instance()->readInstallDir();
  SpriteDesc desc;
  SpriteDefinition def;

  desc.type = type;
  
  DII::Object::MapType::const_iterator fItr = spriteDef.AsMap().find("modes");
  if (fItr == spriteDef.AsMap().end())
    {
      return;
    }
  try
    {
      
      const DII::Object::ListType & modes = fItr->second.AsList();
      DII::Object::ListType::const_iterator mItr;
      
      for(mItr = modes.begin(); mItr != modes.end(); mItr++)
        {    

          try
            {          
              //set the default values each time
              def.filenameSuffix=".png";
              def.loopable = true;
              def.container = false;
              def.isIdleAnimation = false;
              def.reversable = false;
              def.first_frame = 1;
              def.last_frame = 1;
              def.skip =0;
              def.layers = 0;
              def.directions = 1;
              def.contentLayer = 1;
              def.roofLayer = Sprite::MAX_LAYER;
              def.delay = 100;
              def.soundFile = "";
              def.soundSyncFrames.clear();
              
              const DII::Object::MapType & mode = mItr->AsMap();

              //set defaults
              desc.mode ="";
              desc.sex = "";
              desc.size = "";
              desc.status = "";

              DII::Object::MapType::const_iterator cItr;
              if((cItr = mode.find("mode")) != mode.end())
                desc.mode = cItr->second.AsString();
              if((cItr = mode.find("sex")) != mode.end())
                desc.sex = cItr->second.AsString();              
              if((cItr = mode.find("size")) != mode.end())
                desc.size = cItr->second.AsString();
              if((cItr = mode.find("status")) != mode.end())
                desc.status = cItr->second.AsString();
              
              std::string name;
              
              cItr = mode.find("filename");
              def.filenameBase = cItr->second.AsString();
              
              if((cItr = mode.find("suffix")) != mode.end())
                def.filenameSuffix = cItr->second.AsString();
              if((cItr = mode.find("skip"))              != mode.end())
                def.skip = cItr->second.AsInt();
              if((cItr = mode.find("layers")) != mode.end())
                def.layers = cItr->second.AsInt();
              if((cItr = mode.find("contentLayer")) != mode.end())
                def.contentLayer = cItr->second.AsInt();
              if((cItr = mode.find("roofLayer")) != mode.end())
                def.roofLayer = cItr->second.AsInt();
              if((cItr = mode.find("first_frame")) != mode.end())
                def.first_frame = cItr->second.AsInt();
              if((cItr = mode.find("last_frame")) != mode.end())
                def.last_frame = cItr->second.AsInt();
              if((cItr = mode.find("directions")) != mode.end())
                def.directions = cItr->second.AsInt();
              if((cItr = mode.find("delay")) != mode.end())
                def.delay = cItr->second.AsInt();
              
              if((cItr = mode.find("sound")) != mode.end())
                def.soundFile = cItr->second.AsString();
              
              if((cItr = mode.find("syncframes")) != mode.end())
                {
                  const DII::Object::ListType& syncs = cItr->second.AsList();
                  DII::Object::ListType::const_iterator sItr;
                  
                  for(sItr = syncs.begin(); sItr != syncs.end(); sItr++)
                    def.soundSyncFrames.push_back(sItr->AsInt());
                } 
              
              if((cItr = mode.find("container")) != mode.end())
                def.container = (cItr->second.AsInt() != 0);
              if((cItr = mode.find("loopable")) != mode.end())
                def.loopable = (cItr->second.AsInt() != 0);
              if((cItr = mode.find("idleAnimation")) != mode.end())
                def.isIdleAnimation = (cItr->second.AsInt() != 0);
              if((cItr = mode.find("reversable")) != mode.end())
                def.reversable = (cItr->second.AsInt() != 0);
              debugN(8,cerr << "("<<desc.asString()<<") ";);
              
              definitions_[desc.asString()] = def;
            }
          catch(...)
            {
              cerr << "SpriteFactory: error in description for mode "
                   << desc.mode << " of sprite " << desc.type
                   << "; will ignore entry." << endl;
            }
        }
      
      debugN(8,cerr << endl;);
      
    }
  catch(...)
    {
      cerr << "SpriteFactory: error in description for " << desc.type 
           << "; will ignore entry." << endl;
    }
}

bool
SpriteFactory::loadSpriteDefinitions()
{
  std::string install_dir = Registry::instance()->readInstallDir();
  std::string filename = install_dir+"/uclient/sprites.xml";

  spriteFileRoot_ = install_dir + "/forge/";
  spriteCacheRoot_ = ContentRequest::getCacheDir();

  DII::Object obj;
    
  if(!DII::Serialize::read(filename.c_str(), obj))
    {
      debugN(8,cerr<<"SpriteFactory: Error parsing sprites.xml"<<endl;);
      return false;
    };
    
  DII::Object::MapType::iterator fItr = obj.AsMap().find("header");
  if (fItr == obj.AsMap().end())
    {
      debugN(8,cerr<<"SpriteFactory: sprites.xml has no header"<<endl;);
      return false;
    }
  DII::Object::MapType& header = fItr->second.AsMap();
  fItr = obj.AsMap().find("sprites");
  if (fItr == obj.AsMap().end())
    {
      debugN(8,cerr<<"SpriteFactory: sprites.xml has no sprites"<<endl;);
      return false;
    }
  DII::Object::MapType& sprites = fItr->second.AsMap();

  fItr = header.find("version");
  if (fItr == obj.AsMap().end())
    {
      debugN(8,cerr<<"SpriteFactory: sprites.xml has no version"<<endl;);
      return false;
    }
  const std::string & version = fItr->second.AsString();
  if(version != std::string(VERSION))
    {
      cerr << "Warning: sprite definitions are for UClient "<<version
           << ", but you are running Version " << VERSION << endl;
    }

  fItr = obj.AsMap().find("server");
  if (fItr != obj.AsMap().end())
    {
      DII::Object::MapType & serverInfo = fItr->second.AsMap();
      fItr = serverInfo.find("hostname");
      if (fItr != serverInfo.end())
        {
          const std::string & hostname = fItr->second.AsString();
          fItr = serverInfo.find("port");
          if (fItr != serverInfo.end())
            {
              debugN(8,cerr<<"SpriteFactory: Registered media server hostname "
                           <<hostname<<endl;);
              MediaServer::instance()->setServer(hostname,
                                                 fItr->second.AsInt(),
                                                 this);
            }
        }
      fItr = serverInfo.find("baseurl");
      if (fItr != serverInfo.end())
        {
          const std::string & contentBaseURL = fItr->second.AsString();
          ContentRequest::setBaseURL(contentBaseURL);
        }
    }

  debugN(8,cerr << sprites.size() << " sprite definitions found !" << endl;);
  
  DII::Object::MapType::iterator itr;
  for(itr = sprites.begin(); itr != sprites.end(); itr++)
    {
      SpriteDesc desc;
      SpriteDefinition def;

      desc.type = itr->first;
  
      fItr = itr->second.AsMap().find("modes");
      if (fItr == itr->second.AsMap().end())
        {
          continue;
        }
      try
        {
          
          DII::Object::ListType & modes = fItr->second.AsList();
          DII::Object::ListType::iterator mItr;
          
          for(mItr = modes.begin(); mItr != modes.end(); mItr++)
            {    

              try
                {          
                  //set the default values each time
                  def.filenameSuffix=".png";
                  def.loopable = true;
                  def.container = false;
                  def.isIdleAnimation = false;
                  def.reversable = false;
                  def.first_frame = 1;
                  def.last_frame = 1;
                  def.skip =0;
                  def.layers = 0;
                  def.directions = 1;
                  def.contentLayer = 1;
                  def.roofLayer = Sprite::MAX_LAYER;
                  def.delay = 100;
                  def.soundFile = "";
                  def.soundSyncFrames.clear();
                  
                  const DII::Object::MapType & mode = mItr->AsMap();

                  //set defaults
                  desc.mode ="";
                  desc.sex = "";
                  desc.size = "";
                  desc.status = "";

                  DII::Object::MapType::const_iterator cItr;
                  if((cItr = mode.find("mode")) != mode.end())
                    desc.mode = cItr->second.AsString();
                  if((cItr = mode.find("sex")) != mode.end())
                    desc.sex = cItr->second.AsString();              
                  if((cItr = mode.find("size")) != mode.end())
                    desc.size = cItr->second.AsString();
                  if((cItr = mode.find("status")) != mode.end())
                    desc.status = cItr->second.AsString();
                  
                  std::string name;
                  
                  cItr = mode.find("filename");
                  def.filenameBase = cItr->second.AsString();
                  
                  if((cItr = mode.find("suffix")) != mode.end())
                    def.filenameSuffix = cItr->second.AsString();
                  if((cItr = mode.find("skip"))              != mode.end())
                    def.skip = cItr->second.AsInt();
                  if((cItr = mode.find("layers")) != mode.end())
                    def.layers = cItr->second.AsInt();
                  if((cItr = mode.find("contentLayer")) != mode.end())
                    def.contentLayer = cItr->second.AsInt();
                  if((cItr = mode.find("roofLayer")) != mode.end())
                    def.roofLayer = cItr->second.AsInt();
                  if((cItr = mode.find("first_frame")) != mode.end())
                    def.first_frame = cItr->second.AsInt();
                  if((cItr = mode.find("last_frame")) != mode.end())
                    def.last_frame = cItr->second.AsInt();
                  if((cItr = mode.find("directions")) != mode.end())
                    def.directions = cItr->second.AsInt();
                  if((cItr = mode.find("delay")) != mode.end())
                    def.delay = cItr->second.AsInt();
                  
                  if((cItr = mode.find("sound")) != mode.end())
                    def.soundFile = cItr->second.AsString();
                  
                  if((cItr = mode.find("syncframes")) != mode.end())
                    {
                      const DII::Object::ListType& syncs = cItr->second.AsList();
                      DII::Object::ListType::const_iterator sItr;
                      
                      for(sItr = syncs.begin(); sItr != syncs.end(); sItr++)
                        def.soundSyncFrames.push_back(sItr->AsInt());
                    } 
                  
                  if((cItr = mode.find("container")) != mode.end())
                    def.container = (cItr->second.AsInt() != 0);
                  if((cItr = mode.find("loopable")) != mode.end())
                    def.loopable = (cItr->second.AsInt() != 0);
                  if((cItr = mode.find("idleAnimation")) != mode.end())
                    def.isIdleAnimation = (cItr->second.AsInt() != 0);
                  if((cItr = mode.find("reversable")) != mode.end())
                    def.reversable = (cItr->second.AsInt() != 0);
                  debugN(8,cerr << "("<<desc.asString()<<") ";);
                  
                  definitions_[desc.asString()] = def;
                }
              catch(...)
                {
                  cerr << "SpriteFactory: error in description for mode " << desc.mode 
                       << " of sprite " << desc.type << "; will ignore entry." << endl;
                }
            }
          
          debugN(8,cerr << endl;);
          
        }
      catch(...)
        {
          cerr << "SpriteFactory: error in description for " << desc.type 
               << "; will ignore entry." << endl;
        }
    }
  return true;
}
