//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 01/2000

#define NEED_DII_SERVICES

#include "console.h"
#include "tokenise.h"
#include "settings.h"
#include <varconf/Config.h>

#include <uta/debug.h>
#include <uta/rootwindow.h>
#include <uta/application.h>

#include "operation.h"
#include "dii/dii.h"

#include "gui.h"
#include "viewport.h"
#include "player/player.h"

class Console* Console::instance_ = 0;

void
Console::deinit()
{
  if(panel_)
    {
      uta::Rect r = panel_->getRect();
     
      DII::Object::MapType screenarea; 
      screenarea["NAME"] = std::string("screenarea");
      screenarea["section"] = 2;
      screenarea["x"] = r.upperLeft().x;
      screenarea["y"] = r.upperLeft().y;
      screenarea["width"] = r.width();
      screenarea["height"] = r.height(); 

      debugN(3, cerr << "Console: releasing screenarea."<<endl;);
      Gui::instance()->release_screenarea(screenarea);

      delete panel_;
      panel_ = 0;
    }  
}

static DII::Object
parser(const std::string& input)
{

  DII::Object::MapType result;

  // cout << "input: "<< "\""<<input <<"\""<< endl;
  
  std::queue<std::string> tokens;
  
  tokenise(input, tokens);
  
  //get first token (= command ) and drop it from the queue
  result["cmd"] = tokens.front();
  tokens.pop();
  
  DII::Object::ListType param;
  while(!tokens.empty())
    {
      param.push_back(tokens.front());
      tokens.pop();
    }

  result["param"] = param;
  
  return result;
}



bool
Console::init(DII::Module*)
{
  
  DII::Object out;

  //  bool success;

  DII::Object::MapType in;
  in["NAME"] = std::string("screenarea");
  in["section"] = 2; 

  out = Gui::instance()->request_screenarea(in);
 
  if(!doesContain(out,"section"))
     {
      debugN(3,cerr << "Console: failed to obtain screenarea."<<endl;);
      return false;
    }
  else
    {
      debugN(3,cerr << "Console: got granted screenarea."<<endl;);
    }

  //discard old input and use old output as new input
  in = out.AsMap();
  
  //actually bind the screenarea
  out = Gui::instance()->bind_screenarea(in);

  if(!doesContain(out,"section"))
    {
      debugN(3,cerr << "Console: failed to bind screenarea."<<endl;);
      return false;
    }
  else
    {
       debugN(3,cerr << "Console: bound to requested screenarea."<<endl;);
    }
  
  int x = out.AsMap()["x"].AsInt();
  int y = out.AsMap()["y"].AsInt();
  int w = out.AsMap()["width"].AsInt();
  int h = out.AsMap()["height"].AsInt();
	uta::Widget* rootwin = objToPtr<uta::Widget*>(out.AsMap()["parent_widget"]);
    
  assert(rootwin);
  
  panel_ = new Panel(rootwin,x,y,w,h);
  panel_->setAutoDelete(false);
  
  panel_->userinput.connect(slot(this,&Console::slot_userinput));


  //  Gui::instance()->installHandler("userinput", lookupMethod("userinput"));
  Gui::instance()->buttonClicked.connect(SigC::slot(this, &Console::userinput));

  panel_->addMessage("UClient "VERSION " - Console ready.\n");

	assert(panel_);
  debugN(3, cerr << "Console: Ready.\n";);
	
  return true;
}

void
Console::std_out(const std::string &output)
{
  if(panel_) {
    panel_->addMessage(output.c_str());
	}

  if(logWindow_) {
    logWindow_->addText(output.c_str());
	}

  //force immediate screen update
  uapp->rootWindow()->sync();
}  

void
Console::std_err(const std::string &output_in)
{
  std::string output = std::string("Error: ")+output_in;

  cerr << output << flush;

  if(panel_)
    panel_->addMessage(output.c_str());


  if(logWindow_)
    logWindow_->addText(output.c_str());

  //force immediate screen update
  uapp->rootWindow()->sync();
} 

std::string
Console::slot_userinput(const std::string& data)
{
  std::string txt;

//  DII::Module::lock();

  if(data.size()>0)
    {
      if(data.c_str()[0]=='/' && data.size()>1)
        {
          
          DII::Object out = parser(std::string(&(data.c_str()[1])));
        
          std::string cmd = out.AsMap()["cmd"].AsString(); 

          DII::Method service = DII::lookupService(cmd.c_str());

          std::string txt;

          panel_->grabFocus();

          if(service) {
            DII::Object _msg = (*service)(out.AsMap()["param"]);

	    if(_msg.IsString())
	      txt = _msg.AsString();
	    else if(_msg.IsMap()) 
	      {
	        DII::Object msg = _msg.AsMap();
  
	        if(doesContain(msg,"text"))
	          txt = msg.AsMap()["text"].AsString();
	        else if(doesContain(msg,"info"))
	          txt = msg.AsMap()["info"].AsString();
	        else if(doesContain(msg,"message"))
	          txt = msg.AsMap()["message"].AsString();
	      }
	    else
	      return "";
          }
          else
            txt = std::string("unknown command: \"") + cmd + std::string("\"\n");

	  std_out(txt);

          //emit(out);
          
          
        }
      else
        {
          std::string talk;
          if(data.c_str()[0]=='\"')
            talk = std::string(&(data.c_str()[1]));
          else
            talk = data;

          //if the chat command say is available.. go for it !
          Player::instance()->say(talk);
          
        }
    }
 

//  DII::Module::release();

  return txt;
}



bool
Console::userinput(const Atlas::Objects::Operation::RootOperation& op)
{
  const DII::Object::ListType& args = op.GetArgs();

  if(args.size() > 0)
    {  
      DII::Object::MapType arg0 = args.begin()->AsMap();

      if(arg0["what"].AsString()=="clicked")
        {  
          if(arg0["button"].AsString()=="scrollup")
            {  
//               if(panel_)
//                 panel_->history_up();
              showLog();
            }
          else if(arg0["button"].AsString()=="scrolldown")
            {  
//               if(panel_)
//                 panel_->history_down();
              hideLog();
            }
          else if(arg0["button"].AsString()=="side_1")
            {    
              double scale = 
                varconf::Config::inst()->getItem("viewport","scale");
              if(scale <= 0.91)
                {
                  scale += 0.1;
                  varconf::Config::inst()->setItem("viewport","scale",scale);
                  if(Viewport::instance()->hideViewport())
                    Viewport::instance()->showViewport();
                }
                
            }
          else if(arg0["button"].AsString()=="side_2")
            {  
              double scale = 
                varconf::Config::inst()->getItem("viewport","scale");
              if(scale > 0.21)
                {
                  scale -= 0.1;
                  varconf::Config::inst()->setItem("viewport","scale",scale);
                  if(Viewport::instance()->hideViewport())
                    Viewport::instance()->showViewport();
                }
            }   
          else if(arg0["button"].AsString()=="side_4")
            {  
              
              if(varconf::Config::inst()->getItem("world","map_overlay"))
                varconf::Config::inst()->setItem("world","map_overlay",false);
              else
                varconf::Config::inst()->setItem("world","map_overlay",true);
              
            }

        }
    }

  return false;
}



bool
Console::showLog()
{
  if(logWindow_)
    return false;
 
  logWindow_ = new LogWindow(panel_->getText());
  logWindow_->exec();

  delete logWindow_;
  logWindow_ = NULL;
  
  return true; 
}


bool
Console::hideLog()
{
  if(!logWindow_)
    return false;

  logWindow_->accept();

  return true;
}

