//
//  The Worldforge Project
//  Copyright (C) 1998,1999  The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 12/1999

#ifndef _MOD_GUI_H
#define _MOD_GUI_H

#include <sigc++/signal_system.h>

#include <Atlas/Objects/Operation/RootOperation.h>

#include <uta/rect.h>
#include <uta/widget.h>
#include <uta/pushbutton.h>
#include <uta/progressdialog.h>

#include <janus/engine.h>
#include <janus/libuta/init.h>

#include "toolbox/local_slist.h"

#include "dii/object.h"

namespace DII {class Module;}

class GuiCallback : public Janus::Callback
{
  protected:
  Janus::Engine* engine_;

  public:
  GuiCallback(Janus::Engine *engine) : 
    Janus::Callback(),
    engine_(engine)
  {};
  
  virtual ~GuiCallback() {};
  
  virtual void call(const std::string& dialog_id, 
		    const std::string& cmd, 
                    const Janus::Value& val);

};

/** Gui Module.
    In charge of providing visible screenareas to other modules.
    Additionally it provides a bunch of messageboxes.
    The screen is divided into four rectangles as follows:

    --------------------------------------
    |                            |       |
    |                            |       |
    |                            |       |
    |           0                |   1   |
    |                            |       |
    |                            |       |
    |                            |       |
    --------------------------------------
    |                            |       |
    |           2                |   3   |
    |                            |       |
    --------------------------------------

    The paramters have to be an Element* structured like this:

    <MAP>
      <NAME>screenarea</NAME>
      <section>1</section>
      <x>..</x>
      <y>..</y>
      <width>..</width>
      <height>..</height>
    </MAP>

    <section> is the section number following the above figure.
    Using the coordinates the calling module may bind only a part
    of the given screenarea.
*/

class Gui : virtual public SigC::Object
{
 public:
  ///
  Gui();
  ///
  virtual ~Gui() {assert(instance_); instance_ = 0;}

  ///
  bool init(DII::Module*);
  ///
  void deinit();
  ///
  static Gui* instance() {return instance_ ? instance_ : new Gui();}

  /**test if an screenarea is available.
     returns the largest available area which is smaller or equal to 
     the requested area (Element* containing <screenarea>).
  */
  DII::Object request_screenarea(const DII::Object&);
  /**request to actually bind the given area.
     returns the largest available area which is smaller or equal to 
     the requested area (Element* containing <screenarea>).
     The return value contains the actual screenarea and a attribute
     called "parent_widget".
  */
  DII::Object bind_screenarea(const DII::Object&);
  /**release previously bound area.
     returns always NULL.
  */
  bool release_screenarea(const DII::Object&);

  /**popup a message box.
     A message box just contains a message and a "OK" Button
     argument:
     <obj>
         <map>
             <string name="title">optional title</string>
	     <string name="message">message text</string>
	 </map>
     </obj>
     result: <obj></obj>
  */
  void popup_messagebox(const std::string &title, const std::string &message);
  
  /**popup a question box.
     A message box just contains a message and a "YES" and "NO" Button
     argument:
     <obj>
         <map>
             <string name="title">optional title</string>
	     <string name="question">question text</string>
	 </map>
     </obj>
     result: 
     <obj>
         <string>yes|no</string>
     </obj>
  */
  bool Gui::popup_questionbox(const std::string &title,
	const std::string &question);

  /** Create a new dialog object.
      The created dialog is not yet visible. The argument
      is a nested map structure, which just gets translated to 
      a according Janus::Node tree.
      @returns string id of created dialog for further use
  */
  std::string create_dialog(Janus::Node&);
  /** Execute dialog.
      This brings dialog to the front and enters the dialog's eventloop.
      All other widgets are disabled.
      Argument is a string returned from create_dialog.
  */
  Janus::Values run_dialog(const std::string&);
  /** Destroy a dialog.
      All dialogs get destroyed automatically, so this needs not
      to get called. 
  */
  void destroy_dialog(const std::string&);
  /** Show the dialog widget.
      If the created dialog is not modal, 
      use this to just show the dialog on the screen.
  */
  void show_dialog(const std::string&);
  /** Hide the dialog widget.
   */
  void hide_dialog(const std::string&);
  /** Change attributes and/or value of dialog element.
      Argument is expected to be a Map containing:
      "id","sub_id" and optional "attributes" (Map)
      and/or "val" (String)
  */
  void change_dialog(const std::string &id, const std::string &sub_id,
	const Janus::Values *attributes = 0, const std::string *value = 0);
  /** Call dialog method.
      Argument is expected to be a Map containing:
      "id","sub_id","cmd" and "arg" (all type String))
  */
  void call_dialog(const std::string &id, const std::string &sub_id,
	const std::string &cmd, const std::string &arg);
  ///returns string id
  std::string open_progressbar(const std::string&);  
  ///arg is list: id, value
  bool update_progressbar(const std::string&, double);  
  ///arg is string: id
  bool close_progressbar(const std::string&);  
     
  ///
  void buttonClick(int);

  /**returns ID
   */
  unsigned int bind_menubutton();
  /**param is valid button ID
   */
  bool release_menubutton(unsigned int);
  /**param is valid button ID
   */
  bool change_menubutton(const DII::Object&);

  SigC::Signal1<bool, const Atlas::Objects::Operation::RootOperation&,
	SigC::FixedMarshal<bool,false> > buttonClicked;

 protected:
  
  ///
  enum { SCROLLDOWN_BUTTON, SCROLLUP_BUTTON, 
	 SIDE_BUTTON1, SIDE_BUTTON2,
	 SIDE_BUTTON3, SIDE_BUTTON4, MENU_FIRSTBUTTON };
  
  ///
  void create();
 
  typedef LOCAL_SLIST(uta::Rect) UtaRectSList;
  ///
  std::vector<UtaRectSList> available_;
  ///
  std::vector<UtaRectSList> bound_;
  ///
  struct MenuEntry {
    uta::PushButton* button;
    bool used;
  };

  ///
  std::vector<MenuEntry> menu_;

  ///
  uta::PushButton* createMenuButton(int n);

  ///
  uta::Widget* rootwin;

  ///
  Janus::Engine engine_;
  ///
  GuiCallback* callback_;
  ///
  std::map<std::string, uta::ProgressDialog*> progressBars_;

  static Gui *instance_;
};

#endif
