//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 02/2000

#include "viewport.h"

#include <sigc++/convert.h>
#include "dii/method_convert.h"

#include "dii/module.h"
#include "operation.h"
#include "settings.h"
#include <varconf/Config.h>

#include "gui.h"

using namespace DII;


#include <uta/widget.h>
#include <uta/directsurface.h>
#include <uta/painter.h>
#include <uta/debug.h>
#include <uta/resources.h>

using namespace uta;

class ViewportWidget : public Widget
{
protected:
  Point current_pos;
  Point press_pos;
  Point release_pos;

  bool draw_rect;

public:
  ViewportWidget(Widget* parent, int x, int y, int w, int h):
    Widget(parent, x,y,w,h),
    draw_rect(false)
  {
      delete surface_;
      surface_ = new DirectSurface(x,y,w,h);
      surface_->clear();
  };

  DirectSurface* surface() { return (DirectSurface*)surface_; };

  virtual ~ViewportWidget() { };

  /* override these methods from Widget to do 
   * nothing as this widget is just meant
   * as a container for the DirectSurface
   */
  void update() {};
  void blit() {};
  void create() {};

  SigC::Signal1<void, const uta::Rect&> selected;

  SigC::Signal1<void, const uta::Point&> clicked;

  SigC::Signal1<void, const uta::Point&> rightclicked;
     
  SigC::Signal1<void, const uta::Point&> moved;

  virtual bool processEvent(const Event* event)
  {
    if(!Widget::mouseFocus_)
      return false;

    if(event->type() == Event::MOUSE)
      {
        moved.emit(((MouseEvent*)(event))->position());
	return true;
      }

    if(event->type() == Event::BUTTON)
      {
	ButtonEvent* bev = (ButtonEvent*)(event);
	switch(bev->state())
	  {
	  case ButtonEvent::PRESSED:
	    
	    press_pos = bev->position();
	    release_pos = press_pos;
	    draw_rect = true;
	    break;

	  case ButtonEvent::RELEASED:
	    release_pos = bev->position();  

	    if(!(press_pos == release_pos))
	      {
		//the user has selected an area
		selected.emit(uta::Rect(press_pos, release_pos));
	      }
	    else
	      {
		//the user "just" clicked the mouse at his position
		if(bev->buttons()==ButtonEvent::BUTTON_LEFT)
		  clicked.emit(release_pos);
		else if(bev->buttons()==ButtonEvent::BUTTON_RIGHT)
		  rightclicked.emit(release_pos);

	      }
	    draw_rect = false;
	    break;
	  }
	return true;
      }
    return false;
  };
};  

Viewport* Viewport::instance_ = 0;

void
Viewport::deinit()
{
  hideViewport();
}

bool
Viewport::hideViewport()
{
  if(!widget_)
    return false;

  Rect r = widget_->getRect();

  DII::Object::MapType screenarea;

  screenarea["NAME"] = std::string("screenarea");
  screenarea["section"] = 0;
  screenarea["x"] = r.upperLeft().x;
  screenarea["y"] = r.upperLeft().y;
  screenarea["width"] = r.width();
  screenarea["height"] = r.height(); 
  
  Gui* gui = Gui::instance();
  assert(gui);
  
  DII::Object out;
  
  debugN(13,cerr << "Viewport: releasing screenarea."<<endl;);
  out = gui->release_screenarea(screenarea);

  if(background_)
    delete background_;
  
  if(widget_)
    delete widget_;

  widget_ = NULL;
  background_ = NULL;

  signalExposeHide.emit(false, 0);

  return true;
}

static bool
wrap_expose_hide(SigC::Callback1<bool,const DII::Object&> *s, bool expose,
  uta::DirectSurface *surface)
{ 
  Operation op = Operation::Instantiate();
  op.SetParents(DII::Object::ListType(1,std::string("info")));

  op.SetFrom(std::string("Viewport"));
  DII::Object::ListType args;
  args.push_back(expose ? std::string("expose") : std::string("hidden"));
  args.push_back(ptrToObj<uta::DirectSurface*>(surface));
  op.SetArgs(args);

  return s->call(op.AsObject());
}

static bool
wrap_clicked(SigC::Callback1<bool,const DII::Object&> *s, int button,
  const uta::Point &p)
{
  Operation op = Operation::Instantiate();
  op.SetParents(DII::Object::ListType(1,std::string("userinput")));

  op.SetFrom(std::string("Viewport"));
  DII::Object::ListType args;
  args.push_back(button ? std::string("rightclicked") : std::string("clicked"));
  args.push_back(pointToObj(p));
  op.SetArgs(args);

  return s->call(op.AsObject());
}

static bool
wrap_moved(SigC::Callback1<bool,const DII::Object&> *s, const uta::Point &p)
{
  Operation op = Operation::Instantiate();
  op.SetParents(DII::Object::ListType(1,std::string("userinput")));

  op.SetFrom(std::string("Viewport"));
  DII::Object::ListType args;
  args.push_back(std::string("mouse"));
  args.push_back(pointToObj(p));
  op.SetArgs(args);

  return s->call(op.AsObject());
}

static bool
wrap_selected(SigC::Callback1<bool,const DII::Object&> *s, const uta::Rect &r)
{
  Operation op = Operation::Instantiate();
  op.SetParents(DII::Object::ListType(1,std::string("userinput")));

  op.SetFrom(std::string("Viewport"));
  DII::Object::ListType args;
  args.push_back(std::string("selected"));
  args.push_back(rectToObj(r));
  op.SetArgs(args);

  return s->call(op.AsObject());
}

bool
Viewport::init(DII::Module *module)
{
  signalExposeHide.connect(SigC::convert(module->signal.slot(), wrap_expose_hide));
  signalClicked.connect(SigC::convert(module->signal.slot(), wrap_clicked));
  signalMoved.connect(SigC::convert(module->signal.slot(), wrap_moved));
  signalSelected.connect(SigC::convert(module->signal.slot(), wrap_selected));
 
  return showViewport();
}

bool 
Viewport::showViewport()
{
  if(widget_)
    return false;

  DII::Object out;
  DII::Object::MapType in;
  in["NAME"] = std::string("screenarea");
  in["section"] = 0; 

  out = Gui::instance()->request_screenarea(in);

  if(!out.IsMap() || out.AsMap().find("section") == out.AsMap().end())
    {
      debugN(13,cerr << "Viewport: failed to obtain screenarea."<<endl;);
      return 0;
    }
  else
    {
      debugN(13,cerr << "Viewport: got granted screenarea."<<endl;);
    }

  //discard old input and use old output as new input
  in = out.AsMap();
  
  //actually bind the screenarea
  assert(Gui::instance());
  out = Gui::instance()->bind_screenarea(in);
 
  if(!out.IsMap() || out.AsMap().find("section") == out.AsMap().end())
    {
      debugN(13,cerr << "Viewport: failed to bind screenarea."<<endl;);
      return false;
    }
  else
    {
      debugN(13,cerr << "Viewport: bound to requested screenarea."<<endl;);
    }
  
  /* FIXME: dangerous !!!!!!!!!!!!! */
  int x = out.AsMap()["x"].AsInt();
  int y = out.AsMap()["y"].AsInt();
  int w = out.AsMap()["width"].AsInt();
  int h = out.AsMap()["height"].AsInt();
  
  /* hmmm... background_ and vw are actual at the same level in the
     widget screen hierachy and overlap. Usually in such a case it is 
     unpredictable which widget is visible.
     Since however vw uses a DirectSurface it will never be 
     hidden by background_ 
  */

  /* FIXME: dangerous !!!!!!!!!!!!! */
  Widget* rootwin = objToPtr<Widget*>(out.AsMap()["parent_widget"]);
  assert(rootwin);
  ViewportWidget* vw;

  //  double scale = readState("viewport_scale").asFloat();

  double scale = varconf::Config::inst()->getItem("viewport","scale");

  if(scale > 0.2 && scale <= 1.000)
    {
      background_ = new Widget(rootwin,x,y,w,h);
      background_->setBackground(SURFACE_RES("background"));
      background_->setAutoDelete(false);
      //scale the viewport and center it    
      int w_old = w;
      int h_old = h;
      w = (int)(w * scale);
      h = (int)(h * scale);
      x += (w_old - w )/2;
      y += (h_old - h )/2;  
    }
  
  vw = new ViewportWidget(rootwin,x,y,w,h);
  vw->setAutoDelete(false);

  widget_ = vw;
   
  vw->selected.connect(slot(this, &Viewport::slot_selected));
  vw->clicked.connect(slot(this, &Viewport::slot_clicked));
  vw->rightclicked.connect(slot(this, &Viewport::slot_rightclicked));
  vw->moved.connect(SigC::slot(this, &Viewport::slot_moved));

  signalExposeHide.emit(true, ((ViewportWidget*)(widget_))->surface());

  return true;
}

void
Viewport::slot_clicked(const uta::Point& p)
{
  //cerr << "Viewport click @ " << p << endl;

  signalClicked.emit(0, p);
}

void
Viewport::slot_rightclicked(const uta::Point& p)
{
  //cerr << "Viewport rightclick @ " << p << endl;

  signalClicked.emit(1, p);
}

void
Viewport::slot_moved(const uta::Point& p)
{
// DEBUG
//cerr << "Viewport slot_moved to " << p << endl;

  signalMoved.emit(p);
}

void
Viewport::slot_selected(const uta::Rect& r)
{
  //cerr << "Viewport selected area: " << r << endl;

  signalSelected.emit(r);
}

uta::DirectSurface*
Viewport::getScreen()
{
  return widget_ ?((ViewportWidget*)(widget_))->surface() : 0;
}
