//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 01/2000

#include "connection.h"
#include "ftos.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <uta/application.h>  // for sleep

extern "C" {
    #include <stdio.h>
    #include <stdlib.h>
    #include <unistd.h>
}

#if defined(__WIN32__) || defined(WIN32)
  #define Win32_Winsock
  #include <windows.h>



#else /* UNIX */   
  extern "C" {
    #include <sys/time.h>
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <netinet/in.h>
    #include <netdb.h>  
  }   
#endif /* WIN32 */

/* directly include Atlas internal Codec header as
 * the static template class factory is not initialized when linking 
 * statically.
 */
// removed to test 0.3.11 #include "XML.h"

#include <skstream/skstream.h>
  
   
Connection::Connection(Atlas::Bridge* bridge):
  netStream_(NULL),
  codec_(NULL),
  encoder_(NULL),
  state_(NOT_CONNECTED),
  decoder_(bridge),
  lag_(0.3),
  timeoutSec_(0),
  timeoutUsec_(0)
{

}

Connection::~Connection()
{
  disconnect();
  if(netStream_)
    delete netStream_;
}


void 
Connection::setTimeout(unsigned sec, unsigned usec)
{
  timeoutSec_ = sec;
  timeoutUsec_ = usec;
  if(netStream_)
    netStream_->setTimeout(timeoutSec_, timeoutUsec_);
}

void
Connection::poll()
{
  int socket_error = 0;

  if(netStream_)
    {
      /* it is important to check if there is input data pending !
       * codec_->Poll() blocks until new data has arrived, and we do 
       * not want this in a single threaded app.
       */
	    
      socket_error = netStream_->getLastError();
      if(socket_error != 0)
	{
#ifdef DEBUG
	  std::cerr << "Pre-Poll SOCKET ERROR :" << strerror(socket_error) << std::endl;
#endif
	}

      SOCKET_TYPE cli_fd = netStream_->getSocket();

      if(cli_fd == INVALID_SOCKET)
        return;

      fd_set infds;
      struct timeval tv;
      
      FD_ZERO(&infds);
      
      FD_SET(cli_fd, &infds);
 
      tv.tv_sec = 0;
      tv.tv_usec = 10;
      
      int retval = select(cli_fd+1, &infds, NULL, NULL, &tv);
      
      if (retval) 
	{
	  if (FD_ISSET(cli_fd, &infds)) 
	    {
	      if (netStream_->peek() == -1) 
		{
#ifdef DEBUG
		  std::cerr << "Connection: got socket error !" << std::endl
		       << "probably the server has disconnected ! " << std::endl;
#endif
		  // tidy up:
		  disconnect();
		  
		}
	      else
		{
		  if(codec_)
		    codec_->Poll();
		}
	    }  
	}

      // We may have lost it already ! So we _need_ to check again !
      if(netStream_)
      {
	socket_error = netStream_->getLastError();
	if(socket_error != 0)
	  {
#ifdef DEBUG
	    std::cerr << "Post-Poll SOCKET ERROR :" << strerror(socket_error) << std::endl;
#endif
	  }
      }
    }
}

void 
Connection::connect(const std::string& servname, unsigned int portnum)
{
  if(connected())
    return;

  //open the network stream
  state_ = CONNECTING;

  if(netStream_)
    delete netStream_;

  // create new socket stream object
  netStream_ = new tcp_socket_stream;

  assert(netStream_);

  // set timeout
  netStream_->setTimeout(timeoutSec_, timeoutUsec_);
  // try to open connection to remote machine, timeout after 3s
#ifdef HAVE_NONBLOCKING_CONNECT
  netStream_->open(servname, portnum, true);
  for(int i = 0; i < 60; ++i) {
    // The uta mainloop timestep is 50ms, use a shorter length
    // so that the mainloop can catch up, even if it's running slow
    if(netStream_->is_ready(40))
      break;
    uapp->keepAlive();
  }
  if(!netStream_->is_ready())
    netStream_->close();
#else
  netStream_->open(servname, portnum);
#endif
  
  // successfull ?
  if(!netStream_->is_open())
    {
#ifdef DEBUG
      std::cerr << "Connection: Failed to connect to " << servname << ":"<<portnum 
		  << std::endl;
#endif
      state_= NOT_CONNECTED;
      return;
    }
  peerName_ = netStream_->getRemoteHost();

  //negotiating ...
  state_ = NEGOTIATING;

  //cerr << "negotiating protocol with server ... " << std::endl;
  Atlas::Net::StreamConnect c("uclient-"VERSION, *netStream_, decoder_);

  unsigned int timeout = 200;  // this corresponds to 5 seconds 
  while (c.GetState() == Atlas::Negotiate<std::iostream>::IN_PROGRESS && timeout) 
    {
      c.Poll();
      timeout--;
      uta::Application::sleep(25);
    }      

  if(!timeout)
    {
#ifdef DEBUG
      std::cerr << "Connection: Negotiating timed out ! " << std::endl;
#endif
      return;
    }

  if(c.GetState() != Atlas::Negotiate<std::iostream>::SUCCEEDED)
    {
#ifdef DEBUG
      std::cerr << "Connection: Failed to negotiate protocol with server!"<<std::endl;
#endif
      return ;
    }

  //std::cerr << "succesfully negotiated protocol with server !" << std::endl;

  codec_ = c.GetCodec();
  encoder_ = new Atlas::Objects::Encoder(codec_);

  // This should always be sent at the beginning of a session
  codec_->StreamBegin();
  
  state_ = CONNECTED;
  
  //std::cerr << "successfully established stream connection to server :)" << std::endl;
}

void
Connection::disconnect()
{
  if(connected())
    {
      // std::cerr << "Connection::disconnect()" << std::endl;
      codec_->StreamEnd();
      if(netStream_)
        *netStream_ << std::flush;

      delete netStream_;
      netStream_ = NULL;

      // ... not sure about this:
      delete encoder_;
      encoder_ = NULL;
      delete codec_;
      codec_ = NULL;

      state_ = NOT_CONNECTED;

      disconnected.emit();
    }
}


void
Connection::send(const Atlas::Objects::Root* obj)
{
  if(encoder_)
    {
      encoder_->StreamMessage(obj);
      if(netStream_)
	*netStream_ << std::flush;
    }
}

std::string
Connection::getPeerName() const
{
  return peerName_;
}



