#include "wf_net.h"
#include <skstream/skstream.h>
#include "metaserver_protocol.h"
#include "connection.h"
#include "msgdispatch.h"
#include "ftos.h"
#include "dii/object.h"

#include <iostream>
#include <uta/application.h>

#ifdef DEBUG
#define INFO(prg) { prg }
#else
#define INFO(prg)
#endif

using namespace std;

string WFNetwork::metaServer = "metaserver.worldforge.org";
int WFNetwork::metaPort = WFNetwork::METASERVER_PORT;

#ifndef uint32_t
/* WIN32 hack ...
   this is only true for 32bit machines but WIN64 is far ahead !! */
typedef unsigned int uint32_t;
#endif

char *pack_uint32(uint32_t data, char *buffer, unsigned int *size)
{
  uint32_t netorder;
 
  netorder = htonl(data);
  memcpy(buffer, &netorder, sizeof(uint32_t));
  *size += sizeof(uint32_t);
  return buffer+sizeof(uint32_t);
}
 
char *unpack_uint32(uint32_t *dest, char *buffer)
{
  uint32_t netorder;
 
  memcpy(&netorder, buffer, sizeof(uint32_t));
  *dest = ntohl(netorder);
  return buffer+sizeof(uint32_t);
} 


void 
WFNetwork::useMetaServer(const string& hostname, 
			 int port)
{
  metaServer = hostname;
  metaPort = port;
}

#define MAXLINE 1024

list<string>
WFNetwork::queryGameServers()
{
  list<string> result;
  char           mesg[MAXLINE];
  char          *mesg_ptr;
  unsigned int   packet_size;
  string data;

  // create UDP stream
  udp_socket_stream stream;  
  
  INFO(std::cerr << "WFNetwork: trying to send packets to metaserver at "
    << metaServer << std::endl;)

  stream.setTimeout(20);

  // open connection to meta server
  stream.setOutpeer(metaServer, metaPort);

  if(!stream.is_open())
    {
      std::cerr << "WFNetwork: unable to etstablish route to metaserver at " 
	   <<metaServer<< ":"<<metaPort<< std::endl;
      return result;
    }    
  
  INFO(cerr << "WFNetwork: established route to " << stream.getRemoteHost() << endl;)

  uint32_t handshake=0, command=0, total, packed;   
  uint32_t address;
  
  packet_size = 0;
  mesg_ptr = pack_uint32(CKEEP_ALIVE, mesg, &packet_size);

  data = string(mesg, packet_size);
  stream << data << flush;

  // Only really need to check in_avail() on first read,
  // generally we're either connected or not
  
  // 20 seconds
  for(int timeout = 400; true /* Break out elsewhere */; --timeout) {
    SOCKET_TYPE fd = stream.getSocket();
    fd_set fds;
    FD_ZERO(&fds);
    FD_SET(fd, &fds);
    // 40ms, slightly less than libuta's 50ms mainloop tick
    struct timeval select_timeout = {0, 40000};
    int select_val = select(fd+1, &fds, 0, 0, &select_timeout);
    if(select_val == 1)
      break;
    if(select_val < 0 || timeout <= 0) {
      INFO(cerr << "WFNetwork: metaserver did not respond ! " << endl;);
      return result;
    }
    uapp->keepAlive();
  }
  
  mesg_ptr = mesg;
  stream.get(*mesg_ptr++);
  stream.get(*mesg_ptr++);
  stream.get(*mesg_ptr++);
  stream.get(*mesg_ptr++);

  unpack_uint32(&command, mesg);
 
  if(command == HANDSHAKE)
    {   
      mesg_ptr = mesg;

      stream.get(*mesg_ptr++);
      stream.get(*mesg_ptr++);
      stream.get(*mesg_ptr++);
      stream.get(*mesg_ptr++);
      unpack_uint32(&handshake, mesg);
      INFO(cerr << "WFNetwork: got reply from metaserver !" << endl;)
      
      packet_size = 0;
      mesg_ptr = pack_uint32(CLIENTSHAKE, mesg, &packet_size);
      mesg_ptr = pack_uint32(handshake, mesg_ptr, &packet_size);
      data = string(mesg, packet_size);
      stream << data << flush;
    }
  else
    {
      INFO(cerr << "WFNetwork: metaserver did not respond ! " << endl;);
      return result;
    }

  packet_size = 0;
  mesg_ptr = pack_uint32(LIST_REQ, mesg, &packet_size);
  mesg_ptr = pack_uint32(0, mesg_ptr, &packet_size);

  data = string(mesg, packet_size);
  stream << data << flush;
 
  mesg_ptr = mesg;

  stream.get(*mesg_ptr++);
  stream.get(*mesg_ptr++);
  stream.get(*mesg_ptr++);
  stream.get(*mesg_ptr++);
  
  unpack_uint32(&command, mesg);

  
  if(command == PROTO_ERANGE)
    {
      INFO(cerr << "WFNetwork: metaserver said range error. " << endl;);
      return result;
    }
  else if (command == LIST_RESP)
    {
      INFO(cerr << "WFNetwork: received server list packet from metaserver." << endl;)
      mesg_ptr = mesg;

      stream.get(*mesg_ptr++);
      stream.get(*mesg_ptr++);
      stream.get(*mesg_ptr++);
      stream.get(*mesg_ptr++);;
      unpack_uint32(&total, mesg);
      mesg_ptr = mesg;

      stream.get(*mesg_ptr++);
      stream.get(*mesg_ptr++);
      stream.get(*mesg_ptr++);
      stream.get(*mesg_ptr++);
      unpack_uint32(&packed, mesg);

      string hostname;

      INFO(cerr << "WFNetwork: received " <<packed<<" servers of "<<total<<"." << endl;)
      for(unsigned long n = packed; n ; n --)
	{      
	  mesg_ptr = mesg;

	  stream.get(*mesg_ptr++);
	  stream.get(*mesg_ptr++);
	  stream.get(*mesg_ptr++);
	  stream.get(*mesg_ptr++);
	  unpack_uint32(&address, mesg);
	  hostname = 
	    itos((address & 0x000000FF)      )+"."+ 
	    itos((address & 0x0000FF00) >> 8)+"."+
	    itos((address & 0x00FF0000) >> 16)+"."+
	    itos((address & 0xFF000000) >> 24);
	  INFO(cerr << "      running server at " << hostname << endl << flush;)
	  result.push_back(hostname);
	}
    }
  return result;
}


class InfoDecoder : virtual public SigC::Object
{
public:
  InfoDecoder() : pending_(false) {};
  virtual ~InfoDecoder() {};
  
  void decodeInfoOp(const Atlas::Objects::Operation::Info& info)
  {
    INFO( cerr << "WFNetwork: An info operation has arrived." << endl;);
    try 
      { 
	msg_ = info.GetArgs().front().AsMap();  // yeah !
	pending_ = true;
      }
    catch(...)
      {
        cerr << "WFNetwork FAILURE:  decoding of info operation caused exception !" << endl;
      } 
  }

  bool pending() const { return pending_; };

  const DII::Object& decodedMsg() { pending_ = false; return msg_; };

protected:
  bool pending_;
  DII::Object msg_;

};

DII::Object
WFNetwork::queryServer(const string& hostname, int port)
{
  /* set up local decoding system */
  InfoDecoder decoder;
  MsgDispatch dispatcher;
  dispatcher.OpInfoArrived.connect(SigC::slot(&decoder,&InfoDecoder::decodeInfoOp));

  /* create connection object using this newly created system */
  Connection connection(&dispatcher);
  
  /* set hard timeout (10 seconds) ... maybe the host has gone done already ... */
  connection.setTimeout(10);

  /* connect to remote host */
  connection.connect(hostname, port);
  if(!connection.connected() )
    return false; 
   
  /* send a Get operation */
  INFO(cerr << "WFNetwork: querying server "<<hostname<<" for game information ..." << flush;);
  Atlas::Objects::Operation::Get g = 
    Atlas::Objects::Operation::Get::Instantiate();
  connection.send(&g);

  int ping = 0;
  unsigned timeout = 100;  // this corresponds to 2,5 seconds 

  while(!decoder.pending() && timeout)
    {
      connection.poll();
      timeout --;  
      ping += uta::Application::sleep(25);
    }
  
  DII::Object result = false;

  if(timeout != 0)
    { 
      result = decoder.decodedMsg();
      if(!result.IsMap())
	{
	  cerr << "WFNetwork: server replied to query (Get) with a non-Map object !" 
	       << endl
	       << "           I do not know how to handle this !" << endl;
	  result = false;
	}
      else
	{
	  result.AsMap()["ping"] = ping;

	  // print out received data when debugging is on
	  INFO(  DII::Object::MapType::const_iterator itr = 
		 result.AsMap().begin();
		 string txt = "ok.\n";
		 while(itr != result.AsMap().end())
                {
		  if(itr->second.IsString())
		    txt += itr->first+" : "+itr->second.AsString()+"\n";
		  else if(itr->second.IsInt())
		    txt += itr->first+" : "+itos(itr->second.AsInt())+"\n";
		  else if(itr->second.IsFloat())
		    txt += itr->first+" : "+ftos(itr->second.AsFloat())+"\n";
		  itr++;
		}
	     
	     cerr << txt << endl;
		 );  // INFO (
       }
    }
  else
    {
      INFO(cerr << " timed out ! " << endl;);
      
    }
  
  return result;
}
