// The Worldforge Project
// Copyright (C) 1998, 1999 The Worldforge Project
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// 
// For information about Worldforge and its authors, please contact
// the Worldforge Web Site at http://www.worldforge.org.
// 
// created by Karsten Laux 01/2000

#include "player.h"

#include <sigc++/convert.h>

#include "operation.h"
#include "settings.h"
#include "isodefs.h"
#include "ftos.h"
#include "media/sprite.h"
#include "media/sprite_cache.h"
#include "loader/loader.h"

#include <varconf/Config.h>

#include <uta/application.h>
#include <uta/directsurface.h>
#include <uta/font.h>
#include <uta/resources.h>
#include <uta/debug.h>
#include <uta/mouse.h>

#include "dii/serialize.h"
#include "dii/module.h"
#include "misc.h"
#include "network/wf_net.h"

#include "../renderer/renderer.h"
#include "../terrain/terrain.h"
#include "../modules/inventory.h"
#include "../modules/console.h"

#include "pointer.xpm"

// #include <math.h>

#include <Atlas/Objects/Entity/GameEntity.h>

using namespace uta;
using namespace DII;
using namespace Atlas;

#define EMIT_MESSAGE(event)\
DII::Object::MapType msg;\
msg["from"]="Player";\
msg["event"]=event;\
msg["args"]=DII::Object::ListType();\
msgSignal.emit(msg);

Player* Player::instance_ = 0;

Player::Player():
connection_(NULL),
timer_(100),
slowTimer_(1000),
draggedIcon_(uta::Mouse::MAXX, uta::Mouse::MAXY, uta::Pixelformat::BGR888),
state_(INIT), error_(0)
{
    assert(!instance_);
    instance_ = this;

    // install hotkey filter
    uta::Application::keystroke.connect(slot(this, &Player::keyfilter));

    timer_.alarm.connect(slot(this, &Player::tick));
    slowTimer_.alarm.connect(slot(this, &Player::slowTick));

    // connect signals of dispatcher to this object
    dispatcher_.OpSoundArrived.connect(SigC::slot(this, &Player::OpSound));
    dispatcher_.OpSightArrived.connect(SigC::slot(this, &Player::OpSight));
    dispatcher_.OpDisappearanceArrived.
        connect(SigC::slot(this, &Player::OpDisappearance));
    dispatcher_.OpAppearanceArrived.
        connect(SigC::slot(this, &Player::OpAppearance));

    dispatcher_.OpInfoArrived.connect(SigC::slot(this, &Player::OpInfo));
    dispatcher_.OpErrorArrived.connect(SigC::slot(this, &Player::OpError));

}



Player::~Player()
{
    assert(instance_);
    instance_ = 0;

    // ensure that draggedIcon_ is no longer the mousepointer !
    uta::Application::mouse()->setPointer(NULL);
}

void Player::deinit()
{
    if (connection_) {
        delete connection_;
        connection_ = NULL;
    }
}

static bool
op_convert(SigC::Callback1<bool, const DII::Object&> *s,
	   const Atlas::Objects::Operation::RootOperation &o)
{
  return s->call(o.AsObject());
}

bool Player::init(DII::Module *module)
{
  msgSignal.connect(module->signal.slot());
  opSignal.connect(SigC::convert(module->signal.slot(), op_convert));

    // if(varconf::Config::inst()->getItem("flags","standalone"))
    // {
    // DII::callService("editCharacter");
    // }
    // else if(DII::callService("login") == DII::SUCCESS)
    // {
    // if(DII::callService("chooseCharacter") != DII::SUCCESS)
    // {
    // /* if player did not choose existing character
    // * create new one ..
    // */
    // DII::callService("editCharacter");
    // }
    // }

    /*
     * load mapfile (sometime in a future version the server will tell us 
     * what map to load ...  but not yet ... so read from settings file
     * what map to load) 
     */
    // string mapfile = varconf::Config::inst()->getItem("map","name");
    // string mapfile = "agrilan";
    // if(!mapfile.empty())
    // if(bool(varconf::Config::inst()->getItem("map","delayed_load")) ==
    // false)
    // if(Terrain::instance())
    //   Terrain::instance()->readMapFile(mapfile);

    return true;
}

void
 Player::configure()
{
    // nothing here yet 
}

void
 Player::tick()
{
    if (connection_)
        connection_->poll();
}

void
 Player::slowTick()
{
    try {
        if (!currentCharacter())
            return;

        static unsigned int
         itemCount =
            0;
        // need a better way to access inventory !
        std::list<std::string> items = currentCharacter()->contains();
        if (items.size() != itemCount) {
            debugN(11, cerr << "Player: updating inventory..." << endl;);
            itemCount = items.size();
            Inventory::instance()->update(items);
        }
    }
    catch(...) {
        cerr << "unhandled exception in Player::slowTick() ! " << endl;
    }
}

bool
Player::keyfilter(const uta::KeyEvent * key)
{
    // only accept key press events
    if (key->state() != uta::KeyEvent::PRESSED)
        return false;

    // only if there is already a character
    if (!currentCharacter())
        return false;

    float
     dv = 1.0;
    float
     dphi = 3.1415 / 4.0;

    bool handled = false;

    // we swallow all known keys, so no one else
    // may react upon these
    handled = true;
    Atlas::Message::Object::MapType obj;
    obj["id"] = currentCharacter()->id();
    obj["loc"] = currentCharacter()->ref();

    switch (key->key()) {
    case uta::KEY_UP:
    case uta::KEY_KP8:
        currentCharacter()->accelerate(dv);
        obj["velocity"] = coordToObj(currentCharacter()->velocity());
        handled = true;
        break;
    case uta::KEY_DOWN:
    case uta::KEY_KP2:
        currentCharacter()->accelerate(-dv);
        obj["velocity"] = coordToObj(currentCharacter()->velocity());
        handled = true;
        break;
    case uta::KEY_LEFT:
    case uta::KEY_KP4:
        currentCharacter()->turn(-dphi);
        obj["orientation"] = quaternionToObj(currentCharacter()->orientQuat());
        obj["velocity"] = coordToObj(currentCharacter()->velocity());
        handled = true;
        break;
    case uta::KEY_RIGHT:
    case uta::KEY_KP6:
        currentCharacter()->turn(dphi);
        obj["orientation"] = quaternionToObj(currentCharacter()->orientQuat());
        obj["velocity"] = coordToObj(currentCharacter()->velocity());
        handled = true;
        break;
    case uta::KEY_END:
    case uta::KEY_KP5:
        currentCharacter()->halt();
        obj["velocity"] = coordToObj(currentCharacter()->velocity());
        handled = true;
        break;
    case uta::KEY_HOME:
        // currentCharacter()->centerViewport();
        break;
    default:
        handled = false;        // unknown key ... let others handle it
        break;
    }

    if (handled) {

        Atlas::Objects::Operation::Move moveOp =
            Atlas::Objects::Operation::Move::Instantiate();
        moveOp.SetFrom(currentCharacter()->id());
        moveOp.SetArgs(Atlas::Message::Object::ListType(1, obj));

        if (connection_) {
            // moveOp.SetFutureSeconds(connection_->getLag());
            connection_->send(&moveOp);
        }

        if (varconf::Config::inst()->getItem("flags", "standalone")) {
            Atlas::Objects::Operation::Sight sightOp =
                Atlas::Objects::Operation::Sight::Instantiate();
            sightOp.
                SetArgs(Atlas::Message::Object::
                        ListType(1, moveOp.AsObject()));

            World::World::inst()->processSightOp(sightOp);
        }
    }

    return handled;
}

/// maybe we want to control more than one later ...
World::Entity* 
Player::currentCharacter() 
{ 
    if(characterID_.empty())
        return NULL;
    return World::World::inst()->lookupEntity(characterID_); 
}

DII::Object Player::getTime()
{
    return World::World::inst()->getTime();
}


DII::Object Player::queryMemory(const DII::Object & obj)
{
    std::string id;

    if (obj.IsString())
        id = obj.AsString();
    else if (obj.IsList() && obj.AsList().size() > 0)
        if (obj.AsList().begin()->IsString())
            id = obj.AsList().begin()->AsString();

    return World::World::inst()->queryObject(id);
}

bool Player::sendOperation(const DII::Object & arg)
{
    if (!connection_)
        return false;

    if (!arg.IsList() || arg.AsList().size() != 2)
        return false;

    DII::Object::ListType::const_iterator itr = arg.AsList().begin();
    std::string op_name = itr->AsString();
    itr++;
    const
     DII::Object::MapType &
        attr =
        itr->
        AsMap();

    Atlas::Objects::Operation::RootOperation op;
    if (op_name == "Set")
        op = Atlas::Objects::Operation::Set::Instantiate();
    else if (op_name == "Move")
        op = Atlas::Objects::Operation::Move::Instantiate();
    else if (op_name == "Touch")
        op = Atlas::Objects::Operation::Touch::Instantiate();
    else if (op_name == "Divide")
        op = Atlas::Objects::Operation::Divide::Instantiate();
    else if (op_name == "Combine")
        op = Atlas::Objects::Operation::Combine::Instantiate();
    else if (op_name == "Get")
        op = Atlas::Objects::Operation::Get::Instantiate();
    else if (op_name == "Create")
        op = Atlas::Objects::Operation::Create::Instantiate();
    else if (op_name == "Delete")
        op = Atlas::Objects::Operation::Delete::Instantiate();
    else if (op_name == "Look")
        op = Atlas::Objects::Operation::Look::Instantiate();
    else if (op_name == "Imaginary")
        op = Atlas::Objects::Operation::Imaginary::Instantiate();
    else {
        debugN(11,
               cerr << "Player::sendOperation() unknown operation :" <<
               op_name << endl;);
        return false;
    }

    DII::Object::MapType::const_iterator attrItr = attr.begin();
    while (attrItr != attr.end()) {
        op.SetAttr(attrItr->first, attrItr->second);
        attrItr++;
    }

    connection_->send(&op);

    return true;
}


std::string Player::currentCharacterID()
{
    if (currentCharacter())
        return currentCharacter()->id();
    else
        return std::string("");
}

bool
Player::connect(const std::string & host, unsigned int port)
{
    if (state_ != INIT)
        return false;

    if (!connection_) {
        connection_ = new Connection(&dispatcher_);
        connection_->disconnected.
            connect(slot(this, &Player::disconnected));
    }

    connection_->connect(host, port);

    if (!connection_->connected()) {
        delete connection_;
        connection_ = NULL;
        return false;
    }

    EMIT_MESSAGE("connected");

    state_ = CONNECTED;
    return true;
}

std::list<std::string> Player::queryServerList()
{
    // check if an alternate metaserver address is set:
    std::string metaserver;
    int
     metaserver_port =
        WFNetwork::METASERVER_PORT;

    // this checks for player::metaserver and configures WFNetwork
    // appropriately 
    if(DII::GetOption("Player", "metaserver", metaserver)) {
      DII::GetOption("Player", "metaserver_port", metaserver_port);
      WFNetwork::useMetaServer(metaserver, metaserver_port);
    }

    /*
     * get the list of running WF servers 
     */
    return WFNetwork::queryGameServers();
}

DII::Object Player::queryServer(const std::string& hostname, int port)
{
    return WFNetwork::queryServer(hostname, port);
}

bool Player::login(const std::string &_id, const std::string &_password,
		   const std::string &_hostname)
{
    if (!connection_ || !connection_->connected())
        if (!connect(_hostname, 6767)) {
            if(Console::instance())
	      Console::instance()->std_out(std::string("Player.cc: Unable to connect to ") +
                             _hostname + std::string(":6767\n"));
            return false;
        }

    if(Console::instance())
      Console::instance()->std_out(std::string("Player: connected to ") +
                     connection_->getPeerName() + "\n");

    Atlas::Objects::Entity::Account account =
        Atlas::Objects::Entity::Account::Instantiate();
    account.SetAttr("username", _id);
    account.SetPassword(_password);

    Atlas::Objects::Operation::Login l =
        Atlas::Objects::Operation::Login::Instantiate();
    Atlas::Message::Object::ListType args(1, account.AsObject());
    l.SetArgs(args);

    if(Console::instance())
      Console::instance()->std_out(std::string("trying to log in using existing account ..."));

    connection_->send(&l);
    error_ = NO_ERR;

    while (state_ == CONNECTED && error_ == NO_ERR)
        uta::Application::keepAlive();

    if (state_ == LOBBY) {
        if(Console::instance())
	  Console::instance()->std_out(std::string("  successfully logged in\n"));
        writeState("server_logged_in", 1);

        // An anonymous get to query server data
        serverGet();
	// An anonymous look to get server-state
	playerLook();

        EMIT_MESSAGE("loggedin");

        return true;
    }

    if(Console::instance())
      Console::instance()->std_out(std::string("failed to login.\n"
                                 "Please check if your password and login are correct.\n"));

    return false;

}

bool Player::logout()
{
    if (connection_ && connection_->connected()) {
        if(Console::instance())
	  Console::instance()->std_out( std::string("logging out ... \n"));
        Atlas::Objects::Operation::Logout l =
            Atlas::Objects::Operation::Logout::Instantiate();
        l.SetId(accountID_);

        connection_->send(&l);

        unsigned
         timeout =
            20;
        while (state_ != INIT && timeout--)
            uta::Application::keepAlive();

        if (state_ != INIT) {
            if(Console::instance())
	      Console::instance()->std_out(std::string("no reply from server ... disconnecting."));
            connection_->disconnect();
        }

        return true;
    }

    return false;
}

bool Player::createAccount(const std::string &_id, const std::string &_password,
		   	   const std::string &_hostname)
{
    if (!connection_ || !connection_->connected())
        if (!connect(_hostname, 6767)) {
            if(Console::instance())
	      Console::instance()->std_out(std::string("Unable to connect to ") +
                             _hostname + std::string(":6767\n"));
            return false;
        }

    if(Console::instance())
      Console::instance()->std_out(std::string("connected to ") +
                     connection_->getPeerName() + "\n");

    Atlas::Objects::Entity::Account account =
        Atlas::Objects::Entity::Account::Instantiate();
    account.SetAttr("username", _id);
    account.SetPassword(_password);

    Atlas::Message::Object::ListType args(1, account.AsObject());

    Atlas::Objects::Operation::Create c =
        Atlas::Objects::Operation::Create::Instantiate();

    c.SetArgs(args);

    connection_->send(&c);

    error_ = NO_ERR;
    while (state_ == CONNECTED && error_ == NO_ERR)
        uta::Application::keepAlive();

    if (state_ == LOBBY) {
        if(Console::instance())
	  Console::instance()->std_out(std::string("\nsuccessfully logged in (ID = ") +
                         accountID_ + ")\n");
        writeState("server_logged_in", 1);

        // An anonymous get to query server data
        serverGet();
        // This is an anonymous look.
        playerLook();

        EMIT_MESSAGE("loggedin");

        return true;
    }

    if(Console::instance())
      Console::instance()->std_out(std::string("unable to create Account.\n"
                                 "This account name is already in use.\n"));

    return false;

}

bool Player::takeOverCharacter(const std::string& id)
{
    Atlas::Objects::Operation::Look l =
        Atlas::Objects::Operation::Look::Instantiate();

    l.SetFrom(id);
    Atlas::Message::Object::MapType what;
    what["id"] = id;
    Atlas::Message::Object::ListType args(1, what);
    l.SetArgs(args);

    if (!connection_)
        return false;

    connection_->send(&l);

    error_ = NO_ERR;
    while (state_ == LOBBY && error_ == NO_ERR) {
        uta::Application::keepAlive();
    }

    if (state_ != INGAME) {
	if(Console::instance())
          Console::instance()->std_err(std::string("failed to take control of character !\n"));
        return false;
    }
    // we have got it !
    characterID_ = id;

    writeState("server_have_character", 1);

    // tell World to observ this entity
    World::World::inst()->observ(characterID_);

    // tell Renderer to always keep this entity inside the viewport
    if(Renderer::instance())
      Renderer::instance()->observ(characterID_);

    if(Console::instance())
      Console::instance()->std_out(std::string("successfully taken over character ") +
                     id + "\n");

    // initial look around
    look("");

    EMIT_MESSAGE("gotcharacter");

    return true;
}

bool Player::createCharacter()
{
    bool standalone =
        varconf::Config::inst()->getItem("flags", "standalone");
    DII::Object::MapType character_ = readState("character").AsMap();

    Atlas::Objects::Entity::GameEntity character =
        Objects::Entity::GameEntity::Instantiate();
    Atlas::Objects::Operation::Create c =
        Objects::Operation::Create::Instantiate();

    Atlas::Message::Object::MapType::iterator itr = character_.begin();
    while (itr != character_.end()) {
        character.SetAttr(itr->first, itr->second);
        itr++;
    }

    if (standalone) {
        character_["pos"] =
            coordToObj(Coord(-2 + 4 * drand48(), -2 + 4 * drand48(), 0));
        character_["loc"] = std::string("");
        character_["id"] = std::string("playerID");

        if (World::World::inst()->createEntity("playerID") == "playerID") {
            state_ = INGAME;
        } else {
            error_ = UNKNOWN_ERR;
            cerr << "Player: Standalone world refuses to create my character ! " <<
                endl;
        }
    } else {

        if (!connection_ || !connection_->connected()) {
            Console::instance()->std_err(std::string(
		"You first must login to a game server !\n"));
            return false;
        }

        Atlas::Message::Object::ListType args(1, character.AsObject());
        c.SetArgs(args);
        c.SetFrom(accountID_);

        connection_->send(&c);

        error_ = NO_ERR;
        while (state_ == LOBBY && error_ == NO_ERR) {
            uta::Application::keepAlive();
        }

        if (state_ != INGAME) {
	    Console::instance()->std_err(std::string("failed to create character !\n"));
            return false;
        }
    }

    // tell World to observ this entity
    World::World::inst()->observ(characterID_);

    // tell Renderer to always keep this entity inside the viewport
    Renderer::instance()->observ(characterID_);

    Console::instance()->std_out(std::string("successfully created character ! (ID = ") +
                     characterID_ + ")\n");

    writeState("server_have_character", 1);

    look("");

    EMIT_MESSAGE("gotcharacter");

    return true;
}



bool Player::selectEntity(const std::string& id)
{
    if (!currentCharacter())
        return false;

    std::string myID = currentCharacter()->id();

    // get state of modifier keys, because we react different
    // on mouseclicks depending on CTRL,SHIFT and ALT
    bool shiftPressed = (SDL_GetModState() & KMOD_SHIFT) != 0;
    bool ctrlPressed = (SDL_GetModState() & KMOD_CTRL) != 0;

    debugN(11, cerr << "Player::selectEntity(" << id << ")" << endl;);

    if (shiftPressed) {

        // get infos about this object
        DII::Object ent = World::World::inst()->queryObject(id);

        if (!(ent == DII::FAILURE)) {
            std::string name, desc;

            name = ent.AsMap()["name"].AsString();
            desc = ent.AsMap()["description"].AsString();

            if (doesContain(ent, "mass"))
                desc +=
                    ", mass=" + ftos(ent.AsMap()["mass"].AsFloat());
            if (doesContain(ent, "status"))
                desc +=
                    ", status=" + ftos(ent.AsMap()["status"].AsFloat());


            std::string info = name + ", " + desc + ".";


            std::string parent, _id, pos;
            // ent.get("parent",parent);
            _id = ent.AsMap()["id"].AsString();
            pos = ent.AsMap()["position"].AsString();

            debug(info += "[id=" + _id + " ,pos=" + pos + "]";);

            if(Console::instance())
	      Console::instance()->std_out( info + "\n");
        }
    } else if (ctrlPressed) {
        if (id != myID) {
            // take it !
            DII::Object::MapType obj;
            obj["loc"] = myID;
            obj["pos"] = coordToObj(Coord(0, 0, 0));
            obj["velocity"] = coordToObj(Coord(0, 0, 0));
            obj["id"] = id;

            Atlas::Objects::Operation::Move moveOp =
                Atlas::Objects::Operation::Move::Instantiate();
            moveOp.SetFrom(myID);
            moveOp.SetArgs(DII::Object::ListType(1, obj));

            if (connection_)
                connection_->send(&moveOp);
        }
    } else {
        if (id != myID) {
            // touch it !
            Atlas::Objects::Operation::Touch touchOp =
                Atlas::Objects::Operation::Touch::Instantiate();
            touchOp.SetFrom(myID);
            touchOp.SetTo(id);
            Atlas::Message::Object::MapType ent;
            ent["id"] = id;
            touchOp.SetArgs(Atlas::Message::Object::ListType(1, ent));

            if (connection_)
                connection_->send(&touchOp);

            // give immediate feedback
            World::Entity * entity =
                World::World::inst()->lookupEntity(id);
            if (entity)
                entity->touch(touchOp.AsObject());

	    say(std::string("I am hitting ") + id);
        }
    }


    return true;
}

bool Player::moveTo(const DII::Object & pos)
{
    if (!pos.IsList()) {
        cerr << "Player::moveTo only accepts a single coord ! " << endl;
        return false;
    }
    if (!currentCharacter())
        return false;

    DII::Object::MapType obj;
    obj["loc"] = currentCharacter()->ref();
    obj["pos"] = pos;
    obj["velocity"] = coordToObj(Coord(1, 0, 0));
    obj["id"] = currentCharacter()->id();

    Atlas::Objects::Operation::Move moveOp =
        Atlas::Objects::Operation::Move::Instantiate();
    moveOp.SetFrom(currentCharacter()->id());
    moveOp.SetArgs(DII::Object::ListType(1, obj));

    if (connection_)
        connection_->send(&moveOp);

    return true;
}

bool Player::rightclick(const DII::Object & obj)
{
    if (!obj.IsMap())
        return false;
    if (!currentCharacter())
        return false;

    try {
        // if(obj.get("entity",id))
        if (doesContain(obj, "entity")) {
            // right-clicked at entity ... show info about it:
            std::string id = obj.AsMap().find("entity")->second.AsString();
            // get infos about this object
            DII::Object ent = World::World::inst()->queryObject(id);

            if (!(ent == DII::FAILURE)) {
                std::string name, desc;

                name = ent.AsMap()["name"].AsString();
                desc = ent.AsMap()["description"].AsString();

                if (doesContain(ent, "mass"))
                    desc +=
                        ", mass=" +
                        ftos(ent.AsMap()["mass"].AsFloat());
                if (doesContain(ent, "status"))
                    desc +=
                        ", status=" +
                        ftos(ent.AsMap()["status"].AsFloat());

                std::string info = name + ", " + desc + ".";

                std::string parent, _id, pos;
                // ent.get("parent",parent);
                _id = ent.AsMap()["id"].AsString();
                pos = ent.AsMap()["position"].AsString();

                debug(info += "[id=" + _id + " ,pos=" + pos + "]";);

                if(Console::instance())
		  Console::instance()->std_out( info + "\n");
            }
        } else if (doesContain(obj, "coords")) {
            // no entity pointed at .. move to this position
            moveTo(obj.AsMap().find("coords")->second);
        }
    }
    catch(...) {
        debugN(11,
               cerr << "Player::rightclick() threw exception ! " << endl;);
        return false;
    }

    return true;
}

bool Player::leftclick(const DII::Object & obj)
{
    if (!obj.IsMap())
        return false;
    if (!currentCharacter())
        return false;

    DII::Object coords;
    DII::Object offset;
    std::string id;

    // if(obj.get("entity",id))
    if (doesContain(obj, "entity")) {
        id = obj.AsMap().find("entity")->second.AsString();

        if (selectedIDs_.empty()) {
            selectEntity(id);
        } else {
            if (doesContain(obj, "offset"))
                offset = obj.AsMap().find("offset")->second;

            std::list < std::string >::iterator itr;
            for (itr = selectedIDs_.begin(); itr != selectedIDs_.end();
                 itr++) {
                DII::Object::ListType param;
                param.push_back(*itr);
                param.push_back(id);
                param.push_back(offset);
                transfer(param);
            }
            selectedIDs_.clear();
            uta::Application::mouse()->setPointer(NULL);
        }
    } else if (doesContain(obj, "coords") && !selectedIDs_.empty()) {
        coords = obj.AsMap().find("coords")->second;

        DII::Object::MapType obj;
        obj["loc"] = currentCharacter()->ref();
        obj["pos"] = coords;
        obj["velocity"] = coordToObj(Coord(0, 0, 0));

        Atlas::Objects::Operation::Move moveOp =
            Atlas::Objects::Operation::Move::Instantiate();
        moveOp.SetFrom(currentCharacter()->id());

        std::list < std::string >::iterator itr;
        for (itr = selectedIDs_.begin(); itr != selectedIDs_.end(); itr++) {
            obj["id"] = *itr;
            moveOp.SetArgs(DII::Object::ListType(1, obj));

            if (connection_)
                connection_->send(&moveOp);
        }

        selectedIDs_.clear();
        uta::Application::mouse()->setPointer(NULL);
    }

    return true;
}


// transfer <what> <to> [<offset>]
bool Player::transfer(const DII::Object & obj)
{
    if (!obj.IsList() || obj.AsList().size() < 2)
        return false;
    if (!currentCharacter())
        return false;

    std::string to, target;
    DII::Object offset = coordToObj(Coord(0, 0, 0));

    DII::Object::ListType::const_iterator itr = obj.AsList().begin();
    target = itr->AsString();
    itr++;
    to = itr->AsString();
    itr++;

    if (itr != obj.AsList().end())
        offset = *itr;

    DII::Object::MapType what;
    what["loc"] = to;
    what["pos"] = offset;
    what["velocity"] = coordToObj(Coord(0, 0, 0));
    what["id"] = target;

    Atlas::Objects::Operation::Move moveOp =
        Atlas::Objects::Operation::Move::Instantiate();
    moveOp.SetFrom(currentCharacter()->id());
    moveOp.SetArgs(DII::Object::ListType(1, what));

    if (connection_)
        connection_->send(&moveOp);

    return true;
}

bool Player::drag(const DII::Object & obj)
{
    if (!obj.IsList())
        return false;
    if (!currentCharacter())
        return false;
    // if either given list is empty or if the first id matches
    // unselect the stuff
    if (obj.AsList().empty() ||
        (!selectedIDs_.empty() &&
         selectedIDs_.front() == obj.AsList().front().AsString())) {
        selectedIDs_.clear();
        uta::Application::mouse()->setPointer(NULL);
        return true;
    }
    // remember the selected ids
    DII::Object::ListType::const_iterator itr;
    selectedIDs_.clear();
    for (itr = obj.AsList().begin(); itr != obj.AsList().end(); itr++)
        selectedIDs_.push_back(itr->AsString());


    // figure out what has been clicked at
    World::Entity * entity =
        World::World::inst()->lookupEntity(selectedIDs_.front());
    assert(entity);

    // request an image for this
    DII::Object::MapType req;
    DII::Object result;
    int
     gid;
    try {

        SpriteDesc desc;
        desc.type = entity->parent();
        gid = SpriteCache::inst()->request(desc);
        SpriteCache::inst()->lock(gid);
        Sprite *
            spritePtr =
            SpriteCache::inst()->
            lookup(gid);
        const
         uta::Surface *
            icon =
            spritePtr->
            getLayer(0, 0, 0);

        // and copy it to the draggedIcon_ surface
        draggedIcon_.readFromXPM(pointer_xpm);
        draggedIcon_.convert(uta::Pixelformat::BGR888);
        // draggedIcon_.fill(black);
        centerBlit(icon, &draggedIcon_);

        SpriteCache::inst()->release(gid);

        uta::Painter * painter = new uta::Painter(&draggedIcon_);
        uta::Color col = painter->getPixel(uta::Point(0, 0));
        delete painter;

        draggedIcon_.setTransColor(col);
        draggedIcon_.setTransparency(true);

        uta::Application::mouse()->setPointer(&draggedIcon_);
    }
    catch(...) {
        cerr << "unhandled exception in Player::drag(..) " << endl;
    }

    return true;
}

bool Player::drop(const DII::Object & obj)
{
    std::string target;
    unsigned
     amount =
        0;
    if (!currentCharacter())
        return false;

    if (obj.IsList() && obj.AsList().size() > 0) {
        amount = obj.AsList().size();
        target = obj.AsList().front().AsString();
    } else
        target = obj.AsString();

    if (target.empty())
        return false;

    Atlas::Objects::Operation::Move moveOp =
        Atlas::Objects::Operation::Move::Instantiate();
    moveOp.SetFrom(currentCharacter()->id());

    if (amount == 0) {
        DII::Object::MapType what;
        what["loc"] = currentCharacter()->ref();
        what["pos"] =
            coordToObj(currentCharacter()->position() +
                       Coord(-2 + drand48(), -2 + drand48(), 0));
        what["velocity"] = coordToObj(Coord(0, 0, 0));

        what["id"] = target;
        moveOp.SetArgs(DII::Object::ListType(1, what));

        if (connection_)
            connection_->send(&moveOp);
    } else {
        DII::Object::ListType::const_iterator itr;
        for (itr = obj.AsList().begin(); itr != obj.AsList().end(); itr++) {
            DII::Object::MapType what;
            what["loc"] = currentCharacter()->ref();
            what["pos"] =
                coordToObj(currentCharacter()->position() +
                           Coord(-2 + drand48(), -2 + drand48(), 0));
            what["velocity"] = coordToObj(Coord(0, 0, 0));

            what["id"] = itr->AsString();
            moveOp.SetArgs(DII::Object::ListType(1, what));

            if (connection_)
                connection_->send(&moveOp);
        }
    }
    return true;
}

bool Player::inventory()
{
    if (!currentCharacter())
        return false;

    DII::Object me = currentCharacter()->asObject();
    DII::Object::ListType contains;
    std::string info = "You are carrying ";

    try {
        contains = me.AsMap().find("contains")->second.AsList();

        World::Entity * entity;
        for (DII::Object::ListType::iterator itr = contains.begin();
             itr != contains.end(); itr++) {
            entity = World::World::inst()->lookupEntity(itr->AsString());
            assert(entity);

            info += entity->id() + " (" + entity->name() + ") ";
        }
    }
    catch(...) {
        info += "nothing !";
    }
    if(Console::instance())
      Console::instance()->std_out( info + "\n");

    return true;
}

void Player::serverGet()
{
    Objects::Operation::Get g = Objects::Operation::Get::Instantiate();

    if (connection_)
        connection_->send(&g);
}

void Player::playerLook(const std::string &id)
{
    Objects::Operation::Look l = Objects::Operation::Look::Instantiate();

    if (!id.empty()) {
        Atlas::Message::Object::MapType what;
        what["id"] = id;
        Atlas::Message::Object::ListType args(1, what);
        l.SetArgs(args);
        l.SetTo(id);
    }

    l.SetFrom(accountID_);

    if (connection_)
        connection_->send(&l);
}

void Player::action(const std::string &what)
{
    Objects::Operation::Action a =
        Objects::Operation::Action::Instantiate();

    debugN(11, cerr << "Character performs action: " << what << endl;);
    Atlas::Message::Object::MapType arg;
    arg["action"] = what;
    Atlas::Message::Object::ListType args(1, arg);
    a.SetArgs(args);

    a.SetFrom(characterID_);

    /*
     * HACK !!!!!!!!!!! local loopback for testing .... 
     */
    Objects::Operation::Sight sight =
        Objects::Operation::Sight::Instantiate();
    sight.SetArgs(Atlas::Message::Object::ListType(1, a.AsObject()));
    World::World::inst()->processSightOp(sight);

    if (connection_)
        connection_->send(&a);
}

void Player::look(const std::string &id)
{
    Objects::Operation::Look l = Objects::Operation::Look::Instantiate();

    if (!id.empty()) {
        debugN(11, cerr << "Character looks at " << id << endl;);
        Atlas::Message::Object::MapType what;
        what["id"] = id;
        Atlas::Message::Object::ListType args(1, what);
        l.SetArgs(args);
    } else {
        debugN(11, cerr << "Character looks around " << endl;);
    }

    l.SetFrom(characterID_);

    if (connection_)
        connection_->send(&l);
}

void
 Player::OpError(const Atlas::Objects::Operation::Error & err)
{
    error_ = UNKNOWN_ERR;
    debugN(11, cerr << "An error operation has arrived." << endl;);

    try {
        // emit signal containing this message
        opSignal.emit(err);

        if(Console::instance())
	  Console::instance()->std_out( std::string("<Server> FAILURE: "));

        Atlas::Message::Object::ListType args = err.GetArgs();
        if (args.size() < 1) {
            cerr << "Player::OpError() WARNING: args is empty !!" << endl;
            return;
        }
        Atlas::Message::Object & arg = args.front();
        if (arg.IsString()) {
            if(Console::instance())
	      Console::instance()->std_out( arg.AsString() + "\n");
        }
        else if (doesContain(arg, "message")) {
            if(Console::instance())
	      Console::instance()->std_out(arg.AsMap()["message"].AsString() + "\n");
	}
        else if(Console::instance())
	  Console::instance()->std_out(std::string("(unknown server error)\n"));

        debugN(11, {
               cerr << "Player::OpError() error was caused by: " << endl;
               if (args.size() > 1)
               DII::Serialize::write(args[1], cerr);}
        );

        // check for special cases:
        if (state_ == CONNECTED)
            error_ = NOACCOUNT_ERR;
    }
    catch(...) {
        cerr <<
            "    FAILURE: processing of error operation caused exception !"
            << endl;
    }
}

void
 Player::OpInfo(const Atlas::Objects::Operation::Info & info)
{
    debugN(11, cerr << "An info operation has arrived." << endl;);

    const
     Atlas::Message::Object::ListType &
        args =
        info.
        GetArgs();

    try {
        // emit signal containing this message
        opSignal.emit(info);

        if (args.size() < 1) {
            cerr << "Player::OpInfo() WARNING: args is empty !!" << endl;
            return;
        }
        if (state_ == INGAME) {
            return;
        } else if (state_ == CONNECTED) {
            const
             Atlas::Message::Object &
                account =
                args.
                front();
            writeState("account", account);

            Atlas::Objects::Entity::Account obj =
                Atlas::Objects::Entity::Account::Instantiate();

            Atlas::Message::Object::MapType::const_iterator I;
            for (I = account.AsMap().begin(); I != account.AsMap().end();
                 I++)
                obj.SetAttr(I->first, I->second);

            accountID_ = obj.GetId();

            DII::Object::MapType msg;
            msg["from"] = "Player";
            msg["event"] = "StateChange";
            msg["state"] = "Lobby";
            msgSignal.emit(msg);

            state_ = LOBBY;
        } else if (state_ == LOBBY) {
            if (args.front().IsMap()) {
                const Atlas::Message::Object::MapType & character =
                    args.front().AsMap();

                std::string argparent;
                Atlas::Message::Object::MapType::const_iterator I = character.find("parents");
                if (I != character.end())
                  {
                    argparent = I->second.AsList().front().AsString();
                  }
                if (argparent == "server")
                  {
                    I = character.find("ruleset");
                    if (I != character.end() && I->second.IsString())
                      {
                        DII::Loader::loadScript(I->second.AsString());
                      }
                  }
                else
                  {
                    writeState("character", character);

                    characterID_ = character.find("id")->second.AsString();
                    // cerr << "got char [" << characterID_ << "]" << endl <<
                    // flush;
    
                    if (World::World::inst()->createEntity(character) ==
                        characterID_)
                      {
                        state_ = INGAME;

                        DII::Object::MapType msg;
                        msg["from"] = "Player";
                        msg["event"] = "StateChange";
                        msg["state"] = "InGame";
                        msgSignal.emit(msg);

                      }
                    else
                      {
                        error_ = UNKNOWN_ERR;
                        cerr <<
                            "Player: World refuses to create my character ! "
                            << endl;
                      }
                  }
            }
        }
    }
    catch(...) {
        cerr <<
            "    FAILURE: processing of info operation caused exception :"
            << endl;
        DII::Serialize::write(info.AsObject(), cerr);
    }

}


void
 Player::OpSight(const Atlas::Objects::Operation::Sight & sight)
{
    debugN(11, cerr << "A sight operation has arrived." << endl;);


    try {
        // emit signal containing this message
        opSignal.emit(sight);

        if (sight.GetTo() == accountID_)
            lobby_.processSightOp(sight);
        else
            World::World::inst()->processSightOp(sight);
    }
    catch(...) {
        cerr <<
            "    FAILURE: processing of sight operation caused exception :"
            << endl;
        DII::Serialize::write(sight.AsObject(), cerr);
    }
}

void
 Player::OpSound(const Atlas::Objects::Operation::Sound & sound)
{
    debugN(11, cerr << "A sound operation has arrived." << endl;)

        try {
        // emit signal containing this message
        opSignal.emit(sound);

        if (sound.GetTo() == accountID_)
            lobby_.processSoundOp(sound);
        else
            World::World::inst()->processSoundOp(sound);
    }
    catch(...) {
        cerr <<
            "    FAILURE: processing of sound operation caused exception :"
            << endl;
        DII::Serialize::write(sound.AsObject(), cerr);
    }
}

void
 Player::OpAppearance(const Atlas::Objects::Operation::Appearance & o)
{
    debugN(11, cerr << "A appearance operation has arrived." << endl;);
    try {
        // emit signal containing this message
        opSignal.emit(o);

        if (o.GetTo() == accountID_)
            lobby_.processAppearanceOp(o);
        else
            World::World::inst()->processAppearanceOp(o);
    }
    catch(...) {
        cerr <<
            "    FAILURE: processing of appereance operation caused exception :"
            << endl;
        DII::Serialize::write(o.AsObject(), cerr);
    }
}

void
 Player::OpDisappearance(const Atlas::Objects::Operation::
                         Disappearance & o)
{
    debugN(11, cerr << "A disappearance operation has arrived." << endl;);
    try {
        // emit signal containing this message
        opSignal.emit(o);

        if (o.GetTo() == accountID_)
            lobby_.processDisappearanceOp(o);
        else
            World::World::inst()->processDisappearanceOp(o);
    }
    catch(...) {
        cerr <<
            "    FAILURE: processing of disappereance operation caused exception :"
            << endl;
        DII::Serialize::write(o.AsObject(), cerr);
    }
}

void Player::say(const std::string &txt)
{
    Objects::Operation::Talk t = Objects::Operation::Talk::Instantiate();

    Atlas::Message::Object::MapType what;
    what["say"] = txt;
    Atlas::Message::Object::ListType args(1, what);

    t.SetArgs(args);

    if (state_ == INIT) {
	Console::instance()->std_out(std::string(">> ") + txt + std::string("\n"));
    } else if (state_ == INGAME) {
        t.SetFrom(characterID_);
        connection_->send(&t);
    } else if (state_ == LOBBY) {
        t.SetFrom(accountID_);
        connection_->send(&t);
    }
}

void
 Player::disconnected()
{
    state_ = INIT;
    writeState("server_logged_in", 0);
    writeState("server_have_character", 0);

    EMIT_MESSAGE("disconnected");

    if(Console::instance())
      Console::instance()->std_out(std::string("\n- lost connection to server ! -\n"));
}
