//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.1
//
// created by Karsten Laux 01/2000

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "renderer.h"

#include <sigc++/convert.h>
#include "dii/method_convert.h"

#include "../environment/environment.h"
#include "../modules/viewport.h"
#include "../player/player.h"
#include "../terrain/terrain.h"

#include "dii/module.h"

#include <varconf/Config.h>

#include <uta/directsurface.h>
#include <uta/font.h>
#include <uta/resources.h>
#include <uta/painter.h>
#include <uta/debug.h>
#ifndef HAVE_UTA_LIGHTEN_DARKEN
#include <uta/blitters.h>
#endif

#include "clouds.xpm"
// #include "shadow.xpm"

using namespace uta;

//size of scrolling trigger border
#define SCROLL_BORDER_J 100
#define SCROLL_BORDER_I 100

//how many pixels out of sight are graced until an entity
//gets dropped
#define GRACE 800

//radius around a mouse click where entities are accepted
#define SEARCH_DELTA 400

/* if defined the Renderer will remove all objects which are outside
   the viewport
*/
#undef FLUSH_OUTOFSIGHT

/* Renderer keeps a hash of all objects, sorted as shown below.
 * 
 *          Z  
 *          .           
 * Y(-j)    .        X(-i)
 *    .     .     .
 *       .  .  .  
 *          0
*/      
#define RESOLUTION 1.0
#define HASH_COORD(f) (int)(-f / RESOLUTION + 0.45)

#include <cmath>
#ifndef M_PI
#define M_PI 3.14159265358979323844
#endif

uta::Surface* renderShadow(const Coord& bbox, const Coord& bmedian, int w)
{

  int h = static_cast<int>((bmedian.x + bmedian.y + bbox.x + bbox.y ) * TILEHEIGHT);

  // avoid too big shadows
  if(w > 200)
    return NULL;
  if(bbox.x > 2 || bbox.y > 2)
    return NULL;

  // workaround for missing bmedian data ...
  if(fabs(abs(bmedian)) < 0.1)
    h = static_cast<int>(h * 1.5);

  if(h < 2)
    return NULL;

  uta::Surface* shadow = new uta::Surface(w,h,uta::Pixelformat::RGBA8888);
  uta::Painter painter(shadow);

  uta::Point p;
  uta::Point center((int)((bmedian.x - bmedian.y)*TILEWIDTH/2 + w/2),(int)(h/2));

  int max_alpha = 128;
  uta::Color c(32,32,32,max_alpha);

  int r = static_cast<int>(((bbox.x > bbox.y) ? bbox.x : bbox.y) * TILEWIDTH);

  int steps = r * 3;
 
//   cerr << "renderer::renderShadow: bmedian="<<bmedian<<", r="<<r<<", w="<<w<<", h="<<h<<endl;

  double al = 0.0;
  double rx, ry;

  for(int n=0; n<r; n++)
    {
      rx = n*bbox.x*TILEWIDTH / r;
      ry = n*bbox.y*TILEHEIGHT / r;
      c.a = max_alpha - (n*max_alpha)/r;
      for(int w=0; w < steps; w++)
        {
          al = static_cast<double>(w)*2*M_PI/static_cast<double>(steps);
          p.x = center.x + static_cast<int>(rx*cos(al));
          p.y = center.y + static_cast<int>(ry*sin(al));
          painter.setPixel(p,c);
        }
    }

  return shadow;
}

Renderer* Renderer::instance_ = 0;

Renderer::Renderer():
  viewport_(0),
  entities_(NULL),
  viewHeight_(0),
  screen_(NULL),
  buffer_(NULL),
  gamma_(0),
  initialized_(false),
  scroll_border_i(SCROLL_BORDER_I),
  scroll_border_j(SCROLL_BORDER_J),
  grace(GRACE),
  smoothScroll_(true),
  smallMap_(NULL),
  clouds_(NULL),
  use_clouds_(false),
  idleRespawnProb_(0.05),
  cast_shadows_(false),
  daylight_(true)
{
  assert(!instance_);
  instance_ = this;

  viewCenter_.i = 0;
  viewCenter_.j = 0;

  //guessed at this point, assuming 400x200 pixel viewport
  viewRect_.imin = -200;
  viewRect_.imax = 200;
  viewRect_.jmin = -100;
  viewRect_.jmax = 100;

  cloudSpeed_.x = 1.5;
  cloudSpeed_.y = 0.5;

#ifdef DEBUG
  updates_ = 0;
  updateTime_ = 0;
#endif

}

void 
Renderer::configure()
{  
  DII::GetOption("Renderer", "smooth_scrolling",smoothScroll_);

  bool tmp = false;
  if(DII::GetOption("Renderer", "showbounds",tmp))
    flags_.set(SHOWBOUNDS, tmp);
  if(DII::GetOption("Renderer", "highlight",tmp))
    flags_.set(HIGHLIGHT, tmp);
  if(DII::GetOption("Renderer", "showlabels",tmp))
      flags_.set(SHOWLABELS, tmp);
  
  DII::GetOption("Renderer", "use_clouds",use_clouds_);
  DII::GetOption("Renderer", "idle_respawn_prob",idleRespawnProb_);
  
  DII::GetOption("Renderer", "cast_shadows",cast_shadows_);
  DII::GetOption("Renderer", "daylight",daylight_);

}

static bool
convert_clicked(SigC::Callback2<bool,const std::string&,const uta::Point&> *s,
	int button, const uta::Point &p)
{
  return s->call(button ? "rightclicked" : "clicked", p);
}

static bool
convert_moved(SigC::Callback2<bool,const std::string&,const uta::Point&> *s,
	const uta::Point &p)
{
  return s->call("mouse", p);
}

bool
Renderer::init(DII::Module *module)
{
//   shadow_.readFromXPM(shadow_xpm);
//   shadow_.convert(uta::Pixelformat::RGB888);
//   shadow_.setAlpha(40); // libuta sets per surface alpha in percent !

  signal.connect(module->signal.slot());

  if(!clouds_)
    {
      clouds_ = new uta::GammaFunction;
      clouds_->offset_x = 0;
      clouds_->offset_y = 0;
      clouds_->values = new char[256*256];

      uta::Surface temp;
      temp.readFromXPM(clouds_xpm);  // from #include "clouds.xpm"

      uta::Pixelformat newformat = uta::Pixelformat::IND8;
      newformat.palette.resize(32);
        for(int i=0; i<32; i++)
          newformat.palette[i] = uta::Color(i*8, i*8, i*8);
      
      // dither
//       temp.convert(newformat,true);

      temp.lock();
      
      unsigned char* sPtr = (unsigned char*)temp.pixels();
     
      for(unsigned w = 0; w<256; w++)
        for(unsigned h= 0; h<256; h++)
          {
            clouds_->values[w+h*256] = (int)sPtr[w*4 + h*temp.pitch()] - 127;
          }

      temp.unlock();
    }
  
  viewport_ = Viewport::instance();
  if(!viewport_)
    {
      cerr << "Renderer: Error, viewport not found !";
      return false;
    }

//  fpGetScreen = viewport_->lookupMethod("getScreen");
//  if(!fpGetScreen)
//    return false;

  
  //viewport_->installHandler("*", lookupMethod("slot_viewport"));
  viewport_->signalExposeHide.connect(SigC::slot(this, &Renderer::slot_viewport_expose));
  viewport_->signalClicked.connect(SigC::convert(SigC::slot(this,
	&Renderer::slot_viewport_point), convert_clicked));
  viewport_->signalMoved.connect(SigC::convert(SigC::slot(this,
	&Renderer::slot_viewport_point), convert_moved));

  // fpGetScreen is function pointer, we pass a dummy 0, call
//  screen_ = objToPtr<uta::DirectSurface*>((*fpGetScreen)(0));
  screen_ = viewport_->getScreen();

  if(screen_)
    buffer_ = new uta::Surface(screen_->width(), 
                               screen_->height(),
                               uta::Pixelformat::RGB565);

//   if(screen_)
//     buffer_ = new uta::Surface(screen_->width(), 
//                                screen_->height(),
//                                screen_->pixelformat()());
  
//   if(screen_)
//     buffer_ = new uta::Surface(screen_->width(), 
//                                screen_->height(),
//                                uta::Pixelformat::RGBA8888);

  initialized_ = true;  

  //lock viewport
//  viewport_->lock();


  // force sane state ... setting may get overridden by configure()
  DII::SetOption("Renderer", "use_clouds",false);
  flags_.set(HIGHLIGHT, true);

  configure();

  const uta::Surface* nw = SURFACE_RES("corner_nw");
  const uta::Surface* ne = SURFACE_RES("corner_ne");
  const uta::Surface* se = SURFACE_RES("corner_se");
  const uta::Surface* sw = SURFACE_RES("corner_sw");
  
  if(nw && ne && se && sw)
    {
      Overlay overlay1;
      overlay1.age = -1;
      overlay1.type = IMAGE_OVERLAY;
      overlay1.image.image = nw;
      overlay1.x = 0;
      overlay1.y = 0;
      overlays_.push(overlay1);

      Overlay overlay2;
      overlay2.age = -1;
      overlay2.type = IMAGE_OVERLAY;
      overlay2.image.image = ne;
      overlay2.x = -ne->width();
      overlay2.y = 0;
      overlays_.push(overlay2);

      Overlay overlay3;
      overlay3.age = -1;
      overlay3.type = IMAGE_OVERLAY;
      overlay3.image.image = sw;
      overlay3.x = 0;
      overlay3.y = -sw->height();
      overlays_.push(overlay3);

      Overlay overlay4;
      overlay4.age = -1;
      overlay4.type = IMAGE_OVERLAY;
      overlay4.image.image = se;
      overlay4.x = -se->width();
      overlay4.y = -se->height();
      overlays_.push(overlay4);
      
    }
  return true;
}

void
Renderer::deinit()
{
  if(clouds_)
    {
      delete[] clouds_->values;
      delete clouds_;
      clouds_ = NULL;
    }

  if(viewport_)
    {
//      Viewport::instance()->release();
    }

  initialized_ = false;

  clearDisplayList();
  
  DII::SetOption("Renderer", "smooth_scrolling",smoothScroll_);
  DII::SetOption("Renderer", "use_clouds",use_clouds_);
  DII::SetOption("Renderer", "idle_respawn_prob",idleRespawnProb_);
  DII::SetOption("Renderer", "cast_shadows",cast_shadows_);
  DII::SetOption("Renderer", "daylight",daylight_);
}

void 
Renderer::highlight_backend(ScreenEntity *entity, int timeout, RLE::Color color)
{
  if(entity)
    {
      entity->timeout = timeout;
      entity->hightlightColor = color;
    }
}


SpriteID
Renderer::requestSprite(const World::Entity* entity)
{
  SpriteDesc desc;

  desc.mode = entity->mode();
  desc.type = entity->parent();
  
  DII::Object::MapType::const_iterator itr = entity->attributes().begin();
  while(itr != entity->attributes().end())
    {
      //check for special string values (what about "size", "status" ?)
      if(itr->first == "sex")
        {
          if(itr->second.IsString())
            desc.sex = itr->second.AsString();
        }
      else if(itr->second.IsNum())
        {
          //all others are treated as float values..
          //weight, status, ...
          desc.values[itr->first]=itr->second.AsNum();
        }
      itr++;
    }

  //request this sprite from the sprite database
  return SpriteCache::inst()->request(desc);
}


void
Renderer::setMap(const uta::Surface *map)
{
  smallMap_ = map;
} 

void 
Renderer::setGamma(double gamma)
{
  gamma_ = (int)gamma;
} 

void
Renderer::setClouds(int i)
{
  use_clouds_ = (i > 0);
} 

void 
Renderer::clearDisplayList()
{
  debugN(10,cerr << "Renderer::clearDisplayList() ... ";);

//   //delete all DB entries
//   ScreenEntity::iterator itr = entities_.begin();
//   while(itr != entities_.end())
//     {
//       delete *itr;
//       itr++;
//     }
//   entities_.clear();
  
  //clear displaylist
  grid_.clear();

//   scanlines_.clear();
  displayList_.clear();

  debugN(10, cerr << " [OK]."<<endl;);
}

void
Renderer::updateVisibleRange()
{
  //  debugN(10,cerr << "Renderer::updateVisibleRange()"<<endl;);

  int sw;
  int sh;

  if(screen_)
    {
      sw = screen_->width();
      sh = screen_->height();
      scroll_border_i = sw / 5;
      scroll_border_j = sh / 5;
      grace = scroll_border_i * 4;
    }
  else
    {
      //if there is currently no valid screen object
      //calculate width and height from old values:
      sw = viewRect_.imax - viewRect_.imin;
      sh = viewRect_.jmax - viewRect_.jmin;
    }

  int delta_i = viewCenter_.i - sw/2 - viewRect_.imin;
  int delta_j = viewCenter_.j - sh/2 - viewRect_.jmin;
    
  viewRect_.imin = viewCenter_.i - sw/2;
  viewRect_.imax = viewCenter_.i + sw/2;
  viewRect_.jmin = viewCenter_.j - sh/2;
  viewRect_.jmax = viewCenter_.j + sh/2;

  if(clouds_)
    {
      clouds_->offset_x += delta_i;
      clouds_->offset_y += delta_j;
    }

  Terrain::instance()->setVisibleRange(viewRect_);
  Terrain::instance()->setVisibleHeight(viewHeight_);
 
  //  rebuildDisplayList();
}



/* add entities from given grid node to the display list
 * by iterating through already added entities and checking for
 * correct order.
 */
void 
Renderer::addEntityList(EntityList* List)
{
  uta::Rect screen = uta::Rect(uta::Point(viewRect_.imin, viewRect_.jmin),
                               uta::Point(viewRect_.imax, viewRect_.jmax)); 
  EntityList::iterator zItr;
  DisplayList::iterator itr;
  zItr = List->begin();
  while(zItr != List->end())
    {
      if(screen.intersect((*zItr)->rect).isValid())
        {
          itr = displayList_.begin();
          while(itr != displayList_.end() && ScreenEntity::behind_of(*itr, *zItr))
            itr++;
          
          displayList_.insert(itr, *zItr);
          
          // we now need the sprite data to get locked
          (*zItr)->lockSprite();


        }
      zItr++;
    }
}

void
Renderer::rebuildDisplayList()
{
  if(!screen_)
    return;

  // release all sprites before clearing displaylist
  DisplayList::iterator itr2 = displayList_.begin();
  while(itr2 != displayList_.end())
    {
      // release this sprites sprite data
      (*itr2)->releaseSprite();
      itr2++;
    }
  displayList_.clear();

  while(!labels_.empty())
    labels_.pop();
  

  Pixel tmp;
  tmp.i = viewRect_.imax+grace;
  tmp.j = viewRect_.jmin-grace;
  Coord begin_i = p2w(tmp);

  tmp.i = viewRect_.imin-grace;
  tmp.j = viewRect_.jmax+grace;
  Coord end_i = p2w(tmp);

  tmp.i = viewRect_.imin - grace;
  tmp.j = viewRect_.jmin - grace;
  Coord begin_j = p2w(tmp);

  tmp.i = viewRect_.imax + grace;
  tmp.j = viewRect_.jmax + grace;
  Coord end_j = p2w(tmp);

    
  Grid::iterator itr(&grid_, HASH_COORD(begin_i.x), HASH_COORD(begin_j.y));
 
  do
    {      
      itr = Grid::iterator(&grid_, HASH_COORD(begin_i.x), itr.j());
      do
        {
          if(itr.valid())
            addEntityList(reinterpret_cast<EntityList*>(itr.value()));
          itr.next_i();
        }     
      while(itr.valid_i() && itr.i() < HASH_COORD(end_i.x));
      itr = Grid::iterator(&grid_, HASH_COORD(begin_i.x), itr.j());
    
      itr.next_j();
    }
  while(itr.valid_j() && itr.j() < HASH_COORD(end_j.y));
  
}

bool
Renderer::slot_viewport_expose(bool expose, uta::DirectSurface *surface)
{
      screen_ = surface;
      
      if(screen_) 
        {
          if(!buffer_)
            buffer_ = new uta::Surface(screen_->width(), 
                                       screen_->height(),
                                       screen_->pixelformat()());
          else if(screen_->width() != buffer_->width() ||
                   screen_->height() != buffer_->height())
            {
              delete buffer_;
              buffer_ = new uta::Surface(screen_->width(), 
                                         screen_->height(),
                                         screen_->pixelformat()());
            }
          
          Terrain::instance()->setTarget(buffer_);
          Environment::instance()->setTarget(buffer_);
        }

      
      if(screen_)
        {
          updateVisibleRange();
          schedule(REBUILD_LIST);
        }

  return false;
}

bool
Renderer::slot_viewport_point(const std::string &event, const uta::Point &p)
{
          int sw = screen_->width();
          int sh = screen_->height();

          //got a simple leftclick at position pixel:
          Pixel pixel;
          pixel.i = p.x+viewCenter_.i-sw/2-screen_->rect().upperLeft().x;
          pixel.j = p.y+viewCenter_.j-sh/2-screen_->rect().upperLeft().y;
          
          ScreenEntity* entity = getBoundingEntity(pixel);
          
          // hack for terrain height when beeing outside ... FIXME !
          if(!entity)
            {
              World::Entity* pc = World::World::inst()->lookupEntity(idToFollow_);
              if(pc)
                pixel.j += World::World::inst()->getGroundLevel(pc->absPosition().x, 
                                                                pc->absPosition().y);
            }

          Coord coords = p2w(pixel);

          
          DII::Object::MapType msg;

          // world coordinates
          msg["coords"] = coordToObj(coords);
          
          if(entity)
            {
              msg["entity"] = entity->id;
              // offset inside container
              Coord offset = p2w(pixel)-p2w(entity->pixel);
              msg["offset"] = coordToObj(offset);
            }

          msg["event"] = event;
          msg["from"] = "Renderer";
          
          /* Emit signal carrying this message.
           * If one connected object accepts this message it may return "true"
           * in order to "swallow" the event.
           * If no connected object returns true the default handlers get called.
           */
          if(!signal.emit(msg))
            {
              if(event == "rightclicked")
                {
                  Player::instance()->rightclick(msg); 
                }
              else if(event == "clicked")
                {
                  Player::instance()->leftclick(msg);
		}
              else if(event == "mouse" && flags_.test(HIGHLIGHT) )
                {
                  static ScreenEntity* old_one = NULL;
                  if(old_one)
                    {
                      highlight_backend(old_one, 0);
                      old_one = NULL;
                    }
                  if(entity)
                    {
                      old_one = entity;
                      highlight_backend(entity, -1);
                   }
                }
              else
                {
                  cerr << "Renderer: unhandled userinput :" << event << endl;
                }
            }

  return false;
}

ScreenEntity*
Renderer::getBoundingEntity(const Pixel &pixel)
{ 
          uta::Point theClick = uta::Point(pixel.i, pixel.j);
          
          /* iterate through displayList_ reversely and check if mouse click 
           *    is inside of one of the bounding boxes
           */
          uta::Point thePixel;
//           uta::Surface* frame;
//          uta::Color col;
//           uta::Painter* painter;
          ScreenEntity* entity = NULL;
         
          
          DisplayList::reverse_iterator itr = displayList_.rbegin();
          while(itr != displayList_.rend())
            {
              if((*itr)->rect.contains(theClick) || flags_.test(SHOWLABELS) )
                {
                  if(!(*itr)->empty())
                    {
                      debugN(10,cerr << "searching inside "<<(*itr)->id<<" for the clicked object ..." << endl;);
                      
                      // check for nested items:
                      Tree::iterator nitr = (*itr)->begin();
                      while(nitr != (*itr)->end())
                        {
                          ScreenEntity* nested = 
                            dynamic_cast<ScreenEntity*>(*nitr);
                          
                          Point delta = Point((*itr)->pixel.i, (*itr)->pixel.j);
                          if(flags_.test(SHOWLABELS))
                            {
                              if(uta::Rect(nested->rect.upperLeft().x, 
                                           nested->rect.upperLeft().y,
                                           nested->label->width(), 
                                           nested->label->height()).contains(theClick-delta))
                                {
                                  entity = nested;
                                  break;
                                }
                            }
                          else if(nested->rect.contains(theClick-delta))
                            {
                              debugN(10, cerr << "click on "<<nested->id << endl;);                  
                              // this should never happen !
                              assert(nested->sprite);
                              
                              int w = nested->sprite->width();
                              int h = nested->sprite->height();
                              if(nested->sprite->isOpaquePixel(pixel.i-delta.x-(nested->pixel.i-w/2),
                                                               pixel.j-delta.y-(nested->pixel.j-h)))
                                {
                                  entity = nested;
                                  break;
                                }
                              
                            } // if(nested->
                          nitr++;
                        } // while
                    }
                  if(entity)
                    break;
                  
                  //need to check for transparent pixel here
                  // this should never happen !
                  assert((*itr)->sprite);
                  
                  if(flags_.test(SHOWLABELS))
                    {
                      if(uta::Rect((*itr)->rect.upperLeft().x, 
                                   (*itr)->rect.upperLeft().y,
                                   (*itr)->label->width(), 
                                   (*itr)->label->height()).contains(theClick))
                        {
                          entity = (*itr);
                          break;
                        }
                    }
                  else
                    {
                      int w = (*itr)->sprite->width();
                      int h = (*itr)->sprite->height();
                      if((*itr)->sprite->isOpaquePixel(pixel.i-((*itr)->pixel.i-w/2),
                                                       pixel.j-((*itr)->pixel.j-h)))
                        {
                          entity = *itr;
                          break;
                        }
                    }
                }
              itr++;
            }

  return entity;
}

void
Renderer::drawOverlays(std::queue<Overlay>& overlays, uta::Surface* target)
{
  if(!target)
    return;

  std::queue<Overlay> tmp;
  Overlay overlay;
  while(!overlays.empty())
    {
      overlay = overlays.front();
      overlays.pop();
      if(overlay.age > 0)
        {
          // reschedule 
          overlay.age--;
          tmp.push(overlay);
        }
      else if(overlay.age < 0)
        tmp.push(overlay);

      if(overlay.type == SPRITE_OVERLAY)
        {
          SpriteCache::inst()->lock(overlay.sprite.spriteID);
          Sprite* sprite = SpriteCache::inst()->lookup(overlay.sprite.spriteID);
          if(overlay.sprite.semitrans)
            sprite->blit_semitrans(overlay.sprite.orientation,
                                   overlay.sprite.frameNo,
                                   overlay.sprite.layerNo,
                                   target,
                                   overlay.x, overlay.y);
          else
            sprite->blit(overlay.sprite.orientation,
                         overlay.sprite.frameNo,
                         overlay.sprite.layerNo,
                         target,
                         overlay.x, overlay.y);
          
          SpriteCache::inst()->release(overlay.sprite.spriteID);
        }
      else if(overlay.type == IMAGE_OVERLAY)
        {
          int w = overlay.image.image->width();
          int h = overlay.image.image->height();
          int x = (overlay.x >= 0) ? overlay.x : target->width()+overlay.x;
          int y = (overlay.y >= 0) ? overlay.y : target->height()+overlay.y;
          
          overlay.image.image->blit(target,uta::Rect(x,y,w,h));
        }
      else if(overlay.type == LABEL_OVERLAY)
        {
          int w = overlay.label.image->width();
          int h = overlay.label.image->height();
          int x = overlay.x;
          int y = overlay.y;
          
          overlay.label.image->blit(target,uta::Rect(x,y,w,h));
        }
    } // while(!overlays.empty())

  overlays = tmp;
}

void
Renderer::drawEntity(ScreenEntity* entity, const Pixel& offset)
{
  assert(entity);
  if(!entity->sprite)
    return; // Missing media

  int w,h;

  /* --------------------------
   * care for frame counting
   * 
   */
  if(!entity->sprite->reversable())
    {
      if(entity->numFrames > 1 && 
         (entity->sprite->loopable() || entity->frame < (entity->numFrames - 1)))
        {
          //if animated, increment framecounter
          entity->frame++;
          if(entity->frame >= entity->numFrames)
            //reset frame counter
            entity->frame = 0; 
        }
      else if(entity->sprite->isIdleAnimation() && 
              (entity->frame == entity->numFrames-1))
        {
          // this sprite is non loopable, but an idle animation
          // restart it an random times
          //
          if(drand48() > 1.0-idleRespawnProb_)
            entity->frame = 0;
        }
    }
  else if(entity->numFrames > 1)
    {
      // care for reversable, animated sprite
      if(entity->forwards)
        {
          if(entity->frame < entity->numFrames-1)
            entity->frame++;
          else
            {
              if(!entity->sprite->isIdleAnimation() ||
                 (drand48() > 1.0-idleRespawnProb_))
                {
                  entity->forwards = false;
                }
            }
        }
      else
        {
          // we are running backwards ...
          
          if (entity->sprite->loopable())
            {
              // make turn at frame 0
              entity->frame--;
              if(entity->frame == 0)
                entity->forwards = true;
            }
          else 
            {
              // stop at 0
              if(entity->frame > 0)
                entity->frame--;
              else if(entity->sprite->isIdleAnimation())
                {
                  // this sprite is non loopable, but an idle animation
                  // restart it an random times
                  //
                  if(drand48() > 1.0-idleRespawnProb_)
                    entity->forwards = true;
                }
            }
        }
    }

  /* clear dirtry flag.
   */
  entity->dirty = false;
 
  //we are assuming all layers to have the same size
  w = entity->sprite->width();
  h = entity->sprite->height();

  //offset for nested sprites 
  Pixel _offset;
  _offset.i = offset.i + entity->pixel.i;
  _offset.j = offset.j + entity->pixel.j;

  //offset for sprite (reference point is center of baseline)
  int x = _offset.i - w/2;
  int y = _offset.j - h;

  if(flags_.test(SHOWBOUNDS))
    {
      uta::Painter* painter = new uta::Painter(buffer_);
      painter->setColor(uta::blue);
      uta::Point delta(offset.i, offset.j);
         
      int h = 0;
      World::Entity* dbent = World::World::inst()->lookupEntity(entity->id);
      if(dbent)
        h = World::World::inst()->getGroundLevel(dbent->absPosition().x, 
                                                        dbent->absPosition().y);
            
      painter->box(entity->rect.upperLeft()+delta, entity->rect.lowerRight()+delta);
      float dx = entity->bounding_box.x;
      float dy =  entity->bounding_box.y;
      //  float dz =  entity->bounding_box.z;
      Pixel p0 = w2p(entity->position );
      Pixel p1 = w2p(entity->position + Coord(dx,0,0));
      Pixel p2 = w2p(entity->position + Coord(dx, dy, 0));
      Pixel p3 = w2p(entity->position + Coord(0, dy, 0));
      
      painter->setColor(uta::red);
      painter->line(uta::Point(p0.i+offset.i, p0.j+offset.j-h), 
                    uta::Point(p1.i+offset.i, p1.j+offset.j-h));
      painter->line(uta::Point(p1.i+offset.i, p1.j+offset.j-h), 
                    uta::Point(p2.i+offset.i, p2.j+offset.j-h));
      painter->line(uta::Point(p2.i+offset.i, p2.j+offset.j-h), 
                    uta::Point(p3.i+offset.i, p3.j+offset.j-h));
      painter->line(uta::Point(p3.i+offset.i, p3.j+offset.j-h), 
                    uta::Point(p0.i+offset.i, p0.j+offset.j-h));
      
      
      
      delete painter;
      //   string position;
      //   entity->position.toString(position);
      
      //   FONT_RES("entity_font")->blitString(position, screen_, 
      //                                       uta::Point(py2.i+offset.i, py2.j+offset.j));
    } // if(showbounds_)

  if(entity->underlay)
    {
      int _w = entity->underlay->width();
      int _h =  entity->underlay->height();
      entity->underlay->blit(buffer_, uta::Rect(_offset.i - _w/2,
                                                _offset.j - _h, _w, _h));
    }
  
    if(cast_shadows_ && entity->shadow)
    {
      int _w = entity->shadow->width();
      int _h =  entity->shadow->height();
      entity->shadow->blit(buffer_, uta::Rect(_offset.i - _w/2,
                                              _offset.j - _h, _w, _h));
    }
//   shadow_.blit(buffer_, uta::Rect(_offset.i - shadow_.width()/2,
//                                   _offset.j - shadow_.height(), 
//                                   shadow_.width(),
//                                   shadow_.height()));
               

  //  frame->blit(buffer_, uta::Rect(x,y,w,h));
  ScreenEntity::DisplayList::const_iterator citr;
  unsigned int layers = entity->sprite->numLayers();
  //blit optional layers:
  for(unsigned int l=0; l < layers; l++)
    {
      //is this the first layer in front of/above the any content ?
      if(l == entity->sprite->contentLayer())
        {
          //if so
          //check for contained entities, iff the sprite supports it:
          if(entity->sprite->isContainer() && !entity->empty())
            {
              citr = entity->displayList().begin();
              while(citr != entity->displayList().end())
                {
                  drawEntity(*citr, _offset);
                   citr++;
                }
            }        
        }
      if(l == entity->sprite->roofLayer())
        {
          //if the container is not empty
          //skip the blitting of the rest of the layers
          //which are in fact the roof or cover
//           if(entity->sprite->isContainer() && !entity->empty())
//             break;
          if(entity->find(idToFollow_) != NULL) 
            break;
        }

      if( entity->ghost || (l >= entity->sprite->contentLayer()) && (entity->find(idToFollow_) != NULL) )
        {
          // semitransparent blit
          entity->sprite->blit_semitrans(entity->orientation,entity->frame,l,buffer_, x, y);
        }
      else
        {
          /* if this is our character and if it is not inside a building 
           * schedule this sprite for a semitransparent overlay :
           */
          if(entity->id == idToFollow_ && entity->root() == &entities_)
            {
              Overlay overlay;
              overlay.x = x;
              overlay.y = y;
              overlay.type = SPRITE_OVERLAY;
              overlay.age = 0;
              overlay.sprite.orientation = entity->orientation;
              overlay.sprite.frameNo = entity->frame;
              overlay.sprite.layerNo = l;
              overlay.sprite.spriteID = entity->spriteID;
              overlay.sprite.semitrans = true;
              overlays_.push(overlay);
            }
          //go on with normal blitting of the layers

          if(entity->timeout == 0)
            entity->sprite->blit(entity->orientation,entity->frame,l,buffer_, x, y);
          else 
            entity->sprite->blit_highlighted(entity->orientation,entity->frame,l,buffer_, x, y,entity->hightlightColor);
        }
    }

  if(flags_.test(SHOWLABELS))
    {
    
      std::vector<uta::Color> pal = entity->label->pixelformat().palette;
      if(entity->timeout)
        pal[0] = uta::Color(220,220,220);
      else
        pal[0] = uta::Color(180,180,180);
      entity->label->setPalette(pal);

      if(entity->timeout)
        entity->label->setAlpha(100);
      else
        entity->label->setAlpha(50);

      Overlay overlay;
      overlay.x = x;
      overlay.y = y;
      overlay.type = LABEL_OVERLAY;
      overlay.age = 0;
      overlay.label.image = entity->label;
      overlay.label.entity = entity;

      labels_.push(overlay);
        
    }
  
  if(entity->overlay)
    {
      int _w = entity->overlay->width();
      int _h =  entity->overlay->height();
      entity->overlay->blit(buffer_, uta::Rect(_offset.i - _w/2,
                                                _offset.j - _h, _w, _h));
    }

  if(entity->timeout > 0)
    entity->timeout--;

}


bool
Renderer::redraw()
{  
#ifdef DEBUG
  my_time start, end;
  start.set_now();
#endif

  if(!screen_)
    return false;

  // check if a rebuild of the display list was requested
  if(isScheduled(REBUILD_LIST))
    {
      rebuildDisplayList();
      done(REBUILD_LIST);
    }

  // delete all items which were scheduled for deletion
  flush();

  //let modul Terrain render background
  Terrain::instance()->render();

  int sw = screen_->width();
  int sh = screen_->height();

  //pixel offset of upper left corner of the screen
  Pixel offset;
  offset.i = sw/2 - viewCenter_.i;
  offset.j = sh/2 - viewCenter_.j;

  DisplayList::iterator itr = displayList_.begin();
  while(itr != displayList_.end())
    {
      std::string &root_id = dynamic_cast<ScreenEntity*>((*itr)->root())->id;
      if(root_id.empty())
        drawEntity(*itr, offset);
      ++itr;
    }


  //now care for whole screen overlays 
  //like weather, rain or radius of view ...
  Environment::instance()->render();


  if(smallMap_)
    {
      Rect dst(0,0,smallMap_->width(), smallMap_->height());
      dst.warp(Point(sw-smallMap_->width(),sh-smallMap_->height()));
      smallMap_->blit(buffer_, dst);
    }

  //tell the directsurface that we are about to modify its content
  screen_->beginModify();

  if(clouds_)
    {
      clouds_->offset_x = static_cast<unsigned char>(round((double)clouds_->offset_x + cloudSpeed_.x));
      clouds_->offset_y = static_cast<unsigned char>(round((double)clouds_->offset_y + cloudSpeed_.y));
    }

#ifndef HAVE_UTA_LIGHTEN_DARKEN
  if(use_clouds_)
    uta::gammablit(screen_, buffer_, clouds_, uta::Rect(0,0,sw,sh), uta::Point(0,0));  
  else if(daylight_ && gamma_)
    uta::gammablit(screen_, buffer_, gamma_, Rect(0,0,sw,sh), Point(0,0));
  else
    buffer_->blit(screen_);
#else
  buffer_->blit(screen_);
  if(use_clouds_)
    screen_->lighten_darken(clouds_);  
  else if(daylight_ && gamma_)
    screen_->lighten_darken(gamma_);
#endif

  //
  drawOverlays(overlays_, screen_);
  
  if(flags_.test(SHOWLABELS))
    drawOverlays(labels_, screen_);

  //update the visible screen content 
  screen_->update();
  //tell the surface we are done modifying its content
  screen_->endModify();
 
#ifdef DEBUG  
  end.set_now();
  updateTime_ += end()-start();
  updates_++;
  if(updates_ == 100)
    {
      debugN(10,
             cerr << "Renderer::redraw() took " << updateTime_/updates_ 
             << "ms."<<endl;
             );
      updates_ = 0;
      updateTime_ = 0;
    }
#endif

  return true;
}




bool
Renderer::observ(const std::string &str)
{
  idToFollow_ = str;

  ScreenEntity* entity = 
    dynamic_cast<ScreenEntity*>(entities_.find(idToFollow_));
  
  if(!entity)
    return false;
 
  // recenter viewport:
  debugN(10, cerr << "Renderer: recentering viewport. " << endl;);
  

  viewCenter_ = entity->getAbsPixel();
  viewCenter_.j -= 100;               
    
  updateVisibleRange();
    
  return true;
}


bool
Renderer::move(const World::Entity *dbEntity)
{
  if(!initialized_)
    return false;

  std::string id = dbEntity->id();

  if (id == idToFollow_)
    {
      // I can only work out how to do this here, where we have access to
      // the entities world coords, and so can find its height
      Coord _coord = dbEntity->absPosition();
      viewHeight_ = w2p(_coord.z) + World::World::inst()->getGroundLevel(_coord.x, _coord.y);
    }

  //obtain pointer to entity:
  ScreenEntity *entity = 
    dynamic_cast<ScreenEntity*>(entities_.find(id)); 

  if(!entity)
    {
      debugN(10,cerr << "Renderer::move() nothing known about "<<id<<endl;);
      return false;
    }

  //extract new position information
  Pixel pixel = w2p(dbEntity->position());

  // apply ground level if not inside another entity
  if(entity->root() == &entities_)
    pixel.j -= World::World::inst()->getGroundLevel(dbEntity->position().x, dbEntity->position().y);

  entity->position = dbEntity->position();
  entity->ghost = dbEntity->unreal();

  uta::Rect screen = uta::Rect(uta::Point(viewRect_.imin, viewRect_.jmin),
                               uta::Point(viewRect_.imax, viewRect_.jmax));
  
  bool updateList = screen.intersect(entity->rect).isValid();
 
  /*calculate the area which this entity occupies on the screen
   *(this goes for entities outside the visible screen as well)
   */
  int w = entity->rect.width();
  int h = entity->rect.height();
  entity->rect = uta::Rect(pixel.i - w/2, pixel.j - h, w, h);

  updateList |= screen.intersect(entity->rect).isValid();

  //moveEntity must only be called for root-entities
  if(entity->root() == &entities_)
    {
      //move Entity, check if it is still visible, 
      // if not, schedule it for deletion
      moveEntity(id, entity, pixel);
   
    }
  else
    {
      entity->pixel = pixel;
    }

  recenterIfNeeded(entity);

  if(updateList)
    schedule(REBUILD_LIST);

  return true;
}

bool
Renderer::remove(const std::string &id)
{
  if(!initialized_)
    return false;
 
  ScreenEntity* entity = dynamic_cast<ScreenEntity*>(entities_.find(id));
  if(!entity)
    {
      debugN(10,cerr << "Renderer::remove() nothing known about "<<id<<endl;);
      return false;
    }

  // release the sprite (if it is locked)
  entity->releaseSprite();

  uta::Rect screen = uta::Rect(uta::Point(viewRect_.imin, viewRect_.jmin),
                               uta::Point(viewRect_.imax, viewRect_.jmax));
  
  bool updateList = screen.intersect(entity->rect).isValid();

  // remove from root entity
  assert(dynamic_cast<ScreenEntity*>(entity->root()));
  dynamic_cast<ScreenEntity*>(entity->root())->remove(id);
  entity->setRoot(NULL);

  // schedule all contained elements for deletion:
  Tree::LeafMap::iterator litr = entity->leaves().begin();
  while(litr != entity->leaves().end())
    {
      // remove connection to parent
      litr->second->setRoot(NULL);
      // schedule for deletion
      outOfSight_.push(dynamic_cast<ScreenEntity*>(litr->second));
      litr++;
    }

  // remove connectio to all contained elements 
  entity->clear();

  // 1. get last position:
  Pixel pixel = entity->pixel;
    
  //remove entry from display list:
  // 2. search old entry in displaylist:
  
  //remove at old position
  Coord oldpos = p2w(entity->pixel);
  Grid::iterator itr(&grid_, HASH_COORD(oldpos.x), HASH_COORD(oldpos.y));
  if(itr.valid())
    {
      reinterpret_cast<EntityList*>(itr.value())->remove(entity);
      if(reinterpret_cast<EntityList*>(itr.value())->empty())
        {
          delete reinterpret_cast<EntityList*>(itr.value());
          grid_.clear(itr.i(), itr.j());
        }
    }

  /* schedule rebuild of display list 
   */
  if(updateList)
    schedule(REBUILD_LIST);

  /* finally schedule this entity for deletion ...
   * it must not get deleted now, as it may be still in the displaylist ...
   */
  outOfSight_.push(entity);

  return true;
}

bool
Renderer::update(const World::Entity *dbEntity)
{
  if(!initialized_)
    return false;
  
  std::string id = dbEntity->id();

  if(id == "world_0")
      return true;

  ScreenEntity* entity = dynamic_cast<ScreenEntity*>(entities_.find(id)); 

  //extract position information
  Pixel pixel = w2p(dbEntity->position());

  bool created = false;
  //do we need to create new entry ?
  if(!entity)
    {
      created = true;

      //an entity for the given id could not be found ...
      //so we check if we should create a new one:
      if(id == idToFollow_)
        {
          //we are supposed to observe this id, so recenter
          //viewport on the entity's position to show it on the screen:
          
          debugN(10, cerr << "Renderer: recentering viewport. " << endl;);

          // use absolute position !
          Coord _coord = dbEntity->absPosition();
          Pixel _pixel = w2p(_coord);
          viewCenter_.i = _pixel.i;
          viewCenter_.j = _pixel.j-100;
          viewHeight_ = w2p(_coord.z) + World::World::inst()->getGroundLevel(_coord.x, _coord.y);
          
          updateVisibleRange();
        }
      else
        {
#ifdef FLUSH_OUTOFSIGHT
          //check if entity is outside of viewport
          if((viewRect_.imin-grace > pixel.i) || 
             (pixel.i > viewRect_.imax+grace) ||
             (viewRect_.jmin-grace > pixel.j) || 
             (pixel.j > viewRect_.jmax+grace))
            {
              // if so, forget about it:
              return false;
            }
#endif
        }
      //at this point, we are sure, that the entity is visible on the
      //screen, so we create it:

      entity = new ScreenEntity(NULL);
      //store id inside entity ... we need it for identifying it 
      // on mouseclicks etc...
      entity->id = id;
      
    } // if(!entity..

  // request a sprite for the Entity
  SpriteID spriteID = requestSprite(dbEntity);
  
  /* if sprite has changed ... 
   * Note: the initial value of ScreenEntity.spriteID is 0 which is
   * an invalid SpriteID -> so the initial sprite is safely set
   */
  if(spriteID != entity->spriteID)
    {
      // if old sprite was locked
      if(entity->sprite)
        {
          // release the old sprite
          entity->releaseSprite();
          // set the new sprite ID
          entity->spriteID = spriteID;
          // lock the new sprite
          entity->lockSprite();
        }
      else
        {
          // just set the new sprite ID
          entity->spriteID = spriteID;
        }

      // obtain sprite info
      SpriteCache::SpriteInfo info;
      SpriteCache::inst()->queryInfo(spriteID, info);
      if(info.state == SpriteCache::UNKNOWN)
        {
          /* This sprite has not been seen by the cache:
           * We need to lock it temporary to get infos about it
           */
          // lock the sprite in order to get its data
          SpriteCache::inst()->lock(spriteID);
          // and release it immediately
          SpriteCache::inst()->release(spriteID);
          // try again to query infos about this sprite
          SpriteCache::inst()->queryInfo(spriteID, info);
          // if this does not work out ... we give up !
          assert(info.state != SpriteCache::UNKNOWN);
        }

      entity->numFrames = info.frames;
      // reset frame counter
      entity->frame = 0;
      entity->rect = uta::Rect(0,0,info.width, info.height);
    } // if(spriteID ...
  
  std::string rootID = ((World::Entity*)dbEntity->root())->id();
  if(rootID != "" && rootID != "world_0")
    {
      ScreenEntity* new_root = 
        dynamic_cast<ScreenEntity*>(entities_.find(rootID));
      if(!new_root)
        {
          debugN(10,cerr<<"Renderer:update() cannot find root "
                 <<rootID <<endl;);
          // delete created entity here ... ! FIXME !
          return false;
        }
      
      // if there is an old root
      if(entity->root())
        {
          //remove from old root
          dynamic_cast<ScreenEntity*>(entity->root())->remove(id);
          
          // if the old root had its sprite locked, release the sprite for this entity
          if(dynamic_cast<ScreenEntity*>(entity->root())->sprite)
            entity->releaseSprite();
        }
      else
        {
          // else remove it from tree root
          entities_.remove(id);
        } 
      
      //also check for removal at toplevel display list if the entity
      //has not been created 
      if(!created)
        {
          // 1. get last position:
          Pixel old_pixel = entity->pixel;
          
          //remove entry from display list:
          // 2. search old entry in displaylist:

          Coord oldpos = p2w(entity->pixel);
          Grid::iterator itr(&grid_, HASH_COORD(oldpos.x), HASH_COORD(oldpos.y));
          if(itr.valid())
            {
              reinterpret_cast<EntityList*>(itr.value())->remove(entity);
              if(reinterpret_cast<EntityList*>(itr.value())->empty())
                {
                  delete reinterpret_cast<EntityList*>(itr.value());
                  grid_.clear(itr.i(), itr.j());
                }
            }
          
        } // if(!created)...
      
      //set new root
      debugN(10,cerr<< "Renderer moving "<<id<<" into "<<new_root->id<<endl;); 
      new_root->insert(id, entity);
      entity->setRoot(new_root);

      // if the new root has its sprite locked, so need this entity as well
      if(new_root->sprite)
        entity->lockSprite();
    }
  else
    {
      //if no root path has been provided, add the entity
      //to the root database
      if(created)
        {
          debugN(10,cerr<< "Renderer moving "<<id<<" into world_0"<<endl;); 
          entities_.insert(id, entity);
          entity->setRoot(&entities_);
        }
      else
        {
          /* if during an update no root has been provided
             then the entity is (now) a root-entity, we need to check
             if this has changed since last update: 
          */
          if(entity->root() != &entities_)
            {
              if(entity->root())
                entity->root()->remove(id);
              debugN(10,cerr<< "Renderer moving "<<id<<" from "
                     << ((ScreenEntity*)(entity->root()))->id 
                     << " into world_0"<<endl;); 
              entities_.insert(id, entity);
              entity->setRoot(&entities_);
            }
        }
    }

  if(!entity->shadow)
    entity->shadow = renderShadow(dbEntity->bbox(), 
                                  dbEntity->bboxMedian(),
                                  entity->rect.width());

  entity->ghost = dbEntity->unreal();

  entity->position = dbEntity->position();

  // changed since Acorn 0.4:
  entity->bounding_box = dbEntity->bbox() * 2.0;

  entity->orientation=dbEntity->orientation();


  uta::Rect screen = uta::Rect(uta::Point(viewRect_.imin, viewRect_.jmin),
                               uta::Point(viewRect_.imax, viewRect_.jmax));

  bool updateList = screen.intersect(entity->rect).isValid();

  if(entity->root() == &entities_)
    pixel.j -=    
      World::World::inst()->getGroundLevel(dbEntity->position().x, dbEntity->position().y);
  
  /*calculate the area which this entity occupies on the screen
   *(this goes for entities outside the visible screen as well)
   */
  int w = entity->rect.width();
  int h = entity->rect.height();
  entity->rect = uta::Rect(pixel.i - w/2, pixel.j - h, w, h);
  
  updateList |= screen.intersect(entity->rect).isValid();

  //moveEntity must only be called for root-entities
  if(entity->root() == &entities_)
    {
      //move Entity, check if it is still visible, 
      // if not, schedule it for deletion
      moveEntity(id, entity, pixel);
   
    }
  else
    {
      debugN(10,cerr << "Renderer moving "<<entity->id 
             <<" inside container to "<<pixel.i<<","<<pixel.j<<"." <<endl;);
      entity->pixel = pixel;
    }

  recenterIfNeeded(entity);

  if(updateList)
    schedule(REBUILD_LIST);

  return true;
}


void
Renderer::moveEntity(const std::string id, ScreenEntity* ent, const Pixel& pixel)
{
        
  //move the MapEntity if  necessary
  if((ent->pixel.j != pixel.j) ||
     (ent->pixel.i != pixel.i))
    {
      //pixelposition has changed !

      //search old entry and remove it
      Coord oldpos = p2w(ent->pixel);
      //Grid::iterator itr(&grid_, ent->pixel.i, ent->pixel.j);
  
      Grid::iterator itr(&grid_, HASH_COORD(oldpos.x), HASH_COORD(oldpos.y));
      if(itr.valid())
        {
          reinterpret_cast<EntityList*>(itr.value())->remove(ent);
          //if position is empty, remove it completely
          if(reinterpret_cast<EntityList*>(itr.value())->empty())
            {
              delete reinterpret_cast<EntityList*>(itr.value());
              grid_.clear(HASH_COORD(oldpos.x), HASH_COORD(oldpos.y));
            }
        }
         
      //check if entity is viewable :
      if((viewRect_.imin-grace>pixel.i) || (pixel.i>viewRect_.imax+grace) ||
         (viewRect_.jmin-grace>pixel.j) || (pixel.j>viewRect_.jmax+grace))
        {

          if(id !=  idToFollow_)
            {
#ifdef FLUSH_OUTOFSIGHT
              //if not ... remove entity from database
              debugN(10, cerr << "Renderer: entity " << id 
                     <<" moved out of sight !" << endl;);
              
              //move id to the outOfSight list for later removal
              outOfSight_.push(ent);
              //do the same for all nested entities !

              //FIXME !!!!!!!!!! this should be done recursively !
              ScreenEntity::iterator itr = ent->begin();
              while(itr != ent->end())
                {
                  outOfSight_.push(dynamic_cast<ScreenEntity*>(*itr));
                  itr++;
                }
              return;
#endif

            }
        }
      else
        {
          //if entity is player character and it is visible, check if
          //viewport needs to follow
          if(id ==  idToFollow_)
            {
              follow(ent);
            }
        }     

      //update pixel entry
      ent->pixel = pixel;
      //and mark entry dirty
      ent->dirty = true;

      /// if we are a root entity we must not get added to the grid !
      if(ent->root() != &entities_)
        return;

      //add at new position
      Coord newpos = p2w(ent->pixel);
  
      itr = Grid::iterator(&grid_, HASH_COORD(newpos.x), HASH_COORD(newpos.y));
      if(itr.valid())
        reinterpret_cast<EntityList*>(itr.value())->push_front(ent);
      else
        {
          grid_.set(HASH_COORD(newpos.x), HASH_COORD(newpos.y), new EntityList);
          reinterpret_cast<EntityList*>(grid_.at(HASH_COORD(newpos.x), HASH_COORD(newpos.y)))
            ->push_front(ent);
        }

      //scanlines_[ent->pixel.j][ent->pixel.i].push_front(ent);
    }
    
}

void 
Renderer::recenterIfNeeded(ScreenEntity* ent)
{
  std::string id = ent->id;

  if(id ==  idToFollow_)
    {
//       if(smoothScroll_)
//         { 
//           viewCenter_ = ent->getAbsPixel();
//           viewCenter_.j -= 100;               
          
//           updateVisibleRange();
//         }
//       else
//         {
//           follow(ent);
//         }
      follow(ent);
    }
}

void
Renderer::follow(ScreenEntity* entity)
{
  if(entity)
    {
      Pixel delta;
      Pixel pix = entity->getAbsPixel();

      if(pix.i - viewRect_.imin < scroll_border_i)
        {
          if(smoothScroll_)
            delta.i = pix.i - viewRect_.imin - scroll_border_i;
          else
            delta.i = - scroll_border_i;
        }
      else if(viewRect_.imax - pix.i < scroll_border_i)
        {
          if(smoothScroll_)
            delta.i = scroll_border_i - (viewRect_.imax - pix.i);
          else
            delta.i = scroll_border_i;
        }
      else
        delta.i = 0;

      if(pix.j - 100 - viewRect_.jmin < scroll_border_j)
        {
          if(smoothScroll_)
            delta.j = pix.j - 100 - viewRect_.jmin - scroll_border_j;
          else
            delta.j = - scroll_border_j;
        }
      else if(viewRect_.jmax - pix.j < scroll_border_j)
        {
          if(smoothScroll_)
            delta.j = scroll_border_j - (viewRect_.jmax - pix.j);
          else
            delta.j = scroll_border_j;
        }
      else
        delta.j = 0;     
    
      
      if(delta.i || delta.j)
        {
          //only update visible range, if needed
          viewCenter_.i += delta.i;
          viewCenter_.j += delta.j;

          updateVisibleRange();
        }
    }
}

void
Renderer::flush()
{
  ScreenEntity* entity;

  while(!outOfSight_.empty())
    {
      entity = outOfSight_.front();
      outOfSight_.pop();

      debugN(10,cerr << "Renderer::flush() deleting " << entity->id << " ... ";);
      //remove entity from displaylist
      if(entity->root())
        entity->root()->remove(entity->id);
      delete entity;
      debugN(10,cerr << "[ok]"<<endl;);
    }
}

