//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 01/2000

/* 
   coordinates:
       
         Z  
         .           
Y        .        X
   .     .     .
      .  .  .  
         0
      
         
*/


#ifndef _RENDERER_H_
#define _RENDERER_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sigc++/object.h>
#include <uta/timer.h>
#include <uta/font.h>
#include <uta/resources.h>
#include "media/sprite_cache.h"

#include "coord.h"
#include "my_time.h"
#include "isodefs.h"
#include "../media/sprite.h"
#include "../world/entity.h"

class Viewport;

#include <uta/surface.h>
#ifndef HAVE_UTA_LIGHTEN_DARKEN
#include <uta/blitters.h>
#endif

#include "tree.h"
#include "grid.h"

#include "toolbox/local_slist.h"
#include <list>
#include <bitset>

namespace DII {class Module;}

#define MAX(a,b) (a > b ? a : b)
#define MIN(a,b) (a < b ? a : b)

/**Screen object.
   Renderer::redraw() blits the members in following order:
     1. underlay
     2. all contained ScreenEntities
     3. sprite
     4. overlay
     
   Contained ScreenEntities are recursively blitted, meaning these
   may contain ScreenEntities again. Renderer expects all entities
   (the contained ones as well) to have _global_ pixel coordinates.
   The common entity will only have a sprite, more advanced ones
   will also have an underlay (shadow ?) and perhaps an overlay
   (name, status ?).
*/
class ScreenEntity : public Tree
{
public:
  ///
  typedef LOCAL_SLIST(ScreenEntity*) DisplayList;
  // return true of a is behind of b
  static bool behind_of(ScreenEntity* a, ScreenEntity* b)
    {
      // first for the easy and clear situations
      if(a->position.x >= b->position.x && a->position.y >= b->position.y)
	return true;
      if(a->position.x < b->position.x && a->position.y < b->position.y)
	return false;
      
      float _min, _max;
      
      _max = MIN(b->position.y+b->bounding_box.y, a->position.y+a->bounding_box.y);
      _min = MAX(b->position.y, a->position.y);
      
      if(_max > _min)
	{
	  // entities occupy common y range
	  return (a->position.x > b->position.x);
	}
      
      return true;
    }
 protected:
  ///
  DisplayList displayList_;
  ///
  void updateDisplayList()
    {
      displayList_.clear();
      if(Tree::empty())
	return;
      
      ScreenEntity* ent;
      DisplayList::iterator citr; 
      Tree::LeafMap::iterator itr = Tree::leaves().begin();      
      while(itr != Tree::leaves().end())
	{
	  ent = dynamic_cast<ScreenEntity*>(itr->second);
	  citr = displayList_.begin();
	  while(citr != displayList_.end() && behind_of(*citr, ent))
	    citr++;
	  displayList_.insert(citr, ent);

	  itr++;
	} 
      
    }
 public:
  ///
  ScreenEntity(ScreenEntity* root) : 
    Tree(root),
    dirty(true), overlay(NULL), underlay(NULL), label(NULL), sprite(NULL), spriteID(0),
    frame(0), forwards(true), numFrames(1), orientation(0), ghost(false), timeout(0), shadow(NULL)
    {
      pixel.i = 0;
      pixel.j = 0;
    }
  ///
  virtual ~ScreenEntity() 
    {
      
      // if we still have our sprite locked, release it
      if(sprite)
	SpriteCache::inst()->release(spriteID);
      if(label)
	delete label;
      if(shadow)
	delete shadow;
    };
  ///
  DisplayList& displayList() { return displayList_;};
  ///  
  virtual void insert(const std::string& _id, Tree* leaf)
    {      
      Tree::insert(_id,leaf);
      updateDisplayList();
    }
  ///
  virtual void remove(const std::string& _id)
    {
      Tree::remove(_id);
      updateDisplayList();
    }
  ///
  virtual void clear() 
    {
      // hmmm ... this does not delete the leafes !
      Tree::clear();
      updateDisplayList();
    }
  ///
  void releaseSprite()
    {
      if(!sprite)
	return;
      SpriteCache::inst()->release(spriteID);
      sprite = NULL;

      Tree::LeafMap::iterator itr = Tree::leaves().begin();
      while(itr != Tree::leaves().end())
	{
	  dynamic_cast<ScreenEntity*>(itr->second)->releaseSprite();
	  itr++;
	}
    }
  ///
  void lockSprite()
    {
      if(sprite)
	return;

      SpriteCache::inst()->lock(spriteID);
      sprite = SpriteCache::inst()->lookup(spriteID);
      iterator itr = begin();
      while(itr != end())
	{
	  dynamic_cast<ScreenEntity*>(*itr)->lockSprite();
	  itr++;
	}

      if(!label)
	{
	  static const uta::Font* font = NULL;
	  if(!font)
	    font = FONT_RES("entity_font");
	  
	  if(font)
	    label = font->getString(id);
	}
    }
  /// returns absolute pixel coordinates
  Pixel getAbsPixel() 
    {
      if(!root_)
	return pixel;
      
      Pixel result = dynamic_cast<ScreenEntity*>(Tree::root())->getAbsPixel();
      result.i += pixel.i;
      result.j += pixel.j;

      return result;
    }
  ///
  std::string id;                      //who am I ?
  ///
  bool dirty;                     // needs to get redrawn
  ///
  const uta::Surface* overlay;
  ///
  const uta::Surface* underlay;
  ///
  uta::Surface* label;
  ///
  Sprite* sprite;
  ///
  SpriteID spriteID;
  ///
  unsigned int frame;             // framecounter
  ///
  bool forwards;
  ///
  unsigned int numFrames;
  ///
  unsigned int orientation;
  ///
  Pixel pixel;     // screenposition in pixel or offset inside parent
  ///
  uta::Rect rect;  //the screen space this entity occupies
  ///
  Coord bounding_box; // the space this entity occupies
  ///
  Coord position;
  ///
  bool ghost;
  /// -1 means infinite
  int timeout;
  ///
  RLE::Color hightlightColor;
  ///
  uta::Surface* shadow;
};


///
typedef ScreenEntity* MapEntity;
///
typedef LOCAL_SLIST(MapEntity) EntityList;

///
class Renderer : virtual public SigC::Object
{
  
 public:
  
  ///
  Renderer();
  ///
  virtual ~Renderer() {assert(instance_); instance_ = 0;}

  static Renderer* instance() {return instance_ ? instance_ : new Renderer();}

  ///
  bool init(DII::Module*);
  ///
  void deinit();

  /** add/update sprite to displaylist.
  */
  bool update(const World::Entity*);
  /* move the given sprite
   */
  bool move(const World::Entity*);
  /** remove sprite from display list.
   *  @param is string id
  */
  bool remove(const std::string&);
  /// accepts const uta::Surface*
  void setMap(const uta::Surface*); 
  /// accepts int (valid range: -127 ... 128 
  void setGamma(double gamma);
  /// takes string agument: id of entity the viewport should follow
  bool observ(const std::string&);  
  ///
  void clearDisplayList();  
  ///
  void setClouds(int i); // probably should take bool
  ///
  void highlight(const std::string &id, int timeout = -1, RLE::Color color=0x00FF)
    {highlight_backend(dynamic_cast<ScreenEntity*>(entities_.find(id)), timeout, color);}

  ///
  bool slot_viewport_expose(bool, uta::DirectSurface*);
  ///
  bool slot_viewport_point(const std::string&, const uta::Point&);
  ///
  bool redraw();

  ///
  void follow(ScreenEntity*);

  SigC::Signal1<bool, const Atlas::Message::Object&,
	SigC::FixedMarshal<bool,false> > signal;

 protected:
  ///
  void highlight_backend(ScreenEntity *entity, int timeout = -1, RLE::Color color=0x00FF);

  /** draw entity to screen using given offset.
      This method is called recursively for each nested entity.
  */
  void drawEntity(ScreenEntity* entity, const Pixel& offset);
  ///
  ///
  void updateVisibleRange();
  ///
  void flush();
  ///
  void addEntityList(EntityList* List);
  ///
  void recenterIfNeeded(ScreenEntity* ent);

  /**move Entity and  check if it is still visible. 
     If not, schedule it for deletion.
  */
  void moveEntity(const std::string id, ScreenEntity* ent, const Pixel& pixel);
  ///
  void configure();
  ///
  void rebuildDisplayList();
  /** Request a sprite from the SpriteCache.
      This only request the sprite ID, the sprite data is not yet
      available as the sprite is not yet locked.
  */
  SpriteID requestSprite(const World::Entity* entity);
  ///
  ScreenEntity* getBoundingEntity(const Pixel &pixel);
  ///
  Viewport* viewport_;
  ///
//  DII::Method fpGetScreen;

  ///
  ScreenEntity entities_;
  
  ///
  ViewRect viewRect_;
  ///
  Pixel viewCenter_;
  ///
  pixel_t viewHeight_;

  ///
  typedef std::list<ScreenEntity*> DisplayList;
  ///
  DisplayList displayList_;
  ///
  std::queue<ScreenEntity*> outOfSight_;
  ///
  uta::DirectSurface* screen_;
  ///
  uta::Surface* buffer_;
  ///
  char gamma_;

  ///
  bool initialized_;
  ///
  std::string idToFollow_;
  ///
  int scroll_border_i;
  ///
  int scroll_border_j;
  ///
  int grace;
  ///
  bool smoothScroll_;
  ///
  const uta::Surface* smallMap_;
  ///
  Grid grid_;

#ifdef DEBUG
  long int updates_;
  long int updateTime_;
#endif 

  ///
  uta::GammaFunction* clouds_;
  ///
  bool use_clouds_;
  ///
  Coord cloudSpeed_;
  ///
  double idleRespawnProb_;
  ///
  uta::Surface shadow_;
  ///
  bool rebuildDisplayList_;
  ///
  enum RenderCmd { REBUILD_LIST };
  ///
  std::bitset<32> commands_;
  ///
  void schedule(RenderCmd cmd) { commands_.set(cmd); }
  ///
  bool isScheduled(RenderCmd cmd) { return commands_.test(cmd); }
  ///
  void done(RenderCmd cmd) { commands_.reset(cmd); }

  ///
  enum overlay_t { SPRITE_OVERLAY, IMAGE_OVERLAY, LABEL_OVERLAY };

  ///
  typedef struct 
  {
    unsigned char orientation;
    unsigned char frameNo;
    unsigned char layerNo;
    SpriteID spriteID;
    bool semitrans;
  } SpriteCmd;

  
  ///
  typedef struct
  {
    const uta::Surface* image;
    ScreenEntity* entity;
  } LabelCmd;

  ///
  typedef struct
  {
    const uta::Surface* image;
  } ImageCmd;

  ///
  typedef struct
  {
    int x,y;
    overlay_t type;
    int age;
    union
    {
      SpriteCmd sprite;
      ImageCmd image;
      LabelCmd label;
    };
  } Overlay;

  ///
  std::queue<Overlay> overlays_;
  ///
  std::queue<Overlay> labels_;

  void drawOverlays(std::queue<Overlay>& overlays, uta::Surface* target);
  ///
  bool cast_shadows_;
  ///
  bool daylight_;
  ///
  enum FLAG_NAMES { SHOWBOUNDS, HIGHLIGHT, SHOWLABELS } ;
  ///
  std::bitset<32> flags_;

  static Renderer* instance_;
};

#endif
