//  The Worldforge Project
//  Copyright (C) 1998, 1999,2000,2001 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 03/2001

#include "lua_script.h"

extern "C" 
{
#include "lualib.h"
#include "luasocket.h"
}

using namespace Lua;

//static map of all active scripts
Script::ScriptMap Script::scripts;

Script::Script(int flags, int stacksize) :
  flags_(0)
{
  /* create new instance of the LUA virtual machine
   * default stacksize is 1024
   */
  handle_ = lua_open(stacksize);

  //register this script
  Script::scripts[handle_] = this;
  
  initExtensions(flags);
  
}  

/** */
void 
Script::initExtensions(int flags)
{
  /* open additional libraries ... */
  if(flags & INIT_BASELIB && !(flags_ & INIT_BASELIB))
    lua_baselibopen(handle_); // basic stuff
  if(flags & INIT_STRINGS && !(flags_ & INIT_STRINGS))
    lua_strlibopen(handle_); // string manipulation
  if(flags & INIT_MATH && !(flags_ & INIT_MATH))
    lua_mathlibopen(handle_);  // math functions
  if(flags & INIT_FILEIO && !(flags_ & INIT_FILEIO))
    lua_iolibopen(handle_); // file io (might be dangerous ....)
  if(flags & INIT_SOCKETS && !(flags_ & INIT_SOCKETS))
    lua_socketlibopen(handle_); // tcp socket library
  if(flags & INIT_DEBUG && !(flags_ & INIT_DEBUG))
    lua_dblibopen(handle_);  // lua debug interface

  flags_ |= flags;

}

Script::~Script()
{
  //shutdown LUA VM
  lua_close(handle_);

  //deregister this script
  Script::scripts.erase(handle_);
}

Script*
Script::getRef(Script::Handle handle)
{
  ScriptMap::iterator itr = Script::scripts.find(handle);
  if(itr == Script::scripts.end())
    return NULL;
  return itr->second;
}

int
Script::doFile(const char* filename)
{
  /* execute LUA code from file, ensuring the stack is restored afterwards */
  int oldtop = lua_gettop(handle_);
  int result = lua_dofile(handle_, filename);
  lua_settop(handle_, oldtop);
  return result;
}
 
 
int
Script::doBuffer(const char* buffer, size_t size, const char* name)
{
  /* execute LUA code from mem buffer, ensuring the stack is 
   * restored afterwards*/
  int oldtop = lua_gettop(handle_);
  int result = lua_dobuffer(handle_, buffer, size, name);
  lua_settop(handle_, oldtop);
  return result;
} 

void
Script::call(const std::string& funcname)
{
  // remember stack state
  int oldtop = lua_gettop(handle_);

  //get function to be called (and put it onto lua stack)
  lua_getglobal(handle_, funcname.c_str());

  if(!lua_isfunction(handle_, -1))
    throw UnknownFunctionException();

  //call function
  lua_call(handle_,0,0);
 
  //restore lua stack
  lua_settop(handle_, oldtop);
}
  
Value 
Script::call(const std::string& funcname, const ListType& args)
{
  int oldtop = lua_gettop(handle_);
  Value result;

  //get function to be called (and put it onto lua stack)
  lua_getglobal(handle_, funcname.c_str());

  if(!lua_isfunction(handle_, -1))
     throw UnknownFunctionException();

  std::list<Value>::const_iterator itr = args.begin();
  while(itr != args.end())
    pushValue(*itr++);

  //call function: args.size() arguments, 1 result
  lua_call(handle_,args.size(),1);
 
  //get result from the stack
  result = popValue();

  //restore lua stack
  lua_settop(handle_, oldtop);
 
  return result;      
}

  
Script::ListType 
Script::call(const std::string& funcname, 
	     const ListType& args, int nresults)
{
  int oldtop = lua_gettop(handle_);
  ListType result;

  //get function to be called (and put it onto lua stack)
  lua_getglobal(handle_, funcname.c_str());

  if(!lua_isfunction(handle_, -1))
     throw UnknownFunctionException();

  std::list<Value>::const_iterator itr = args.begin();
  while(itr != args.end())
    pushValue(*itr++);


  //call function: args.size() arguments, nresults results
  lua_call(handle_,args.size(),nresults);
 
  //get result from the stack
  for(int n=0; n < nresults; n++)
    result.push_front(popValue());

  //restore lua stack
  lua_settop(handle_, oldtop);
 
  return result;    
}
 
void 
Script::registerFunc(const std::string& funcname, Value::FunctionType func)
{
  lua_register(handle_,funcname.c_str(), func);
}

 
void 
Script::setGlobal(const std::string& name, const Value& val)
{
  pushValue(val);
  lua_setglobal(handle_, name.c_str());
}
  
Value 
Script::getGlobal(const std::string& name)
{
  lua_getglobal(handle_, name.c_str());
  return popValue();
}


  
Value 
Script::peekValue() throw (StackOverflowException, StackUnderrunException)
{
  // check if stack is not empty
  if(!lua_gettop(handle_))
    throw StackUnderrunException();

  Value result;
  switch(lua_type(handle_, -1))
    {
    case LUA_TNONE:
      throw StackUnderrunException();
      break;
    case LUA_TNIL:
      break;
    case LUA_TNUMBER:
      result = lua_tonumber(handle_, -1);
      break;
    case LUA_TSTRING:
      result = std::string(lua_tostring(handle_, -1), lua_strlen(handle_, -1));
      break;
    case LUA_TTABLE:
      {
	int table_index = lua_gettop(handle_);
	Value::TableType table;
	//get tag of the lua object
	Value::TagType tag = lua_tag(handle_, -1);
	/* check if there is at least 1 entry free on the stack,
	 * which is needed for table traversal 
	 */
	if(!lua_stackspace(handle_))
	  throw StackOverflowException();
	
	lua_pushnil(handle_);  /* push first 'dummy' key */
	Value key, value;
	while (lua_next(handle_, table_index) != 0) 
	  {
	    /* `key' is at index -2 and `value' at index -1 */
	    value = popValue(); // note the recursion ! (nested tables ...)
	    key = peekValue(); // keep key on the stack for next iteration !
	    table.push_back(std::pair<Value, Value>(key, value));
	  }
	result = table;
	if(tag != LUA_TTABLE)
	  result.setTag(tag);
      }
      break;
    case LUA_TFUNCTION:
      result = lua_tocfunction(handle_, -1);
      break;
    case LUA_TUSERDATA:
      {
	//get tag of the lua object
	Value::TagType tag = lua_tag(handle_, -1);
	result = lua_touserdata(handle_, -1);
	if(tag != LUA_TUSERDATA)
	  result.setTag(tag);
      }
      break;
    default:
      break;
    }

  return result;
}

Value
Script::popValue()
{
  // read value at top of stack
  Value result = peekValue();

  // actually pop the element from the stack
  lua_pop(handle_,1);
  
  return result;
}  
  
void 
Script::pushValue(const Value& val) throw (StackOverflowException)
{
  if(!lua_stackspace(handle_))
    throw StackOverflowException();

  switch(val.type())
    {
    case Value::NIL:
      lua_pushnil(handle_);
      break;
    case Value::NUMBER:
      lua_pushnumber(handle_, val.asNumber());
      break;
    case Value::STRING:
      lua_pushlstring(handle_, val.asString().c_str(), val.asString().size());
      break;
    case Value::USERDATA:
      {      
	lua_pushusertag(handle_, val.asUserdata(), val.tag());
      }
      break;
    case Value::FUNCTION:
      lua_pushcfunction(handle_, val.asFunction());
      break;
    case Value::TABLE:
      {
	//create a new empty table and push it onto the stack
	lua_newtable(handle_);
       
	//if table owns special tag, tag the lua object appropriately
	if(val.tag() != LUA_ANYTAG)
	  lua_settag(handle_, val.tag());

	//remember the stack index of the table
	int table_index = lua_gettop(handle_);
	Value::TableType::const_iterator itr = val.asTable().begin();
	while(itr != val.asTable().end())
	  {
	    /* push key and value onto the stack ... */
	    pushValue(itr->first);
	    pushValue(itr->second); // note the recursion ! (nested tables ...)
	    /* and set the table's entry: */
	    lua_settable(handle_, table_index);
	    itr++;
	  }
      }
      break;
      
    }
}

void
Script::registerType(const std::string& type)
{
  TagMap::iterator itr = typeTags_.find(type);
  if(itr == typeTags_.end())
    {
      //not yet known type
      typeTags_[type] = lua_newtag(handle_);
      //make tag known to LUA
      setGlobal(type+"_tag",typeTags_[type]);
    }
  
}

Value::TagType
Script::getTypeTag(const std::string& type) const
{
  TagMap::const_iterator itr = typeTags_.find(type);
  
  //throw exception if type not found
  if(itr == typeTags_.end())
    throw UnknownTypeException();
  
  return itr->second;
}

Value
Script::createObject(const std::string& type, const Value::TableType& val)
{
  Value::TagType tag = getTypeTag(type);
  return Value(tag, val); 
}

Value
Script::createObject(const std::string& type, void* val)
{
  Value::TagType tag = getTypeTag(type);
  return Value(tag, val); 
}



