//  The Worldforge Project
//  Copyright (C) 1998, 1999,2000,2001 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 03/2001

#ifndef _LUA_SCRIPT_H_
#define _LUA_SCRIPT_H_

#include "lua_value.h"
#include <list>
#include <map>
#include <stdexcept>

namespace Lua {

class StackUnderrunException : public std::underflow_error {
 public: 
  StackUnderrunException() : 
    std::underflow_error("LuaScript stack underrun") {};
};

class StackOverflowException  : public std::overflow_error {
 public: 
  StackOverflowException() : 
    std::overflow_error("LuaScript stack overflow") {};
};

class UnknownTypeException : public std::runtime_error {
 public: 
  UnknownTypeException() : 
    std::runtime_error("LuaScript got unknown type.") {}; 
};
 
class UnknownFunctionException : public std::runtime_error {
 public: 
  UnknownFunctionException() : 
    std::runtime_error("LuaScript unknown function was called") {};
};

///
class Script
{
 public:
  
  ///
  typedef lua_State* Handle;
  ///
  typedef std::list<Value> ListType;
  ///
  enum  { 
    INIT_BASELIB = 1 << 0,
    INIT_STRINGS = 1 << 1, 
    INIT_MATH    = 1 << 2,
    INIT_FILEIO  = 1 << 3,  
    INIT_SOCKETS = 1 << 4,
    INIT_DEBUG   = 1 << 5,
    INIT_ALL     = 0xFF
  };
  
  ///
  Script( int flags = INIT_BASELIB | INIT_STRINGS | INIT_MATH,
	  int stacksize = 1024 );
  ///
  virtual ~Script();   

  /** */
  void initExtensions(int flag);

  /** execute source or compiled lua code from a file . */
  int doFile(const char* filename);
 
  /** execute source or compiled lua code from buffer.
      "name" is only used for error messages. (just like filename)
  */
  int doBuffer(const char* buffer, size_t size, const char* name = NULL);   

  /** call function with no args and no results */
  void call(const std::string& funcname);

  /** call function with n args and a single value result */
  Value call(const std::string& funcname, const ListType& args);
  
  /** call function with n args and nresult results */
  ListType call(const std::string& funcname, const ListType& args, int nresults);
  
  /** register a C-Function (static C++ function) for calling from LUA */
  void registerFunc(const std::string& funcname, Value::FunctionType func);
  
  ///
  void setGlobal(const std::string& name, const Value& val);
  ///
  Value getGlobal(const std::string& name);
  
  ///
  Value createObject(const std::string& type, const Value::TableType& );
  ///
  Value createObject(const std::string& type, void*);
  ///
  void registerType(const std::string& id);
  ///
  Value::TagType getTypeTag(const std::string& id) const;

  ///
  Value peekValue() throw (StackOverflowException, StackUnderrunException);
  ///
  Value popValue();
  ///
  void pushValue(const Value&) throw (StackOverflowException);

  /** get reference to script object.
   *  This static member is used in callback functions:
   *  
   *  static int callback(Lua::Script::Handle handle)
   *  {
   *    Lua::Script* script = Lua::Script::getRef(handle);
   */
  static Script* getRef(Handle);
  /** get size of the argument stack.
   *  Inside a callback function this value is equal to the number of arguments.
   */
  int getStacksize() const { return lua_gettop(handle_); };

 protected:
  ///
  Handle handle_;
  ///
  typedef std::map<std::string, Value::TagType> TagMap;
  ///
  TagMap typeTags_;
  ///
  typedef std::map<Handle, Script*> ScriptMap;
  ///
  static ScriptMap scripts;
  ///
  int flags_;

  
};

} // namespace

#endif
