//  The Worldforge Project
//  Copyright (C) 1998, 1999,2000,2001 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 03/2001

#ifndef _LUA_VALUE_H_
#define _LUA_VALUE_H_

extern "C"
{
  #include "lua/lua.h"
}

#include <string>
#include <vector>

namespace Lua {
      
class NoCastException { };  
class NoTableException { };
class NoSuchElementException { };

class Script;

class Value
{
 public:
  typedef double NumberType;
  typedef std::string StringType;
  typedef std::vector< std::pair<Value, Value> > TableType;
  typedef int TagType;
  typedef void* UserdataType;
  typedef int(*FunctionType)(lua_State*);
  enum type_t { NIL, NUMBER, STRING, TABLE, USERDATA, FUNCTION };

  Value() : type_(NIL) {};
  Value(NumberType val):type_(NUMBER), tag_(LUA_ANYTAG) 
    { n = new NumberType(val); };
  Value(int val):type_(NUMBER), tag_(LUA_ANYTAG) 
    { n = new NumberType(val); }; 
  Value(const char* val):type_(STRING), tag_(LUA_ANYTAG)  
    { s = new StringType(val); };
  Value(const StringType& val):type_(STRING), tag_(LUA_ANYTAG)  
    { s = new StringType(val); };
  Value(const TableType& val):type_(TABLE), tag_(LUA_ANYTAG)  
    { t = new TableType(val); };
  Value(const UserdataType& val):type_(USERDATA), tag_(LUA_ANYTAG)  
    { u = val; };
  Value(FunctionType val):type_(FUNCTION), tag_(LUA_ANYTAG)  
    { f = val; };

  Value(const Value& val)
    {
      //actually do the assignment
      type_ = val.type_;
      tag_ = val.tag_;
      
      switch(type_) 
	{
	case NIL:
	case FUNCTION:
	  f = val.f;
	  break;
	case NUMBER: 
	  n = new NumberType(*val.n);
	  break;
	case STRING: 
	  s = new StringType(*val.s);
	  break;
	case TABLE: 
	  t = new TableType(*val.t);
	  break;
	case USERDATA: 
	  u = val.u;
	  break;
	};    
    }

  Value& operator=(const Value& val)
  {
    //check for self assignment
    if(&val == this)
      return *this; 
    //clear existing data
    clear();
    //actually do the assignment
    type_ = val.type_;
    tag_ = val.tag_;
    
    switch(type_) 
	{
	case NIL:
	case FUNCTION:
	  f = val.f;
	  break;
	case NUMBER: 
	  n = new NumberType(*val.n);
	  break;
	case STRING: 
	  s = new StringType(*val.s);
	  break;
	case TABLE: 
	  t = new TableType(*val.t);
	  break;
	case USERDATA: 
	  u = val.u;
	  break;
	};    
    return *this;
  }
  
  Value& operator[](const char* id) throw(NoSuchElementException, NoTableException)
    {
      if(type_ != TABLE) throw NoTableException();
      TableType::iterator itr = t->begin();
      std::string _id = std::string(id);

     while(itr != t->end())
	{
	  if(itr->first.isString() && itr->first.asString() == _id)
	    break;
	  else
	    itr++;
	}

      if(itr == t->end())
	{
	  //create new NIL entry
    std::pair<Value, Value> entry;
	  entry.first = _id;
	  t->push_back(entry);
	};
      return itr->second;
    }

  const Value& operator[](const char* id) const throw(NoSuchElementException, NoTableException)
    {
      if(type_ != TABLE) throw NoTableException();
      TableType::const_iterator itr = t->begin();
      std::string _id = std::string(id);

      while(itr != t->end())
	{
	  if(itr->first.isString() && itr->first.asString() == _id)
	    break;
	  else
	    itr++;
	}
      if(itr == t->end()) throw NoSuchElementException();
      return itr->second;
    }    
  
  bool has(const char* id) const throw(NoTableException)
    {
      if(type_ != TABLE) throw NoTableException();
      TableType::const_iterator itr = t->begin();
      std::string _id = std::string(id);

      while(itr != t->end())
	{
	  if(itr->first.isString() && itr->first.asString() == _id)
	    break;
	  else
	    itr++;
	}

      return (itr != t->end());
    }

  Value& operator[](int _n) throw(NoTableException)
    {
      unsigned n = (unsigned) _n;
      if(type_ != TABLE) throw NoTableException();
      if(n >= t->size())
	{
	  t->resize(n+1);
	  (*t)[n].first = (int)(n+1);
	}
      return (*t)[n].second;
    }
  
  const Value& operator[](int _n) const throw(NoTableException)
    {
      unsigned n = (unsigned) _n;
      if(type_ != TABLE) throw NoTableException();
      if(n >= t->size()) throw NoSuchElementException();
      return (*t)[n].second;
    }
  
 protected:

  friend class Script;

  Value(TagType tag, void* val) : type_(USERDATA), tag_(tag) 
    { u = val; }; 
  Value(TagType tag, const TableType& val) : type_(TABLE), tag_(tag) 
    { t = new TableType(val); };
  
  void setTag(TagType tag) { tag_ = tag; };

 public:
  
  virtual ~Value() { clear(); }


  bool isNil() const { return type_ == NIL; };
  bool isNumber() const { return type_ == NUMBER; };
  bool isString() const { return type_ == STRING; };
  bool isTable() const { return type_ == TABLE; };
  bool isUserdata() const { return type_ == USERDATA; };
  bool isFunction() const { return type_ == FUNCTION; };

  type_t type() const { return type_; };
  TagType tag() const { return tag_; };

  NumberType asNumber() const  throw(NoCastException)
    {
      if(type_ != NUMBER) throw NoCastException();
      return *n;
    }
  NumberType& asNumber()  throw(NoCastException)
    {
      if(type_ != NUMBER) throw NoCastException();
      return *n;
    }

  const StringType& asString() const throw(NoCastException)
    {
      if(type_ != STRING) throw NoCastException();
      return *s;
    }
  StringType& asString() throw(NoCastException)
    {
      if(type_ != STRING) throw NoCastException();
      return *s;
    }
  
  const TableType& asTable() const  throw(NoCastException)
    {
      if(type_ != TABLE) throw NoCastException();
      return *t;
    }
  TableType& asTable()  throw(NoCastException)
    {
      if(type_ != TABLE) throw NoCastException();
      return *t;
    }

  const UserdataType& asUserdata() const  throw(NoCastException)
    {
      if(type_ != USERDATA) throw NoCastException();
      return u;
    }
  UserdataType& asUserdata()  throw(NoCastException)
    {
      if(type_ != USERDATA) throw NoCastException();
      return u;
    }

  const FunctionType& asFunction() const  throw(NoCastException)
    {
      if(type_ != FUNCTION) throw NoCastException();
      return f;
    }

  FunctionType& asFunction()  throw(NoCastException)
    {
      if(type_ != FUNCTION) throw NoCastException();
      return f;
    }
  
 protected:

  type_t type_;
  TagType tag_;

  void clear()
    {
      switch(type_) 
	{
	case NIL:
	case FUNCTION:
	  break;
	case NUMBER: 
	  delete n;
	  break;
	case STRING: 
	  delete s;
	  break;
	case TABLE: 
	  delete t;
	  break;
	case USERDATA: 
	  break;
	};
    }

  union 
  {
    NumberType* n;
    StringType* s;
    TableType* t;
    UserdataType u;
    FunctionType f;
  };

};

} // namespace

#endif



  
