-- Simple HTML Browser
--
-- Recognized tags: <br> <p> <a> <img> <h1 .. h6> <b> <i>
-- For links only the href attribute is evaluated, for
-- images only the src attribute and in both cases these
-- attributes must be the first inside the tag
--
-- Currently all relative links (especially img links) must
-- be relative to $(INSTALLDIR)/share, usually /opt/forge/share.
--
-- written by Karsten Laux, August 2001
--
-- Note:
--
-- The libuta widget which is used to actually render to content
-- to the screen currently (version 0.3.42) works with a fixed
-- lineheight, thus when embedding images the author must ensure that
-- there are enough free lines above (e.g. using <br>),
-- otherwise the image will be drawn across existing text.
-- (see help.html for an example)
--

-- globals
img_count = 0
img_table = {}
link_count = 0
link_table = {}

function addImage(img_src)
   if img_src == nil then mydebug("browser: null image source !") return end
   img_count = img_count + 1
   -- add to image link table
   img_table[img_count] = img_src
   -- be verbose
   mydebug("image link "..tostring(img_count).." : "..img_src)  

   return "\127\123"..strchar(img_count).."\126"
end

function addLink(link_src)
   if link_src == nil then mydebug("browser: null link !") return end
   link_count = link_count + 1
   -- add to link table
   link_table[link_count] = link_src
   -- be verbose
   mydebug("link "..tostring(link_count).." : "..link_src)  

    return "\127\122"..strchar(link_count)
end

function browse(arg)
   
   -- reset globals
   img_count = 0
   img_table = {}
   link_count = 0
   link_table = {}

   local prefix = DII_callService("readRegistryEntry", "/install_dir")     

   -- extract the url from the command line parameters
   --
   if type(arg) ~= "table" then return 0 end
   local url = arg[1]
   if url == nil then return 0 end

   local content = ""

   if strfind(url,"http",1,1) then
      DII_callService("std_out", "getting "..url.."... ")
      content = http_get(url)
      DII_callService("std_out", " OK.\n")
   else
      DII_callService("std_out", "reading from "..url.."...")
      -- open file for reading
      local handle = DII_callService("file_open",DII_newList({url,"r"}))
      if handle == nil then 
          DII_callService("std_out"," failed !\n")
	  return 0
      end
      -- read the complete file into 'content'
      content = DII_callService("file_read",DII_newList({handle,"*a"}))
      -- close the file again
      DII_callService("file_close",handle)
      DII_callService("std_out", " OK.\n")
   end

   -- convert all tags to lower case
   content = gsub(content,"<(%w+)", function(w) return "<"..strlower(w) end)
   content = gsub(content,"</(%w+)",function(w) return "</"..strlower(w) end)

   -- erase all newline (inject space if needed)
   content = gsub(content,"\n+","\n")
   content = gsub(content,"([.,;:!?])\n","%1 ")
   content = gsub(content,"([%w])\n","%1 ")
   content = gsub(content,"\n","");

   -- convert paragraph and break tags to newline   
   content = gsub(content,"</%a*:*p>","\n\n")
--   content = gsub(content,"<%a*:*p>", "\n")
   content = gsub(content,"<br>","\n")
      
   -- extract body 
   local dummy, body_start, body_end

   dummy, body_start = strfind(content,"<body")
   if dummy == nil then
      mydebug("browser.lua: HTML parse error !\n")
      return 0
   end
   dummy, body_start = strfind(content,">",body_start)
   if dummy == nil then
      mydebug("browser.lua: HTML parse error !\n")
      return 0
   end

   body_end, dummy = strfind(content,"</body",body_start,1)
   if body_end ~= nil then
      body_end = body_end - 1
   else
      body_end = -1
   end
   local body = strsub(content, body_start+1, body_end)

   -- IMPORTANT NOTE: lua takes decimal values for "\ddd" !!!
   --                 (in contrast to octal values in C/C++)

   -- enclose headings with empty lines and use font #1, color #1
   -- (using libuta markup sequences)
   --
   body = gsub(body,"<h(%d+)>","\n\n\127\124\001\127\125\001")
   body = gsub(body,"</h(%d+)>","\126\126\n\n")

   -- for italics use font #2, color #2
   body = gsub(body,"<i>","\127\124\002\127\125\002")
   body = gsub(body,"</i>","\126\126")

   -- for bold use font #3, color #3
   body = gsub(body,"<b>","\127\124\003\127\125\003")
   body = gsub(body,"</b>","\126\126")

   -- extract image sources and build table of images
   -- (the capture marked by () inside the pattern is passed
   --  as the argument to function addImage, which in return 
   --  return the correct escape sequence)
   body = gsub(body, '<img%s+src="([^"]+)"[^>]*>', addImage)

   -- extract links and build table of links
   -- (the capture marked by () inside the pattern is passed
   --  as the argument to function addLink, which in return 
   --  returns the correct escape sequence)
   body = gsub(body, '<a%s+href="([^"]+)"[^>]*>', addLink)
   body = gsub(body, '</a>', "\126")

   -- finally remove all remaining tags from the body text
   body = gsub(body,"<[^>]+>","")
	
   -- need to add more entities here ......
   body = gsub(body,"&nbsp;"," ")
   body = gsub(body,"&lt;","<")
   body = gsub(body,"&bg;",">")

   -- finally remove empty lines
   body = gsub(body,"\n (%s+)\n","\n")

   -- just popup a simple dialog and display the downloaded body text
   --
   style = newJanusItem("style")
   data = newJanusItem("data")
   structure = newJanusItem("structure")
   logic = newJanusItem("logic")

   data.grey = newJanusItem("Color", {id="grey", data="128 128 128 255"})
   data.darkgrey = newJanusItem("Color", {id="darkgrey", data="64 64 64 255"})
   data.black= newJanusItem("Color", {id="black", data="16 16 16 255"})
   data.blue = newJanusItem("Color", { id="blue", data="0 0 200 255"})
   
   structure.browser  = newBorderedDialog(50,50,600,450)
   structure.browser.clientArea.background = "background"

   structure.browser.clientArea.label1 = newJanusItem("Label", 
            {
              x=10, y=10, width=580, height=400, 
              scrollbar="left",
              text = body,
	      color = "grey",
              text_color = "black",
	      fade_color = "grey",
	      text_color1 = "black",
              text_color2 = "blue",
              text_color3 = "black",
              font = "text_font",
              font1 = "button_font",
              font2 = "text_font",
              font3 = "text_font",
              text_background="background",
	      arrow_up="triangle_up",
              arrow_down="triangle_down",
	      arrow_up_pressed="triangle_up_pressed",
              arrow_down_pressed="triangle_down_pressed"
            })

    -- actually load the images
    local img = 1
    while img <= img_count do
       local img_id = "image"..tostring(img)
       data[img_id] = newJanusItem("Image",{ type="file",
                                             id=img_id,
                                             source=prefix.."/"..img_table[img] })
       structure.browser.clientArea.label1[img_id]=img_id
       mydebug("loading "..prefix.."/"..img_table[img].." as "..img_id)
       img = img + 1
    end    

    structure.browser.clientArea.button1 = newJanusItem("PushButton", 
            {
              x=490, y=420, width=100, height=25, 
              text="Close"
            })
     
   logic.conn1 = newJanusItem("connection", 
                            { sender="button1",   signal="clicked",
                              receiver="browser", slot="accept" })
   logic.conn2 = newJanusItem("connection", 
                             { sender="label1",   signal="linkActivated",
                              receiver="browser", slot="browser:link_activated" })

     
   dialog = newJanusItem("janus", 
                        { style=style, 
              		  data=data,  
              		  structure=structure, 
              		  logic=logic })
   
   -- create and run the dialog
   --
   id = DII_callMethod("Gui","create_dialog", dialog)

   if id == nil then return end

   DII_callMethod("Gui","show_dialog", id)	

   local call ={ id=id, sub_id="label1", cmd="scroll2Top", arg="" }
   settag(call, Map_tag)
   DII_callMethod("Gui","call_dialog",call)

   DII_callMethod("Gui","run_dialog", id)

   DII_callMethod("Gui","destroy_dialog", id)

end

-- this function gets called when user clicks on a hyper link
--
function link_activated(arg)
   mydebug("browser: link "..tostring(arg.value).." activated.")
   local url = link_table[arg.value]
   if url == nil then mydebug("browser: internal error !") return 0 end
   
   mydebug("browser: going to open "..url.." ...")
   return 1
end

-- global
window_id = nil

-- Retrieve content from the given url.
-- supported protocols: file, http
--
-- Note: retrieve() returns the preprocessed body: all
-- tags are lowercase and all newline characters have
-- already been removed.
--
function retrieve(url)

   local content = ""

   if strfind(url,"http",1,1) then
      DII_callService("std_out", "getting "..url.."... ")
      content = http_get(url)
      DII_callService("std_out", " OK.\n")
   else
      DII_callService("std_out", "reading from "..url.."...")
      -- open file for reading
      local handle = DII_callService("file_open",DII_newList({url,"r"}))
      if handle == nil then 
          DII_callService("std_out"," failed !\n")
	  return 0
      end
      -- read the complete file into 'content'
      content = DII_callService("file_read",DII_newList({handle,"*a"}))
      -- close the file again
      DII_callService("file_close",handle)
      DII_callService("std_out", " OK.\n")
   end
   -- convert all tags to lower case
   content = gsub(content,"<(%w+)", function(w) return "<"..strlower(w) end)
   content = gsub(content,"</(%w+)",function(w) return "</"..strlower(w) end)

   -- erase all newline (inject space if needed)
   content = gsub(content,"\n+","\n")
   content = gsub(content,"([.,;:!?])\n","%1 ")
   content = gsub(content,"([%w])\n","%1 ")
   content = gsub(content,"\n","");

   -- convert paragraph and break tags to newline   
   content = gsub(content,"</%a*:*p>","\n\n")
--   content = gsub(content,"<%a*:*p>", "\n")
   content = gsub(content,"<br>","\n")
      
   -- extract body 
   local dummy, body_start, body_end

   dummy, body_start = strfind(content,"<bod",1)
   dummy, body_start = strfind(content,">",body_start)
   if dummy == nil then
      mydebug("browser.lua: HTML parse error !\n")
      return 0
   end

   body_end, dummy = strfind(content,"</body",body_start,1)
   if body_end ~= nil then
      body_end = body_end - 1
   else
      body_end = -1
   end

   local body = strsub(content, body_start+1, body_end)

   return body
end

-- Create the window to show the content
--
function create_window()
   style = newJanusItem("style")
   data = newJanusItem("data")
   structure = newJanusItem("structure")
   logic = newJanusItem("logic")

   data.grey = newJanusItem("Color", {id="grey", data="128 128 128 255"})
   data.darkgrey = newJanusItem("Color", {id="darkgrey", data="64 64 64 255"})
   data.black= newJanusItem("Color", {id="black", data="16 16 16 255"})
   data.blue = newJanusItem("Color", { id="blue", data="0 0 200 255"})
   
   structure.browser  = newBorderedDialog(50,50,600,450)
   structure.browser.clientArea.background = "background"
   structure.browser.clientArea.label1 = newJanusItem("Label", 
            {
              x=10, y=10, width=580, height=400, 
              scrollbar="left",
              text = "getting url ...",
	      color = "grey",
              text_color = "black",
	      fade_color = "grey",
	      text_color1 = "black",
              text_color2 = "blue",
              text_color3 = "black",
              font = "text_font",
              font1 = "button_font",
              font2 = "text_font",
              font3 = "text_font",
              text_background="background",
	      arrow_up="triangle_up",
              arrow_down="triangle_down",
	      arrow_up_pressed="triangle_up_pressed",
              arrow_down_pressed="triangle_down_pressed"
            })
    structure.browser.clientArea.button1 = newJanusItem("PushButton", 
            {
              x=490, y=420, width=100, height=25, 
              text="Close"
            })
     
   logic.conn1 = newJanusItem("connection", 
                            { sender="button1",   signal="clicked",
                              receiver="browser", slot="accept" })
   logic.conn2 = newJanusItem("connection", 
                             { sender="label1",   signal="linkActivated",
                              receiver="browser", slot="browser:link_callback" })

     
   dialog = newJanusItem("janus", 
                        { style=style, 
              		  data=data,  
              		  structure=structure, 
              		  logic=logic })
   -- create and run the dialog
   --
   id = DII_callMethod("Gui","create_dialog", dialog)

   return id
end

-- Update window content.
-- The second argument is meant to contain the prepared
-- html code of the body. 
-- (so: yes, the translating is done here !)
--
function update_window(window_id, body)

   -- reset globals
   img_count = 0
   img_table = {}
   link_count = 0
   link_table = {}

   -- IMPORTANT NOTE: lua takes decimal values for "\ddd" !!!
   --                 (in contrast to octal values in C/C++)

   -- enclose headings with empty lines and use font #1, color #1
   -- (using libuta markup sequences)
   --
   body = gsub(body,"<h(%d+)>","\n\n\127\124\001\127\125\001")
   body = gsub(body,"</h(%d+)>","\126\126\n\n")

   -- for italics use font #2, color #2
   body = gsub(body,"<i>","\127\124\002\127\125\002")
   body = gsub(body,"</i>","\126\126")

   -- for bold use font #3, color #3
   body = gsub(body,"<b>","\127\124\003\127\125\003")
   body = gsub(body,"</b>","\126\126")

   -- extract image sources and build table of images
   -- (the capture marked by () inside the pattern is passed
   --  as the argument to function addImage, which in return 
   --  return the correct escape sequence)
   body = gsub(body, '<img%s+src="([^"]+)"[^>]*>', addImage)

   -- extract links and build table of links
   -- (the capture marked by () inside the pattern is passed
   --  as the argument to function addLink, which in return 
   --  returns the correct escape sequence)
   body = gsub(body, '<a%s+href="([^"]+)"[^>]*>', addLink)
   body = gsub(body, '</a>', "\126")

   -- finally remove all remaining tags from the body text
   body = gsub(body,"<[^>]+>","")
  
   -- need to add more entities here ......
   body = gsub(body,"&nbsp;"," ")
   body = gsub(body,"&lt;","<")
   body = gsub(body,"&bg;",">")

   -- finally remove empty lines
   body = gsub(body,"\n (%s+)\n","\n")

   -- need to load the references images 
   -- (by using a Janus description only containing a data entry)

   -- need to tell the hypertext widget about the updated
   -- id <-> imagename mapping
   -- (libJanus can't do this currently !)

end

-- User has clicked on a hyperlink !
-- Get the selected url and update the window content.
--
function link_callback(arg)
   mydebug("browser: link "..tostring(arg.value).." activated.")
   local url = link_table[arg.value]
   if url == nil then mydebug("browser: internal error !") return 0 end
   
   -- get new content ...
   local content = retrieve(url)
   -- and update the window
   update_window(window_id, content)
   return 1
end

-- Execute the dialog.
-- As we have constructed a modal dialog, this function
-- returns when the user closes the window.
--
function exec_dialog()
   DII_callMethod("Gui","run_dialog", id)
   DII_callMethod("Gui","destroy_dialog", id)
end

-- Usercommand, called from the commandline
--
function browser(arg)

   -- extract the url from the command line parameters
   if type(arg) ~= "table" then return 0 end
   local url = arg[1]
   if url == nil then return 0 end

   -- create window
   window_id = create_window()
   -- get content
   local content = retrieve(url)
   -- update content to window
   update_window(content) 
   -- run the dialog window
   exec_dialog()
   -- everything is fine :)
   return 1
end

-- init this module
function init()

   loadExtension("http.lua")
   loadExtension("dii.lua")
   loadExtension("janus.lua")

   return 1
end

-- shutdown the module
function deinit()
   return 1
end
		
-- all following lines are "global" code and executed as soon
-- as the script has been loaded and compiled
--
-- the semantics of the calls are the same as for binary modules,
-- just the name of the module is omitted
-- (see README.modules)
--
REGISTER_METHOD("init")
REGISTER_METHOD("deinit")
REGISTER_METHOD("browse")
REGISTER_METHOD("link_activated")

REGISTER_AS_SERVICE("browse")

NEEDS_MODULE("Gui")















