 //  The Worldforge Project
//  Copyright (C) 1998, 1999, 2000 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 03/2000


#include "tilestretch.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "terrain.h"
#include "coord.h"
#include "my_time.h"
#include "world.h"

#include "media/sprite_cache.h"
#include "media/media_index.h"

#include "registry/registry.h"
#include "modules/console.h"
#include "modules/viewport.h"

#include "dii/module.h"

#include <varconf/Config.h>

#include <uta/debug.h>
#include <uta/progressdialog.h>
#include <uta/painter.h>
#include <uta/application.h>
#include <uta/directsurface.h>
#ifndef HAVE_UTA_LIGHTEN_DARKEN
#include <uta/blitters.h>
#endif

#include "operation.h"
#include "dii/serialize.h"
#include "toolbox/isodefs.h"

#include <Atlas/Codecs/XML.h>
#include <Atlas/Objects/Decoder.h>
#include <Atlas/Objects/Encoder.h>

#include <fstream>


/* maximum map area 
 * 500 tiles * 500 tiles -> 500m * 500m -> 250000 sqrmeters
 * */
#define MAP_SIZE_MAX 500
#define SLOPE_SCALE 1.5

/* we support tiles consisting of max MAX_SUBTILES * MAX_SUBTILES standard tiles */
#define MAX_SUBTILES 8
#define MAX_EDGE (MAX_SUBTILES+1)*(MAX_SUBTILES+1)

/* how to retrieve the height at position x,y ?
 */
#define GET_HEIGHT(x,y) World::World::inst()->getGroundLevel(x,y);

// ISOEDIT hacks ...
#undef ISOEDIT_IGNORE_OFFSET 
#define ISOEDIT_FACTOR 10000.0

using namespace DII;
using namespace uta;

//the optimized scroller does not yet work ...
//once it works, we will undef this permanently :)
#define DUMB_SCROLL

/* how many pixels outside viewport are we looking for tiles ? */
#define GRACE 300

/* The size of our viewport window, in pixels. */
#define OVERVIEW_WIDTH 600
#define OVERVIEW_HEIGHT 400

/* Trigraphs can be useful! */
#define MAX(a,b) (((a) >= (b)) ? (a) : (b))
#define MIN(a,b) (((a) <= (b)) ? (a) : (b))


uta::Color fillColor(78,104,65);

Terrain* Terrain::instance_ = 0;

Terrain::Terrain():
  cache_(NULL),
  target_(NULL),
  prepared_(false),
  height_(0),
  defaultTile_(NULL),
  shadows_(NULL),
  use_shadows_(true),
  heightmap_(NULL),
  terrain_mode_(32),
  terrain_shadow_(1.5)
{
  assert(!instance_);
  instance_ = this;
  
  range_.imin = -100;
  range_.jmin = -100;
  range_.imax = 100;
  range_.jmax = 100;

  splitX_ = 0;
  splitY_ = 0;

  aw = 0;
  ax = 0;
  
  bw = cw = ah = bh = ch = 0; 

  overview_ = new uta::Surface(OVERVIEW_WIDTH, OVERVIEW_HEIGHT);
  overview_->convert(uta::Pixelformat::DISPLAY);
  overview_->setTransparency(false);
  
  markedView_ = new uta::Surface(OVERVIEW_WIDTH, OVERVIEW_HEIGHT);
  markedView_->convert(uta::Pixelformat::DISPLAY);
  markedView_->setTransColor(uta::black);
  markedView_->setTransparency(true);

  terrainSlope_ = new char[(MAP_SIZE_MAX+1)*(MAP_SIZE_MAX+1)];
  assert(terrainSlope_);

}



Terrain::~Terrain()
{
  assert(instance_);
  instance_ = 0;

  if(overview_)
    delete overview_;
  if(markedView_)
    delete markedView_;

  /* free loaded tiles */
  std::map<std::string, uta::Surface*>::iterator itr = tiles_.begin();
  while(itr != tiles_.end())
    {
      delete itr->second;
      itr++;
    }

  delete[] terrainSlope_;
}

void 
Terrain::configure()
{
  bool refresh = false;
  DII::GetOption("Terrain", "use_shadows",use_shadows_);
  
  refresh |= DII::GetOption("Terrain", "terrain_mode",terrain_mode_);
  refresh |= DII::GetOption("Terrain", "terrain_shadow",terrain_shadow_);
    
  /* if terrain drawing options have changed schedule an imediately redraw
   */
  if(refresh)
    prepare();
}

void
Terrain::deinit()
{
  if(cache_)
    {
      delete cache_;
      cache_ = NULL;
    }

  if(shadows_)
    {
      delete[] shadows_->values;
      delete shadows_;
      shadows_ = NULL;
    }

  if(defaultTile_)
    {
      delete defaultTile_;
      defaultTile_ = NULL;
    }

  DII::SetOption("Terrain", "use_shadows",use_shadows_);
  DII::SetOption("Terrain", "terrain_mode",terrain_mode_);
  DII::SetOption("Terrain", "terrain_shadow",terrain_shadow_);
}

bool
Terrain::init(DII::Module*)
{ 
  if(!shadows_)
    {
      /* uta::GammaFunction only supports gamma maps at 256*256 pixels ! */
      shadows_ = new uta::GammaFunction;
      shadows_->offset_x = 0;
      shadows_->offset_y = 0;
      shadows_->values = new char[256*256];

      char* ptr = shadows_->values;
      for(size_t n=0; n < 256*256; n++)
          *ptr++ = 0;
    }

  std::string install_dir = Registry::instance()->readInstallDir();
  std::string filename = install_dir + "/uclient/types.xml";
  
  
  debugN(12,cerr << "Terrain: parsing " << filename << "... ";);
  if(DII::Serialize::read(filename.c_str(), typesDB_))
    {
      debugN(12,
             cerr <<"\t"<<typesDB_.AsMap().size()
             << " sprite definitions." 
             << endl;
             );
    }
  else
    {
      debugN(12,cerr << "Error on parsing data file !" << endl;);
      typesDB_ = DII::Object::MapType();
    }


  configure();

  return true;
}

const uta::Surface* 
Terrain::getTerrainMap()
{
  return overview_;
}

const uta::Surface* 
Terrain::getMap()
{
  if(overview_)
    {
      overview_->blit(markedView_);
      
      //mark current viewport in overview

      Painter* painter = new Painter(markedView_);
      
      Point p1,p2,p3,p4;
      
      p1 = Point(range_.imin/(TILEWIDTH/2)+overview_->width()/2, 
                 range_.jmin/TILEHEIGHT+overview_->height()/2);
      p2 = Point(range_.imin/(TILEWIDTH/2)+overview_->width()/2, 
                 range_.jmax/TILEHEIGHT+overview_->height()/2);
      p3 = Point(range_.imax/(TILEWIDTH/2)+overview_->width()/2, 
                 range_.jmax/TILEHEIGHT+overview_->height()/2);
      p4 = Point(range_.imax/(TILEWIDTH/2)+overview_->width()/2, 
                 range_.jmin/TILEHEIGHT+overview_->height()/2);

      painter->setColor(blue);
      painter->line(p1,p2);
      painter->line(p2,p3);
      painter->line(p3,p4);
      painter->line(p4,p1);
       
      delete painter;

      return markedView_;
    }
  return 0;
}


void
Terrain::setTarget(uta::Surface* target)
{
  
  target_ = target;
  if(target_)
    {
      if(cache_)
        delete cache_;

      //create identically surface
      //same size, same pixelformat as target
      
      /* though I have written libuta I am unsure 
         what happens here if target_ points
         to a "DirectSurface" - meaning if its pixeldata is in VRAM (?!)
      
         cache_ = new Surface(*target_);
      
      -> this should be safer, as we do not want to copy the content: */
      cache_ = new Surface(target_->width(), target_->height(), 
                           uta::Pixelformat::RGB565);
      assert(cache_);

      //we want to kill original content of target, thus
      //we disable transparency for cache_
      cache_->setTransparency(false);

      prepared_ = false;
    }
}

bool
Terrain::prepare()
{
  
  if(!cache_)
    return false;

#ifdef DEBUG
  my_time start; start.set_now();
#endif
  
  Rect rect;
  ViewRect range;
  pixel_t w = cache_->width();
  pixel_t h = cache_->height();

  //if no dirtyRects are specified, redraw everything
  if(dirtyRects_.empty())
    drawArea(range_, Rect(0,0,w,h));
  else
    while(!dirtyRects_.empty())
      {
        rect = dirtyRects_.front();
        dirtyRects_.pop();
        range = dirtyRanges_.front();
        dirtyRanges_.pop();

//         range.imin = range_.imin + rect.upperLeft().x;
//         range.imax = range.imin + rect.width();
//         range.jmin = range_.jmin + rect.upperLeft().y;
//         range.jmax = range.jmin + rect.height();

        debugN(12,cerr << "rendering rect: " << rect << endl;);
        debugN(12,cerr << "upperleft corner of rect     : ("
               <<range.imin <<","<<range.jmin<<")" <<endl;);
        debugN(12,cerr << "upperleft corner of viewport : " 
               <<"("<<range_.imin <<","<<range_.jmin<<")"<<endl;);

        debugN(12,cerr << "ax,aw,ay,ah"<<endl
               <<ax<<","<<aw<<","<<ay<<","<<ah<<endl
               <<"bx,bw,by,bh"<<endl
               <<bx<<","<<bw<<","<<by<<","<<bh<<endl
               <<"cx,cw,cy,ch"<<endl
               <<cx<<","<<cw<<","<<cy<<","<<ch<<endl;);  

        drawArea(range, rect);
        
  }

#ifdef DEBUG
  my_time end; end.set_now(); 
  debugN(12,cerr << "Terrain::prepare() took: "<<end() - start()
         << " ms."<<endl;);
#endif

  prepared_ = true;

  return true;
  
}

/* Streching blit from surface tile to surface target.
 *
 * Note: all coordinates are given in game world coordinates, but
 *       converted to pixel coordinates following the isometric
 *       perspective.
 *
 * i,j                are the coordinates of the tile's baseline.
 * offset_i, offset_j are the pixel coordinates referring to the
 *                    target's pixel at (0,0)
 *
 */
void
Terrain::textureTile(const uta::Surface* tile, uta::Surface* target, 
                     pixel_t i, pixel_t j, pixel_t offset_i, pixel_t offset_j)
{
  
  // this function only works on 16bit pixel formats (RGB555, RGB565)
  // bpp is BYTES per pixel, not bit
  assert(tile);
  if (tile->pixelformat().bpp() != 2) {
          std::cerr << "Bytes Per Pixel of " << tile->width() << "x"
		<< tile->height() << " tile " << (void*)tile << " is "
		<< tile->pixelformat().bpp() << std::endl;
         }
  assert(tile->pixelformat().bpp() == 2);
  assert(target->pixelformat().bpp() == 2);
  assert(tile->width());
  assert(tile->height());

  // calculate screen destination of tile base center
  Pixel dest;
  dest.i = i - offset_i;
  dest.j = j - offset_j;
   
  /* Check for the size of this tile ...
   * We also need to check for irregular tiles. Those tiles
   * have objects on them (e.g. grass) which needs to get treated
   * seperately when stretching the tile.
   *
   * This assumes that the tiles always have the standard tile width.
   */
  int n = tile->width() / (TILEWIDTH-2);  
  
  // normalize to standard tilesize
  int w = tile->width();
  int h = n*TILEHEIGHT;  

  /* if dh > 0 then there is something on the tile (e.g. stones, grass)
   * if dh == 0 then this is a regular tile
   * if dh < 0 then this tile a smaller than usual tile 
   */
  int dh = tile->height() - h;

  /* check for non-regular tiles: 
   * if unsure set dh = 0
   */
  if(dh < 0)
    dh = 0;
    

  /* check if the desired target area is inside the target surface,
   * if not, we have nothing to do and may leave ... 
   */
  if(dest.i+w/2 < 0)
    return;
  if(dest.i-w/2 > target->width())
    return;
  
  /* Convert pixel coordinates back to world coordinates (s.a. toolbox/isodefs.h)
   * x,y are the absolute game world coordinates
   * At this point we are not aware of any containers   
   */
  int x = static_cast<int>(round((i / (TILEWIDTH/2.0) - j / (TILEHEIGHT/2.0)) / 2.0));
  int y = static_cast<int>(round((-i / (TILEWIDTH/2.0) - j / (TILEHEIGHT/2.0)) / 2.0));   
  
  /* Arrays to hold the values for the tile edges. All values are
   * relative to the center of the tile baseline are have the same
   * orientation as the pixel coordinate on the screen.
   * (meaning x running left to right, y running top to bottom)
   *
   * xn, bn : x,y of unstretched tile
   * xn, hn : x,y of stretched tile
   */
     
  static int xn[MAX_EDGE];
  static int bn[MAX_EDGE];
  static int hn[MAX_EDGE];
  static char sn[MAX_EDGE];

  assert((n+1)*(n+1) <= MAX_EDGE);

  /* calculate these values for all edges of the tile itself or 
   * the subtiles.
   */
  assert(n>0);

  int h_min = 0;
  int h_max = 0;
  int h_temp, h_save = GET_HEIGHT(x,y);
  bool plain = true;

  /* Calculate the edge of the subtiles
   */ 
  for(int sx=0; sx<n+1; sx++)
    for(int sy=0; sy<n+1; sy++)
      {
        xn[sx+sy*(n+1)] = 34 + (sx-sy+(n-1))*36;  
        bn[sx+sy*(n+1)] = h - (sx+sy)*18 - 1;
        h_temp = GET_HEIGHT(x+sx, y+sy);
        hn[sx+sy*(n+1)] = h - (sx+sy)*18 - h_temp - 1;

        sn[sx+sy*(n+1)] = terrainSlope_[(x+sx+MAP_SIZE_MAX/2)+(y+sy+MAP_SIZE_MAX/2)*MAP_SIZE_MAX];

        /* check if this tile is really plain */
        if(h_temp != h_save)
          plain = false;

        if(h_temp < h_min)
          h_min = h_temp;
        if(h_temp > h_max)
          h_max = h_temp;
      }

  /* check if the heightmapped tile is still visible ...
   * if not return immediately
   */
  if(dest.j-h-h_max-dh > target->height())
    return;
  if(dest.j < 0)
    return;
  
  /* do the actual stretching ... */
  if(terrain_mode_ & 32)
    {
      static Stretch_SrcSurface sourceSurface;
      static Stretch_Args args;
      static Stretch_TargetSurface targetSurface;

      /* calculate screen offset using the normalized tile sizes
       * dx,dy are the upper left corner of a rectangle inside the target
       * surface which will embrace the stretched tile.
       */
      int dx = dest.i - w/2;
      int dy = dest.j - h;
      
      Stretch_setShadingTarget(shadows_->values);

      /* With SDL (and libuta) we need to lock the surfaces in order
       * to gain access to the pixel data.
       */
      target->lock();
      tile->lock();
 
      sourceSurface.pixels = (uint8*)tile->pixels();
      sourceSurface.pitch = tile->pitch();
      
      // extract transparent color from the tile - using color of pixel(0,0)
      sourceSurface.transColor =  *((Uint16*)tile->pixels());
      
      
    
      targetSurface.pixels = (uint8*)target->pixels();
      targetSurface.pitch = target->pitch();
      targetSurface.width = target->width();
      targetSurface.height = target->height();

      Stretch_setTarget(&targetSurface);
      
      Stretch_setSource(&sourceSurface);
 
      /* treat every subtile seperately
       */
      for(int tx=0; tx<n; tx++)
        {
          for(int ty=0; ty<n ; ty++)
            {
              // x0, b2
              sourceSurface.offset_x = xn[tx+(ty+1)*(n+1)]+2;
              sourceSurface.offset_y = bn[tx+1+(ty+1)*(n+1)]+1+dh;
              
              
              args.h0 = hn[tx+(ty)*(n+1)];
              args.h1 = hn[tx+1+(ty)*(n+1)];
              args.h2 = hn[tx+1+(ty+1)*(n+1)]+1;
              args.h3 = hn[tx+(ty+1)*(n+1)];
              
              args.sl0 = sn[tx+(ty)*(n+1)];
              args.sl1 = sn[tx+1+(ty)*(n+1)];
              args.sl2 = sn[tx+1+(ty+1)*(n+1)];
              args.sl3 = sn[tx+(ty+1)*(n+1)];
              
              args.offset_x = sourceSurface.offset_x + dx;
              args.offset_y = dy;
              
              Stretch_blit(&args);

              if(tx == n-1)
                RightOverlap_blit(&args, dh);

            } // for (int ty= ...)

          LeftOverlap_blit(&args, dh);

        } // for(int tx=..)
  
      /* We are done !
       * Unlock the surface so that others may access the pixel data again.
       */
      target->unlock();
      tile->unlock();

      
    }  // if (terrain_mode_ & 32)
  else
    {
      
      // draw tile on baselevel
      Rect rect(dest.i-tile->width()/2,dest.j-tile->height(),tile->width(),tile->height()); 
      tile->blit(target,rect);
    }

  
  if(terrain_mode_ & 16)
    {
      // draw wire frame     
      Painter* painter = new Painter(target);
            
      // calculate screen offset using the normalized tile sizes
      int dx = dest.i - w/2;
      int dy = dest.j - h;
      
      
      if(terrain_mode_ & 1)
        {
          // read out the height map and add the apropriate pixel offset for this tile
          int h0 = h - GET_HEIGHT(x,y);
          int h1 = h/2 - GET_HEIGHT(x+n,y);
          int h2 = - GET_HEIGHT(x+n,y+n);
          int h3 = h/2 - GET_HEIGHT(x,y+n);
          /* x-values
           */
          int x0 = w/2;
          int x1 = w;
          int x2 = w/2;
          int x3 = 0;
          /* y-values for baselevel
           */
          int b0 = h;
          int b1 = h/2;
          int b2 = 0;
          int b3 = h/2;

          // draw heightbar at the tile corners (blue)
          painter->setColor(Color(16,16,128));
          painter->line(Point(x2+dx, b2+dy), Point(x2+dx, h2+dy));
          painter->setColor(Color(32,32,255));
          painter->line(Point(x0+dx, b0+dy), Point(x0+dx, h0+dy));
          painter->line(Point(x1+dx, b1+dy), Point(x1+dx, h1+dy));
          painter->line(Point(x3+dx, b3+dy), Point(x3+dx, h3+dy));
          
        }

//       // draw boundings of stretched tile (white)
//       painter->setColor(Color(200,255,255));
//       // h3 - h2
//       painter->line(Point(x3+dx, h3+dy), Point(x2+dx, h2+dy));
//       // h3 - h0
//       painter->line(Point(x3+dx, h3+dy), Point(x0+dx, h0+dy));
//       // h2 - h1
//       painter->line(Point(x2+dx, h2+dy), Point(x1+dx, h1+dy));
//       // h0 - h1
//       painter->line(Point(x0+dx, h0+dy), Point(x1+dx, h1+dy));

      if(terrain_mode_ & 2)
        {
          
          // draw cut lines of sub-tiles (white)
          painter->setColor(Color(150,150,150));
         
          for(int tx=0; tx<n; tx++)
            for(int ty=0; ty<n ; ty++)
              {
                int x0 = xn[tx+(ty)*(n+1)];
                int x1 = xn[tx+1+(ty)*(n+1)]-1;
                int x2 = xn[tx+1+(ty+1)*(n+1)];
                int x3 = xn[tx+(ty+1)*(n+1)]+2;
                int h0 = hn[tx+(ty)*(n+1)];
                int h1 = hn[tx+1+(ty)*(n+1)];
                int h2 = hn[tx+1+(ty+1)*(n+1)]+1;
                int h3 = hn[tx+(ty+1)*(n+1)];

                // h3 - h2
                painter->line(Point(x3+dx, h3+dy), Point(x2+dx, h2+dy));
                // h3 - h0
                painter->line(Point(x3+dx, h3+dy+1), Point(x0+dx, h0+dy));
                // h2 - h1
                painter->line(Point(x2+dx, h2+dy), Point(x1+dx, h1+dy));
                // h0 - h1
                painter->line(Point(x0+dx, h0+dy), Point(x1+dx, h1+dy+1));                            
              }
        }

      delete painter;
    }
  
}


void
Terrain::drawArea(ViewRect range, Rect rect)
{
  //clear area
  cache_->fill(rect, fillColor);
  cache_->setTransparency(false);
      
  pixel_t offset_i;
  pixel_t offset_j;

  offset_i = range.imin - rect.upperLeft().x;
  offset_j = range.jmin - rect.upperLeft().y;
  
  //now blit the tiles to the cache surface

  TileStripe::iterator i_itr, i_begin, i_end;
  TileMap::iterator j_itr, j_begin, j_end;
  TileStack::iterator z_itr;

  const uta::Surface* tile;
  //uta::Rect src;
  int i,j,w,h,u,v;


  //get pointers to begin-2 and end+2 of visible j range
  
  j_begin = tileMap_.lower_bound(range.jmin-GRACE);

  j_end = tileMap_.upper_bound(range.jmax+GRACE+height_);

  j_itr = j_begin;
  
  uta::Rect _rect = uta::Rect(rect.upperLeft().x - GRACE,
                              rect.upperLeft().y - GRACE,
                              rect.width() + GRACE,
                              rect.height() + GRACE + height_);

  while(j_itr != j_end)
    {  
      //get pointers to begin/end of visible i range
      i_begin = j_itr->second.lower_bound(range.imin-GRACE);
      
      i_end = j_itr->second.upper_bound(range.imax+GRACE);
 
     
      i_itr = i_begin;
      
      while(i_itr != i_end)
        {
          z_itr = i_itr->second.begin();
          while(z_itr != i_itr->second.end())
            {
              tile = z_itr->second; 
        
              //width and height of tile
              w = tile->width(); 
              h = tile->height();
              //position
              i = i_itr->first - offset_i - w/2;
              j = j_itr->first - offset_j - h;
              if(terrain_mode_)
                {
                  if (w + h > 0) 
                  {
                          textureTile(tile, cache_, i_itr->first,j_itr->first, offset_i, offset_j);
                  }
                }
              else
                if(uta::Rect(i,j,w,h).intersect(_rect).isValid())
                  {
 
                    {      
                      // if(i >= 0 && j >= 0 && i+w<rw && j+h<rh )
                      if(i >= 0 && j >= 0)
                        {
                          tile->blit(cache_, uta::Rect(i,j,w,h));
                        }
                      else
                        {
                          u = v = 0;
                          //need clipping for upper and/or left border
                          if(i < 0)
                            {
                              w = w + i;
                              u = -i;
                              i = 0;
                            }
                          if(j < 0)
                            {
                              h = h + j;
                              v = -j;
                              j = 0;
                            }
                          tile->blit(cache_, uta::Rect(i,j,w,h),uta::Rect(u,v,w,h));
                        } // if( ... isValid())
                    }
                } // if(terrain_mode_)
              z_itr++;
            }
          i_itr++;
        }

      j_itr++;
    }
  
}



bool
Terrain::render()
{
  
  if(!target_ || !cache_)
    return false;

  if(!prepared_)
    prepare();

  // target, destination, source
//   if(ah)
//     {
//       if(aw)
//         cache_->blit(target_,Rect(0,0,aw,ah), Rect(ax,ay,aw,ah));
//       if(bw)
//         cache_->blit(target_,Rect(aw,0,bw,ah), Rect(bx,ay,bw,ah));
//       if(cw)
//         cache_->blit(target_,Rect(aw+bw,0,cw,ah), Rect(cx,ay,cw,ah));
//     }

//   if(bh)
//     {
//       if(aw)
//         cache_->blit(target_,Rect(0,ah,aw,bh), Rect(ax,by,aw,bh));
//       if(bw)
//         cache_->blit(target_,Rect(aw,ah,bw,bh), Rect(bx,by,bw,bh));
//       if(cw)
//         cache_->blit(target_,Rect(aw+bw,ah,cw,bh), Rect(cx,by,cw,bh));
//     }

//   if(ch)
//     {
//       if(aw)
//         cache_->blit(target_,Rect(0,ah+bh,aw,ch), Rect(ax,cy,aw,ch));
//       if(bw)
//         cache_->blit(target_,Rect(aw,ah+bh,bw,ch), Rect(bx,cy,bw,ch));
//       if(cw)
//         cache_->blit(target_,Rect(aw+bw,ah+bh,cw,ch), Rect(cx,cy,cw,ch));
//     }


#ifndef HAVE_UTA_LIGHTEN_DARKEN
  uta::gammablit(target_, cache_, shadows_, 
                 uta::Rect(0,0,cache_->width(), cache_->height()), uta::Point(0,0));
#else
  cache_->blit(target_);
  target_->lighten_darken(shadows_);
#endif


  return true;
}

void Terrain::view(const Pixel &pix)
{
      DII::Object::ListType ul, lr;
      
      ul.push_back((int)(pix.i-cache_->width()/2));
      ul.push_back((int)(pix.j-cache_->height()/2));
  
      lr.push_back((int)(pix.i+cache_->width()/2));
      lr.push_back((int)(pix.j+cache_->height()/2));
      
      DII::Object::MapType param;
      param["upperleft"] = ul;
      param["lowerright"] = lr;
}   

void
Terrain::setVisibleRange(const ViewRect& range)
{
  ViewRect oldRange_ = range_;

  range_ = range;

  pixel_t w = 0;
  pixel_t h = 0;
  
#ifdef DUMB_SCROLL
  if(cache_)
    {
      w = cache_->width();
      h = cache_->height();
    }
  aw = w; ax = 0;
  ah = h; ay = 0;
  prepared_ = false;

  return;
#endif

  if(prepared_ && cache_)
    {
      //we have already drawn something ...
      //calculate delta... 
 
      // deltaX > 0 means viewport has scrolled to the right
      deltaX_ = range_.imin - oldRange_.imin;
      // deltaY > 0 means viewport has scrolled down
      deltaY_ = range_.jmin - oldRange_.jmin;

      w = cache_->width();
      h = cache_->height();
      
      aw = w; ax = 0;
      ah = h; ay = 0;
      
      bw = cw = bh = ch = 0;

      debugN(12,cerr << "deltaX: "<<deltaX_<<"  deltaY: "<<deltaY_<<endl;);

      //out of range
      if(abs(deltaX_) >= w || abs(deltaY_) >= h)
        {                
          splitX_ = 0;
          deltaX_ = w;
          splitY_ = 0;
          deltaY_ = h;
          debugN(12,cerr << "Terrain: cache flushed."<<endl;);
        }
      else
        {

          //warp around...
          oldSplitX_ = splitX_;
          splitX_ += deltaX_;

          if(splitX_ < 0)
            splitX_ += w;
          if(splitX_ >= w)
            splitX_ -= w;
          
          oldSplitY_ = splitY_;
          splitY_ += deltaY_;

          if(splitY_ < 0)
            splitY_ += h;
          if(splitY_ >= h)
            splitY_ -= h;

          debugN(12,cerr << "splitX: "<<splitX_<<"   splitY: "<<splitY_<<endl;);

          /* now for the hard part:
             1) calculate the 9 rectangles,
             2) check which need to get redrawn
             3) calculate target positions.
          */
          
          Rect dirtyX, dirtyY;
          ViewRect rangeX, rangeY;

          if(deltaX_)
            {
              if(deltaX_ > 0)
                {
                  //realize warp around buffer for x
                  if(oldSplitX_ < splitX_)
                    { 
                      ax = splitX_;
                      aw = w - splitX_;
                      bx = 0;
                      bw = oldSplitX_;
                      
                      //C-area needs to be redrawn !
                      cx = oldSplitX_;
                      cw = splitX_ - oldSplitX_;
                      
                      rangeX.imin = range_.imin + aw + bw;
                      rangeX.imax = range_.imin + aw + bw + cw;
                      rangeX.jmin = range_.jmin;
                      rangeX.jmax = range_.jmax;
                      dirtyX = Rect(cx,0,cw,h);
                    }
                  else
                    { 
                      //A-area needs to be redrawn !
                      ax = splitX_;
                      aw = oldSplitX_ - splitX_;
                      
                      rangeX.imin = range_.imin + cw;
                      rangeX.imax = range_.imin + cw +aw;
                      rangeX.jmin = range_.jmin;
                      rangeX.jmax = range_.jmax;
                      dirtyX = Rect(ax,0,aw,h);
                      
                      bx = oldSplitX_;
                      bw = w - oldSplitX_;
                      cx = 0;
                      cw = splitX_;
                    }
                }
              else  //(if deltaX_ < 0)
                {
                  //realize warp around buffer for x
                  if(oldSplitX_ < splitX_)
                    { 
                      ax = splitX_;
                      aw = w - splitX_;
                      bx = 0;
                      bw = oldSplitX_;
                      
                      //C-area needs to be redrawn !
                      cx = oldSplitX_;
                      cw = splitX_ - oldSplitX_;
                      
                      rangeX.imin = range_.imin + aw + bw;
                      rangeX.imax = range_.imin + aw + bw + cw;
                      rangeX.jmin = range_.jmin;
                      rangeX.jmax = range_.jmax;
                      dirtyX = Rect(cx,0,cw,h);
                    }
                  else
                    { 
                      //A-area needs to be redrawn !
                      ax = splitX_;
                      aw = oldSplitX_ - splitX_;
                      
                      rangeX.imin = range_.imin + cw;
                      rangeX.imax = range_.imin + cw +aw;
                      rangeX.jmin = range_.jmin;
                      rangeX.jmax = range_.jmax;
                      dirtyX = Rect(ax,0,aw,h);
                      
                      bx = oldSplitX_;
                      bw = w - oldSplitX_;
                      cx = 0;
                      cw = splitX_;
                    }
                } //(if deltaX_ < 0)                
              dirtyRanges_.push(rangeX);
              dirtyRects_.push(dirtyX);
            }
          if(deltaY_)
            {
              //same for y
              if(oldSplitY_ < splitY_)
                { 
                  ay = splitY_;
                  ah = h - splitY_;
                  by = 0;
                  bh = oldSplitY_;
                  
                  //c needs redraw
                  cy = oldSplitY_;
                  ch = splitY_ - oldSplitY_;
                  
                  rangeY.imin = range_.imin;
                  rangeY.imax = range_.imax;
                  rangeY.jmin = range_.jmin + ah + bh;
                  rangeY.jmax = range_.jmax + ah + bh + ch;
                  dirtyY = Rect(0,cy,w,ch);
                  
                }
              else
                { 
                  //a needs redraw
                  ay = splitY_;
                  ah = oldSplitY_ - splitY_;
                  rangeY.imin = range_.imin;
                  rangeY.imax = range_.imax;
                  rangeY.jmin = range_.jmin + ch;
                  rangeY.jmax = range_.jmax + ch + ah;
                  dirtyY = Rect(0,ay,w,ah);
                  
                  by = oldSplitY_;
                  bh = h - oldSplitY_;
                  cy = 0;
                  ch = splitY_;
                }
              //FIXME dirtyX and dirtyY will surely overlap
              // thus this overlap region will get drawn twice !!!!!
              dirtyRects_.push(dirtyY);
              dirtyRanges_.push(rangeY);

            }
        }
    }

  prepared_ = false;
}

void
Terrain::setVisibleHeight(pixel_t height)
{
  height_ = height;
}

// a little helper ...
void skipEOL(std::ifstream& fin){
  int mark = fin.tellg();
  char nl;
  fin.get(nl);
  while(nl=='\n'){
    mark = fin.tellg();
    fin.get(nl);
  }
  fin.seekg(mark);
}   

bool
Terrain::readMapFile(const std::string& name, const Coord &offset)
{
        // TODO :

        // This is where we will be loading the map with COAL...
  std::vector<uta::Surface*> _tiles;

  // always load sprites 
  bool loadSprites_ = true;

//   if(varconf::Config::inst()->findItem("map","omit_sprites"))
//     loadSprites_ = 
//       bool(varconf::Config::inst()->getItem("map","omit_sprites")) == false;
 
  int num_tiles;
  int mark;

  std::string install_dir = Registry::instance()->readInstallDir();
  std::string filename = install_dir +"/uclient/maps/"+name+".map";

  // now static in order to get unique sprite IDs across several map file
  static int spriteID=0;  
  std::ifstream fin ;
  fin.open(filename.c_str());
  if(!fin) {
    return false;
		debugN(12, std::cerr << "Terrain::readMapFile couldn't open sprite " <<
				filename << std::endl;);
	}
  char* head = new char[15] ;
  
  if(!fin.read(head,sizeof(char)*5)) {
		delete []head;
		if(Console::instance())
		  Console::instance()->std_err(std::string("readMapFile: could not read header from ") + name + std::string(".map!\n"));
		return 0;
	}

  if(Console::instance())
    Console::instance()->std_out(std::string("- parsing map file ")+filename+std::string("\n"));
  debugN(12,std::cerr << "Terrain:readMapFile- parsing "<<filename<< std::endl;);

  int map_size = 0;

  if(strstr(head,"Map: "))
    {
      /* hide viewport to speed up map loading.
       * (if the viewport is visible the display list gets recalculated everytime
       *  a single object is added !)
       */
      Viewport::instance()->hideViewport();

      int map_width, map_height, xs, ys ;
      
      //max size, tile size
      fin >> map_width >> map_height >> xs >> ys ;
      
      //get Name of the map
      char *n = new char[256] ;
      fin.getline(n,255) ;

      // for now we only accept iso-sized (square) maps !
      assert(map_width == map_height);
      map_size = map_width; 
      
      debugN(12,
             cerr << endl;
             cerr << "Reading Map \"" << n << "\" with a size of " 
             << map_width <<" * "<< map_height <<" squaremeters." 
             << endl;
             );
      
      do 
        {
          skipEOL(fin);
          mark = fin.tellg() ;
          if(!fin.read(head,sizeof(char)*10)) { delete []head; return false;}
          
          if(strstr(head,"TileList: "))
            fin >> num_tiles;
          
        } while (!strstr(head,"TileGfx: "));
      
      fin.seekg(mark);
      
      debugN(12,cerr << "Going to load " << num_tiles <<" TileGfx images." << endl;);

      //    char cname[64];
      int d1, d2, d3;
      char space;
      char texturename[256];
      std::string typname;
      int p1, p2;

      std::map<int, uta::Color> colors_;
      uta::Painter* peeker;
      uta::Point src,dest;
      std::string texturefile;

      _tiles.resize(num_tiles);
      sprites_.resize(num_tiles);

      if(Console::instance())
	Console::instance()->std_out(std::string("- loading textures ...\n"));

      uta::Application::keepAlive();
      uta::ProgressDialog progress; 
      std::string info = name+" : loading textures...";
      progress.info(info.c_str()); 
      progress.exec(); 

#ifdef USE_ATLAS_MAP
      std::fstream tl_file((name+".tl").c_str(),std::ios::out);
#endif     
			debugN(12,std::cerr << "Processing " << num_tiles << "tiles." <<  std::endl << std::flush;);
      
      for(int i=0; i < num_tiles; i++)
        {
          progress.setValue(i*100/num_tiles);
 
          fin.get(head, sizeof(char)*15,' ');
          fin >> d1 >> d2 >> d3 ;
          //skip space
          fin >> space;
          //rest of line is texturename
          fin.getline(texturename, 255);
          
          skipEOL(fin);
          
          texturefile = std::string(texturename);
          typname = std::string(texturename);
          p1 = typname.rfind("/");
          p2 = typname.rfind(".");
              
          //remember only the pure filename without path and ".png"
          typname=typname.substr(p1+1,p2-p1-1);

          //read only the ground and floor tiles
          if(
             ((texturefile.find("ground") != std::string::npos) ||
              (texturefile.find("floor") != std::string::npos) ||
              (texturefile.find("tiles") != std::string::npos) ||
              (texturefile.find("road") != std::string::npos) ||
              (texturefile.find("liquid") != std::string::npos) ||
              (texturefile.find("bridge") != std::string::npos)  ||
              (texturefile.find("stairs") != std::string::npos) ||
              (texturefile.find("field") != std::string::npos) ||
              (texturefile.find("pebbles") != std::string::npos) ||
	      (texturefile.find("grass") != std::string::npos) ||
              (texturefile.find("meadow") != std::string::npos )
              )
             &&
             ((texturefile.find("grass_strip") == std::string::npos) &&
              (texturefile.find("shrub_bank") == std::string::npos) &&
              (texturefile.find("waterplants") == std::string::npos) &&
              (texturefile.find("resources") == std::string::npos) &&
              (texturefile.find("field_potatos") == std::string::npos) &&
              (texturefile.find("summer") == std::string::npos) &&
              (texturefile.find("harvest") == std::string::npos) &&
              (texturefile.find("field_lavender") == std::string::npos)
              )
             )
            {
              
              //it is a tile !
              sprites_[i] = "";

#ifdef USE_ATLAS_MAP
              tl_file <<install_dir+"/forge/"+texturename<<endl;
#endif
              debugN(12, cerr << "loading "<<install_dir+"/forge/"+texturename<<endl;);
              
              /* check if this tile has already been loaded for another map file */
              std::map<std::string, uta::Surface*>::iterator itr = tiles_.find(texturename);
              if(itr != tiles_.end())
                {
                  /* if so use this pointer */
                  _tiles[i] = itr->second;
                }
              else
                {
                  /* if not load it now ! */
                  _tiles[i]=new uta::Surface();
                  if(!_tiles[i]->load(install_dir+"/forge/"+texturename))
                    {
                      //if we fail to load the provided file, let
                      //MediaIndex search for it:
                      //_tiles[i]->load(MediaIndex::lookupFilename(typname));
                      std::cout << "MEDIA INDEX!" << std::endl <<std::flush;
                    };              
                                    
                  /* convert to our internal pixelformat
                   * also set transparency flag for using the SDL blitters 
                   */
                  _tiles[i]->convert(Pixelformat::RGB565);
                  _tiles[i]->setTransparency(true);
                  
                  /* and store its reference */
                  tiles_[texturename] = _tiles[i];
                }

              //extract color
              peeker = new uta::Painter(_tiles[i]);
              src = uta::Point(_tiles[i]->width()/2,
                               _tiles[i]->height()/2);
              colors_[i] = peeker->getPixel(src);
              debugN(12,cerr << "choosing color : "<<colors_[i]<<endl;);
              delete peeker;

            }
          else if(loadSprites_)
            {
              // so far ... we know this one is a sprite and not a tile
              _tiles[i] = NULL;
              DII::Object::MapType::iterator itr = typesDB_.AsMap().find(typname);
              if(itr != typesDB_.AsMap().end())
                {
                  // not needed anmore ... I think ... FIXME: ENSURE THIS !
                  //itr->second.AsMap()["spritefile"] = install_dir+"/forge"+texturefile;
                  itr->second.AsMap()["name"] = typname;

                  //store "pointer" to the description, but ignore hills
                  //if(typname.find("hill_") == string::npos)
                  
                  sprites_[i] = typname;
                }
              else
                {
                  //create simple object def
                  DII::Object::MapType t;
                  DII::Object::ListType parent;
                  parent.push_back(typname);
                  t["parents"] = parent;
                  t["description"] = std::string("some thing");
                  t["name"] = typname;
                  // t["spritefile"] = install_dir+texturefile;
                

                  //store "pointer" to the description, but ignore hills
                  //if(typname.find("hill_") == string::npos)
                  
                  sprites_[i] = typname;
                
                  typesDB_.AsMap()[typname] = t;
                }
            }
          debugN(12,cerr << ".";);
          
        }

#ifdef USE_ATLAS_MAP
      tl_file.close();
#endif
#ifdef DEBUG
cerr << "Loaded "<< num_tiles <<" textures."<<endl<<flush;
#endif
//      progress.info(info.c_str()); 
 //     progress.exec(); 
 
      debugN(12,cerr << endl;);
      
      int id;
      int count;
      
      Pixel pixel;
      Coord pos;
      uta::Painter* painter = NULL;
      int ox,oy;

      if(overview_)
        {
          painter = new uta::Painter(overview_);
        }
      
      debugN(12,cerr << "going to load entities: " << endl;);
      
 

      /* coordinate ranges:
       * */
      int x_start = static_cast<int>(offset.x - map_size/2);
      int x_end = static_cast<int>(offset.x + map_size/2);
      int y_start = static_cast<int>(offset.y - map_size/2);
      int y_end = static_cast<int>(offset.y + map_size/2);
      
      /* Load the height map (if any is available)
       * */
      std::string heightmap_name = install_dir +"/uclient/maps/"+name+"_hm.png";
      uta::Surface hm;
      if(hm.load(heightmap_name))
        {
	  if(Console::instance())
	    Console::instance()->std_out("- loading terrain height map\n");
          info = name + " : loading terrain height map";
          progress.info(info.c_str());
          if (hm.width() == map_size && hm.height() == map_size)
            {
              int bpp = hm.pixelformat().bpp();
              if (bpp == 1 || bpp == 3 || bpp == 4)
                {
                  hm.lock();      
                  unsigned char* sPtr = (unsigned char*)hm.pixels();
        
                  for(int w = 0; w<map_size; w++)
                    for(int h= 0; h<map_size; h++)
                      World::World::inst()->setGroundLevel(w+x_start, h+y_start, 
                                    ((int)sPtr[w*bpp + (map_size-h-1)*hm.pitch()] - 127) * 5);

                  hm.unlock();
                }
              else if(Console::instance())
                {
                  Console::instance()->std_out("  Error: heightmap must use 8, 24 or 32 bits per pixel !\n");
                }
            }
          else if(Console::instance())
            {
	      Console::instance()->std_out("  Error: heightmap must be of same size");
            }
        }
      /* Calculate the terrain slope for the loaded area 
       * */
      if(Console::instance())
	Console::instance()->std_out("- calculating terrain slope\n");

      info = name + " : Calculating terrain slope";
      progress.info(info.c_str());
      
      for(int _y = y_start; _y <= y_end; _y++)
        {
          progress.setValue((_y + map_size/2)*100/map_size);
          int slope;
          for(int _x = x_start;  _x <= x_end; _x++)
            {
              slope = static_cast<int>(
                                       round((World::World::inst()->getGroundLevel(_x, _y) -
                                              (World::World::inst()->getGroundLevel(_x+1, _y) / 2.0 +
                                               World::World::inst()->getGroundLevel(_x, _y-1) / 2.0 )
                                              )*SLOPE_SCALE));
              if(slope > 127)
                slope = 127;
              else if(slope < -127)
                slope = -127;
              terrainSlope_[(_x+MAP_SIZE_MAX/2) + (_y+MAP_SIZE_MAX/2)*MAP_SIZE_MAX] = slope;
            }
        }

      if(Console::instance())
	Console::instance()->std_out(std::string("- Loading map texture data ...\n"));

      uta::Application::keepAlive();
      progress.setValue(0);
      progress.forceUpdate(1);
      info = name + " : Loading map texture data";
      progress.info(info.c_str());
      uapp->keepAlive();

      int num =0;
      DII::Object coords;
      DII::Object desc;
      DII::Object loc;

      std::string idStr;

#ifdef USE_ATLAS_MAP
      std::fstream out_file((name+".xml").c_str(),std::ios::out);
      Atlas::Objects::Decoder d;
      Atlas::Codecs::XML codec(out_file,&d);
      Atlas::Message::Encoder encoder(&codec);
      codec.StreamBegin();

      Atlas::Message::Object::ListType world_contains;
      
#endif // USE_ATLAS_MAP

//      for(int y=j-1; y>=0; y --)
      for(int y=0; y < map_height; y++)
      {
          debugN(6,cerr << ".";);
          num++;
              progress.setValue(static_cast<int>((100.0*num) / map_height));

          // allow sprite cache to flush the buffers
          SpriteCache::inst()->check();

//          for(int x=i-1; x>=0 ; x--)
          for(int x=0; x < map_width; x++)
          {
              
              fin.get(head, sizeof(char)*15,' ');
              fin >> count ;
              
              for(int n=0; n < count; n++)
                {
                  // d1 is z-order
                  // d2:d3 is dx:dy (offsets by easting and northing)
                  fin >> id >> d1 >> d2 >> d3 ;
                  
                  // these are optional offset values
                  //                 int dx = d2;
                  //                 int dy = d3;
                  
                  //convert isoedit grid to world coordinatoes ...
//                   pos.x = (float)(i/2-x) + float(d2) / 70.0;
//                   pos.y = (float)(j/2-y) - float(d2) / 70.0;
#ifdef ISOEDIT_IGNORE_OFFSET
                  pos.x = (float)(map_width / 2-x);
                  pos.y = (float)(map_height / 2-y);
#else
                  pos.x = (float)(map_width / 2-x) + float(d2) / 70.0;
                  pos.y = (float)(map_height / 2-y) - float(d2) / 70.0;
#endif
                  //d1 is not a real coordinate value, but it is the
                  //integer z-ordering ... so we do a bit of cheating here ...
                  pos.z =  d1 / ISOEDIT_FACTOR - float(d3) / 36.0;
                  
                  
                  if(id >= 0) 
                    {
                      //if id is not in typesDN, then it is
                      //neither a floor nor a ground tile,
                      //but maybe it is a sprite ?!
                      if(_tiles[id])
                        {
                          //convert world to pixel coordinates
                          //using given offset
                          pixel = w2p(pos + offset);

                          //add this tile to the map
                          tileMap_[pixel.j][pixel.i][d1] = _tiles[id];

                          //setPixel in overview_ map
                          if(painter)
                            {
//                               ox = (int) (pos.x+offset.x);
//                               oy = (int) -(offset.y+pos.y);
                              ox = pixel.i / (TILEWIDTH/2);
                              oy = pixel.j / TILEHEIGHT;

                              dest = uta::Point(ox+overview_->width()/2,
                                                oy+overview_->height()/2);
                             

                              std::map<int, uta::Color>::iterator itr;
                              itr = colors_.find(id);
                                
                              if(itr->second.r && 
                                 itr->second.g && 
                                 itr->second.b)
                                {
                                  painter->setPixel(dest,itr->second);
                                  painter->setPixel(dest+Point(-1,-1),itr->second);
                                  painter->setPixel(dest+Point(+1,-1),itr->second);
                                  painter->setPixel(dest+Point(0,-1),itr->second);
                                }
                            }
                        }
                      else if(loadSprites_ && !sprites_[id].empty())
                        {

                          DII::Object::MapType::iterator itr;
                          itr = typesDB_.AsMap().find(sprites_[id]);
                          if(itr != typesDB_.AsMap().end() && itr->second.IsMap())
                            {
                              DII::Object desc = itr->second;
                              
                                    //and add information about where to put it
                              desc.AsMap()["pos"] = coordToObj(pos+offset);
                              desc.AsMap()["loc"] = std::string("");
                              
                              idStr = itr->first +"_"+ itos(spriteID) + "_map";
                              spriteID++;
                              desc.AsMap()["id"] = idStr;

                              desc.AsMap()["loc"] = std::string("world_0");

                              desc.AsMap()["stamp"] = 0.0f;
                              
                              World::World::inst()->createEntity(desc);
                              
#ifdef USE_ATLAS_MAP
                              itr = desc.AsMap().find("parents");
                              if ((itr != desc.AsMap().end()) && itr->second.IsList()) {
                                  desc.AsMap()["name"] = itr->second.AsList().front();
                              }
                              encoder.StreamMessage(desc);
                              world_contains.push_back(idStr);
#endif // USE_ATLAS_MAP

                            }

                        } // else if(loadSprites ...
                        
                    }
                }
              //skipEOL(fin);
            }
        }
      if(painter)
        delete painter;

#ifdef USE_ATLAS_MAP
      Atlas::Message::Object::MapType map_world;
      map_world["id"] = "world_0";
      map_world["contains"] = world_contains;
      encoder.StreamMessage(Atlas::Message::Object(map_world));
      codec.StreamEnd();
      out_file.close();
#endif // USE_ATLAS_MAP


      Viewport::instance()->showViewport();
    }
  else
    {
			std::cerr << "no valid map file...." << endl;
      return false;
    }
  
  
  fin.close();

  debugN(12,cerr << endl;);
  debugN(12, if(loadSprites_) cerr << " ... the mapfile contained " 
         << spriteID << " sprites." <<endl;);

  return true;
}

class FileDecoder : public Atlas::Message::DecoderBase {
  private:
    std::fstream m_file;
    Atlas::Codecs::XML m_codec;
    Object::MapType m_entities;
    int m_count;
    std::string m_topLevelId;

    virtual void ObjectArrived(const Object & obj) {
        const Object::MapType & omap = obj.AsMap();
        Object::MapType::const_iterator I;
        if (((I = omap.find("id")) == omap.end()) ||
            (!I->second.IsString()) || (I->second.AsString().empty())) {
            std::cerr << "WARNING: Object in file has no id. Not stored."
                      << std::endl << std::flush;
            return;
        }
        const std::string & id = I->second.AsString();
        if (((I = omap.find("loc")) == omap.end()) ||
            (!I->second.IsString()) || (I->second.AsString().empty())) {
            // Entity with loc attribute must be the root entity
            if (!m_topLevelId.empty()) {
                std::cerr << "ERROR: File contained TLE with id = "
                          << m_topLevelId << " and also contains another id = "
                          << id << std::endl << std::flush;
            }
            m_topLevelId = id;
        }
        m_entities[id] = obj;
        m_count++;
    }
  public:
    FileDecoder(const std::string & filename) :
                m_file(filename.c_str(), std::ios::in),
                m_codec(m_file, this), m_count(0)
    {
    }

    void read() {
        while (!m_file.eof()) {
            m_codec.Poll();
        }
    }

    void report() {
        debugN(12,std::cout << m_count << " objects read from file and stored."
                            << std::endl << std::flush;);
        if (m_topLevelId.empty()) {
            debugN(12,std::cerr << "WARNING: No TLE found" << std::endl
		                << std::flush;);
        }
    }

    const Object::MapType & getEntities() const {
        return m_entities;
    }

    const std::string & getTopLevel() const {
        return m_topLevelId;
    }
};

void Terrain::insertEntityContents(const std::string & container,
                                   const Atlas::Message::Object::MapType & ent,
                                   const Atlas::Message::Object::MapType & entities)
{
    Atlas::Message::Object::MapType::const_iterator I = ent.find("contains");
    if ((I == ent.end()) || !I->second.IsList() || I->second.AsList().empty()) {
        return;
    }
    const Atlas::Message::Object::ListType & contents = I->second.AsList();
    Atlas::Message::Object::ListType::const_iterator J = contents.begin();
    for(; J != contents.end(); ++J) {
        if (!J->IsString()) { continue; }
        Atlas::Message::Object::MapType::const_iterator K = entities.find(J->AsString());
        if (K == entities.end()) { continue; }
        Atlas::Message::Object::MapType newEnt = K->second.AsMap();
        K = newEnt.find("id");
        if ((K == newEnt.end()) && !K->second.IsString()) { continue; }
        const std::string & id = K->second.AsString();
        newEnt["loc"] = container;
        World::World::inst()->createEntity(newEnt);
        insertEntityContents(id, newEnt, entities);
    }
}

bool
Terrain::readXmlMapFile(const std::string& name)
{
  std::string install_dir = Registry::instance()->readInstallDir();
  std::string filename = install_dir +"/uclient/maps/"+name+".xml";

  debugN(12, std::cerr << "LOADER: Being asked to load " << filename
	               << std::endl << std::flush;);

  FileDecoder fd(filename);
  fd.read();
  fd.report();
  const Atlas::Message::Object::MapType & fileEnts = fd.getEntities();
  const std::string & topId = fd.getTopLevel();
  Atlas::Message::Object::MapType::const_iterator I = fileEnts.find(topId);
  if (I == fileEnts.end()) {
      debugN(12,std::cerr << "ERROR: Failed to find top level entity " << topId
	                  << " in file" << std::endl << std::flush;);
      return false;
  } else {
      debugN(12,std::cerr << "Top level entity " << topId << " found in file"
                          << std::endl << std::flush;);
  }
  insertEntityContents("world_0", I->second.AsMap(), fileEnts);


  return true;
}
