//  The Worldforge Project
//  Copyright (C) 1998, 1999, 2000 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 03/2000


#ifndef _TERRAIN_H_
#define _TERRAIN_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sigc++/object.h>

#include "isodefs.h"

#include <uta/surface.h>
#ifndef HAVE_UTA_LIGHTEN_DARKEN
#include <uta/blitters.h>
#endif

#include <map>
#include <vector>

#include "dii/object.h"
namespace DII {class Module;}

/**Terrain Renderer.
   Based on isometric tiles.

   The tiles are blitted to a cache surface ("cache_") everytime, 
   the method prepare() is called. render() does a very fast job
   of just copying the cache content to the target surface.
   
   prepare() is optimized by the following idea:
   The viewport usually scrolls by some dozens of pixels, so most of
   the cache content can be reused, it only needs to be "moved" to an other 
   part of the surface. These rects show the viewport moved over the 
   gamemap and is the image the target shall contain after a call
   to render().
      
      target
   ------------- 
   |           |            target
   |           |         -------------
   |xxxxxxxxxxx|  --->   |xxxxxxxxxxx|
   |xxxxxxxxxxx| scroll  |xxxxxxxxxxx|
   -------------         |...........| 
    ...........          |...........|
               		 -------------
                             target

  The area containing the 'x' is identical in both viewports. One way
  would be to move parts of the content inside the cache surface, but this
  takes as well a lot of time. It would be better if the rendered surface
  data could stay at the same memory location.
  The idea is that the cache gets splitted up into nine rects of varying
  sizes:

   ------------- 
   |           |
   |           |        
   |xxxxxxxxxxx|    
   |xxxxxxxxxxx| 
   -------------       
      cache       

          | prepare  
	  V
  
     ------------- 
     |...........|
     |...........|         -------------
     |xxxxxxxxxxx|  --->   |xxxxxxxxxxx|
     |xxxxxxxxxxx| render  |xxxxxxxxxxx|
     -------------         |...........| <-- splitline
        cache              |...........|
                 	   -------------
                               target
			      
   (Here for ease of reading we assume that there are only two rects, 
    which are sufficient for just vertical scrolling and if we neglect
    any warp arounds )
   The renderer is requested to render a scene for which the viewport has
   scrolled down. The cache contains still the "old" content of the viewport.
   prepare() will now just render the upper half of the viewport (the '.')
   which has to have new content and render() will then 
   blit the cache content in reverse order to the target !
   The 'x' area inside the cache has not moved, but shows at a different 
   position inside the viewport.
  

*/
class Terrain : virtual public SigC::Object
{
  
 public:
  
  ///
  Terrain();
  ///
  virtual ~Terrain();

  static Terrain* instance() {return instance_ ? instance_ : new Terrain();}

  ///
  void deinit();
  ///
  bool init(DII::Module*);

  /**set target.
     @param: (long)uta::Surface* target
  */
  void setTarget(uta::Surface*);
  
  ///precalculate graphics internally
  bool prepare();
  ///actually transfer data to target
  bool render();

  /// How high are we looking above sea level
  void setVisibleHeight(pixel_t);
  /** sets visible range.
      @param map: "upperleft" (list of 2 ints)
                  "lowerright" (list of 2 ints)
     "upperleft" and "lowerright" refer to the edges of the
     target surface.
  */
  void setVisibleRange(const ViewRect&);
  ///returns const uta::Surface* pointing to overview map
  const uta::Surface* getMap();
  ///
  const uta::Surface* getTerrainMap();   

  ///load a mapfile
  bool readMapFile(const std::string& name, const Coord &offset = Coord(0,0,0));
  ///load an atlas xml file
  bool readXmlMapFile(const std::string& name);
  ///view the given position
  void view(const Pixel&);
    
 protected:
  ///
  void configure();
  ///
  void drawArea(ViewRect range, uta::Rect rect);
  ///
  void textureTile(const uta::Surface* tile, uta::Surface* target, 
		   pixel_t i, pixel_t j, pixel_t offset_i, pixel_t offset_j);
  ///
  void insertEntityContents(const std::string & container,
                            const Atlas::Message::Object::MapType & ent,
                            const Atlas::Message::Object::MapType & entities);
  ///
  uta::Surface* cache_;
  ///
  uta::Surface* target_;
  
  ///
  typedef std::map<int, const uta::Surface*> TileStack;
  ///
  typedef std::map<int, TileStack> TileStripe;
  ///
  typedef std::map<int, TileStripe> TileMap;
  ///
  TileMap tileMap_;

  ///
  typedef int TileID;
  ///
  typedef std::map<TileID, uta::Surface*> TileDB;


  ///
  std::map<std::string,uta::Surface*> tiles_;

  ///
  DII::Object typesDB_;
  ///
  std::vector<std::string> sprites_;

  ///
  bool prepared_;
  ///
  ViewRect range_;
  ///
  pixel_t height_;

  ///
  uta::Surface* overview_;
  ///
  uta::Surface* markedView_;

  ///
  pixel_t splitX_;
  ///
  pixel_t oldSplitX_;
  ///
  pixel_t deltaX_;
  ///
  pixel_t splitY_;
  ///
  pixel_t oldSplitY_;
  ///
  pixel_t deltaY_;
  ///
  pixel_t ax, aw, bx, bw, cx, cw;
  ///
  pixel_t ay, ah, by, bh, cy, ch;
  ///
  std::queue<uta::Rect> dirtyRects_;
  ///
  std::queue<ViewRect> dirtyRanges_;

  ///
  uta::Surface* defaultTile_;
  ///
  uta::GammaFunction* shadows_;
  ///
  bool use_shadows_;
  ///
  char* heightmap_;
  /** Set the drawing mode of the terrain renderer.
   * The value is interpreted bitwise, the first
   * 4 bit nibble chooses the mode, the lower nibble
   * enables mode specific options:
   *
   * 7654 3210
   * 0000 0000   plain tile blitting
   * ***1 ****   wireframe
   *      ***1   draw height bars
   *      **1*   draw sub-tile dividing lines  
   *
   * **1* ****   tile stretching height mapping
   *
  */
  int terrain_mode_;
  ///
  double terrain_shadow_;
  ///
  char *terrainSlope_;
 
  static Terrain* instance_; 
};

#endif
