/*
  SOLID - Software Library for Interference Detection
  Copyright (C) 1997-1998  Gino van den Bergen

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  Please send remarks, questions and bug reports to gino@win.tue.nl,
  or write to:
                  Gino van den Bergen
		  Department of Mathematics and Computing Science
		  Eindhoven University of Technology
		  P.O. Box 513, 5600 MB Eindhoven, The Netherlands
*/

#include "solid.h"

#include <algorithm>

#include "Box.h"
#include "BoxList.h"
#include "Encounter.h"
#include "Response.h"

#include <map>
#include <set>

typedef map<DtObjectRef, Object *> ObjectList;
typedef set<Encounter> ProxList;

ObjectList objectList;
Response defaultResp;
ProxList proxList;

Object *currentObject = 0;
bool caching = true;

// Object instantiation

void dtCreateBox(DtObjectRef object, DtScalar x, DtScalar y, DtScalar z) {
  currentObject = objectList[object] = new Object(object, Vector(x, y, z)); 
}

void dtSelectObject(DtObjectRef object) {
  ObjectList::iterator i = objectList.find(object);
  if (i != objectList.end()) {
    currentObject = (*i).second;
  }
}

void dtDeleteObject(DtObjectRef object) {
  ObjectList::iterator i = objectList.find(object);
  if (i != objectList.end()) {
    if (currentObject == (*i).second) currentObject = 0;
    delete (*i).second;
    objectList.erase(i);
  }
}

void dtTranslate(DtScalar x, DtScalar y, DtScalar z) {
  if (currentObject) currentObject->translate(Vector(x, y, z));
}

void dtLoadIdentity() { 
  if (currentObject) currentObject->setIdentity();
}

// Response

void dtSetDefaultResponse(DtResponse response, void *client_data) {
  defaultResp = Response(response, client_data);
}

// Runtime

void dtEnableCaching() { caching = true; }

void dtDisableCaching() { caching = false; }

void addPair(ObjectPtr object1, ObjectPtr object2) {
  proxList.insert(Encounter(object1, object2));
}

void removePair(ObjectPtr object1, ObjectPtr object2) {
  proxList.erase(Encounter(object1, object2)); 
}

bool object_test(const Encounter& e) {
  static Point p1, p2;

  if (intersect(*e.obj1, *e.obj2)) { 
    defaultResp(e.obj1->ref, e.obj2->ref);
    return true; 
  }

  return false;
}

DtCount dtTest() {
  DtCount count = 0;
  if (caching) {
    for (ProxList::iterator i = proxList.begin(); i != proxList.end(); ++i) {
      if (object_test(*i)) ++count;
    }
  }
  else {
    for (ObjectList::const_iterator j = objectList.begin(); 
	 j != objectList.end(); ++j)
      for (ObjectList::const_iterator i = objectList.begin(); 
	   i != j; ++i) {
	Encounter e((*i).second, (*j).second);
	if (object_test(e)) ++count;
      } 
  }
  return count;
}

