//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 05/2000

/* 
world coordinates: (right hand coordinate system)
       
         Z  
         .           
Y        .        X
   .     .     .
      .  .  .  
         0
      
         
*/

#ifndef _ISODEFS_H_
#define _ISODEFS_H_


#include "coord.h"
#include "Quaternion.h"

#define TILEWIDTH 70
//#define TILEWIDTH 72
#define TILEHEIGHT 36

///
typedef int GID;
/**pixel coordinate typ.
   2^32 pixels are roughly 60000 km. Seems to be enough for a reasonable
   gameworld.
 */
typedef int pixel_t;

#define SCALE 1.0

///
typedef struct
{
  pixel_t i;
  pixel_t j;
} Pixel;

///
typedef struct
{
  pixel_t imin;
  pixel_t imax;
  pixel_t jmin;
  pixel_t jmax;
} ViewRect;


/**transform world coordinates to pixel coordinates.
 */
inline Pixel w2p (const Coord& w)
{
  Pixel p;
  p.i = (pixel_t)(TILEWIDTH/2*SCALE*(w.x-w.y));
  p.j = (pixel_t)(TILEHEIGHT/2*SCALE*(-w.y-w.x+w.z));

  return p;
}

inline pixel_t w2p(const double& l)
{
  return (pixel_t)(40.0*SCALE*l);
}

/**transform pixel coordinates to world coordinates.
   Since pixel coordinates are only 2D, the z-coordinate
   is always 0. It cannot be calculated from the pixel coordinates
*/
inline Coord p2w (const Pixel& p)
{
  return Coord((p.i/(TILEWIDTH/2.0)-p.j/(TILEHEIGHT/2.0))/2.0/SCALE,
	       (-p.i/(TILEWIDTH/2.0)-p.j/(TILEHEIGHT/2.0))/2.0/SCALE,
	       0.0);
}

/**transform vector to discrete orientation.
   returns 0 for E, 1 for NE, 2 for N, etc.
   by Bryce Harrington ...
*/
inline int getIsoOrientation(const Coord& v)
{ 
  int h;
  int orientation;
  // Ideally, instead of dividing by 2.0, should multiply by tan(pi/8) = 0.4142...
  // This is probably close enough, and faster
  double xval = fabs(v.x / 2.0) , yval = fabs(v.y / 2.0);

  if (v.x > yval) 
    h = -1;
  else if (v.x < -yval) 
    h = +1;
  else
    h = 0;
  
  if (v.y > xval) 
    orientation = 2+h;
  else if (v.y < -xval) 
    orientation = 6-h;
  else 
    orientation = 2 + 2*h;

  return orientation;
}

inline int getIsoOrientation (const Quaternion& q)
{
  Coord orient;
  
  // When cyphesis sends us a quaternion to represent 
  // orientation, an unrotated vector is pointing due 
  // east.  We consider this to be the vector (1, 0, 0)

  // The following is a simplified quaternion rotation of
  // the vector (1, 0, 0)
  orient.x = 2 * (q[3] * q[3] + q[0] * q[0]) - 1;
  orient.y = 2 * (q[1] * q[0] + q[3] * q[2]);
  orient.z = 0;

  return getIsoOrientation(orient);  

}
#endif
