//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 06/2000

#ifndef _TREE_H_
#define _TREE_H_

#include "toolbox/local_hash_map.h"
#include <stack>
#include <string>


/**nested tree class.
   internally it uses a hash map for quick access to its leaves.
   Tree::local_find(id) searches inside the current tree, while
   Tree::find(id) does a recursive search through all leaves.
   Both methods return a valid Tree* or NULL if failed.

   Note: Tree does not free any sub-trees when deleted !
         Nor does it delete a leaf when the entry
	 is removed or deleted !


 */
class Tree
{
 public:
  
  ///
  typedef STRING_HASH_MAP(Tree*) LeafMap;

  ///
  class iterator
    {
    public:
      ///
      iterator():tree_(NULL) {};
      ///
      iterator(Tree::LeafMap::iterator i, 
	       Tree* c)
	{ current_ = i; tree_ = c;};
 
      ///
      ~iterator() {};
     
      ///
      Tree* operator*();
      ///
      iterator& operator++();
      ///
      iterator& operator++(int);
      ///comparism is not exact!
      bool operator==(const iterator& ); 
       ///comparism is not exact!
      bool operator!=(const iterator& );
      
    protected:
      ///
      Tree* tree_;
      ///
      std::stack <Tree*> trees_;
      ///
      Tree::LeafMap::iterator current_;
      ///
      std::stack<  Tree::LeafMap::iterator > visited_;
    };

  friend class Tree::iterator;

  ///
  iterator begin()
    { 
      if(!leaves_)
	leaves_ = new LeafMap();
      return iterator(leaves_->begin(),this);
    };
  ///
  iterator end()
    { 
      if(!leaves_)
	leaves_ = new LeafMap();
      return iterator(leaves_->end(),this);
    };
  ///
  bool empty() const { return (leaves_) ? leaves_->empty() : false; };
  ///
  size_t size() const { return (leaves_) ? leaves_->size(): 0; };

  ///create root Tree
  Tree();
  ///
  Tree(Tree* root);
  ///
  virtual ~Tree();

  ///
  virtual void insert(const std::string& id, Tree* leaf);
  ///
  virtual void remove(const std::string& id);
  ///
  virtual void clear() 
    { 
      if(leaves_)
	{
	  leaves_->clear(); 
	  delete leaves_;
	  leaves_ = NULL;
	}
    };

  ///direct lookup
  Tree* local_find(const std::string& id);
  ///recursively lookup
  Tree* find(const std::string& id); 

  ///
  Tree* root() const { return root_; };
  ///
  void setRoot(Tree* root) { root_ = root; };
  ///recursively lookup absolute root
  Tree* absRoot() 
    {
      if(root_)
	return root_->absRoot();
      return this;
    }
  ///
  const LeafMap& leaves() const
    { 
      if(!leaves_)
	return Tree::emptyMap_;
      
      return *leaves_; 
    };
  ///
  LeafMap& leaves()
    { 
      if(!leaves_)
	leaves_ = new LeafMap();
      return *leaves_; 
    };
 protected:
  ///
  Tree* root_;
 private:
  ///
  LeafMap *leaves_;
  ///
  static LeafMap emptyMap_;

};

#endif
