//
//  The Worldforge Project
//  Copyright (C) 1998,1999  The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 01/1999

#define NEED_DII_SERVICES


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "res_loader.h"
#include "dii/dii.h"
#include "loader/loader.h"
#include "media/media_index.h"
#include "media/sprite_cache.h"
#include "world/world.h"
#include "world/world_adaptor_renderer.h"
#include "app.h"
#include "registry/registry.h"

#include "media/content_request.h"

#include <varconf/Config.h>

#include <uta/application.h>
#include <uta/debug.h>
#include <uta/resources.h>
#include <uta/multilineedit.h>
#ifndef HAVE_UTA_LIGHTEN_DARKEN
#include <uta/blitters.h>
#endif
#include <uta/version.h>

using namespace uta;

//need these for mkdir()
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h> 

#if defined(HAVE_EXECINFO_H) && defined(HAVE_SIGNAL_H)
extern "C" {
#include <execinfo.h>
#include <signal.h>
}
#define INSTALL_BACKTRACE
#else
#undef INSTALL_BACKTRACE
#endif

#include "uclient_icon.h"
 
#include <fstream>
#include <stdexcept>

#ifdef INSTALL_BACKTRACE

#define MAX_BACKTRACE 50
static void* backtraceTable[MAX_BACKTRACE];
typedef void (*sighandler_t)(int);  
static sighandler_t default_handler = NULL;

void createBacktrace(int sig)
{
  
  cerr << endl
       << "*******************************************************"<<endl
       << "SEGMENTATION FAULT " << endl 
       << "Program Version " << VERSION << "    Build " << __DATE__ <<" - "<<__TIME__ << endl
       << endl
       << "dumping backtrace ... " << endl << endl;
  
  // get the symbols
  int n = backtrace( backtraceTable, MAX_BACKTRACE );
  char** symbols = backtrace_symbols( backtraceTable, n);
  
  /* use offset of 2, so that the backtrace starts at
   * the point where the segmentatio fault occured.
   */
  if(n > 2)
    for(int i=2; i < n; i++)
      cerr << i-2 << " : " << symbols[i] << endl; 

  
  cerr << endl
       << "Please include this information when sending a " << endl
       << "bug report to <malcolm@worldforge.org>."<<endl
       << "*******************************************************"<<endl
       << endl;
  
  /* call default handler in order to shutdown "cleanly"
   */
  if(default_handler && default_handler != SIG_ERR)
    default_handler(sig);

  SDL_Quit();
  exit(1);
}
#endif

class WaitForSth : virtual public SigC::Object
{
protected:
  bool flag;
public:
  WaitForSth() : flag(false) {};
  bool key_slot(const uta::KeyEvent*) { flag = true; return true; };
  bool click_slot(const uta::ButtonEvent*) { flag = true; return true; };
  bool operator() (void) { return !flag; };
};

/* Monitor the global settings.
   Via the callback function which gets connected to signal of varconf::Config
   this object monitors the global and sets for exampe the debuglevel
   appropriately.
   This allows the changing of the debug level and other options at runtime.
*/
class Globals : virtual public SigC::Object
{
public:
  Globals()
  {
    varconf::Config::inst()->
      sigv.connect(SigC::slot(this, &Globals::configCallback));
  }

  void configCallback(const std::string& section, const std::string& key)
  {
    if(!section.empty())
      return;
    if(key == "hardware_cursor")
      {
        int flag=int(varconf::Config::inst()->getItem("","hardware_cursor"));
        uta::Application::useHardwareCursor(flag);
      }
    else if(key == "gamma")
      {
        float gamma= 
          double(varconf::Config::inst()->getItem("","gamma"));
        uta::ScreenSurface::setGamma(gamma, gamma, gamma);
      }
    else if(key == "debug")
      {
#ifdef DEBUG
        int debug_level = varconf::Config::inst()->getItem("","debug");
#endif
        DEBUG_INIT(debug_level);
        debug(cout << "debug_level = " << debug_level << endl << flush;);
      }
      else if(key == "fullscreen")
      {
         if(SDL_GetVideoSurface())
            SDL_WM_ToggleFullScreen(SDL_GetVideoSurface());
      }
  }
} globals;

static std::string getConfDir()
{ 
  std::string homedir;
  std::string confdir;

#ifdef  __WIN32__
    const char* homedrive;
    const char*  homepath;
    
    homedrive = getenv("HOMEDRIVE");
    homepath = getenv("HOMEPATH");
   
    if(homepath == 0)
      {
        // we're homeless...
        debugN(4,
               cerr << "can't find STARTDIR, looking for ";
               cerr << "config files in current directory."<<endl;
               );
        homepath=".";
      }
    
    if(homedrive == 0)
      homedrive = "";
    
    homedir = std::string(homedrive)+std::string(homepath);

    homedir = ".";

    debug(cout << "Using " << homedir << " to store settings. " << endl;);

#else   //unix
    homedir = std::string(getenv("HOME"));
    if ( homedir.empty() ) 
      {
        // we're homeless...
        debugN(4,
               cerr << "can't find user's HOME, looking for ";
               cerr << "config file in current directory."<<endl;
               );
        homedir = ".";
      }
#endif  // ifdef __WIN32__

  debugN(4,cerr << "OK." << endl;);

#ifdef __WIN32__
  //use no subdir in WIN32
  confdir = homedir;
  // mkdir(confdir.c_str());
#else
  confdir = homedir+"/.uclient";
  //ensure that the directory exists
  mkdir(confdir.c_str(),0755);
#endif

  return confdir;
}
static const std::string checkSprite("/forge/media/media-2d/char-collection-redman/animations/paladin_female/paladin_female_1_walk_dw_SW_1.png");
static const std::string checkRawSprite("/forge/media/media-2d/char-collection-redman/animations/paladin_female/paladin_female_1_walk_dw.walking.female.._.raw");

static inline bool
testSpriteDir(const std::string & test_dir)
{
  struct stat testStat;
  int test  = stat(std::string(test_dir+"/forge").c_str(), &testStat)
            + stat(std::string(test_dir+"/uclient").c_str(), &testStat);
  if(test != 0)
    {
      return false;
    }

  test = stat(std::string(test_dir+checkSprite).c_str(), &testStat);
  if(test != 0)
    {
      
      // maybe this is a raw-file based distribution:
      test = stat(std::string(test_dir+checkRawSprite).c_str(), &testStat);
      if(test != 0)
        {
          return false;
        }
    }

  return true;
}

static void
getInstDir(std::string & input_dir)
{
#ifdef HAVE_READLINE
  char * input = readline(0);
#else
  char input[256];
  fgets(input,255,stdin);
#endif

  if(strlen(input)>1)
    {
      input[strlen(input)-1]='\0';
      input_dir = std::string(input);
#ifdef HAVE_READLINE
      free(input);
#endif
    }
}

static bool
validateInstallation()
{
  debug(cout << "Validating installation ..." << endl;);

  DII::Object::MapType req;

  std::string install_dir;
  int test;
  struct stat testStat;
  
  install_dir = Registry::instance()->readInstallDir();
  
  if(install_dir.empty())
    {
      install_dir = std::string(INSTALLDIR) + "/share";
      if (!testSpriteDir(install_dir))
	{
          cout << endl 
               << "Configuration Error:" << endl
               << "You seem to have started UClient for the first time,"<<endl;

#ifdef __WIN32__
          install_dir = "./share";
	  cout << "please check if the settings are correct :" << endl
               << "directory containing worldforge/uclient data files: "
	       << install_dir << endl;
#else
          cout << "and the installation directory cannot be found,"<<endl
               << "please give the share directory : " << endl
               << "installation directory :" << endl;
          getInstDir(install_dir);
	}
#endif // __WIN32__

      //registry.writeEntry("install_dir",install_dir.c_str());
      req["path"]=std::string("/");
      req["key"]=std::string("install_dir");
      req["value"]=install_dir;
      DII::callService("writeRegistryEntry",req);

#ifdef __WIN32__
      cout << " skipping tests..." << endl;
      return true;
#endif // __WIN32__
    }
  else
    {
      // Check for existance of our subdirectories in the install dir
      test  = stat(std::string(install_dir+"/forge").c_str(), &testStat);
      test += stat(std::string(install_dir+"/uclient").c_str(), &testStat);
      if(test != 0)
        {
          cerr << endl << "Installation Error: " << endl
               << "Cannot find the UClient data directories in "
	       << install_dir  << " [" << test << "]\n";
	  cerr << "Checked:\n"
	       << std::string(install_dir + "/forge") << endl
               << std::string(install_dir + "/uclient") << endl;
          cerr << "PLEASE INSTALL ALL NEEDED MEDIA / SUPPORT FILES ! "
               << "See INSTALL for details." << endl << endl;          
          cerr << "Please check if the settings are correct : " << endl
               << "(just press enter to accept the given values)" << endl
               << endl
               << "installation directory :" << install_dir << endl;

#ifdef __WIN32__
          install_dir = ".";
#else

	  getInstDir(install_dir);

#endif // __WIN32__
          //          registry.writeEntry("install_dir",install_dir.c_str());
          req["path"]=std::string("/");
          req["key"]=std::string("install_dir");
          req["value"]=install_dir;
          DII::callService("writeRegistryEntry",req);
          
        }
    }

   // Check for animation graphics.
   // TODO: Fetch errno and print out if stat fails
   if (!testSpriteDir(install_dir))
     {
       cerr << endl
            << "The needed character animation files are not installed in "
            << endl << install_dir 
            << "/forge/media/media-2d/char-collection-redman/animations !! "
            << endl << endl;
       cerr << "PLEASE INSTALL ALL NEEDED MEDIA FILES ! "
            << "See INSTALL for details." << endl << endl;  
       return false;
     }

   return true;
}

/* -------------------     MAIN      ---------------------------*/   
int main (int argc, char **argv)
{

  ios::sync_with_stdio(true);

  cout  << endl 
        << "UClient - an isometric client for Worldforge" << endl
        << endl
        << "Version " << VERSION << endl 
#ifdef _WIN32
        << "Win32 - "
#endif
        << "Build " << __DATE__ <<" - "<<__TIME__ << endl
        << endl
        << "Copyright 1998-2002 by The Worldforge Project."<< endl
        << "This program is licensed under the GPL, see COPYING for details." << endl 
	<< "Original code by Karsten-O. Laux <klaux@rhrk.uni-kl.de>" << endl
	<< "Mason version by Malcolm Walker <malcolm@worldforge.org>" << endl;

#ifndef HAVE_UTA_LIGHTEN_DARKEN
  if(!uta::hasMMXExtension())
    {
      cerr << "WARNING !!" << endl
           << "The CPU does not feature the MMX extensions, this" << endl
           << "might result in very bad performance." << endl
           << "You might consider to turn off all advanced renderings."<< endl
           << "(e.g. Weather effects, Terrain Rendering)" << endl << endl;
    }
#endif

#ifdef DEBUG
#ifdef INSTALL_BACKTRACE
  cerr << "installing backtrace dump ..." << endl;
  default_handler = signal(SIGSEGV, createBacktrace);
#endif
#endif

  int exitcode = 0;
 
  DII::Module* registry;
  std::string inifile;
  std::string globalconf;
  
  try 
    {
      //getConfDir determines where the ini files resides
      //in works on Win32 as well as Unix
      inifile = getConfDir()+"/settings.ini";
      globalconf = getConfDir()+"/registry.xml";
      
      //register commandline switches
      varconf::Config::inst()->setParameterLookup('w',"width",true);
      varconf::Config::inst()->setParameterLookup('h',"height",true);
      varconf::Config::inst()->setParameterLookup('b',"bpp",true);
      varconf::Config::inst()->setParameterLookup('f',"fullscreen",true);
      varconf::Config::inst()->setParameterLookup('d',"debug",true);
      varconf::Config::inst()->setParameterLookup('i',"installdir",true);

      debug(cout << "parsing settings ..." << endl << flush;);
      
      //parse ini file
      varconf::Config::inst()->readFromFile(inifile);
      
      //parse command line
      varconf::Config::inst()->getCmdline(argc, argv);
      
      debugN(3,cerr << "restoring registry ....";);
      //open $HOME/.uclient/registry.ini
   
      //load cache dir entry
      std::string cacheDir = getConfDir()+"/cache/";
      if(varconf::Config::inst()->findItem("media","cachedir"))
        {
          varconf::Variable vc_cacheDir = varconf::Config::inst()->getItem("media","cachedir");
	  if(vc_cacheDir.is_string()) cacheDir = std::string(vc_cacheDir);
        }
      ContentRequest::setCacheDir(cacheDir);

      //load proxy settings
      int proxyPort = 0;
      std::string proxyHost;
      if(varconf::Config::inst()->findItem("media","proxy_host"))
        {
          varconf::Variable vc_proxyHost = varconf::Config::inst()->getItem("media","proxy_host");
	  if(varconf::Config::inst()->findItem("media","proxy_port"))
            {
              varconf::Variable vc_proxyPort = varconf::Config::inst()->getItem("media","proxy_port");
	      if(vc_proxyHost.is_string() && vc_proxyPort.is_int())
                {
                  proxyHost = std::string(vc_proxyHost);
		  proxyPort = int(vc_proxyPort);
                }
            }
        }
      ContentRequest::setProxy(proxyHost,proxyPort);

      //load the registry module
      if(varconf::Config::inst()->findItem("","installdir"))
        {
          varconf::Variable installdir = 
            varconf::Config::inst()->getItem("","installdir");
          if(installdir.is_string())
            {
              cerr << "using user provided installation directory \""
                   << std::string(installdir) << "\"" << endl;
              DII::Loader::setSearchPath(std::string(installdir)+"/lib");
            }
        }
      DII::Loader::loadModule("Registry");
      registry = DII::lookupModule("Registry");
      assert(registry);
      
      //and restore its content
      if(registry->callMethod("open",globalconf)!=DII::SUCCESS)
        {
          debugN(3,cerr << " error reading file " << globalconf << endl;);
        }
      else
        {
          debugN(3,cerr << "OK." << endl;);
        } 
      
      //purge registry:/runtime, where all runtime information is stored
      DII::Object::MapType runtime;
      DII::Object::MapType req;
      req["path"]=std::string("/");
      req["key"]=std::string("runtime");
      req["value"]=runtime;
      DII::callService("writeRegistryEntry",req);
      
      debugN(3,cerr << "successfully purged old runtime entries." << endl;);
    }
  catch(const std::exception& e)
    {
      cout << "UClient: unhandled exception in initialization:"<<e.what() << endl;
      exit(1);
    }

  //check if needed data files are installed
  if(!validateInstallation())
    {
      cout << "UClient: failed to validate installation of data files." << endl;
      exit(1);
    }

  debug(cout << "Installation verified." << endl <<endl;);

  std::string install_dir;
  int x,y,bpp;
  bool fullscreen;
  Surface icon;
  icon.readFromHeader(::header_data,::width,::height);

  //create application instance
  UClientApp app(argc, argv);

  try 
    {
      
      //get path to installation directory
      install_dir = Registry::instance()->readInstallDir();
        DII::Loader::setScriptPath(install_dir+"/uclient/scripts");
      
      //--------------------------------------------------------------
      
      debug(cout << "UClient starting up..." << endl;);
      
      //load App module
      DII::Loader::loadModule("App");

      //read screen options
      x = varconf::Config::inst()->getItem("","width");
      y = varconf::Config::inst()->getItem("","height");
      bpp = varconf::Config::inst()->getItem("","bpp");
      fullscreen = varconf::Config::inst()->getItem("","fullscreen");

      
      if(x < 800)
        {
          x = 800;
          varconf::Config::inst()->setItem("","width",x);
        }
      if(y < 600)
        {        
          y = 600;
          varconf::Config::inst()->setItem("","height",y);
        }
      
#ifdef __WIN32__
      //on windows machine there is no default depth ...
      //FIXME !!!
      if(!bpp)
        bpp = 16;
      varconf::Config::inst()->setItem("","bpp",bpp);
#endif
      
      
      debugN(3, cout << "initializing graphic mode ... ";);
      
      //init Application setting resolution and pixel depth
      //FIXME error checking ????????????????????????????????????

      uapp->init(x,y,bpp,fullscreen,false, &icon);
      debugN(3,cerr << "OK." << endl;);
      
      DII::Object::MapType request;
      
      request["path"] = std::string("/runtime/modules");
      request["key"] =  std::string("RootWindow");
      request["value"] = ptrToObj<Widget*>(uapp->rootWindow());
      DII::callService("writeRegistryEntry",request);

      request["path"] = std::string("/runtime/modules");
      request["key"] =  std::string("Resources");
      request["value"] = ptrToObj<Resources*>(Resources::instance);
      DII::callService("writeRegistryEntry",request);
     
      //set Window Title
      uapp->rootWindow()
        ->setTitle("Worldforge-UClient  loading resources ...");
      
			// Hardcoding this is necessary - we can't use the resources before 
			// they are loaded!
      Surface logo;
      logo.readFromFile(std::string(install_dir+"/forge/media/media-2d/collection-malcolm/ui/mason_splash_1_686x481_mw_vv.png").c_str());
      int logox = logo.width();
      int logoy = logo.height();
      int xpos = (uapp->rootWindow()->width() - logox) / 2;
      int ypos = (uapp->rootWindow()->height()- logoy) / 4;
			
			Widget* titlebg = new Widget(uapp->rootWindow(), 0, 0,
					uapp->rootWindow()->width(), uapp->rootWindow()->height());
			titlebg->setColor(uta::white);
      Widget* title = new Widget(titlebg, xpos,ypos,logox,logoy);
      title->setBackground(&logo,false);
      
      
      //global eventloop not yet running... so we have to update
      //manually
      uapp->rootWindow()->sync();
      
      //initialize MediaIndexer
      MediaIndex::initialize(install_dir);
            
      // call resourceDII::Loader do load all surfaces, samples, songs etc
      exitcode = 
        resourceLoader(std::string(install_dir+"/uclient/resources.ini").c_str());
      
      debugN(3, cerr << "done loading resources. removing title ..." << endl;);
      
      //delete the title pic
      delete title;
			delete titlebg;
      
      debugN(3, cerr << "ok. now fire up the gui ..." << endl;);
    }
  catch(const std::exception& e)
    {
      cout << "UClient: unhandled exception while loading resources :"<<e.what() << endl;
      exit(1);
    }
    catch(...)
    {
      cout << "UClient: unhandled unknown exception while loading resources ."<< endl;
      exit(1);
    }
  
  // initialize subsystems:
  World::World* world = 0;
  World::RendererAdaptor* world_adaptor = 0;

  // if loading/configuration went fine... fire it up
  if(exitcode == 0)
    {
    
      try
        {
          //set Window Title
          uapp->rootWindow()->setTitle("Worldforge-UClient");

          // initialize subsystems:
          world = World::World::inst();
          world_adaptor = new World::RendererAdaptor(world);

          /* load script fileaccess.lua which offers
           * a chrooted fileaccess via the DII interface
           */
          DII::Loader::loadScript("fileaccess","trusted");

          /* load init script (which loads additional modules, scripts, ...) */
          DII::Loader::loadScript("Init");
          
          DII::callMethod("main_menu","main");

          debugN(3,cerr << "entering uta::Application::exec()." << endl;);
        }
      catch(const std::exception& e)
        {
          cout << "UClient: unhandled exception while loading modules :"<< e.what() << endl;
          exit(1);
        }
    catch(...)
        {
          cout << "UClient: unhandled unknown exception while loading modules."<< endl;
          exit(1);
        }

      // End of initialization
      //*******************************************

      try 
        {
          //run the application
          exitcode = uapp->exec();
        }
      catch(const std::exception& e)
        {
          cout << "UClient: unhandled exception in main loop :" << e.what() << endl;
          exit(1);
        }
    catch(...)
        {
          cout << "UClient: unhandled unknown exception in main loop."<< endl;
          exit(1);
        }

      // ******************************************
      // Begin of shutdown
      
      try
        {
          debugN(3,cerr << "returned from Application.exec()." << endl;);
          
          uapp->rootWindow()->setTitle("Worldforge-UClient shutting down ...");
          
          //write registry back 
          debugN(3,cerr << "storing configure information ... ";);
          registry->callMethod("flushRegistry");
          
          //write back settings
          varconf::Config::inst()->writeToFile(inifile);
          
          debugN(3,cerr << "OK."<<endl;);
        }
      catch(const std::exception& e)
        {
          cout << "UClient: unhandled exception in first stage shutdown:"<<e.what() << endl;
          exit(1);
        }
    catch(...)
        {
          cout << "UClient: unhandled unknown exception in first stage shutdown."<< endl;
          exit(1);
        }
    }

  try
    {
      DII::Module* appMod = DII::lookupModule("App");
      if(appMod)
        appMod->lock();

      //try to unload all modules (this checks for dependencies !)
      DII::Loader::releaseAll();

      //World modules
      world->deinit(); // this deletes instance as well !
      delete world_adaptor;

      //VarConf library
      // delete varconf::Config::inst();
      
      // Sprite Cache
      delete SpriteCache::inst();
      
      /* calling init again is possible since libuta 0.3.32
         it reinitializes the graphic mode, it also allows changing
         the resolution
         all existing widgets including the rootwindow get deleted !
      */ 
      uapp->init(x,y,bpp,fullscreen,&icon);  


      uapp->rootWindow()->setTitle("Worldforge-UClient  Credits");
          
      MultiLineEdit* title = new MultiLineEdit(uapp->rootWindow(), 0,0,
                                 uapp->rootWindow()->width(),
                                 uapp->rootWindow()->height());
      title->setTextColor(uta::black, uta::grey);
      title->setBackground(SURFACE_RES("credits_screen"));
      
      title->setText("UClient "VERSION"\n"      
                     "Copyright 1998-2002 by The Worldforge Project.\n"
                     "\n"
                     "This program is free software; you can redistribute it and/or modify "
                     "it under the terms of the GNU General Public License as published by "
                     "the Free Software Foundation; either version 2 of the License, or "
                     "(at your option) any later version.\n"
                     "This program is distributed in the hope that it will be useful, "
                     "but WITHOUT ANY WARRANTY; without even the implied warranty of "
                     "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the "
                     "GNU General Public License for more details.\n"
                     "You should have received a copy of the GNU General Public License "
                     "along with this program; if not, write to the Free Software "
                     "Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n"
                     "For information about Worldforge and its authors, please contact "
                     "the Worldforge Web Site at http://www.worldforge.org.\n"
                     "\n"
                     "written by Karsten-O. Laux <klaux@rhrk.uni-kl.de>\n"
                     "layout and default theme by Pegasus  <szymanek@rhrk.uni-kl.de>\n"
                     "win32 porting by Malcolm Walker <walker05@islandnet.com>\n"
                     "\n"
                     "using libraries and code from:\n"
                     "FreeSockets by Rafael Guterres Jeffman <rafasgj@exatas.unisinos.br>, "
                     "SOLID by Gino van den Bergen <gino@win.tue.nl>, "
                     "libuta by Karsten-O. Laux <klaux@rhrk.uni-kl.de>, Stefanus Du Toit <sdt@gmx.net>, "
                     "SDL by Sam Lantinga  <slouken@devolution.com>, "
                     "Varconf by Joseph A. Zupko <jaz147@psu.edu>, Stefanus Du Toit <sdt@gmx.net>, "
                     "libAtlas-C++ by Aloril <aloril@iki.fi>, Dmitry Derevyanko <dmitryd@linuxfreak.com>,"
                     "Michael Day <mikeday@corplink.com.au>, Stefanus Du Toit <sdt@gmx.net>,"
                     "Alistair Riddoch <alriddoch@zepler.org> (maintainer), "
                     "libsigc++ by Karl E. Nelson <kenelson@ece.ucdavis.edu>, Tero Pulkkinen <terop@students.cc.tut.fi>, "
                     "libxml by Daniel Veillard <Daniel.Veillard@w3.org>, "
                     "libpng  by G. Randers-Pehrson <randeg@alum.rpi.edu>,Andreas E. Dilger <adilger@enel.ucalgary.ca>,"
                     "Guy Eric Schalnat <gschal@infinet.com>, "
                     "libz by Jean-loup Gailly <jloup@gzip.org>, Mark Adler <madler@alumni.caltech.edu>, "
"libjanus by Karsten-O. Laux <klaux@rhrk.uni-kl.de>, "
"LUA by Waldemar Celes, Roberto Ierusalimschy and Luiz Henrique de Figueiredo. <lua@tecgraf.puc-rio.br> ");

      //global eventloop not yet running... so we have to update
      //manually
      uapp->rootWindow()->sync();
      
      WaitForSth _wait;
      uta::Application::keystroke.connect(slot(&_wait, &WaitForSth::key_slot));
      uta::Application::mouseclick.connect(slot(&_wait, &WaitForSth::click_slot));
      while(_wait())
               uta::Application::keepAlive();

      if(appMod)
        {
          //just to be sure...
          appMod->release();
          appMod->release();
          appMod->release();
        }
      DII::Loader::releaseModule("App");

    }
  catch(...)
    {
      cout << "UClient: unhandled exception in final shutdown." << endl;
      exit(1);
    }

  cout << endl 
       << endl 
       << "visit http://www.worldforge.org for more ..." << endl << endl;

  return exitcode;
}

