//
//  The Worldforge Project
//  Copyright (C) 1998,1999,2000,2001  The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 06/2001

#define NEED_DII_SERVICES


#include "config.h"

#include "dii/dii.h"
#include "loader/loader.h"
#include "media/media_index.h"
#include "media/sprite_cache.h"
#include "registry/registry.h"

#include <varconf/Config.h>

#include <uta/application.h>
#include <uta/debug.h>
#include <uta/resources.h>

using namespace uta;

//need these for mkdir()
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h> 

#if defined(HAVE_EXECINFO_H) && defined(HAVE_SIGNAL_H)
extern "C" {
#include <execinfo.h>
#include <signal.h>
}
#define INSTALL_BACKTRACE
#else
#undef INSTALL_BACKTRACE
#endif
 
#include <fstream>
#include <stdexcept>

#include "sprites.h"

#ifdef INSTALL_BACKTRACE

#define MAX_BACKTRACE 50
static void* backtraceTable[MAX_BACKTRACE];
typedef void (*sighandler_t)(int);  
static sighandler_t default_handler = NULL;

void createBacktrace(int sig)
{
  
  std::cerr << std::endl
       << "*******************************************************"<<std::endl
       << "SEGMENTATION FAULT " << std::endl 
       << "Program Version " << VERSION << "    Build " << __DATE__ <<" - "<<__TIME__ << std::endl
       << std::endl
       << "dumping backtrace ... " << std::endl << std::endl;
  
  // get the symbols
  int n = backtrace( backtraceTable, MAX_BACKTRACE );
  char** symbols = backtrace_symbols( backtraceTable, n);
  
  /* use offset of 2, so that the backtrace starts at
   * the point where the segmentation fault occured.
   */
  if(n > 2)
    for(int i=2; i < n; i++)
      std::cerr << i-2 << " : " << symbols[i] << std::endl; 

  
  std::cerr << std::endl
       << "Please include this information when sending a " << std::endl
       << "bug report to <malcolm@worldforge.org>." << std::endl
       << "*******************************************************" << std::endl
       << std::endl;
  
  /* call default handler in order to shutdown "cleanly"
   */
  if(default_handler && default_handler != SIG_ERR)
    default_handler(sig);

  SDL_Quit();
  exit(1);
}
#endif


/* Monitor the global settings.
   Via the callback function which gets connected to signal of varconf::Config
   this object monitors the global and sets for exampe the debuglevel
   appropriately.
   This allows the changing of the debug level and other options at runtime.
*/
class Globals : virtual public SigC::Object
{
public:
  Globals()
  {
    varconf::Config::inst()->
      sigv.connect(SigC::slot(this, &Globals::configCallback));
  }

  void configCallback(const std::string& section, const std::string& key)
  {
    if(!section.empty())
      return;
    if(key == "hardware_cursor")
      {
        int flag=int(varconf::Config::inst()->getItem("","hardware_cursor"));
        uta::Application::useHardwareCursor(flag);
      }
    else if(key == "gamma")
      {
        float gamma= 
          double(varconf::Config::inst()->getItem("","gamma"));
        uta::ScreenSurface::setGamma(gamma, gamma, gamma);
      }
    else if(key == "debug")
      {
        int debug_level = varconf::Config::inst()->getItem("","debug");
        DEBUG_INIT(debug_level);
        debug(std::cout << "debug_level = " << debug_level << std::endl << flush;);
      }
  }
} globals;

std::string getConfDir()
{ 
  std::string homedir;
  std::string confdir;

#ifdef  __WIN32__
    const char* homedrive;
    const char*  homepath;
    
    homedrive = getenv("HOMEDRIVE");
    homepath = getenv("HOMEPATH");
   
    if(homepath == 0)
      {
        // we're homeless...
        debugN(4,
               std::cerr << "can't find STARTDIR, looking for ";
               std::cerr << "config files in current directory."<<std::endl;
               );
        homepath=".";
      }
    
    if(homedrive == 0)
      homedrive = "";
    
    homedir = std::string(homedrive)+std::string(homepath);

    homedir = ".";

    debug(std::cout << "Using " << homedir << " to store settings. " << std::endl;);

#else   //unix
    homedir = std::string(getenv("HOME"));
    if ( homedir.empty() ) 
      {
        // we're homeless...
        debugN(4,
               std::cerr << "can't find user's HOME, looking for ";
               std::cerr << "config file in current directory."<<std::endl;
               );
        homedir = ".";
      }
#endif  // ifdef __WIN32__

  debugN(4,std::cerr << "OK." << std::endl;);

#ifdef __WIN32__
  //use no subdir in WIN32
  confdir = homedir;
  // mkdir(confdir.c_str());
#else
  confdir = homedir+"/.uclient";
  //ensure that the directory exists
  mkdir(confdir.c_str(),0755);
#endif

  return confdir;
}

void usage()
{
  std::cout << "Usage: uedit [-d <debugmask>] [-s]|[-m <atlasmap>] [-h] " << std::endl
       << std::endl
       << "-s    : uedit will parse the installed sprites.xml file, usually " << std::endl
       << "        /opt/forge/share/uclient/sprites.xml, and compresses all " << std::endl
       << "        listed sprites." <<std::endl
       << "-m <> : uedit will read the given mapfile and compress the sprites " << std::endl
       << "        for all object types contained in this file." << std::endl
       << "        Note: -s and -m are to be used exclusively ! " << std::endl
       << "-h    : print this help " << std::endl
       << std::endl;

}
/* -------------------     MAIN      ---------------------------*/   
int main (int argc, char **argv)
{
  
#ifdef INSTALL_BACKTRACE
  default_handler = signal(SIGSEGV, createBacktrace);
#endif

  ios::sync_with_stdio(true);

  std::cout  << std::endl 
        << "UEdit - sprite compiler for UClient" << std::endl
        << std::endl
        << "Version " << VERSION << std::endl 
#ifdef _WIN32
        << "Win32 - "
#endif
        << "Build " << __DATE__ <<" - "<<__TIME__ << std::endl
        << std::endl
        << "Copyright 1998,1999,2000,2001 by The Worldforge Project."<< std::endl
        << "This program is licensed under the GPL, see COPYING for details."
        << std::endl 
        << "written by Karsten-O. Laux <klaux@rhrk.uni-kl.de>" 
        << std::endl
        << std::endl;


  int exitcode = 0;
 
  DII::Module* registry;
  std::string inifile;
  std::string globalconf;
  
  try 
    {
      //getConfDir determines where the ini files resides
      //in works on Win32 as well as Unix
      inifile = getConfDir()+"/settings.ini";
      globalconf = getConfDir()+"/registry.xml";
      
      //register commandline switches
      varconf::Config::inst()->setParameterLookup('w',"width",true);
      varconf::Config::inst()->setParameterLookup('h',"height",true);
      varconf::Config::inst()->setParameterLookup('b',"bpp",true);
      varconf::Config::inst()->setParameterLookup('f',"fullscreen",true);
      varconf::Config::inst()->setParameterLookup('d',"debug",true);
      varconf::Config::inst()->setParameterLookup('i',"installdir",true);
      varconf::Config::inst()->setParameterLookup('m',"mapfile",true);
      varconf::Config::inst()->setParameterLookup('s',"sprites",true);
      varconf::Config::inst()->setParameterLookup('h',"help",true);

      debug(std::cout << "parsing settings ..." << std::endl << flush;);
      
      //parse command line
      varconf::Config::inst()->getCmdline(argc, argv);
      
      if(varconf::Config::inst()->findItem("","help"))
        {
          usage();
          exit(1);
        }

      debugN(3,std::cerr << "restoring registry ....";);
      //open $HOME/.uclient/registry.ini
      
      //load the registry module
      std::string installdir = std::string(INSTALLDIR);

      if(varconf::Config::inst()->findItem("","installdir"))
        {
          varconf::Variable _installdir = 
            varconf::Config::inst()->getItem("","installdir");
          if(_installdir.is_string())
            {
              std::cerr << "using user provided installation directory \""
                   << std::string(_installdir) << "\"" << std::endl;
              installdir = std::string(_installdir);
            }
        }
      
      // do not cache, but flush immediately
      varconf::Config::inst()->setItem("spritecache","max_memcache_size",1);
      
#if !defined(__WIN32__)
      DII::Loader::setSearchPath(installdir+std::string("/lib"));
      DII::Loader::setScriptPath(installdir+std::string("/share/uclient/scripts"));
#else
      DII::Loader::setSearchPath("");
      DII::Loader::setScriptPath(std::string("./share/uclient/scripts"));
#endif

      DII::Loader::loadModule("Registry");
      registry = DII::lookupModule("Registry");
      assert(registry);
      
      //and restore its content
      if(registry->callMethod("open",globalconf)!=DII::SUCCESS)
        {
          debugN(3,std::cerr << " error reading file " << globalconf << std::endl;);
        }
      else
        {
          debugN(3,std::cerr << "OK." << std::endl;);
        } 
      
    }
  catch(const std::exception& e)
    {
      std::cout << "UEdit: unhandled exception in initialization:"<<e.what() << std::endl;
      exit(1);
    }


  std::string install_dir;
  int x,y,bpp;
  bool fullscreen;

  //create application instance
  uta::Application app(argc, argv);

  try 
    {
      
      //get path to installation directory
      install_dir = Registry::instance()->readInstallDir();
      //preserve old execution path
      if(install_dir == "")
        throw Atlas::Message::WrongTypeException();
      
      // need to load these fonts so that libuta works ...
      std::string resource = install_dir+"/forge/fonts/flatbrush.ttf, 16, 0x202020, 0x80808080";
      REG_FONT_RES("text_font",resource.c_str());
      resource = install_dir+"/forge/fonts/flatbrush.ttf, 18, 0x202020, 0x80808080";
      REG_FONT_RES("button_font",resource.c_str());

      //--------------------------------------------------------------
      
      debug(std::cout << "UEdit starting up..." << std::endl;);

      //read screen options
      x = varconf::Config::inst()->getItem("","width");
      y = varconf::Config::inst()->getItem("","height");
      bpp = varconf::Config::inst()->getItem("","bpp");
      fullscreen = varconf::Config::inst()->getItem("","fullscreen");

      
      if(x < 800)
        {
          x = 800;
          varconf::Config::inst()->setItem("","width",x);
        }
      if(y < 600)
        {        
          y = 600;
          varconf::Config::inst()->setItem("","height",y);
        }
      
#ifdef __WIN32__
      //on windows machine there is no default depth ...
      //FIXME !!!
      if(!bpp)
        bpp = 16;
      varconf::Config::inst()->setItem("","bpp",bpp);
#endif
      
      
      //      debugN(3, std::cout << "initializing graphic mode ... ";);
      
      //init Application setting resolution and pixel depth
      //FIXME error checking ????????????????????????????????????

      //   uapp->init(x,y,bpp,fullscreen);
      debugN(3,std::cerr << "OK." << std::endl;);
      
      //initialize MediaIndexer
      MediaIndex::initialize(install_dir);
      
      // dummy call to create instance
      SpriteCache::inst();

      debugN(3, std::cerr << "ok. now fire up the gui ..." << std::endl;);
    }
  catch(const std::exception& e)
    {
      std::cout << "UEdit: unhandled exception while loading resources :"<<e.what() << std::endl;
      exit(1);
    }
  
  // if loading/configuration went fine... fire it up
  if(exitcode == 0)
    {
    
      try
        {
          //set Window Title
          //  uapp->rootWindow()->setTitle("Worldforge-UEdit");

//           /* load init script (which loads additional modules, scripts, ...) */
//           DII::Loader::loadScript("Init_uedit");
          
//           /* call main */
//           DII::callMethod("Init_uedit","main");

          if(varconf::Config::inst()->findItem("","mapfile"))
            {
              varconf::Variable mapfile = 
                varconf::Config::inst()->getItem("","mapfile");
              if(mapfile.is_string())
                readAtlasXML(std::string(mapfile));
            }          
          else if(varconf::Config::inst()->findItem("","sprites"))
            {
              readSpritesXML();
            }
          else
            {
              usage();
            }
        }
      catch(const std::exception& e)
        {
          std::cout << "UEdit: unhandled exception while loading modules :"<< e.what() << std::endl;
          exit(1);
        }

      // End of initialization
      //*******************************************

 //      try 
//         {
//           debugN(3,std::cerr << "entering uta::Application::exec()." << std::endl;);
//           //run the application
//           exitcode = uapp->exec();
//         }
//       catch(const std::exception& e)
//         {
//           std::cout << "UEdit: unhandled exception in main loop :" << e.what() << std::endl;
//           exit(1);
//         }

      // ******************************************
      // Begin of shutdown
      
      try
        {
          debugN(3,std::cerr << "returned from Application.exec()." << std::endl;);
          
          //uapp->rootWindow()->setTitle("Worldforge-UEdit shutting down ...");
          
          //write registry back 
          debugN(3,std::cerr << "storing configure information ... ";);
          registry->callMethod("flushRegistry");
          
          //write back settings
          varconf::Config::inst()->writeToFile(inifile);
          
          debugN(3,std::cerr << "OK."<<std::endl;);
        }
      catch(const std::exception& e)
        {
          std::cout << "UEdit: unhandled exception in first stage shutdown:"
						        << e.what() << std::endl;
          exit(1);
        }
    }

  try
    {

      //try to unload all modules (this checks for dependencies !)
      //DII::Loader::releaseAll();

      // Sprite Cache
      delete SpriteCache::inst();
      
      //VarConf library
      // delete varconf::Config::inst();
      
    }
  catch(...)
    {
      std::cout << "UEdit: unhandled exception in final shutdown." << std::endl;
      exit(1);
    }

  std::cout << std::endl 
       << std::endl 
       << "visit http://www.worldforge.org/ for more ..." << std::endl << std::endl;

  return exitcode;
}

