//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 05/2000

#include "my_time.h"
#include "dii/object.h"
#include "isodefs.h"

#include "tree.h"

#include <list>

#ifndef _WORLD_ENTITY_H_
#define _WORLD_ENTITY_H_

#ifndef HAVE_DRAND48
#define drand48() rand()/((double)RAND_MAX)
#endif


namespace World {


/**Entity.
   This class stores all basic entity information. 
*/
class Entity : public Tree
{
 public: 
  ///
  typedef struct 
  {
    Entity* root;
    DII::Object createArg;
  } Parameters;

  /// expects create-operation argument
  Entity(const Parameters&);
  ///
  virtual ~Entity();
  
  /** Store application specific pointer.
   *  libworld makes not use of this entry. It is meant to get used
   *  by the application. Usually a client would store a pointer
   *  to the sprite here.
  */
  void setSpriteRef(void* ref) { spriteRef_ = ref; };
  ///
  void* getSpriteRef() { return spriteRef_; };

  ///
  const std::string& name() const { return name_; };
  ///
  const std::string& id() const { return id_; };
  ///
  bool visible() const { return visible_; };
  ///
  bool unreal() const { return unreal_; };

  ///calculate new position from time since last update and current velocity
  virtual bool move(const my_time&);

  /**Operations.
     When overriding in derived classes, the new method
     should call the implementation of the superclass at the _end_
     of the method, because the baseclass's implementation also
     care for loading a new sprite if necessary.
  */
  //@{
  ///expects move-operation argument
  virtual bool move(const DII::Object&);
  ///expects set-operation argument
  virtual bool set(const DII::Object&);
  ///expects sight-operation argument
  virtual bool sight(const DII::Object&);
  ///expects sound-operation argument (hmm does this make sense)
  virtual void sound(const DII::Object&);
  ///
  virtual bool touch(const DII::Object&);
  //@}


  ///Physics
  //@{
  int orientation() const { return orientation_; };
  ///
  const std::string& mode() const { return mode_; };
  ///
  const std::string& description() const { return description_; };
  ///
  const std::string& parent() const { return parent_; };
  ///
  const std::string& ref() const { return refID_; };
  ///
  double stamp() const { return stamp_; };

  ///how much have we moved since last update ?
  const Coord& delta() const { return delta_;};

  ///position inside the local coordinate system 
  const Coord& position() const { return position_;};
  ///
  Coord absPosition() const 
    { 
      if(Tree::root_) 
	return position_+((Entity*)(Tree::root_))->absPosition();
      return position_;
    }

  ///
//  const Coord& face() const { return face_;};
  ///
  Quaternion orientQuat() const;
  ///
  const Coord& velocity() const { return velocity_; }
  ///
  float speed() const { return abs(velocity_); };
  ///
  const Coord& bbox() const { return bbox_;};
  ///
  const Coord& bboxMedian() const { return bboxMedian_;};

  ///
  bool action(const std::string& action, const my_time& restoreTime);
  ///
  bool tick(const my_time& theTime);
  ///
  bool turn(float);
  ///
  bool accelerate(float);
  ///
  bool halt();
  /** do one step backwards.
      This is intented to get used to resolv a collision as it actually
      reverts the action which caused the collision.
  */
  bool stepback();
  //@}

  /**dump all data to an Object::Map.
   */
  virtual DII::Object asObject() const;
  ///
  const DII::Object::MapType& attributes() const { return attributes_; };
  ///
  std::list<std::string> contains() const;

 protected:

  ///
  virtual void addSpriteProperties(DII::Object&) {};

  ///
  double stamp_;
  ///
  std::string id_;
  ///
  Coord position_;
  ///
  // face_ represents a vector (x, y, z) decribing orientation
//  Coord face_;
  ///
  Coord velocity_;
  ///  
  ///size of the entity, assuming a box as basic form
  Coord bbox_;
  /// center of this bounding box
  Coord bboxMedian_;
  ///
  Coord delta_;

  ///
  std::string mode_;
  ///
  std::string description_;
  ///
  std::string name_;
  ///this should be a list actually
  std::string parent_;
  ///
  std::string refID_;

  ///the sprite handle ...
  int gid_;
  ///
  int orientation_;
  ///
  bool needNewSprite_;

  ///remember time of last position update
  my_time lastUpdate_;

  ///
  bool visible_;
  ///
  bool unreal_;
  ///
  DII::Object::MapType attributes_;
  ///
  std::string savedMode_;
  ///
  long modeRestoreTime_;
  ///
  void* spriteRef_;

};

}

#endif
  
