// The Worldforge Project
// Copyright (C) 1998, 1999 The Worldforge Project
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// 
// For information about Worldforge and its authors, please contact
// the Worldforge Web Site at http://www.worldforge.org.
// 
// created by Karsten Laux 01/2000

#include "world.h"

using namespace DII;

#include "isodefs.h"
#include <varconf/Config.h>

#include "operation.h"
#include "settings.h"
#include <uta/debug.h>
#include <uta/painter.h>

#include "../player/player.h"
#include "../renderer/renderer.h"
#include "../terrain/terrain.h"
#include "../modules/console.h"

using namespace uta;

#include <fstream>

/*
 * FIXME: needs to correspond to terrain/terrain.cc ! maximum map area
 * 500 tiles * 500 tiles -> 500m * 500m -> 250000 sqrmeters 
 */
#define MAP_SIZE_MAX 500

/*
 * Agreed timescale for Acorn 0.4 release: 1 game second = 1 realife
 * second 1 game minute = 20 game seconds 1 game hour = 60 game minutes 1
 * game day = 24 game hours --> 1 game day = 8 realife hours The
 * calculation of daytime may get changed completely in the future... do
 * not rely upon it 
 */
#define SECONDS_PER_MINUTE 20
#define SECONDS_PER_HOUR (SECONDS_PER_MINUTE*60)
#define SECONDS_PER_DAY (SECONDS_PER_HOUR*24)
#define SECONDS_PER_MONTH (SECONDS_PER_DAY*30)
#define SECONDS_PER_YEAR (SECONDS_PER_MONTH*12)

// update world every 100 ms
#define UPDATE_INTERVAL 100
#define SIGHT_RADIUS 10.0

#include "misc.h"
#include "operation.h"

using namespace Atlas;

namespace World {

    World *World::instance_ = NULL;

    World *World::inst() {
        if (!instance_) {
            instance_ = new World();
            instance_->init();
        }
        return instance_;
    }

  World::World():
    root_(NULL),
        timer_(UPDATE_INTERVAL),
        overviewTimer_(2000),
        overview_(NULL),
        testing_(false),
        standalone_(false),
        collisions_(false),
        defaultActionDuration_(5.0), time_(0.0), gameTimer_(1000) {

        sightDecoder_.createOp.
            connect(slot(this, &World::processSightCreate));
        sightDecoder_.moveOp.connect(slot(this, &World::processSightMove));
        sightDecoder_.setOp.connect(slot(this, &World::processSightSet));
        sightDecoder_.deleteOp.
            connect(slot(this, &World::processSightDelete));
        sightDecoder_.actionOp.
            connect(slot(this, &World::processSightAction));

        sightDecoder_.rawObj.connect(slot(this, &World::processSight));

        soundDecoder_.talkOp.connect(slot(this, &World::processSoundTalk));


        timer_.halt();
        timer_.alarm.connect(slot(this, &World::update));

        overviewTimer_.alarm.connect(slot(this, &World::updateOverview));
        gameTimer_.alarm.connect(SigC::slot(this, &World::gameTick));

        overview_ = new uta::Surface(600, 400, Pixelformat::DISPLAY);
        overview_->setTransColor(black);
        overview_->setTransparency(true);

        smallMap_ = new uta::Surface(150, 120, Pixelformat::DISPLAY);
        smallMap_->setAlpha(50);
        smallMap_->setTransColor(black);
        smallMap_->setTransparency(true);

#ifdef DEBUG
        updates_ = 0;
        updateTime_ = 0;
#endif

        if (!varconf::Config::inst()->findItem("world", "playersight"))
            varconf::Config::inst()
                ->setItem("world", "playersight", (double)SIGHT_RADIUS);
        else {
            double tmp =
                varconf::Config::inst()->getItem("world", "playersight");
            if (tmp < 1.0) {
                varconf::Config::inst()
                    ->setItem("world", "playersight",
                              (double)SIGHT_RADIUS);
            }
        }

        varconf::Config::inst()->
            sigv.connect(SigC::slot(this, &World::configCallback));

        /*
         * later we can rig in more than one heightmap if needed ... 
         */
        heightMap_ = new HeightMap(MAP_SIZE_MAX, MAP_SIZE_MAX);
        heightMapOffsetX_ = MAP_SIZE_MAX / 2;
        heightMapOffsetY_ = MAP_SIZE_MAX / 2;
    }



    void
     World::configCallback(const std::string & section,
                           const std::string & key) {
        if (section != "world")
            return;

        if (key == "update_interval") {
            timer_.setInterval((int)varconf::Config::inst()->
                               getItem(section, key));
        } else if (key == "action_duration") {
            defaultActionDuration_ =
                varconf::Config::inst()->getItem(section, key);
        }
    }

    World::~World() {
        if (smallMap_)
            delete smallMap_;

        if (overview_)
            delete overview_;

        if (heightMap_)
            delete heightMap_;

	if (instance_ == this)
	    instance_ = 0;
    }

    HeightMap::level_t World::getGroundLevel(double x, double y) const {
        /*
         * see docs/interpolating_height.fig for details 
         */
        int x0 = (int)floor(x);
        int y0 = (int)floor(y);

        double u = x - x0;
        double v = y - y0;

        double result = 0.0;

        /*
         * important note: the heightMapOffset_ will get added in
         * getGroundLevel(int, int) !!! 
         */
         HeightMap::level_t Az = getGroundLevel(x0, y0);
        double uz = getGroundLevel(x0 + 1, y0) - Az;
        double vz = getGroundLevel(x0, y0 + 1) - Az;
        double wz = getGroundLevel(x0 + 1, y0 + 1) - Az;

        if (u - v < 0)
             result = Az + u * (wz - vz) + v * vz;
        else
             result = Az + u * uz + v * (wz - uz);

         return HeightMap::level_t(result);
    } void
     World::setGroundLevel(double x, double y, HeightMap::level_t level) {
        setGroundLevel(int (round(x)), int (round(y)), level);
    }

    HeightMap::level_t World::getGroundLevel(int x, int y) const {
        int x_ = x + heightMapOffsetX_;
        int y_ = y + heightMapOffsetY_;
        if (x_ >= 0 && x_ < heightMap_->getSizeX() && y_ >= 0
            && y_ < heightMap_->getSizeY())
             return heightMap_->getGroundLevel(x_, y_);
         return HeightMap::DEFAULT_LEVEL;
    } void
     World::setGroundLevel(int x, int y, HeightMap::level_t level) {
        int x_ = x + heightMapOffsetX_;
        int y_ = y + heightMapOffsetY_;
        if (x_ >= 0 && x_ < heightMap_->getSizeX() && y_ >= 0
            && y_ < heightMap_->getSizeY())
            heightMap_->setGroundLevel(x_, y_, level);
    }

    const uta::Surface * World::getOverview() {
        if (overview_ && root_) {
            overview_->fill(black);

            // mark current viewport in overview

            Painter *painter = new Painter(overview_);

            Pixel pixel;
            Point p;
            Entity *entity;

            Entity::iterator itr = root_->begin();
            while (itr != root_->end()) {
                entity = dynamic_cast < Entity * >(*itr);
                assert(entity);

#ifdef DEBUG
                // std::cerr << entity->id() << ", ";
#endif
                // calculate absolute Position and
                // transform to pixels
                pixel = w2p(entity->absPosition());

                p = Point(pixel.i / (TILEWIDTH / 2) +
                          overview_->width() / 2,
                          pixel.j / TILEHEIGHT + overview_->height() / 2);

                // if(dynamic_cast<Creature*>(entity))
                // {
                // //draw small red cross
                // painter->setColor(red);
                // painter->line(p-Point(1,0),p+Point(1,0));
                // painter->line(p-Point(0,1),p+Point(0,1));
                // }
                // else if(dynamic_cast<Building*>(entity))
                // {
                // //draw grey box
                // painter->setColor(lightgrey);
                // painter->line(p+Point(-4,0),p+Point(0,-4));
                // painter->line(p+Point(0,-4),p+Point(4,0));
                // painter->line(p+Point(4,0),p+Point(0,4));
                // painter->line(p+Point(0,4),p+Point(-4,0));
                // }
                // else
                // {
                // draw blue cross for everything else
                painter->setColor(blue);
                painter->line(p - Point(1, 0), p + Point(1, 0));
                painter->line(p - Point(0, 1), p + Point(0, 1));
                // }
                itr++;
            }

#ifdef DEBUG
            // std::cerr << ". " << std::endl;
#endif

            delete painter;

            return overview_;
        }

        return NULL;
    }

    void
     World::init() {

        standalone_ =
            bool(varconf::Config::inst()->getItem("flags", "standalone"));
        collisions_ =
            bool(varconf::Config::inst()->
                 getItem("flags", "collision_detection"));

        timer_.run();

        if (standalone_) {
            std::cerr << "The world is running in standalone mode !" << std::endl;
        }

        if (collisions_) {
            /*
             * initialize SOLID2 - use caching... since we move slowly - 
             * set callback 
             */

            dtEnableCaching();
            dtSetDefaultResponse(&World::collisionCB, this);
        }
        // lat least we need to create root entity:
        DII::Object::ListType parents(1, std::string("worldrouter"));

        Atlas::Objects::Root obj("world_0");
        obj.SetParents(parents);

        root_ = createNewEntity(obj.AsObject());
        root_->setRoot(NULL);

    }

    void
     World::deinit() {

        timer_.halt();

        // traverse tree and delete all entities
        if (root_) {
            delete root_;
            root_ = NULL;
        }

        delete instance_;
        instance_ = NULL;
    }

    void
     World::observ(const std::string & id) {
        // remember this ID
        idToObserv_ = id;


        // tell Renderer to clear its display list
	// if(Renderer::instance())
        // / Renderer::instance()->clearDisplayList(0);

        // now initiate an screen update, so that we are sure,
        // the character is in the viewport...
        Entity *entity = lookupEntity(idToObserv_);
        if (entity)
            rendererUpdateEntity(entity);
    }


    DII::Object World::queryObject(const std::string & id) {
        if (id.empty()) {
            DII::Object::ListType idlist;
            // provide a list of all known IDs
            Tree::iterator itr = root_->begin();
            while (itr != root_->end()) {
                idlist.push_back(((Entity *) (*itr))->id());
                itr++;
            }
            return idlist;
        }

        Entity *entity = lookupEntity(id);
        if (!entity)
            return false;

        return entity->asObject();
    }


    void
     World::processSightOp(const Atlas::Objects::Operation::Sight & sight) {
        debugN(5, std::cerr << "A sight operation arrived." << std::endl;
            );
        const Atlas::Message::Object::ListType & args = sight.GetArgs();
        if (args.empty()) {
            std::cerr << "ERROR World::processSightOp args is not list !" <<
                std::endl;
            return;
        }
        const Atlas::Message::Object & objseen = args.front();
        time_ = sight.GetSeconds();
        sightDecoder_.process(objseen);
    }

    void
     World::processSoundOp(const Atlas::Objects::Operation::Sound & sound) {
        debugN(5, std::cerr << "A sound operation arrived." << std::endl;
            )
        const Atlas::Message::Object::ListType & args = sound.GetArgs();
        if (args.empty()) {
            std::cerr << "ERROR World: sound::args is not list !" << std::endl;
            return;
        }
        const Atlas::Message::Object & objseen = args.front();
        time_ = sound.GetSeconds();
        soundDecoder_.process(objseen);
    }

    void
     World::processAppearanceOp(const Atlas::Objects::Operation::
                                Appearance & appear) {
        debugN(5, std::cerr << "A Appearance operation arrived." << std::endl;
            );
        const Atlas::Message::Object::ListType & args = appear.GetArgs();
        if (args.empty()) {
            std::cerr << "ERROR World::processAppearance args is not list !" <<
                std::endl;

            return;
        }

        std::string id;
        double stamp;
        Entity *ent;
        Atlas::Message::Object::MapType obj;
        Atlas::Message::Object::ListType::const_iterator itr =
            args.begin();
        while (itr != args.end()) {
            obj = itr->AsMap();
            id = obj["id"].AsString();

            /*
             * try to lookup Entity ... this will trigger a look
             * operation if id is unknown 
             */
            ent = lookupEntity(id);
            /*
             * if the entity is already known compare stamp to see whether 
             * it has changed 
             */
            if (ent) {
                /*
                 * Ok, we already know this id. Now we need to test if
                 * its attributes have changed ... We do this by comparing 
                 * the "stamp" attribute 
                 */
                stamp = obj["stamp"].AsFloat();
                if (stamp > ent->stamp()) {
                    /*
                     * It has changed ! --> We need to send a look
                     * operation in order to get updated information. Upon 
                     * reception of the answer the Renderer will get
                     * informed about the update 
                     */
                    if (!idAlreadyQueried_.count(id))
                        idLookList_.insert(id);
                } else {
                    /*
                     * this entity has not changed, but it re-appeared ...
                     * instruct Renderer to display it: 
                     */
                    entityChanged.emit(ent);
                }
            }
            itr++;
        }
    }

    void
     World::processDisappearanceOp(const Atlas::Objects::Operation::
                                   Disappearance & disappear) {
        debugN(5, std::cerr << "A disappearance operation arrived." << std::endl;
            );
        const Atlas::Message::Object::ListType & args =
            disappear.GetArgs();
        if (args.empty()) {
            std::cerr << "ERROR World: sight::args is not list !" << std::endl;
            return;
        }

        Entity *ent;

        Atlas::Message::Object::MapType obj;
        Atlas::Message::Object::ListType::const_iterator itr =
            args.begin();
        while (itr != args.end()) {
            obj = itr->AsMap();

            ent = lookupEntity(obj["id"].AsString());
            if (ent)
                entityDisappeared.emit(ent);
            itr++;
        }
    }

    void
     World::processSightDelete(const Atlas::Objects::Operation::
                               Delete & delOp) {
        DII::Object obj;
        std::string id;
        try {
            obj = delOp.GetArgs().front();
            id = obj.AsMap()["id"].AsString();
        }
        catch(...) {
            std::cerr << "error in operation" << std::endl;
            return;
        }

        debugN(5, std::cerr << "World::sightHandler::delete()" << std::endl;
            );

        Entity *ent = lookupEntity(id);
        if (!ent) {
            debugN(5,
                   std::cerr << "World: could not find entity id " << id <<
                   std::endl;
                );
            return;
        }

        debugN(5,
               std::cerr << "    * removing " << id << " from database" << std::endl;
            );

        if (ent->root()) {
            ent->root()->remove(id.c_str());
        }

        debugN(5,
               std::cerr << "    * removing " << id << " from moving set " <<
               std::endl;
            );
        // be sure to remove it from moving objects
        moving_.erase(ent);

        debugN(5,
               std::cerr << "    * removing " << id << " from inAction set " <<
               std::endl;
            );
        inAction_.erase(ent);

        debugN(5,
               std::cerr << "    * telling Renderer to remove entity." << std::endl;
            );
        entityDeleted.emit(ent);

        debugN(5, std::cerr << "    Finally delete the object " << std::endl;
            );

        if (collisions_)
            dtDeleteObject(ent);

        debugN(5, std::cerr << "    * deleting " << id << std::endl;
            );
        delete ent;
        debugN(5, std::cerr << "    done." << std::endl;
            );

        debugN(5, std::cerr << "    Survived :))" << std::endl;
            );

    }

    void
     World::processSightAction(const Atlas::Objects::Operation::
                               Action & actionOp) {
        std::string action;
        double duration = defaultActionDuration_;
        std::string from = actionOp.GetFrom();

        try {
            assert(!actionOp.GetArgs().empty());
            const DII::Object::MapType & arg0 =
                actionOp.GetArgs().front().AsMap();
            DII::Object::MapType::const_iterator itr;

            if ((itr = arg0.find("action")) != arg0.end())
                action = itr->second.AsString();
            if ((itr = arg0.find("duration")) != arg0.end())
                duration = itr->second.AsNum();

        }
        catch(...) {
            std::cerr << "error in action operation" << std::endl;
            return;
        }

        debugN(5,
               std::cerr << "World::sightHandler: " << from << ".action(" <<
               action << ", " << duration << ")" << std::endl;
            );

        Entity *ent = lookupEntity(from);
        if (!ent) {
            debugN(5,
                   std::cerr << "World: could not find entity " << from << std::endl;
                );
            return;
        }


        my_time theTime;
        theTime.set_now();

        ent->action(action, theTime() + long (duration * 1000));

        inAction_.insert(ent);

        rendererUpdateEntity(ent);
    }

    void
     World::processSoundTalk(const Atlas::Objects::Operation::
                             Talk & talkOp) {
        DII::Object obj;
        std::string say;
        std::string from;
        try {
            obj = talkOp.GetArgs().front();
            say = obj.AsMap()["say"].AsString();
            from = talkOp.GetFrom();
            Entity *entity = lookupEntity(from);

            Renderer::instance()->highlight(from, 50);

            if (entity)
                from = entity->name();
	    if(Console::instance())
	      Console::instance()->std_out(from + std::string(" : ") + say +
                             std::string("\n"));
        }
        catch(...) {
            std::cerr << "error in talk operation" << std::endl;
        }
    }

    void
     World::processSightMove(const Atlas::Objects::Operation::
                             Move & moveOp) {

        DII::Object obj;
        std::string id;
        try {
            obj = moveOp.GetArgs().front();
            id = obj.AsMap()["id"].AsString();
        }
        catch(...) {
            std::cerr << "error in sight-move operation" << std::endl;
            return;
        }

        Entity *entity = lookupEntity(id);
        if (!entity) {
            debugN(5,
                   std::cerr << "World received move-Operation for unknown ID "
                   << id << std::endl;
                );
            createEntity(obj);
            // check if creation of entity worked
            entity = lookupEntity(id);
            assert(entity);
        }
        // call move operation of entity
        if (entity->move(obj)) {
            checkRoot(entity);
            // object has changed -> tell Renderer about it
            rendererUpdateEntity(entity);
        } else {
            checkRoot(entity);
            // object has only moved -> tell Renderer about it
            rendererMoveEntity(entity);
        }

        updateMovingObjects(entity);
        return;
    }

    void
     World::updateMovingObjects(Entity * entity) {
        if (entity->speed() > 0.0) {
            // add entity to list of moving objects
            // (moving_ is a set and thus the reference is only
            // stored once)
            moving_.insert(entity);
        } else {
            // remove it from list of moving objects
            moving_.erase(entity);
        }
    }

    void
     World::processSightSet(const Atlas::Objects::Operation::Set & setOp) {
        DII::Object obj;
        std::string id;
        try {
            obj = setOp.GetArgs().front();
            id = obj.AsMap()["id"].AsString();
        }
        catch(...) {
            std::cerr << "error in sight-set operation" << std::endl;
            return;
        }

        // update position of this object in DB
        Entity *entity = lookupEntity(id);
        if (!entity) {
            debugN(5,
                   std::cerr << "World received set-Operation for unknown ID "
                   << id << std::endl;
                );
            createEntity(obj);
            // check if creation of entity worked
            entity = lookupEntity(id);
            assert(entity);
        }
        // call set operation of entity
        if (entity->set(obj)) {
            checkRoot(entity);
            // object has actually changed -> tell Renderer about it
            rendererUpdateEntity(entity);
        }

        updateMovingObjects(entity);
        return;
    }

    std::string World::createEntity(const DII::Object & obj) {
        if (!obj.IsMap())
            return "";

        // create new entity object
        Entity *newone = createNewEntity(obj);
        assert(newone);

        if (collisions_) {
            /*
             * create a bounding box of appropriate size,
             * bind this shape to the entity object
             */
            dtCreateBox(newone, newone->bbox().x, newone->bbox().y, newone->bbox().z);
            // move object to starting position
            dtTranslate(newone->position().x - 0.25,
                        newone->position().y - 0.25, newone->position().z);

        }
        // check for root entity
        Entity *root;
        if (root_) {
            root = lookupEntity(newone->ref());
            if (!root) {
                debugN(5, std::cerr << "World: could not identify root "
                       << newone->ref() << "." << std::endl
                       <<
                       "creating empty object and querying server about."
                       << newone->ref() << " !" << std::endl;
                    );

                Atlas::Message::Object::MapType root_desc;
                root_desc["id"] = newone->ref();
                root_desc["loc"] = std::string("world_0");

                Atlas::Message::Object::ListType parents(1, "object");
                root_desc["parents"] = parents;

                Atlas::Objects::Operation::Create createOp =
                    Atlas::Objects::Operation::Create::Instantiate();
                DII::Object::ListType args(1, root_desc);
                createOp.SetArgs(args);
                processSightCreate(createOp);

                root = lookupEntity(newone->ref());
                assert(root);

            }

            debugN(5, std::cerr << "World: adding " << newone->id() << " to "
                   << root->id() << std::endl;
                );

            // ok found root, insert the new entity 
            root->insert(newone->id().c_str(), newone);

            // tell newone about it:
            newone->setRoot(root);
        } else {
            // this should not happen !!!!
            assert(false);
            // std::cerr << "World: setting "<<id<<" as root entity."<<std::endl;
            // root_ = newone;
        }

        return newone->id();
    }


    void
     World::processSightCreate(const Atlas::Objects::Operation::
                               Create & createOp) {
        DII::Object obj;
        std::string id;
        try {
            obj = createOp.GetArgs().front();
            id = obj.AsMap()["id"].AsString();
        }
        catch(...) {
            std::cerr << "error in sight-create operation" << std::endl;
            return;
        }

        Entity *newone;

        idAlreadyQueried_.erase(id);

        if ((newone = lookupEntity(id)) != NULL) {
            if (newone->sight(obj))
                rendererUpdateEntity(newone);
            return;
        }
        // create new entity object
        newone = createNewEntity(obj);
        assert(newone);

        if (collisions_) {
            /*
             * create a bounding box of appropriate size,
             * bind this shape to the entity object
             */
            dtCreateBox(newone, newone->bbox().x, newone->bbox().y, newone->bbox().z);
            // move object to starting position
            dtTranslate(newone->position().x - 0.25,
                        newone->position().y - 0.25, newone->position().z);

        }
        // check for root entity
        Entity *root;
        if (root_) {
            root = lookupEntity(newone->ref());
            if (!root) {
                debugN(5, std::cerr << "World: could not identify root "
                       << newone->ref() << "." << std::endl
                       <<
                       "creating empty object and querying server about."
                       << newone->ref() << " !" << std::endl;
                    );

                Atlas::Message::Object::MapType root_desc;
                root_desc["id"] = newone->ref();
                root_desc["loc"] = std::string("world_0");

                Atlas::Message::Object::ListType parents(1, "object");
                root_desc["parents"] = parents;

                Atlas::Objects::Operation::Create createOp =
                    Atlas::Objects::Operation::Create::Instantiate();
                DII::Object::ListType args(1, root_desc);
                createOp.SetArgs(args);
                processSightCreate(createOp);

                root = lookupEntity(newone->ref());
                assert(root);

            }

            debugN(5, std::cerr << "World: adding " << newone->id() << " to "
                   << root->id() << std::endl;
                );

            // ok found root, insert the new entity 
            root->insert(id.c_str(), newone);

            // tell newone about it:
            newone->setRoot(root);
        } else {
            // this should not happen !!!!
            assert(false);
            // std::cerr << "World: setting "<<id<<" as root entity."<<std::endl;
            // root_ = newone;
        }

        if (newone) {
            // add new entity to displaylist
            rendererUpdateEntity(newone);
        } else {
            debugN(5, std::cerr << "Critical: World failed to create Entity "
                   << id << std::endl;
                );
            return;
        }

        updateMovingObjects(newone);
        return;
    }


    void
     World::processSight(const Atlas::Message::Object & obj) {
        std::string id;
        std::string type;


        if (!obj.IsMap())
            return;
        const Atlas::Message::Object::MapType & args = obj.AsMap();
        Atlas::Message::Object::MapType::const_iterator itr;


        try {
            if ((itr = args.find("id")) != args.end())
                id = itr->second.AsString();
        }
        catch(...) {
            std::cerr << "error in sight operation" << std::endl;
            return;
        }



        try {
            if ((itr = args.find("objtype")) != args.end())
                type = itr->second.AsString();
        }
        catch(...) {
            // objtype is not always set ... so continue
        }

        if (type == "op") {
            // handle sight of operation here ...
            return;
        }

        idAlreadyQueried_.erase(id);

        Entity *entity = lookupEntity(id);



        if (entity) {
            // call entity's sight operation
            if (entity->sight(obj)) {
                checkRoot(entity);
                // if actually changed, provide updated object to renderer
                rendererUpdateEntity(entity);
            }
        } else {
            // entity is unknown till now, so we create it 

            Atlas::Objects::Operation::Create createOp =
                Atlas::Objects::Operation::Create::Instantiate();
            DII::Object::ListType args(1, obj);
            createOp.SetArgs(args);

            processSightCreate(createOp);

            // may return NULL ! 
            entity = lookupEntity(id);
            if (!entity)
                std::cerr << "World: failed to create object " << id << " !!" <<
                    std::endl;
            else
                rendererUpdateEntity(entity);
        }


        try {
            // extract any contained ids:
            if ((itr = args.find("contains")) != args.end()) {
                const DII::Object::ListType & contains =
                    itr->second.AsList();

                debugN(5,
                       if (entity) std::cerr << entity->id() << " contains :";);

                for (DII::Object::ListType::const_iterator i =
                     contains.begin(); i != contains.end(); i++) {
                    debugN(5, std::cerr << i->AsString() << " ";
                        );
                    // check if this id is known, if not: ask server via
                    // Player::look
                    Entity *ent = lookupEntity(i->AsString());
                    if (ent) {
                        entityChanged.emit(ent);
                        if (!idAlreadyQueried_.count(ent->id()))
                            idLookList_.insert(ent->id());
                    }

                }
                debugN(5, std::cerr << std::endl;
                    );
            }
        }
        catch(...) {
            std::cerr << "World::processSight() error in contains" << std::endl;
        }


        if (entity)
            updateMovingObjects(entity);

        return;
    }




    void
     World::rendererUpdateEntity(const Entity * entity) {

        if (collisions_) {
            if (!testing_) {
                dtSelectObject((void *)entity);
                dtLoadIdentity();
                dtTranslate(entity->position().x,
                            entity->position().y, entity->position().z);
            }
        }

        entityChanged.emit(entity);

    }

    void
     World::rendererMoveEntity(const Entity * entity) {
        // if(!entity->visible())
        // return;

        if (collisions_) {
            if (!testing_) {
                dtSelectObject((void *)entity);
                dtLoadIdentity();
                dtTranslate(entity->position().x,
                            entity->position().y, entity->position().z);
            }
        }

        entityMoved.emit(entity);

    }


    void
     World::updateOverview() {
        if (!varconf::Config::inst()->getItem("world", "map_overlay")) {
            Renderer::instance()->setMap(0);
            return;
        }

        Entity *centerEnt = lookupEntity(idToObserv_);
        if (centerEnt) {
            Coord center = centerEnt->absPosition();

            const uta::Surface * terrain = NULL;
            try {
                terrain = Terrain::instance()->getTerrainMap();
            }
            catch(...) {
            };

            if (terrain) {
                uta::Rect dest(0, 0, smallMap_->width(), smallMap_->height());

                Pixel m = w2p(center);
                uta::Point mp = uta::Point(m.i / (TILEWIDTH / 2) + terrain->width() / 2,
                                           m.j / TILEHEIGHT + terrain->height() / 2);

                uta::Rect src(mp.x - smallMap_->width() / 2,
                              mp.y - smallMap_->height() / 2,
                              smallMap_->width(), smallMap_->height());

                terrain->blit(smallMap_, dest, src);
            } else {
                smallMap_->clear();
            }

            Painter *painter = new Painter(smallMap_);

            Pixel pixel;
            Point p;
            Entity *entity;

            Entity::iterator itr = root_->begin();
            while (itr != root_->end()) {
                entity = dynamic_cast < Entity * >(*itr);
                assert(entity);

                // calculate absolute Position and
                // transform to pixels
                pixel = w2p(entity->absPosition() - center);

                p = Point(pixel.i / (TILEWIDTH / 2) + smallMap_->width() / 2,
                          pixel.j / TILEHEIGHT + smallMap_->height() / 2);

                // if(dynamic_cast<Creature*>(entity))
                // {
                // //draw small red cross
                // painter->setColor(red);
                // painter->line(p-Point(1,0),p+Point(1,0));
                // painter->line(p-Point(0,1),p+Point(0,1));
                // }
                // else if(dynamic_cast<Building*>(entity))
                // {
                // //draw grey box
                // painter->setColor(lightgrey);
                // painter->line(p+Point(-4,0),p+Point(0,-4));
                // painter->line(p+Point(0,-4),p+Point(4,0));
                // painter->line(p+Point(4,0),p+Point(0,4));
                // painter->line(p+Point(0,4),p+Point(-4,0));
                // }
                // else
                // {
                // draw blue cross for everything else
                painter->setColor(lightblue);
                painter->line(p - Point(1, 0), p + Point(1, 0));
                painter->line(p - Point(0, 1), p + Point(0, 1));
                // }
                itr++;
            }


            delete painter;

            // //this is not very clever ... smallMap_* does not change !
            Renderer::instance()->setMap(smallMap_);
        }
    }

    void
     World::update() {
        if (!root_)
            return;

        /*
         * IMPORTANT: we do not check if the root entity ("world_0") has
         * moved ! 
         */


        // get next id to look
        if (!idLookList_.empty()) {
            std::set < std::string >::iterator itr = idLookList_.begin();
	    if(Player::instance())
              Player::instance()->look(*itr);
            idAlreadyQueried_.insert(*itr);
            idLookList_.erase(itr);
        }

#ifdef DEBUG
        my_time start,
         end;
        start.set_now();
#endif

        my_time theTime;
        theTime.set_now();


        // only care for moving objects !
        std::set < Entity * >::iterator itr = moving_.begin();
        while (itr != moving_.end()) {
            // no test here, since all these entities _are_ moving
            (*itr)->move(theTime);
            rendererMoveEntity(*itr);
            itr++;
        }

        Entity *toberemoved;
        itr = inAction_.begin();
        while (itr != inAction_.end()) {
            /*
             * if tick returns true, the entity has ended the action,
             * remove it from the set 
             */
            if ((*itr)->tick(theTime)) {
                rendererUpdateEntity(*itr);
                toberemoved = *itr;
            }
            itr++;
            inAction_.erase(toberemoved);
        }

        if (collisions_) {
            // actually do collision test (now collision might be called)
            // ... during this the screen position of some entities might
            // get changed
            // again
            testing_ = true;
            dtTest();
            testing_ = false;
        }

        debugN(5,
               end.set_now();
               updateTime_ += end() - start();
               updates_++; if (updates_ == 100) {
               std::cerr << "World::update() took " << updateTime_ /
               updates_ << "ms " << "for " << (unsigned int)root_->
               leaves().size() << " objects." << std::endl; updates_ = 0;
               updateTime_ = 0;}
        );

        // update the screen
        Renderer::instance()->redraw();

    }



    void
     World::sendMoveOp(const Entity * entity, const std::string & from) {

    }

/*
    void
     World::addActor(const DII::Object & obj) {
        return;
    }

    void
     World::removeActor(const DII::Object & obj) {
        return;
    }
*/
    Entity *World::lookupEntity(const std::string & id) {
        // std::cerr << "World: looking up " << id << " ... ";
        Entity *entity = NULL;

        if (root_) {
            if (id.empty()) {
#ifdef DEBUG
                std::cerr << "WARNING: lookupEntity() with empty id called ! "
                    << "Will return entity \"world_0\" ! " << std::endl;
#endif
                return root_;
            }

            if (root_->id() == id)
                entity = root_;
            else
                entity = dynamic_cast < Entity * >(root_->find(id));
        }

        if (!entity) {
            debug(std::cerr << "World::lookupEntity(" << id << ") failed !!" <<
                  std::endl;
                );
            if (id.find("_map") != std::string::npos) {
                std::cerr <<
                    "CRITICAL: client-side and server-side database are out of sync !! "
                    << std::endl <<
                    "Either your client installation needs an update or the server is out of date !"
                    << std::endl <<
                    "The error occured during lookup of id: " << id << std::endl;
            }
            if (!idAlreadyQueried_.count(id))
                idLookList_.insert(id);
        }

        return entity;
    }


    void
     World::checkRoot(Entity * entity) {
        std::string rootID = entity->ref();
        Entity *old_root,
        *new_root;

        old_root = dynamic_cast < Entity * >(entity->root());

        if (old_root && rootID == old_root->id())
            return;

        // we need to correct root pointer

        // 1. search for new root
        new_root = lookupEntity(rootID);

        // 1.5 safely treat aliases (different ids, but actually same
        // Entity)
        if (old_root && old_root == new_root)
            return;
        if (new_root == entity)
            return;

        if (!new_root) {

            debugN(5, std::cerr << "World: could not identify root "
                   << rootID << "." << std::endl
                   << "creating empty object and querying server about."
                   << rootID << " !" << std::endl;
                );

            Atlas::Message::Object::MapType root_desc;
            root_desc["id"] = rootID;
            root_desc["loc"] = std::string("world_0");

            Atlas::Message::Object::ListType parents(1, "object");
            root_desc["parents"] = parents;

            Atlas::Objects::Operation::Create createOp =
                Atlas::Objects::Operation::Create::Instantiate();
            DII::Object::ListType args(1, root_desc);
            createOp.SetArgs(args);
            processSightCreate(createOp);

            new_root = lookupEntity(rootID);
            assert(new_root);

        }
        // 2. check for cyclic dependency
        if (new_root->root() == entity) {
            std::cerr << "World::checkRoot(): cyclic dependency found !" <<
                std::endl;
            std::cerr << "(" << entity->id() << " moves in " << new_root->
                id() << " which is in ";
            std::cerr << entity->id() << ")" << std::endl;
            std::cerr << " ... will ignore move-operation." << std::endl;
            return;
        }
        // 3. remove entity from old root
        if (old_root)
            old_root->remove(entity->id());
        // 4. insert entity into new root
        new_root->insert(entity->id(), entity);
        // 5.tell entity about this
        entity->setRoot(new_root);
    }


    void
     World::collisionCB(void *inst,
                        DtObjectRef e1, DtObjectRef e2) {
#ifdef DEBUG
        // std::cerr << "clash ! "<<std::endl;
#endif

        World *me = reinterpret_cast < World * >(inst);
        if (me)
            me->collision(e1, e2);
    }

    void
     World::collision(DtObjectRef e1, DtObjectRef e2) {
        Entity *ent1 = reinterpret_cast < Entity * >(e1);
        Entity *ent2 = reinterpret_cast < Entity * >(e2);

        if (!ent1 || !ent2) {
            std::cerr << "World::collision objects of unknown type." << std::endl;
            return;
        }
        // ignore any collision with a nested object (e.g. an acorn inside 
        // a pig)
        if (ent1->root() != root_ || ent1->root() != root_) {
            return;
        }

#ifdef DEBUG
        /*
         * std::cerr << "World::collision between " << ent1->name() << " and " 
         * << ent2->name() << std::endl; 
         */
#endif

        // ok, so far, but what now ?

        /*
         * idea : Set the velocity of all entities to zero and do a step
         * backwards. 
         */
        // Coord dist;
        // float delta;
        // float min_dist = ( abs(ent1->bbox()) + abs(ent2->bbox()) ) / 2;

        // if(ent1->speed() > 0.01)
        // {
        // ent1->halt();

        // //vector pointing from ent2 to ent1
        // dist = ent1->position() - ent2->position();

        // if(abs(dist) < min_dist)
        // {
        // //move entities apart
        // DII::Object loc, coords, moveOp;

        // delta = abs(dist);
        // if(delta < 0.001)
        // delta = 0.001;

        // Coord newpos = ent1->position() + dist / delta *
        // (min_dist-delta);

        // coords = DII::Object::mkList();
        // coords.append(newpos.x);
        // coords.append(newpos.y);
        // coords.append(newpos.z);
        // loc = DII::Object::mkMap();
        // loc.set("ref",ent1->ref());
        // loc.set("coords",coords);
        // moveOp.set("loc",loc);

        // ent1->move(moveOp);
        // }

        // if(ent1->id() == idToObserv_)
        // sendMoveOp(ent1);

        // rendererUpdateEntity(ent1);

        // }

        // if(ent2->speed() > 0.01)
        // {
        // ent2->halt();

        // //vector pointing from ent1 to ent2
        // dist = ent2->position() - ent1->position();

        // if(abs(dist) < min_dist)
        // {
        // //move entities apart
        // DII::Object loc, coords, moveOp;

        // delta = abs(dist);
        // if(delta < 0.001)
        // delta = 0.001;

        // Coord newpos = ent2->position() + dist / delta *
        // (min_dist-delta);

        // coords = DII::Object::mkList();
        // coords.append(newpos.x);
        // coords.append(newpos.y);
        // coords.append(newpos.z);
        // loc = DII::Object::mkMap();
        // loc.set("ref",ent2->ref());
        // loc.set("coords",coords);
        // moveOp.set("loc",loc);

        // ent2->move(moveOp);
        // }
        // if(ent2->id() == idToObserv_)
        // sendMoveOp(ent2);
        // rendererUpdateEntity(ent2);

        // }


        // //vector pointing from ent2 to ent1
        // dist = ent1->position() - ent2->position();

        // if(abs(dist) < min_dist)
        // {
        // //move entities apart
        // DII::Object loc, coords, moveOp;

        // delta = abs(dist);
        // if(delta < 0.001)
        // delta = 0.001;


        // Coord newpos = ent1->position() + dist / delta *
        // (min_dist-delta) / 2;

        // coords = DII::Object::mkList();
        // coords.append(newpos.x);
        // coords.append(newpos.y);
        // coords.append(newpos.z);
        // loc = DII::Object::mkMap();
        // loc.set("ref",ent1->ref());
        // loc.set("coords",coords);
        // moveOp.set("loc",loc);

        // ent1->move(moveOp);

        // if(ent1->id() == idToObserv_)
        // sendMoveOp(ent1);

        // rendererUpdateEntity(ent1);

        // newpos = ent2->position() - dist / delta * (min_dist-delta) /
        // 2;

        // coords = DII::Object::mkList();
        // coords.append(newpos.x);
        // coords.append(newpos.y);
        // coords.append(newpos.z);
        // loc = DII::Object::mkMap();
        // loc.set("ref",ent2->ref());
        // loc.set("coords",coords);
        // moveOp.set("loc",loc);

        // ent2->move(moveOp);

        // if(ent2->id() == idToObserv_)
        // sendMoveOp(ent2);

        // rendererUpdateEntity(ent2);
        // }

    }


    Entity *World::createNewEntity(const DII::Object & obj) {
        Entity *entity = NULL;

        Entity::Parameters param;
        param.createArg = obj;
        param.root = NULL;

        entity = new Entity(param);
        entity->sight(obj);

        return entity;

    }

    void
     World::gameTick() {
        time_ += 1.0;
        timeTick.emit();
    }

    DII::Object World::getTime()const {
        DII::Object::MapType result;
        result["tick"] = time_;
        result["second"] = int (time_) % SECONDS_PER_MINUTE;
         result["minute"] =
            int (time_) % SECONDS_PER_HOUR / SECONDS_PER_MINUTE;
         result["hour"] = int (time_) % SECONDS_PER_DAY / SECONDS_PER_HOUR;
         result["day"] = int (time_) % SECONDS_PER_MONTH / SECONDS_PER_DAY;
         result["month"] =
            int (time_) % SECONDS_PER_YEAR / SECONDS_PER_MONTH;
         result["year"] = int (time_) / SECONDS_PER_YEAR;
         return result;
}}                              // namespace
