//  The Worldforge Project
//  Copyright (C) 1998, 1999 The Worldforge Project
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  For information about Worldforge and its authors, please contact
//  the Worldforge Web Site at http://www.worldforge.org.
//
// created by Karsten Laux 01/2000

#ifndef _WORLD_H_
#define _WORLD_H_

#include "dii/object.h"

#include <uta/timer.h>
#include <uta/surface.h>

#include "toolbox/local_hash_map.h"
#include <set>

#include "my_time.h"
#include "operation.h"
#include "decoders.h"
#include "heightmap.h"
#include "entity.h"

#include "solid.h"

namespace World {

/**
 * The World class represents all the objects in our 
 * simulation.  It is a 'mirror' of the objects on the
 * server.
 */
class World : virtual public SigC::Object
{
 
 public:
  
  ///
  virtual ~World();
  ///
  static World* inst();

  /**query a describing object for the given id.
     If id is an empty string, a list of all known ids is returned.
   */
  DII::Object queryObject(const std::string& id);
  ///
  Entity* lookupEntity(const std::string& id);

  ///
  void init();
  ///
  void deinit();

  ///returns id of newly created entity
  std::string createEntity(const DII::Object&);

  ///returns pointer to uta::Surface* showing overview of world
  const uta::Surface* getOverview();
 
  /** add actor to list of valid actors.
      actor=<name of controlling module>
      entity=<id of controlled entity>
      (any information about this entity is passed
       to method "update" of the controlling module)
  */
/*
  void addActor(const DII::Object&);
  ///
  void removeActor(const DII::Object&);
*/
  ///
  void observ(const std::string& id);
  
  /** returns game time object
   *  containing nums: "year", "month", "day", "hour", "minute", "second", "tick"
   */
  DII::Object getTime() const ;
  ///
  void processSightOp(const Atlas::Objects::Operation::Sight& sight);
  ///
  void processSoundOp(const Atlas::Objects::Operation::Sound& sound);
  ///
  void processAppearanceOp(const Atlas::Objects::Operation::Appearance& appear);
  ///
  void processDisappearanceOp(const Atlas::Objects::Operation::Disappearance& disappear);

  ///
  void update();
  ///
  void updateOverview();

  ///
  HeightMap::level_t getGroundLevel(double x, double y) const;
  ///
  void setGroundLevel(double x, double y, HeightMap::level_t level);
  ///
  HeightMap::level_t getGroundLevel(int x, int y) const;
  ///
  void setGroundLevel(int x, int y, HeightMap::level_t level);

  ///
  SigC::Signal1<void, const Entity*>entityDisappeared;
  ///
  SigC::Signal1<void, const Entity*>entityMoved;
  ///
  SigC::Signal1<void, const Entity*>entityChanged;
  ///
  SigC::Signal1<void, const Entity*>entityDeleted;
  ///
  SigC::Signal1<void, const Entity*>entityCreated;
  ///
  SigC::Signal0<void> timeTick;

protected:

  ///
  World();
  ///
  static World* instance_;
  ///
  void gameTick();

  ///
  void configCallback(const std::string& section, const std::string& key);  
  ///
  SightDecoder sightDecoder_;
  ///
  SoundDecoder soundDecoder_;

  ///
  void rendererUpdateEntity(const Entity* entity);
  ///
  void rendererMoveEntity(const Entity* entity);

  ///
  static void collisionCB(void*, DtObjectRef, DtObjectRef);
  ///
  void collision(DtObjectRef, DtObjectRef);

  ///
  void sendMoveOp(const Entity*, const std::string& from="");

  ///
  void processSoundTalk(const Atlas::Objects::Operation::Talk&);
  ///
  void processSightCreate(const Atlas::Objects::Operation::Create&);
  ///
  void processSightMove(const Atlas::Objects::Operation::Move&);
  ///
  void processSightSet(const Atlas::Objects::Operation::Set&);
  ///
  void processSightDelete(const Atlas::Objects::Operation::Delete&);
  ///
  void processSightAction(const Atlas::Objects::Operation::Action&);
  ///
  void processSight(const Atlas::Message::Object&);


  ///
  void checkRoot(Entity*);
  ///
  void updateMovingObjects(Entity*);

  ///
/*   float current_lag() const  */
/*     { return DII::callMethod("Cyphesis","currentLag").asFloat(); } */

  /**root entity.
     All known entities are contained. (replaces objectDB)
  */
  Entity* root_;

  ///
  std::set<std::string> idLookList_;
  ///
  std::set<std::string> idAlreadyQueried_;
  ///
  std::map<std::string, std::string> controllers_;
  ///
  Entity* playerEntity_;

  ///
  uta::Timer timer_;
  ///
  uta::Timer overviewTimer_;

  ///
  uta::Surface* overview_;
  ///
  uta::Surface* smallMap_;
  ///
  Entity* createNewEntity(const DII::Object&);
  ///
  std::string idToObserv_;
  ///
  Coord posToObserv_;
  ///
  const Entity* entityToObserv_;
  
#ifdef DEBUG
  long int updates_;
  long int updateTime_;
#endif
  ///
  bool testing_;

  ///
  std::set<Entity*> moving_;
  ///
  bool standalone_;
  ///
  bool collisions_;
  ///
  HeightMap* heightMap_;
  ///
  int heightMapOffsetX_;
  ///
  int heightMapOffsetY_;
  ///
  std::set<Entity*> inAction_;
  ///
  double defaultActionDuration_;
  /** game time.
   *  Gets updated by every entity update and is directly
   *  taken from the "stamp" attribute.
   */
  double time_;
  ///
  uta::Timer gameTimer_;
 
};
}

#endif
